#!/usr/bin/env python3.2

import sys
import re
import argparse
import logging
from itertools import groupby
from collections import OrderedDict
from xigt import XigtCorpus, Igt, Tier, Item, Metadata, Meta
from xigt.codecs import xigtxml

def blk_to_igt(blk, i):
    assert blk[0].startswith('\\ref'), 'Ref line: {}'.format(blk[0])
    ref = blk[0].replace('\\ref ', '', 1) # \\ref filename.NUM
    attrs = {'ToolboxRef':ref}
    tiers = []
    for line in blk[1:]:
        key, *toks = line.split()
        if key == '\\mph':
            tiers.append(Tier(type='words', id='w',
                items=reattach_morphemes(toks)))
            tiers.append(Tier(type='morphemes', id='m',
                items=make_items('m', toks)))
        elif key == '\\mgl':
            tiers.append(Tier(type='glosses', id='g',
                items=make_items('g', toks)))
        elif key == '\\eng':
            tiers.append(Tier(type='translation', id='t',
                              items=[Item(id='t1', content=' '.join(toks))]))
    return Igt(id='i{}'.format(i), tiers=tiers, attributes=attrs)

reattach_re = re.compile(r'\s*([-=])\s*')
def reattach_morphemes(toks):
    # it's too bad I have to rejoin the string just to split again. oh well
    words = reattach_re.sub(r'\1', ' '.join(toks)).split()
    return make_items('w', words)

def make_items(idprefix, data):
    return [Item(id='{}{}'.format(idprefix, i), content=datum)
            for i, datum in enumerate(data)]

def get_txt_tier(tier, reattach=False):
    string = ' '.join(item.content for item in tier)
    if reattach:
        string = reattach_re.sub(r'\1', string)
    return string

ctn_metadata = Metadata(
    type='xigt-meta',
    content=[
        Meta(
            type='language',
            attributes=OrderedDict([
                ('name', 'Chintang'),
                ('iso-639-3', 'ctn')
            ])
        )
    ]
)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('files', nargs='+', metavar='FILE',
            help='input Toolbox files')
    parser.add_argument('-O', '--odin', action='store_true',
            help='output to ODIN-formatted text instead of XML')
    parser.add_argument('-l', '--old-lang-line', action='store_true',
            help='if --odin is specified, use the old style language line '
                 '(e.g. "stage3_lang_chosen: ...")')
    parser.add_argument('-v', '--verbose', action='count',
            dest='verbosity', default=2,
            help='increase logging verbosity')
    args = parser.parse_args()
    logging.basicConfig(level=50-(args.verbosity*10))

    xc = XigtCorpus(metadata=ctn_metadata)
    clean = lambda f: map(str.strip, open(f, 'r'))
    for f in args.files:
        # get list of IGT blocks (groupby gets adjacent lines until a newline)
        blks = [list(g) for k, g in groupby(clean(f), lambda s: not s)
                if not k]
        for i, blk in enumerate(blks):
            if not blk[0].startswith('\\ref'): continue
            igt = blk_to_igt(blk, i)
            if igt:
                xc.add(igt)
            else:
                raise Exception('\n'.join(['Error processing block:'] + blk))

    if args.odin:
        lang_line = 'language: chintang (ctn)'
        if args.old_lang_line:
            lang_line = 'stage3_lang_chosen: chintang (ctn)\n'\
                        'lang_code: chintang (ctn)\n'\
                        'note: lang_chosen_idx=0'
        for i, igt in enumerate(xc.igts):
            i *= 3 # 3 ODIN-like lines per item
            docstring = 'doc_id={} {} {} L G T'.format(
                    igt.attributes.get('ToolboxRef', '?'), i+1, i+3
            )
            print('{}\n{}\n'
                  'line={} tag=L: {}\n'
                  'line={} tag=G: {}\n'
                  'line={} tag=T: {}\n\n'
                  .format(docstring, lang_line,
                      i+1, get_txt_tier(igt.get('w')),
                      i+2, get_txt_tier(igt.get('g'), reattach=True),
                      i+3, get_txt_tier(igt.get('t'))
                  )
            )
    else:
        print(xigtxml.dumps(xc, pretty_print=True))
