#!/usr/bin/env python3.2

import sys
import os.path
import logging
import re
import unicodedata
import argparse
from collections import OrderedDict, defaultdict
from itertools import chain
from xigt import XigtCorpus, Igt, Tier, Item, Metadata, Meta
from xigt.codecs import xigtxml

### IMPORTING ################################################################

def regenerator(gen):
    for x in gen:
        resend = (yield x)
        while resend is not None:
            yield None
            resend = (yield resend)

doc_re = re.compile(r'doc_id=(?P<id>\S+) '
                    r'(?P<start>\d+) '
                    r'(?P<end>\d+) '
                    r'(?P<linetypes>.*)')
def odin_igts(lines, replacement_char='\uFFFD'):
    i = 0
    j = 0
    line_iterator = regenerator(l for l in lines)
    for line in line_iterator:
        doc = doc_re.match(line)
        if doc is None:
            if 'doc_id=' in line:
                logging.warning('Possible ODIN instance missed: {}'
                                .format(line))
            continue
        i += 1
        header_lines = []
        while not line.startswith('line='):
            header_lines.append(line)
            line = next(line_iterator)
        # we've seen line=... send it back for now
        line_iterator.send(line)
        lang, iso639 = get_best_lang_match(header_lines)
        if lang is None or iso639 is None:
            logging.warning('Failed to get language or language code for '
                            'document {}, lines {}--{}.'
                            .format(doc.group('id'),
                                    doc.group('start'),
                                    doc.group('end')))
        else:
            logging.debug('Document {}, lines {}--{}, Language: {}, '
                          'ISO-639-3: {}'
                          .format(doc.group('id'),
                                  doc.group('start'),
                                  doc.group('end'),
                                  lang, iso639))
        j += 1
        yield {'doc_id': doc.group('id'),
               'line_range': (doc.group('start'), doc.group('end')),
               'line_types': doc.group('linetypes').split(),
               'language': lang,
               'iso-639-3': iso639,
               'lines': odin_lines(line_iterator, replacement_char),
               'header_lines': header_lines
              }

lang_chosen_re = re.compile(r'(?P<name>.*) \((?P<iso639>[^)]+)\)\s*$',
                            re.UNICODE)
stage2_LN_re = re.compile(r'stage2_LN_lang_code: (?P<name>.*) '
                          r'\([^,]+, (?P<iso639>[^)]+)\)')
chosen_idx_re = re.compile(r'lang_chosen_idx=(?P<idx>[-0-9]+)')
def get_best_lang_match(lines):
    lang_lines = dict(l.split(':', 1) for l in lines if ':' in l)
    # find best match
    lang = iso639 = None
    if 'stage3_lang_chosen' in lang_lines:
        match = lang_chosen_re.search(lang_lines['stage3_lang_chosen'])
        if match:
            lang = match.group('name')
            iso639 = match.group('iso639')
    elif 'stage2_lang_chosen' in lang_lines:
        match = lang_chosen_re.search(lang_lines['stage2_lang_chosen'])
        if match:
            lang = match.group('name')
            iso639 = match.group('iso639')
    elif 'stage2_LN_lang_code' in lang_lines:
        first = lang_lines['stage2_LN_lang_code'].split('||', 1)[0]
        match = stage2LN_re.match(first)
        if match:
            lang = match.group('name')
            iso639 = match.group('iso639')
    elif 'lang_code' in lang_lines and \
         'lang_chosen_idx' in lang_lines['note']:
        match = chosen_idx_re.search(lang_lines['note'])
        if match:
            idx = int(match.group('idx'))
            if idx != -1:
                langstring = lang_lines['lang_code'].split('||')[idx]
                match = lang_chosen_re.match(langstring)
                if match:
                    lang = match.group('name')
                    iso639 = match.group('iso639')
    return lang, iso639

line_re = re.compile(r'line=(?P<line>\d+) '
                     r'tag=(?P<tag>[^:]+):'
                     r'(?P<content>.*)')
def odin_lines(line_iterator, replacement_char='\uFFFD'):
    igtlines = []
    for line in line_iterator:
        if line == '':
            return igtlines
        else:
            match = line_re.match(line)
            if match:
                attrs = OrderedDict([('line', match.group('line')),
                                     ('tag', match.group('tag'))])
                igtlines.append(
                    (attrs,
                     replace_invalid_xml_chars(match.group('content'),
                                               replacement_char))
                )
            else:
                logging.warning('Non-empty IGT line could not be parsed:\n{}'
                                .format(line))
    # if we hit the end of file, there might be something left
    return igtlines

#Character replacement based on an answer by "drawnonward" on Stack Overflow
#http://stackoverflow.com/questions/3220031/how-to-filter-or-replace-unicode-characters-that-would-take-more-than-3-bytes

# XML 1.0 valid ranges:
#  \u0009, \u000A, \u000D, \u0020-\uD7FF, \uE000-\uFFFD, \U0001000-\U0010FFFF
# but these are discouraged:
#  \u007F-\u0084, \u0086-\u009F
# XML 1.1 valid ranges:
#  \u0001-\uD7FF, \uE000-\uFFFD, \U00010000-\U0010FFFF
# but these are discouraged:
#  \u0001-\u0008, \u000B-\u000C, \u000E-\u001F, \u007F-\u0084, \u0086-\u009F
# Moreover, these are allowed but are discouraged, and may be "invalid" by
# some parsers:
#  \uFDD0-\uFDDF
#  \U0001FFFE, \U0001FFFF, \U0002FFFE, \U0002FFFF, \U0003FFFE, \U0003FFFF,
#  \U0004FFFE, \U0004FFFF, \U0005FFFE, \U0005FFFF, \U0006FFFE, \U0006FFFF,
#  \U0007FFFE, \U0007FFFF, \U0008FFFE, \U0008FFFF, \U0009FFFE, \U0009FFFF,
#  \U000AFFFE, \U000AFFFF, \U000BFFFE, \U000BFFFF, \U000CFFFE, \U000CFFFF,
#  \U000DFFFE, \U000DFFFF, \U000EFFFE, \U000EFFFF, \U000FFFFE, \U000FFFFF,
#  \U0010FFFE, \U0010FFFF
invalid_char_re = re.compile('[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD'
                             #'\U00010000-\U0010FFFF]'
                             ']'
                             , re.UNICODE)
def replace_invalid_xml_chars(input, replacement_char):
    # \uFFFD is the unicode replacement character
    return invalid_char_re.sub(replacement_char, input)


### CLEANING #################################################################

def clean_items(items):
    # first make copies of the original items
    cln_items = [Item(id=item.id.replace('o','c'),
                      attributes=item.attributes,
                      content=item.content)
                 for item in items]
    # then execute the cleaning steps
    #cln_items = merge_diacritics(cln_items)
    cln_items = merge_lines(cln_items)
    cln_items = join_lines(cln_items)
    cln_items = normalize_lines(cln_items)
    return cln_items

def merge_diacritics(items):
    """
    Use unicodedata's normalize function first to get combining
    diacritics instead of standalone ones, then for any combining
    diacritic see if it combines with a nearby character (where
    "combine" here means that the two can be normalized as a single
    character.
    """
    # this StackOverflow answer by bobince was very helpful:
    # http://stackoverflow.com/a/447047
    newitems = []
    for item in items:
        line = item.content
        if line.strip() == '':
            newitems.append(item)
            continue
        line = unicodedata.normalize('NFKD', line)
        # first remove inserted spaces before diacritics
        max_j = len(line) - 1
        line = ''.join(c for j, c in enumerate(line)
                       if c != ' ' or (j < max_j and
                                       unicodedata.combining(line[j+1])==0))
        # then combine diacritics with previous, then following chars if
        # they can be combined
        chars = [line[0]] # always append first char
        max_j = len(line) - 1 # need to recalc this
        for j, c in enumerate(line[1:]):
            # skipped first char, so counter needs to be incremented
            j += 1
            if unicodedata.combining(c):
                # a character x and combining character c have length 2,
                # but if they combine, they have length 1
                if j < max_j and \
                     len(unicodedata.normalize('NFC', line[j+1] + c)) == 1:
                    # just append
                    chars.append(unicodedata.normalize('NFC', line[j+1] + c))
                elif len(unicodedata.normalize('NFC', chars[-1] + c)) == 1:
                    # need to replace appended previous char
                    chars[-1] = unicodedata.normalize('NFC', chars[-1] + c)
            else:
                chars.append(c)
        # then check if combining diacritics match adjacent chars
        #for j, c in enumerate(line)
        item.content = ''.join(chars)
        newitems.append(item)
    return newitems

def merge_lines(items):
    """
    Return the lines with corrupted and split lines merged.
    Merge corrupted lines if:
      * Both lines have the +CR tag
      * Both lines have one other tag in common
      * The lines are sequential
      * tokens in one line align to whitespace in the other
    TODO:
      * Can we allow some non-whitespace overlap, in which case the
        token would be inserted in the closest match
      * Can we recombine diacritics with the letter it came from?
        E.g. is this usually accents or umlauts on vowels?
      * Is there anything that can be done about intraline corruption?
        E.g. when spaces are inserted w i t h i n words
    """
    n = len(items)
    # nothing to do if there's just 1 line
    if n < 2:
        return items
    newitems = [items[0]]
    for i in range(1,n):
        # lines are pairs of attributes and content
        prev = newitems[-1]
        cur = items[i]
        p_tags = prev.attributes.get('tag','').split('+')
        c_tags = cur.attributes.get('tag','').split('+')
        # if no non-CR tags are shared
        if 'CR' not in c_tags or \
           len(set(p_tags).intersection(c_tags).difference(['CR'])) == 0:
            newitems.append(cur)
            continue
        merged = bit_merge(prev.content, cur.content)
        if merged is not None:
            # there's no OrderedSet, but OrderedDict will do
            tags = OrderedDict((t,1) for t in p_tags + c_tags)
            line_nums = ' '.join([prev.attributes.get('line'),
                                  cur.attributes.get('line')])
            prev.attributes['tag'] = '+'.join(tags)
            prev.attributes['line'] = line_nums
            prev.content = merged
    return newitems

def merge_strings(a, b):
    maxlen = max([len(a), len(b)])
    a = a.ljust(maxlen)
    b = b.ljust(maxlen)
    m = []
    for i in range(maxlen):
        pass
        # do something like this
        #   "a a a"  "a a a"    "a a a"  "a a a"  "a a a"
        # + " b"     " bbb "    " b b "  "  b"    "  bb"
        # = "aba a"  "abbba a"  "ababa"  "a aba"  "a abba"

    return ''.join(m).rstrip(' ')

def bit_merge(a, b):
    """
    Merge two strings on whitespace by converting whitespace to the
    null character, then AND'ing the bit strings of a and b, then
    convert back to regular strings (with spaces).
    """
    if len(b) > len(a): return bit_merge(b, a)
    try:
        # get bit vectors with nulls instead of spaces, and
        # make sure the strings are the same length
        a = a.replace(' ','\0').encode('utf-8')
        b = b.replace(' ','\0').encode('utf-8').ljust(len(a), b'\0')
        c_pairs = zip(a, b)
    except UnicodeDecodeError:
        return None
    c = []
    for c1, c2 in c_pairs:
        # only merge if they merge cleanly
        if c1 != 0 and c2 != 0:
            return None
        else:
            c.append(c1|c2)
    try:
        return bytes(c).decode('utf-8').replace('\0',' ').rstrip(' ')
    except UnicodeDecodeError:
        return None

def join_lines(items):
    return items

def normalize_lines(items):
    for item in items:
        item.content = normalize_line(item.content)
    return items

paren_re = re.compile(r'\([^)]*\)')
paren_num_re = re.compile(
    r"^\(\s*(" # start (X; ( X; group for alternates
    r"[\d.]+\w?|\w|" # 1 1a 1.a 1.23.4b; a b (no multiple chars, except...)
    r"[ivxlc]+)"    # roman numerals: i iv xxiii; end alt group
    r"['.:]*\s*\)[.:]*") # optional punc (X) (X:) (X') (X.) (X. ) (X). (X):
num_re = re.compile(
    r"^([\d.]+\w?|\w|[ivxlc]+)" # nums w/no parens; same as above
    r"['.):]+\s") # necessary punc; 1. 1' 1) 1: 10.1a. iv: etc.
precontent_re = re.compile(r'^\s*\w+(\s\w+)?\s*:\s')
hyphen_re = re.compile(r'\s*-\s*')

def normalize_line(line):
    if line is None or line.strip() == '': return None
    # remove spaces, quotes, and punctuation to make later regexes simpler
    line = line.strip().strip('"\'`')
    # only strip initial parens if on both sides (so we don't turn 'abc (def)'
    # into 'abc (def'
    if line.startswith('(') and line.endswith(')'): line = line[1:-1]
    # one more space, quote, and punc strip, in case the parens grouped them
    line = line.strip().strip('"\'`')
    # re-add a period (to all lines)
    # remove inner parens (with encapsulated content)
    # this seems to to go too far in some cases, see ace.item11
    #line = paren_re.sub('', line)
    # IGT-initial numbers (e.g. '1.' '(1)', '5a.', '(ii)')
    line = paren_num_re.sub('', line).strip()
    line = num_re.sub('', line).strip()
    # precontent tags can be 1 or 2 words ("intended:" or "speaker", "a:")
    # ignore those with 3 or more
    line = precontent_re.sub('', line)
    # there may be morphemes separated by hyphens, but with intervening
    # spaces; remove those spaces (e.g. "dog-  NOM" => "dog-NOM")
    line = hyphen_re.sub('-', line)
    # ignore ungrammatical or questionably lines or those with alternates
    # now we want to include these, separate out the # or * and mark the
    # judgments  for /, it's no longer a bother
    # do this later anyway
    # if line.startswith('*') or line.startswith('#'): or '/' in line:
    #    return None
    return line

### XIGT STRUCTURES  #########################################################


def make_metadata(igt):
    lg_attrs = OrderedDict()
    if igt.get('language'):
        lg_attrs['name'] = igt['language'].strip()
    if igt.get('iso-639-3'):
        lg_attrs['iso-639-3'] = igt['iso-639-3'].strip()
    lg_attrs['tiers'] = 'phrases words morphemes'
    metas = []
    metas.append(Meta(type='language', attributes=lg_attrs))
    metas.append(
        Meta(type='language',
             attributes=OrderedDict([
                 ('name', 'English'),
                 ('iso-639-3', 'eng'),
                 ('tiers', 'translations')
             ])
        )
    )
    metas.append(
        Meta(type='odin-source',
             attributes=OrderedDict([
                 ('doc-id', igt['doc_id']),
                 ('line-range', '-'.join(igt['line_range'])),
                 ('line-types', ' '.join(igt['line_types']))
             ])
        )
    )
    metadata = [Metadata(type='xigt-meta', content=metas)]
    return metadata


def make_raw_tier(raw_lines):
    items = [Item(id='o{}'.format(j+1), attributes=a, content=c)
             for j, (a, c) in enumerate(raw_lines)]
    tier = Tier(id='o', type='odin-raw', items=items)
    return tier


def make_clean_tier(raw_tier):
    cleaned_items = clean_items(raw_tier.items)
    tier = Tier(id='c', type='odin-clean', items=cleaned_items)
    return tier


def make_aligned_tiers(odin_tier):
    # at best, we'll get phrases (p), words (w), morphemes (m),
    # glosses (g), and translations (t), at worst we'll just
    # get p, g, and t (with single items each) for a complete
    # IGT, but incomplete IGTs may have fewer tiers.
    p, w, m, g, t = align_lgt(odin_tier.items)
    tiers = []
    if len(p.items) > 0:
        tiers.append(p)
    if len(w.items) > 0:
        tiers.append(w)
    if len(m.items) > 0:
        tiers.append(m)
    if len(g.items) > 0:
        tiers.append(g)
    if len(t.items) > 0:
        tiers.append(t)
    return tiers


tok_re = re.compile(r'\S+')

def align_lgt(items):
    attrs = [('content', 'c')]
    p = Tier(id='p', type='phrases', attributes=dict(attrs))
    w = Tier(id='w', type='words', attributes=dict([('segmentation','p')]))
    m = Tier(id='m', type='morphemes', attributes=dict([('segmentation','w')]))
    g = Tier(id='g', type='glosses', attributes=dict(attrs))
    t = Tier(id='t', type='translations', attributes=dict(attrs))
    pi = wi = mi = gi = ti = 0 # item counters

    for item in items:
        tags = item.attributes.get('tag','').split('+')
        if not item.content:
            continue
        toks = list(tok_re.finditer(item.content))
        if len(toks) > 0:
            start = toks[0].span()[0]
            end = toks[-1].span()[1]
            if 'L' in tags:
                attrs = {'content':'{}[{}:{}]'.format(item.id, start, end)}
                p.add(Item(id='p{}'.format(pi), attributes=attrs))
                pi += 1
            elif 'G' in tags:
                attrs = {'content':'{}[{}:{}]'.format(item.id, start, end)}
                if pi > 0:
                    attrs['alignment'] = 'p{}'.format(pi)
                g.add(Item(id='g{}'.format(gi), attributes=attrs))
                gi += 1
            elif 'T' in tags:
                attrs = {'content':'{}[{}:{}]'.format(item.id, start, end)}
                if pi > 0:
                    attrs['alignment'] = 'p{}'.format(pi)
                t.add(Item(id='t{}'.format(ti), attributes=attrs))
                ti += 1

    return p, w, m, g, t


def make_path(outdir, lg_code):
    return os.path.join(outdir, lg_code + '.xml')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-C','--replacement-char', metavar='CHAR',
            default='\uFFFD',
            help='xml replacement character (default: \uFFFD)')
    parser.add_argument('-c','--clean', action='store_true',
            help='create "odin-clean" tier with cleaned data')
    parser.add_argument('-a','--align', action='store_true',
            help='attempt to create basic alignment tiers')
    parser.add_argument('-v', '--verbose', action='count',
            dest='verbosity', default=2,
            help='increase logging verbosity')
    parser.add_argument('-o', '--outdir', required=True,
                        help='output directory')
    parser.add_argument('-l', '--by-language', action='store_true')
    parser.add_argument('-d', '--by-doc-id', action='store_true')
    parser.add_argument('-t', '--text', action='store_true')
    args = parser.parse_args()
    logging.basicConfig(level=50-(args.verbosity*10))

    igts = list(odin_igts(map(str.strip, sys.stdin),
                              replacement_char=args.replacement_char))
    lgs = defaultdict(int)
    docs = set()
    num_added = -1
    num_total = 0
    while num_added != 0:
        xc = XigtCorpus()
        num_added = 0
        cur = None
        for i, igt in enumerate(igts):
            lg_code = igt.get('iso-639-3')
            doc_id = igt.get('doc_id')
            if lg_code is None:
                lg_code = 'und'  # undetermined
            if cur is None:
                if args.by_language and lg_code not in lgs:
                    cur = lg_code
                    logging.info('Creating subcorpus for language: {}'
                                 .format(lg_code))
                elif args.by_doc_id and doc_id not in docs:
                    cur = doc_id
                    logging.info('Creating subcorpus for document: {}'
                                 .format(doc_id))
                else:
                    continue
                if args.text:
                    txtfile = open(os.path.join(args.outdir, cur + '.txt'), 'w')
            else:
                if args.by_language and lg_code != cur:
                    continue
                elif args.by_doc_id and doc_id != cur:
                    continue
            if args.text:
                txtlines = igt.get('header_lines',[])
                for attrs, content in igt.get('lines', []):
                    txtline = '{}:{}'.format(
                        ' '.join('%s=%s' % (k,v) for k, v in attrs.items()),
                        content
                    )
                    txtlines.append(txtline)
                print('\n'.join(txtlines) + '\n', file=txtfile)
            else:
                metadata = make_metadata(igt)
                tiers = [make_raw_tier(igt.get('lines', []))]
                if args.clean:
                    tiers.append(make_clean_tier(tiers[0]))
                if args.align:
                    # cleaned, if available, otherwise raw
                    odin_tier = tiers[-1]
                    tiers.extend(make_aligned_tiers(odin_tier))
                xc.add(Igt(id='i{}'.format(i+1),
                           metadata=metadata,
                           tiers=tiers))
            num_added += 1
            num_total += 1
            docs.add(doc_id)
            lgs[lg_code] += 1
        if num_added > 0:
            logging.info('Added {} IGT to subcorpus.'.format(num_added))
            if not args.text:
                path = make_path(args.outdir, cur)
                xigtxml.dump(open(path, 'w'), xc)
        if args.text:
            txtfile.close()
    logging.info('Added {} IGT total.'.format(num_total))

