#!/usr/bin/env Rscript

#Olga Zamaraeva
#K-Means Clustering for AGGREGATION

#Legend: word order integer codes
#SOV = 1
#OSV = 2
#SVO = 3
#OVS = 4
#VSO = 5
#VOS = 6
#Vfinal = 7
#Vinitial = 8
#FreeV2 = 9

# CURRENT VERSION USAGE:
# Provide path to the data file, TRUE if
# the data file's first column is the gold labels,
# FALSE otherwise.

# EXAMPLE:
# Rscript kmeans-cluster.R ../dev1dev2.txt TRUE 

#command line args

myarg <- commandArgs()
print(myarg) #diagnostics
writeLines("")

# parameters
dir<-"."
datafile<-myarg[6] #expected: path to the data file
labels<-myarg[7] #expected: TRUE or FALSE
do_plot<-myarg[8] #TRUE or FALSE

# Definitions

# Convert labels from words to number codes
labelsToNum <- function(orderlabels)
{
  numlabels<-rep(0,length(orderlabels))
  for (i in 1:length(orderlabels)) {
    if (orderlabels[i] == "sov") { # subject-object-verb
      numlabels[i]<-1
    }
    else if (orderlabels[i] == "osv") { # object-subject-verb
      numlabels[i]<-2
    }
    else if (orderlabels[i] == "svo") { # subject-verb-object
      numlabels[i]<-3
    }
    else if (orderlabels[i] == "ovs") { # object-verb-subject
      numlabels[i]<-4
    }
    else if (orderlabels[i] == "vso") { # verb-subject-object
      numlabels[i]<-5
    }
    else if (orderlabels[i] == "vos") { # verb-object-subject
      numlabels[i]<-6
    }
    else if (orderlabels[i] == "vfi") {# verb final
      numlabels[i]<-7
    }
    else if (orderlabels[i] == "vin") { # verb initial
      numlabels[i]<-8
    }
    else if (orderlabels[i] == "fre") { # free
      numlabels[i]<-9
    }
    else { # wrong data format
      numlabels[i]<-NaN
      print("Warning: Gold labels are in wrong format. Accuracy will be meaningless. Clusters are still valid.")
    }
  }
  return(as.matrix(numlabels))
}

# Compare clustering results to gold labels
# Assign canonical labels to centers of clusters
# If no gold labels were provided, will not produce meaningful result
analyzeResults <- function(results, goldlabels, initial_centers, plot)
{
  labeled_data <- results$labeled_data
  X<-labeled_data[,2:4]
  final_centers <- results$centers
  print("Final centers:")
  print(final_centers)
  writeLines("")
  center_labels <- results$cntrlbs
  clus<-labeled_data[,1]
  #Clusters are numbered automatically 1 through k; rename them according to corresponding canonical centers
  closest_canonical <- assignCenters(final_centers,canonical_centers())
  for (i in 1:length(clus))
  {
    clus[i] <- closest_canonical[clus[i]]
  }
  #Compute accuracy
  mistakes <- which(clus != goldlabels)
  num_mistakes <- length(mistakes)
  labeled_data[,1] <- clus
  mistake <- rep(0, length(clus))
  mistake[mistakes] <- 1
  labeled_data <- cbind(goldlabels,labeled_data)
  labeled_data <- cbind(mistake,labeled_data)
  print("Original data with cluster labels, gold labels, and mistakes: ")
  print(labeled_data)
  writeLines("");
  accuracy <- 1- num_mistakes/dim(labeled_data)[1]
  print("Accuracy: total")
  print(accuracy)
  #Plots
  if (plot==TRUE)
  {
    inicent <- scatterplot3d(initial_centers, color="blue",main="Initial (blue) and final (red) centers; all points in green")
    inicent$points3d(final_centers, col="red", pch=as.character(closest_canonical))
    inicent$points3d(X, col="green", pch=as.character(goldlabels))
    #label initial centers
    text(inicent$xyz.convert(0,0,0), labels="Free", col="blue", pos=1)
    text(inicent$xyz.convert(-1,-1,-1), labels="SOV", col="blue", pos=1)
    text(inicent$xyz.convert(-1,-1,1), labels="OSV", col="blue", pos=1)
    text(inicent$xyz.convert(-1,1,-1), labels="SVO", col="blue", pos=1)
    text(inicent$xyz.convert(1,-1,1), labels="OVS", col="blue", pos=1)
    text(inicent$xyz.convert(1,1,-1), labels="VSO", col="blue", pos=1)
    text(inicent$xyz.convert(1,1,1), labels="VOS", col="blue", pos=1)
    text(inicent$xyz.convert(-1,-1,0), labels="Vfinal", col="blue", pos=1)
    text(inicent$xyz.convert(1,1,0), labels="Vinitial", col="blue", pos=1)
    #plot clusters in 2D projection
    clusplot(X,clus,labels=2,span=FALSE, diss=FALSE,lines=0,color=TRUE,main="Clustered languages by word order")
    clusplot(X,goldlabels,labels=2,span=FALSE, diss=FALSE,lines=0,color=TRUE,main="Gold labels of Languages by word order")
    incr_mistakes <- labeled_data[,1]+1
    clusplot(X,clus,labels=2,span=FALSE, diss=FALSE,lines=0,color=TRUE,col.p=incr_mistakes,main="Mistakes: wrong cluster assigned (red)")
    clusplot(X,goldlabels,labels=2,span=FALSE, diss=FALSE,lines=0,color=TRUE,col.p=incr_mistakes,main="Gold labels, the one we cannot get are in red")
  }
}

# Set the canonical centers
canonical_centers <- function()
{
  SOV <- c(-1, -1, -1)
  OSV <- c(-1, -1, 1)
  SVO <- c(-1, 1, -1)
  OVS <- c(1, -1, 1)
  VSO <- c(1, 1, -1)
  VOS <- c(1, 1, 1)
  Vfinal <- c(-1, -1, 0)
  Vinitial <- c(1, 1, 0)
  FreeV2 <- c(0, 0, 0)
  lcenters <- list(SOV, OSV, SVO, OVS, VSO, VOS, Vfinal, Vinitial, FreeV2)
  centers <- do.call(rbind,lcenters)
  rownames(centers)<-c("sov","osv","svo","ovs","vso","vos","vfi","vin","fre")
  return(centers)
}

# custom KMeans
kMeans <- function(centers, data)
{
  #k<-5
  #OPTIONAL: select k starting centers randomly (will override passed centers)
  #centers <- data[sample(nrow(data),k),]
  #or assign centers as canonical values
  k<-dim(centers)[1]
  labels <- assignCenters(data, centers)
  #adjust centers 20 times or until labels stop changing
  i<-1
  while (i <= 20)
  {
    i <- i+1
    newctrs <- newCenters(data, labels)
    centers <- newctrs$centers
    newlabels <- assignCenters(data, centers)
    if (all(newlabels == labels))
    {
      break
    }
    else
    {
      labels <- newlabels 
    }
  }
  labeled_data <- cbind(labels,data)
  return(list(centers=centers, labeled_data=labeled_data))
}

# Find nearest center to a vector (point) x
nearest <- function(x,centers)
{
  m <- rbind (x, centers)
  d <- as.matrix(dist(m))
  nearest <- which.min(d[-1,1])
  return (nearest)
}

# Assign nearest center to each vector in a matrix
assignCenters <- function(data, centers)
{
  labels <- apply(data,1, FUN=nearest, centers)
  return(labels)
}

# Find centers of given clusters
newCenters <- function(data, labels)
{
  centers <- aggregate(data, list(label = labels),mean)
  lbs <- centers[,1]
  centers <- as.matrix(centers[,-1])
  return(list(centers=centers,lbs=lbs))
}

# Find the most frequent gold label in a cluster
 
majority_vote <- function(v)
{
  t <- table(v)
  vote <- as.numeric(names(t[t == max(t)]))
  return(vote)
}

# main program

setwd(dir)
library(cluster)
library(scatterplot3d)
data <- read.table(datafile)
if (labels == TRUE) {
  goldlabels <- as.matrix(data[,1])
  goldlabels <- labelsToNum(goldlabels)
  data <- data[,-1]
  rownames(goldlabels) <- data[,2]
}
X<-data[,-1]
rownames(X)<- data[,1]
initial_centers <- canonical_centers();
results <-kMeans(initial_centers,X)
if (labels == TRUE) {
  analyzeResults(results, goldlabels, initial_centers, do_plot)
} else { 
  print(results)
  print ("Cluster lables here do not correspond to any canonical order codes.")
}


