#!/usr/bin/env perl

## Purpose: align words and project DS in IGT
##
## The file includes four types of functions:
##     - all kinds of functions for word alignment
##     - align sentence pairs using heuristics
##     - project DS using heuristics
##     - the functions to read annotated rich IGT files (w/ Q1-Q6)


## Word alignment is stored in two formats:
##  suppose src pos 1 aligns to tgt position 2 and 3
##
##  format 1: $map{1}="2,3": 
##     used in align_xx_pair(), print_word_align*(), get_reverse_alignment()
##       alignment_composition() 
##  format 2: $map{"1-2"}=type (NULL_link or non_NULL_link), $map{"1-3"}=type 
##     used in read_word_align(), compare_word_align()
##
##  convert_word_align_21() converts from format 2 to format 1

## Similary, dep_tree is stored in two formats:
##   suppose pos 1 depends on position 2 and 3
##  format 1: $map[1]="2,3":   ## note: here we use array, not hash
##     used in project_dependency_tree(),
##             get_dependency_tree_from_ttree(),
##             print_dependency_tree() in ttree.pm
##             str_to_rich_igt_struct*
##
##  format 2: $map{"1-2"}=type (NULL_link or non_NULL_link), $map{"1-3"}=type 
##     used in read_dep_tree(), compare_dep_tree()
##
##  convert_dep_tree_21() converts from format 2 to format 1

use strict;



############ constants that are also defined in compare_annot.pl
my $YES = 1;
my $NO = 0;

my $NON_NULL_LINK = 1;  # a link that both positions are not NULL
my $NULL_LINK = 2;      # a link that one position is NULL.

my $UNALIGNED = -2;     # in a DS, if the dep link is "pos UNALIGNED", it means pos is not aligned.

my $QUEST_NUM = 5;   # Q2-Q6 


my $debug = 1;

### the following are also defined in ab600.exec
my $HEAD_INITIAL = 1;
my $HEAD_POS_UNKNOWN = 0;
my $HEAD_FINAL = -1;

## all the word positions start from 1 (not 0). except for trans_words[] etc. 
## The 1-based position is used in align_map and depend_ptr and leaf_nodes_pts

############################# align src and gloss line

### here, we assume that the two lines have been tokenized, and cleaned
## return success
sub align_src_and_gloss_pair {
    my ($src_words_ptr, $gloss_words_ptr, $align_map_ptr) = @_;

    my @src_words = @$src_words_ptr;
    my @gloss_words = @$gloss_words_ptr;

    my $leng = scalar @src_words;
    if($leng != scalar @gloss_words){
	return 0;
    }

    for(my $i=0; $i<$leng; $i++){
	my $src_word = $src_words[$i];
	my $gloss_word = $gloss_words[$i];
	my @src_parts = split(/\-/, $src_word);
	my @gloss_parts = split(/\-/, $gloss_word);

	my $i1 = $i+1;
	
	$align_map_ptr->{$i1} = $i1;

	my $token_leng = scalar @src_parts;
	if(($token_leng > 1) && ($token_leng == scalar @gloss_parts)){
	    for(my $j=0; $j<$token_leng; $j++){
		my $j1 = $j+1;
		my $key = "$i1.$j1";
		$align_map_ptr->{$key} = $key;
	    }
	}
    }
    return 1;
}



#############################  align gloss and translation line
### given gloss words, trans words, roots of trans words, POS tags of trans words,
###     and translation lexicon (optional)
###
### Find word alignment between gloss and trans line 
###   - for each trans word, add it and its root form to two hash tables
###   - for each morph in the gloss line, try to find its match
###
### Two tables (optional):
###   translation lexicon: the format "gloss_word trans_word"
###   morph_table: 
### 
sub align_gloss_and_trans_pair {
    my ($gloss_words_ptr, $trans_words_ptr, $trans_roots_ptr, $POS_tags_ptr,
	$align_map_ptr, $trans_lex_exist, $trans_lex_ptr,
	$morph_gloss_flag, $morph_table_ptr) = @_;

   
    my @trans_words = @$trans_words_ptr;
    ## my %cased_hash = ();
    my %lc_hash = ();
    my %root_hash = ();

    
    #### step 1: create hash table for words in the trans line
    for(my $i=0; $i<scalar @trans_words; $i++){
	my $wd = $trans_words[$i];

	my $i1 = $i + 1;
	## add_to_hash($wd, \%cased_hash, $i+1);

	my $lc_wd = $wd;
	$lc_wd =~ tr/A-Z/a-z/;
	
	add_to_hash($lc_wd, \%lc_hash, $i+1);

	my $root = $trans_roots_ptr->[$i]; 
	if($lc_wd ne $root){
	    add_to_hash($root, \%root_hash, $i+1);
	}
    }


    #### step 2: go through each word in the gloss line
    my @gloss_words = @$gloss_words_ptr;

    for(my $i=0; $i<scalar @gloss_words; $i++){
	### check each gloss word
	my $gloss_word = $gloss_words[$i];
	my @gloss_parts = split(/\-/, $gloss_word);
	my $i1 = $i + 1;
	my $pos_str = "";

	for(my $j=0; $j < scalar @gloss_parts; $j++){
	    ## check each "morpheme"
	    my $gloss = $gloss_parts[$j];
	    my @tmp_trans_pos = ();

	    my $tmp_morph_gloss_flag = 0;
	    if($j == 0){
		### assuming the 1st morph is the root
		$tmp_morph_gloss_flag = $morph_gloss_flag;
	    }

	    my $num = find_match_for_a_morph($gloss, \%lc_hash, \%root_hash, 
					     $trans_lex_exist, $trans_lex_ptr,
					     $tmp_morph_gloss_flag, $morph_table_ptr,
					     \@tmp_trans_pos);
	    
	    if($num){
		my $j1 = $j + 1;
		my $key = "$i1.$j1";
		my $tmp_pos_str = join(",", @tmp_trans_pos);
		$align_map_ptr->{$key} = sort_pos_str($tmp_pos_str);
		$pos_str .= $tmp_pos_str . ",";
	    }
	} # for each morph
	
	$pos_str =~ s/\,$//;
	if($pos_str =~ /\d/){
	    $align_map_ptr->{$i1} = sort_pos_str($pos_str);
	}
 
    } ## for each gloss word
}



### print out word alignment: word alignment index starts from 1
### $align_map_ptr is a pointer to a hash (not an array)
sub print_word_align {
    my ($fp, $words_ptr1, $words_ptr2, $align_map_ptr) = @_;
    
    my @words1 = @$words_ptr1;
    my @words2 = @$words_ptr2;

    my $leng1 = scalar @words1;
    my $leng2 = scalar @words2;

    my %words2_hash = ();

    ##### step 1: set the hash
    for(my $i=0; $i<$leng2; $i++){
	my $word2 = $words2[$i];
	my $i1 = $i + 1;
	$words2_hash{$i1} = $word2;

	my @tokens = split(/\-/, $word2);
	my $token_num = scalar @tokens;
	if($token_num == 1){
	    next;
	}
	
	######### get morph-level token
	for(my $j=0; $j<$token_num;$j++){
	    my $j1 = $j+1;
	    my $key = "$i1\.$j1";
	    my $token = $tokens[$j];
	    my $tt = "";
	    if($j){
		$tt = "-";
	    }
	    $words2_hash{$key} = $tt . $token;
	}
    }
    
    ##### step 2: print out the hash
    for(my $i=0; $i<$leng1; $i++){
	my $word1 = $words1[$i];
	
	########### print word-level alignment
	my $i1 = $i + 1;
	my $tmp = $align_map_ptr->{$i1};
	if(defined($tmp)){
	    my $tmp_str = get_word_str($tmp, \%words2_hash);
	    print $fp "$i1 $tmp \# $word1 $tmp_str\n";
	}else{
	    print $fp "$i1 0 \# $word1 NULL\n";
	}

	my @tokens = split(/\-/, $word1);
	my $token_num = scalar @tokens;
	if($token_num == 1){
	    next;
	}
	
	######### print morph-level alignment
	for(my $j=0; $j<$token_num;$j++){
	    my $j1 = $j+1;
	    my $key = "$i1\.$j1";
	    my $tmp = $align_map_ptr->{$key};
	    my $token = $tokens[$j];
	    if($j){
		$token = "-" . $token;
	    }
	    if(defined($tmp)){
		my $tmp_str = get_word_str($tmp, \%words2_hash);
		print $fp " $key $tmp \# $token $tmp_str\n";
	    }else{
		print $fp " $key 0 \# $token NULL\n";
	    }
	}
    }
}


### print out word alignment: word alignment index starts from 1
### $align_map_ptr is a pointer to a hash (not an array)
### Also, print out sent_id
sub print_word_align_wo_morph_level {
    my ($fp, $words_ptr1, $words_ptr2, $align_map_ptr, $sent_id) = @_;
    
    my @words1 = @$words_ptr1;
    my @words2 = @$words_ptr2;

    my $leng1 = scalar @words1;
    my $leng2 = scalar @words2;

    ########### print word-level alignment
    ##### we print the line when src_pos=0 as well
    for(my $i=0; $i<=$leng1; $i++){
	my $word1 = "NULL";
	if($i){
	    $word1 = $words1[$i-1];
	}
	my $tmp = $align_map_ptr->{$i};
	if(defined($tmp)){
	    my @tgt_pos_array = split(/\,/, $tmp);
	    foreach my $tgt_pos (@tgt_pos_array){
		my $tgt_word = "NULL";
		if($tgt_pos > 0){
		    $tgt_word = $words2[$tgt_pos-1];
		}
		printf $fp "%04d ", $sent_id;
		print $fp "$i $tgt_pos \# $word1 $tgt_word\n";
	    }
	}else{
	    if($i){
		printf $fp "%04d ", $sent_id;
		print $fp "$i 0 \# $word1 NULL\n";
	    }
	}
    }
}


### given positions and the spelling of words in $hash_ptr, return the words
### e.g., $pos_str is "1,2,4"
###  return "w[1] w[2] w[4]"
sub get_word_str {
    my ($pos_str, $hash_ptr) = @_;

    my @pos = split(/\,/, $pos_str);
    my @words = ();

    for(my $i=0; $i<scalar @pos; $i++){
	my $tmp = $hash_ptr->{$pos[$i]};

	if(defined($tmp)){
	    push(@words, $tmp);
	}else{
	    print STDERR "no word for position $pos[$i]. This should not happen\n";
	}
    }

    return join(",", @words);
}

#### find the match for a gloss morph
####
#### str may contain "." (e.g., get.off), but it does not contain "-"
#### all paras are input, except for $pos_ptr
####
#### return the size of @$pos_ptr
####
sub find_match_for_a_morph {
    my ($str, $lc_hash_ptr, $root_hash_ptr,
	$trans_lex_exist, $trans_lex_ptr,
	$morph_gloss_flag, $morph_table_ptr,
	$pos_ptr) = @_;

    @$pos_ptr = ();
    
    my @parts = ($str);
    if($str =~ /^(.+)\.(.+)/){
	@parts = split(/\./, $str);
    }

    for my $tok (@parts){
	find_match_for_a_min_token($tok, $lc_hash_ptr, $root_hash_ptr,
				   $trans_lex_exist, $trans_lex_ptr,
				   $morph_gloss_flag, $morph_table_ptr,
				   $pos_ptr);
    }
    return scalar @$pos_ptr; 
}


#### don't clear @$pos_ptr
#### $str does not contain "."/"-", etc.
#### return suc
####
sub find_match_for_a_min_token {
    my ($gloss, $lc_hash_ptr, $root_hash_ptr, 
	$trans_lex_exist, $trans_lex_ptr, 
	$morph_gloss_flag, $morph_table_ptr,
	$pos_ptr) = @_;

    my $lc_gloss = $gloss;
    $lc_gloss =~ tr/A-Z/a-z/;
    
    ### step 1: check the spelling first
    my $res = find_token_in_two_hashes($gloss, $lc_hash_ptr, 
				       $root_hash_ptr, $pos_ptr);

    #### step 3: morph the gloss: 
    #### Fei: this used to appear after step 2
    if(($res > 0) || !$morph_gloss_flag){
	return $res;
    }

    my $gloss_root = morph_word($lc_gloss, $morph_table_ptr);
    if($lc_gloss ne $gloss_root){
	## print "lc_gloss=$lc_gloss, root=$gloss_root\n";
	$res = find_token_in_two_hashes($gloss_root, $lc_hash_ptr, 
					$root_hash_ptr, $pos_ptr);
    }


    ### step 2: check some translation lexicon
    if(($res <= 0) && $trans_lex_exist){
	my $poss_trans_str = $trans_lex_ptr->{$lc_gloss};
	if(defined($poss_trans_str)){
	    my @poss_trans = split(/\+/, $poss_trans_str);
	    foreach my $cur_trans (@poss_trans){
		$res += find_token_in_two_hashes($cur_trans, 
						 $lc_hash_ptr, $root_hash_ptr,
						 $pos_ptr);
	    }
	}

    }


    return $res;
}



### return 1 if found a match
### return 0 if there is no match
###
### input: str is a min token, it does not contain "."/"-" etc.
##         $lc_hash_ptr and $root_hash_ptr are hash table for trans words
##
##  output: $pos_ptr stores the positions       
###         don't clear @$pos_ptr
##
sub find_token_in_two_hashes {
    my ($str, $lc_hash_ptr,
	$root_hash_ptr, $pos_ptr) = @_;

    my $suc = 0;

    ## @$pos_ptr = ();
    my $lc_gloss = $str;
    $lc_gloss =~ tr/A-Z/a-z/;

    #### step 1: use the gloss morpheme and trans words
    my $res = find_word_in_hash($lc_gloss, $lc_hash_ptr);
	
    #### step 2: use the gloss morpheme and trans roots
    if($res < 0){
	$res = find_word_in_hash($lc_gloss, $root_hash_ptr);
    }

    if($res >= 0){
	push(@$pos_ptr, $res);
	$suc = 1;
    }

    return $suc;
}




##### return pos if found. if not found, return -1
sub find_word_in_hash {
    my ($wd, $hash_ptr) = @_;

    my $tmp = $hash_ptr->{$wd};
    if(defined($tmp)){
	if($tmp =~ /^(\d+)\,(.+)$/){
	    ## the word appears multiple times, choose the 1st one
	    my $res = $1;
	    $hash_ptr->{$wd} = $2;
	    return $res;
	}else{
	    ## the word appears only one
	    return $tmp;
	}
    }

    return -1;
}

sub add_to_hash {
    my ($wd, $hash_ptr, $idx) = @_;

    my $tmp = $hash_ptr->{$wd};
    if(defined($tmp)){
	$hash_ptr->{$wd} .= "," . $idx;
    }else{
	$hash_ptr->{$wd} = $idx;
    }
}

################################# combine two maps to get a new map
### given a mapping (map_ptr), get the reverse mapping
sub get_reverse_alignment {
    my ($map_ptr, $rev_map_ptr) = @_;

    %$rev_map_ptr = ();

    foreach my $key (keys %$map_ptr){
	my $tmp = $map_ptr->{$key};
	
	my @parts = split(/\,/, $tmp);
	foreach my $part (@parts){
	    my $tmp1 = $rev_map_ptr->{$part};
	    if(defined($tmp1)){
		$rev_map_ptr->{$part} .= "," . $key;
	    }else{
		$rev_map_ptr->{$part} = $key;
	    }
	}
    }

    foreach my $key (keys %$rev_map_ptr) {
	my $val = $rev_map_ptr->{$key};
	$rev_map_ptr->{$key} = sort_pos_str($val);
    }
}


#### res_map(x) = map2(map1(x))
sub alignment_composition {
    my ($map1_ptr, $map2_ptr, $res_map_ptr) = @_;
    
    %$res_map_ptr = ();

    foreach my $key (keys %$map1_ptr){
	my $new_str = "";
	my $tmp = $map1_ptr->{$key};

	my @parts = split(/\,/, $tmp);
	foreach my $part (@parts){
	    my $tmp1 = $map2_ptr->{$part};
	    if(defined($tmp1)){
		$new_str .= "$tmp1,";
	    }else{
		next;
	    }
	}
	if($new_str =~ /\d/){
	    $new_str =~ s/\,$//;
	    $new_str = sort_pos_str($new_str);
	    $res_map_ptr->{$key} = $new_str;
	}
    }

}


#### '1,4,3.1,2,1' becomes '1,2,3.1,4', where the duplicate '1' is removed
## return the sorted string
##
sub sort_pos_str {
    my ($str) = @_;
    $str =~ s/^\,+//;
    $str =~ s/\,+$//;

    my @parts = split(/\,+/, $str);
    my %tmp_hash = ();
    
    foreach my $part (@parts){
	if(!defined($tmp_hash{$part})){
	    $tmp_hash{$part} = 1;
	}
    }

    my $res = "";
    for my $key (sort keys %tmp_hash){
	$res .=$key . ",";
    }

    $res =~ s/\,$//;
    return $res;
}

### sort the array and remove duplicate elements
sub sort_and_remove_dup_in_array {
    my ($array_ptr, $new_array_ptr) = @_;
    @$new_array_ptr = ();

    if(scalar @$array_ptr == 0){
	return;
    }

    my @tmp_array = sort {$a <=> $b} @$array_ptr;
    my $prev = $tmp_array[0];
    push(@$new_array_ptr, $prev);

    for(my $i=1; $i<scalar @tmp_array; $i++){
	my $cur = $tmp_array[$i];
	if($cur != $prev){
	    push(@$new_array_ptr, $cur);
	    $prev = $cur;
	}
    }
}



############################ project dependency trees

## projection is done in the following steps:
##   1. remove the non-aligned words from the DS.
##       If i depends on an non-aligned words, let it depend on its grandparent
##   2. replace index on one side with the index on the other side
##   3. deal with multiple words problem
##   4. choose a unique root of the src DS
##   5. deal with unaligned words
##
## Note: index in align_ptr starts from 1, where the index starts from 0 for
##         leaf_nodes_ptr and depend_ptr and pos_tags_ptr
##
## align_ef is word alignment from e (Eng) to f (src)
## align_fe is word alignment from f (src) to e (Eng)
##
## The function calculates $depend_ptrF
sub project_dependency_tree {
    my ($src_words_ptr, $src_headness, $trans_leng, 
	$depend_E_ptr, $align_EF_ptr, $align_FE_ptr, 
	$depend_F_ptr, $attach_unaligned_option) = @_;
    
    ### step 0: init depend_ptr2
    my $lengE = $trans_leng;
    my $lengF = scalar @$src_words_ptr;

    my @depend_E_new = (-3);  ## the new DS after removing some links.
                              ## add dummy elements so that index starts from 1.

    ### step 1: remove the non-aligned words from the DS1    
    for(my $i=1; $i<=$lengE; $i++){
	my $orig_head = $depend_E_ptr->[$i];
	my $cur_head = $orig_head;

	my $changed = 0;
	while(($cur_head >= 0) && 
	      !defined($align_EF_ptr->{$cur_head})){
	    $cur_head = $depend_E_ptr->[$cur_head];
	    $changed = 1;
	}

	if($changed && $cur_head > 0){
	    ### use the changed head only if the head is not -1 (*TOP*)
	    push(@depend_E_new, $cur_head);
	}else{
	    push(@depend_E_new, $orig_head);
	}

	## print "$i: orig_head=$orig_head, new_head=$cur_head flag1=", 
	## ($orig_head == $cur_head), "\n";
    }


    ### step 2: projection: if f1->e1, e1 depends on e2, e2->f2, and f1 != f2
    ###         then f1 depends on f2 
    my @depend_F_tmp = (-3);  ## put in a dummy node, so that the position
                           ##  starts from 1
    for(my $j=1; $j<=$lengF; $j++){
	### find e1 s.t. f1 -> e1
	my $e1_str = $align_FE_ptr->{$j};
	if(!defined($e1_str) || ($e1_str eq "")){
	    push(@depend_F_tmp, "-2");  ## -2 means unknown, -1 means the head of
                                        ##  the sentence
	    next;
	}

	my @e1s = split(/\,/, $e1_str);
	
	my $res_str = "";

	### find e2 s.t. e1 depends e2
	foreach my $e1 (@e1s){
	    ## my $e2 = $depend_E_ptr->[$e1];
	    my $e2 = $depend_E_new[$e1];
	    my $f2 = -1;
	    if($e2 >= 0){
		## find f2 s.t. e2 -> f2
		my $f2_str = $align_EF_ptr->{$e2};
		my @f2s = split(/\,/, $f2_str);
		foreach my $f2 (@f2s){
		    ## don't allow xx.yy at this stage
		    if(($j ne $f2) && ($f2 !~ /\d.\d/)){
			$res_str .= "$f2,";
		    }
		}
	    }else{
		$res_str .= "-1,";
	    }
	}
	
	$res_str =~ s/,$//;
	if($res_str !~ /\d/){
	    $res_str = "-2";
	}
	push(@depend_F_tmp, $res_str);
    }

    ## print_dependency_tree($fp, $src_words_ptr, $depend_ptrF);
    
    #### step 3: deal with multiple "head"
    @$depend_F_ptr = (-3);  ## add dummy so that the index starts from 1
    my $res = adjust_DS($src_words_ptr, $src_headness,
			$depend_E_ptr, \@depend_E_new, 
			$align_EF_ptr, $align_FE_ptr,
			\@depend_F_tmp, $depend_F_ptr, 
			$attach_unaligned_option);
    return $res;
}

### if y depends on y1, y2, ...., choose one of y_i as y's head
### The only output parameter is $depend_ptrF_new
###
sub adjust_DS {
    my ($src_words_ptr, $src_headness, 
	$depend_ptrE, $depend_ptrE_new, 
	$align_EF_ptr, $align_FE_ptr, 
	$depend_ptrF, $depend_ptrF_new, 
	$attach_unaligned_option) = @_;

    my $lengF = scalar @$src_words_ptr;


    @$depend_ptrF_new = (-3);   ## push in the dummy so the idx starts from 1

    #### step 1: deal with the case when one has multiple heads
    remove_multi_parents_in_DS($src_words_ptr, $src_headness, 
			       $depend_ptrE, $depend_ptrE_new,
			       $align_EF_ptr, $align_FE_ptr,
			       $depend_ptrF, $lengF, $depend_ptrF_new);

    #### step 2: choose a unique root of DS
    my $head_pos = sel_DS_root($src_words_ptr, $src_headness,
			       $depend_ptrF, $lengF);

    ### step 3: link the unaligned word or other "head" words 
    ####         to the root of the DS
    if(!defined($attach_unaligned_option) ||
       ($attach_unaligned_option == 1)){
	print "head_pos=$head_pos\n";
	attach_unaligned_words_to_DS_root($head_pos, $lengF, $depend_ptrF_new);
    }else{
	attach_unaligned_words_to_nearest_neighbor($head_pos, $lengF,
						   $depend_ptrF_new);
    }
}


##### if a node has multiple parents in the DS, remove some of them
sub remove_multi_parents_in_DS {
    my ($src_words_ptr, $src_headness, 
	$depend_ptrE, $depend_ptrE_new, 
	$align_EF_ptr, $align_FE_ptr, 
	$depend_ptrF, $lengF,
	$depend_ptrF_new) = @_;

    for(my $j=1; $j<=$lengF; $j++){
	### f1 depends on f2
	my $f2_str = $depend_ptrF->[$j];
	my @f2s = split(/\,/, $f2_str);
	my $f2_num = scalar @f2s;
	if($f2_num == 1){
	    push(@$depend_ptrF_new, $f2_str);
	    next;
	}

	#### if y depends on y1 and y2 and y1 depends y2, than remove y1
	my $new_f2_str = "";
	if($f2_str =~ /\-1/){
	    ## the head of the whole sentence
	    $new_f2_str = "-1";
	}else{
	    for(my $k1=0; $k1 < $f2_num; $k1++){
		my $y1 = $f2s[$k1];
		my $keep = 1;
		my $y1_depend_str = $depend_ptrF->[$y1];
		
		for(my $k2=0; $k2 < $f2_num; $k2++){
		    if($k1 == $k2){
			next;
		    }
		    my $y2 = $f2s[$k2];
		    if(($y1 == $y2) && ($k1 > $k2)){
			$keep = 0;
			last;
		    }

		    if($y1_depend_str =~ /^(.*)$y2(.*)$/){
			my $t1 = $1;
			my $t2 = $2;
			if(($t1 !~ /(\d|\-)$/) && ($t2 !~ /^(\d|\-)/)){
			    ## y1 depends on y2
			    $keep = 0;
			    last;
			}
		    }
		}
		    
		if($keep){
		    $new_f2_str .= $y1 . ",";
		}
	    }
	}

	$new_f2_str =~ s/\,$//;
	push(@$depend_ptrF_new, $new_f2_str);
    }
}


### return head_pos (-1 when we cannot decide a unique head)
sub sel_DS_root {
    my ($src_words_ptr, $src_headness, 
	$depend_ptrF, $lengF) = @_;


    my $head_pos = -1;

    if($src_headness == $HEAD_FINAL){
	for(my $j=$lengF; $j>=1; $j--){
	    my $f2_str = $depend_ptrF->[$j];
	    if($f2_str =~ /\-1/){
		$head_pos = $j;
		last;
	    }
	}
	if($head_pos < 0){
	    if($src_words_ptr->[$lengF-1] =~ /[a-zA-Z]/){
		$head_pos = $lengF;
	    }else{
		$head_pos = $lengF - 1;  ## the one before the final punct
	    }
	}
    }elsif($src_headness == $HEAD_INITIAL){
	for(my $j=1; $j<=$lengF; $j++){
	    my $f2_str = $depend_ptrF->[$j];
	    if($f2_str =~ /\-1/){
		$head_pos = $j;
		last;
	    }
	}
	if($head_pos < 0){
	    if($src_words_ptr->[0] =~ /[a-zA-Z]/){
		$head_pos = 1;
	    }else{
		$head_pos = 2;  ## the one after the final punct
	    }
	}
    }else{
	## neither head-initial or head-final
	for(my $j=1; $j<=$lengF; $j++){
	    my $f2_str = $depend_ptrF->[$j];
	    if($f2_str =~ /\-1/){
		if($head_pos < 0){
		    $head_pos = $j;
		}else{
		    ## multiple head;
		    $head_pos = -1;
		    last;
		}
	    }
	}
    }

    return $head_pos;
}

#### attach unaligned source words to the root of the DS
sub attach_unaligned_words_to_DS_root {
    my ($head_pos, $lengF, $depend_ptrF_new) = @_;
    if($head_pos > 0){
	for(my $j=1; $j<=$lengF; $j++){
	    my $f2_str = $depend_ptrF_new->[$j];
	    if($j == $head_pos){
		next;
	    }

	    if($f2_str =~ /\-[1|2]/){
		$depend_ptrF_new->[$j] = "$head_pos";
	    }
	}
    }

    return 1;
}


### for each unalign word j, find the nearest neighbors (i, k), s.t.
##  i appears before j and k appears after j, and i depends k (or vice versa).
##  Now attach j to k (or i, respectively)
##
sub attach_unaligned_words_to_nearest_neighbor {
    my ($head_pos, $lengF, $depend_ptrF_new) = @_;

    for(my $j=1; $j<=$lengF; $j++){
	my $f2_str = $depend_ptrF_new->[$j];
	if($j == $head_pos){
	    next;
	}
	
	### j has parent(s) in DS already.
	if($f2_str !~ /\-/){
	    next;
	}

	### j is one of the roots
	if(($f2_str =~ /\-1/) && ($head_pos > 0)){
	    $depend_ptrF_new->[$j] = "$head_pos";	    
	    next;
	}

	
	if($f2_str =~ /\-[1|2]/){
	    my $i = -1;
	    my $k = -1;
	    my $head = -1;
	    my $res = find_closest_neighbors_in_DS($j, $lengF, 
						   $depend_ptrF_new,
						   \$head);
	    if($res){
		$depend_ptrF_new->[$j] = "$head";	    
	    }else{
		if($head_pos > 0){
		    $depend_ptrF_new->[$j] = "$head_pos";
		}
	    }
	}
    }
}


### return 1 if found
### return 0 o.w.
###
sub find_closest_neighbors_in_DS {
    my ($j, $lengF, $depend_ptrF, $head_ptr) = @_;


    ############ case 1: $j is not on the boundary
    ### find the i1, k1, s.t. i1<j<k1 (or i1>j>k1) and i1 depends on k1
    if(($j > 1) && ($j < $lengF)){
	my $min_dist = -1;
	for(my $i=1; $i<=$lengF; $i++){
	    if($i == $j){
		next;
	    }

	    my $str = $depend_ptrF->[$i];
	    if($str =~ /\-/){
		next;
	    }
	    
	    my @parts = split(/\,/, $str);
	    foreach my $k (@parts){
		if( ($i < $j && $k > $j) || 
		    ($i > $j && $k < $j) ){
		    my $dist = abs($k - $i);
		    if($min_dist < 0 || $dist < $min_dist){
			$min_dist = $dist;
			$$head_ptr = $k;
		    }
		}
	    }
	}
		
	if($min_dist > 0){
	    return 1;
	}else{
	    return 0;
	}
    }

    return 0;
}



#############################################################################
## return 1 (when the data is not clean) or the size of answers[]
## answers[] and chunks[] should each have 5 values.
## return 0 if the format is wrong
## 
##  Don't init $$unans_cnt_ptr
sub break_igt_str_to_chunks {
    my ($str, $format, $src_line_ptr, $gloss_line_ptr, $trans_line_ptr,
	$q1_ans_ptr, $chunks_ptr, $answers_ptr, $unans_cnt_ptr) = @_;

    my $fail = 0;

    @$chunks_ptr = ();
    @$answers_ptr = ();

    my @lines = split(/\n/, $str);

    my $line_num = scalar @lines;
    if($line_num < 15){
	print STDOUT "the current IGT has only $line_num lines\n";
	return $fail;
    }

    #### step 1: read the IGT
    $$src_line_ptr = remove_space($lines[0]);
    $$gloss_line_ptr = remove_space($lines[1]);
    $$trans_line_ptr = remove_space($lines[2]);

    my $quest_id = -1;
    my $ans = -1;
    my $l_cnt = 3;   # line_id

    #### step 2: Q1
    my $q1_line = $lines[$l_cnt];

    my $unans = 0;
    my $res = line_to_answer($q1_line, \$quest_id, $q1_ans_ptr, \$unans);

    if(!$res || ($quest_id != 1)){
	print STDOUT "+$q1_line+ is not in the right format for the Q1 line: quest_id=$quest_id\n";
	return $fail;
    }

    
    if($unans){
	print STDOUT "warning: Q1 is not answered: +$q1_line+\n";
	$$unans_cnt_ptr ++;
    }

    if($$q1_ans_ptr eq $NO){
	return 1;  ## IGT is not clean
    }

    $l_cnt ++;


    ### step 3: Q2-Q6
    my $q_cnt = 2;
    while($l_cnt < $line_num){
	####### deal with question line when it appears for the 1st time
	my $q_line = $lines[$l_cnt]; 
	my $unans = 0;

	if($format == 1){
	    $res = line_to_answer($q_line, \$quest_id, \$ans, \$unans);
	    if($unans){
		print STDOUT "warning: Q$q_cnt is not answered: +$q_line+\n";
		$$unans_cnt_ptr ++;
	    }
	    

	    if(!$res || ($quest_id != $q_cnt)){
		print STDOUT "1st_time: +$q_line+ is not in the right format for Q$q_cnt: quest_id=$quest_id, res=$res\n";
		return $fail;
	    }

	    push(@$answers_ptr, $ans);
	}

	$l_cnt ++;

	####### get word align info
	my $str = "";
	while(($l_cnt < $line_num) && 
	      ($lines[$l_cnt] !~ /^\#+\s*(Q\d)\:/)){
	    if($q_cnt > 2 && $lines[$l_cnt] =~ /^\s*\d+(\.\d+)?\s+\-?\d+/){
		## word alignment
		$str .= "$lines[$l_cnt]\n";
	    }else{
		if($debug > 10){
		    print STDOUT "The following line is ignored: +$lines[$l_cnt]+\n";
		}
	    } 
	    $l_cnt ++;
	}
	$str =~ s/\n$//;	    
	push(@$chunks_ptr, $str);

	
	if($l_cnt >= $line_num){
	    next;
	}

	######## Q_i appears twice
	if($format > 1){
	    $q_line = $lines[$l_cnt];
	    my $unans = 0;
	    $res = line_to_answer($q_line, \$quest_id, \$ans, \$unans);
	    if(!$res || ($quest_id != $q_cnt)){
		print STDOUT "2nd time: +$q_line+ is not in the right format for Q$q_cnt: res=$res\n";
		return $fail;
	    }

	    if($unans){
		print STDOUT "warning: Q$q_cnt is not answered: +$q_line+\n";
		$$unans_cnt_ptr ++;
	    }

	    push(@$answers_ptr, $ans);
	    $l_cnt ++;
	}

	$q_cnt ++;
    }
    
    return @$chunks_ptr;
}


#### line is of the format "### Q1: ...? y/n"
## return 1 if the line is of that right format.
## return 0 o.w.
##
sub line_to_answer {
    my ($line, $quest_id_ptr, $answer_ptr, $unans_ptr) = @_;
    
    if($line =~ /^\#+\s*Q(\d)\:[^\?]+\?\s*(Answer\:\s*)?(\S*)/i){ ### Q2: IGT is clean? Answer: y/n/x
	my $quest_id = $1;
	$$quest_id_ptr = $1;
	my $answer_str = $3;
	if(!defined($answer_str) || ($answer_str !~ /^(y|n|x)/i) ){
	    print STDOUT "$quest_id is not answered in +$line+. Assuming the answer is yes.\n";
	    $$unans_ptr = 1;
	    $$answer_ptr = $YES;
	}else{
	    $$unans_ptr = 0;
	    if($answer_str =~ /^y/i){
		$$answer_ptr = $YES;
	    }else{
		## the answer is no.
		$$answer_ptr = $NO;
	    }
	}
	return 1;
    }else{
	$$unans_ptr = 0;
	print STDOUT "wrong format for answer: +$line+\n";
	return 0;
    }
}



## read igt file, and store the strs to an array, one elem per igt.
## ignore the parts after "######### end"

sub read_igt_file {
    my ($fp, $igt_ptr, $id_lines_ptr) = @_;

    my $str = "";
    my $id = -1;
    
    while(<$fp>){
	chomp;
	## s/^\s+//;
	s/\s+$//;

	next if (/^\s*\#*\s*$/);  ## ignore comment lines or empty lines

	### the "#### end" line
	if(/^\s*\#+\s*end(\s|\#)*$/i){
	    last;
	}

	if(/^\s*Igt\_id=(\d+)\s+/){
	    my $tmp_id=$1;
	    s/^\s+//;
	    s/\s+$//;

	    if(defined($id_lines_ptr->{$tmp_id})){
		print STDERR "warning: Igt_id=$tmp_id appears multiple times\n";
	    }

	    $id_lines_ptr->{$tmp_id} = $_;

	    if($str ne ""){
		$igt_ptr->{$id} = $str;
		$str = "";
	    }

	    $id = $tmp_id;
	    
	    next;
	}
	
	### remove user comments
	if(/^\s*\#{1,}\s*(\S+)\s+/){
	    my $wd = $1;
	    if($wd !~ /^Q[0-9]/){
		if($debug > 2){
		    print STDOUT "the following lines are ignored: +$_+\n";
		}
		next;
	    }
	}
	
	$str .= "$_\n";
    }

    if($str ne ""){
	$igt_ptr->{$id} = $str;
    }

    return scalar (keys %$igt_ptr);
}



### input line: "1 3,4" => hash{"1-3"}=1 hash{"1-4"}=1
###  input line could also be "4.1 3.2" => hash{"4.1-3.2"}=1 and hash{"4-3")=1
##
## This is for both src-gloss alignment and gloss-trans alignment, 
## so "src" and "trans" in the names are misleading
##
## str => word align in format2
sub read_word_align {
    my ($str, $src_leng, $trans_leng, 
	$link_wo_null_ptr, $link_w_null_ptr, $hash_ptr)=@_;
    
    %$hash_ptr = ();

    ### print STDERR "\nalign_str=$str\n";

    ############# step 1: set the flag
    my @src_flags = (0);  ## dummy, so that the word position index starts from 1
    my @trans_flags = (0);
    my $cnt = 0;

    for(my $i=0; $i<$src_leng; $i++){
	push(@src_flags, 0);
    }

    for(my $i=0; $i<$trans_leng; $i++){
	push(@trans_flags, 0);
    }

    ############## step 2: check the lines
    my @lines = split(/\n/, $str);
    
    foreach my $line (@lines){
	$line =~ s/\#(.+)$//;  # remove the words
	if($line =~ /^\s*((\d+)(\.\d+)?)\s+(\-?\d+(\.\d+)?(\s*\,\s*\-?\d+(\.\d+)?)*)\s*$/){
	    my $src = $1;
	    my $src_part1 = $2;
	    if($src_part1 > $src_leng){
		print STDOUT "warning: src position is too big: $src_part1 > $src_leng in +$line+\n";
		next;
	    }

	    my $trans_str = $4;
	    my @parts = split(/\s*\,\s*/, $trans_str);
	    my $part_num = scalar @parts;
	    foreach my $trans (@parts){
		my $trans_part1 = $trans;
		$trans_part1 =~ s/\.\d+//;

		my $key = "$src\-$trans";
		if($trans_part1 < 0){
		    print STDOUT "tgt position should >= 0, assume tgt pos is 0 in line +$line+\n";
		    $trans = 0;
		    $trans_part1 = 0;
		}

		if($trans_part1 == 0){
		    if($part_num > 1){
			print STDOUT "0 should not be here: +$line+\n";
		    }
		    next;
		}else{
		    if($trans_part1 > $trans_leng){
			print STDOUT "warning: trans position is too big: $trans_part1 in $trans > $trans_leng\n";
			next;
		    }
		}

		$src_flags[$src_part1] = 1;
		$trans_flags[$trans_part1] = 1;

		if(defined($hash_ptr->{$key})){
		    print STDOUT "warning: word align pair $key appeared before\n";
		}else{
		    $hash_ptr->{$key} = $NON_NULL_LINK;
		    $cnt ++;
		}

		$key = "$src_part1\-$trans_part1";
		$hash_ptr->{$key} = $NON_NULL_LINK;
	    }
	}else{
	    print STDOUT "warning: the line is not a word-align line: +$line+\n";
	}
    }

    $$link_wo_null_ptr = $cnt;

    ########## step 3: deal with unaligned position
    for(my $i=1; $i<=$src_leng; $i++){
	if(!$src_flags[$i]){
	    my $key = "$i\-0";
	    $hash_ptr->{$key} = $NULL_LINK;
	    $cnt ++;
	}
    }
    
    for(my $i=1; $i<=$trans_leng; $i++){
	if(!$trans_flags[$i]){
	    my $key = "0\-$i";
	    $hash_ptr->{$key} = $NULL_LINK;
	    $cnt ++;
	}
    }
 
    $$link_w_null_ptr = $cnt;

    return 1;
}


## return the numbers of agreed links
sub compare_word_align {
    my ($quest_id, $hash1_ptr, $hash2_ptr) = @_;
    
    my $agree1 = 0;   ## excluding NULL links
    my $agree2 = 0;   ## including NULL links

    foreach my $key (keys %$hash1_ptr){
	my $tmp = $hash2_ptr->{$key};
	if(defined($tmp)){
	    if($tmp == $NON_NULL_LINK){
		$agree1 ++;   ## only non-Null links
	    }
	    $agree2 ++;  ## all links
	}else{
	    $tmp = $hash1_ptr->{$key};
	    if($debug){
		$key =~ s/(\d)\-/$1 /;
		print STDOUT "DISAGREE in Q$quest_id: file1 has +$key+, file2 does not, type=$tmp\n";
	    }
	}
    }


    if($debug){
	foreach my $key (keys %$hash2_ptr){
	    my $tmp = $hash1_ptr->{$key};	
	    if(!defined($tmp)){
		my $tmp = $hash2_ptr->{$key};	
		$key =~ s/(\d)\-/$1 /;
		print STDOUT "DISAGREE in Q$quest_id: file1 does not have +$key+, file2 does, type=$tmp\n";
	    }
	}
    }	

    return ($agree1, $agree2);
}



### convert word alignment format from 1 to 2.
## $hash1_ptr->{"m"}="n1,n2" becomes $hash2_ptr->{"m-n1"} = 1
##   and hash2_ptr->{"m-n2"} = 1
## return $suc
sub convert_word_align_12 {
  my ($hash1_ptr, $hash2_ptr) = @_;
  
  foreach my $key (%$hash1_ptr){
      my $tmp_str = $hash1_ptr->{$key};
      my @parts = split(/\,/, $tmp_str);
      foreach my $part (@parts){
	  my $new_key = "$key-$part";
	  $hash2_ptr->{$new_key} = 1;
      }
  }

  return 1;
}


### convert word alignment format from 2 to 1.
## $hash2_ptr->{"m-n"} becomes $hash1_ptr->{"m"} = "n"
##
## return $suc
sub convert_word_align_21 {
    my ($hash2_ptr, $hash1_ptr) = @_;

    %$hash1_ptr = ();

    foreach my $key (keys %$hash2_ptr){
	my @parts = split(/\-/, $key);
	if(@parts != 2){
	    return 0;
	}
	my $src = $parts[0];
	my $tgt = $parts[1];

	if($tgt <= 0){
	    next;
	}
	if(defined($hash1_ptr->{$src})){
	    $hash1_ptr->{$src} .= ",$tgt";
	}else{
	    $hash1_ptr->{$src} = $tgt;
	}
    }

    return 1;
}




##### read the dep_tree: the format is "pos pos1{,pos2,...} \# ...."
##
## if a word is not aligned to anything, assume it is unknown: a.k.a. align to -2
## Those unaligned links will be included in link_w_null, but not in link_wo_null
##
## if x depends on y, hash_ptr->{"x-y"}=type (format 2)
## x can depend on multiple y's.
##
sub read_dep_tree {
    my ($str, $sent_leng, $link_wo_null_ptr, $link_w_null_ptr, $hash_ptr)=@_;
    
    %$hash_ptr = ();

    ############# step 1: set the flag
    my @flags = (0);  ## dummy, so that the word position index starts from 1
    my $cnt = 0;

    for(my $i=0; $i<$sent_leng; $i++){
	push(@flags, 0);
    }


    ############## step 2: check the lines
    my @lines = split(/\n/, $str);
    
    foreach my $line (@lines){
	$line =~ s/\#(.+)$//;  # remove the words
	if($line =~ /^(\d+)\s+((\-?\d+)(\s*\,\s*\-?\d+)*)\s*$/){
	    my $src = $1;
	    if($src > $sent_leng){
		print STDOUT "warning: position in DS is too big: $src > $sent_leng\n";
		next;
	    }

	    $flags[$src] = 1;
	    
	    my $trans_str = $2;
	    my @parts = split(/\s*\,\s*/, $trans_str);
	    my $part_num = scalar @parts;
	    foreach my $trans (@parts){
		my $key = "$src\-$trans";
		if($trans == 0){
		    if($part_num > 1){
			print STDOUT "0 should not be here: +$line+\n";
		    }
		    next;
		}else{
		    if($trans > $sent_leng){
			print STDOUT "warning: trans position is too big: $trans > $sent_leng\n";
			next;
		    }
		}

		if($trans == $src){
		    print STDOUT "error: a word should not depend on itself: $src=$trans\n";
		    next;
		}

		if($trans > 0){
		    $flags[$trans] = 1;
		}

		if(defined($hash_ptr->{$key})){
		    print STDERR "warning: word align pair $key appeared before\n";
		}else{
		    $hash_ptr->{$key} = $NON_NULL_LINK;
		    $cnt ++;
		}
	    }
	}else{
	    print STDERR "warning: the line is not a dependency link: +$line+\n";
	}
    }

    $$link_wo_null_ptr = $cnt;

    ########## step 3: deal with unaligned position
    for(my $i=1; $i<=$sent_leng; $i++){
	if(!$flags[$i]){
	    my $key = "$i\-$UNALIGNED";
	    $hash_ptr->{$key} = $NULL_LINK;
	    $cnt ++;
	    if($debug){
		print STDOUT "position $i is not aligned\n";
	    }
	}
    }
    
    $$link_w_null_ptr = $cnt;

    return 1;
}


### assuming format 2
sub compare_dep_tree {
    my ($quest_id, $hash1_ptr, $hash2_ptr) = @_;

    return compare_word_align($quest_id, $hash1_ptr, $hash2_ptr);
}


### convert word alignment format from 2 to 1.
## $hash2_ptr->{"m-n"} becomes $array1_ptr[m] = "n"
##
## return $suc
sub convert_dep_tree_21 {
    my ($sent_leng, $hash2_ptr, $array1_ptr) = @_;

    @$array1_ptr = ();
    for(my $i=0; $i<=$sent_leng; $i++){
	push(@$array1_ptr, -2);
    }

    foreach my $key (keys %$hash2_ptr){
	if($key =~ /^(\d+)\-(\-?\d+)$/){
	    my $src = $1;
	    my $tgt = $2;

	    if($tgt < -1){
		next;
	    }
	    if($array1_ptr->[$src] > 0){
		$array1_ptr->[$src] .= ",$tgt";
	    }else{
		$array1_ptr->[$src] = $tgt;
	    }
	}else{
	    print STDERR "wrong format in conv_dep_tree_21: +$key+\n";
	    return 0;
	}
    }

    return 1;
}



sub str_to_rich_igt_struct_w_parse_line {
    my ($igt_str, $format, 
	$q1_ans_ptr, $un_ans_cnt_ptr,
	$src_line_ptr, $gloss_line_ptr, $trans_line_ptr,
	$align_12_map_ptr, $align_23_map_ptr, 
	$depend_E_ptr, $depend_F_ptr, $parse_line_ptr) = @_; 

    my @answers = ();

    return 
	str_to_rich_igt_struct_w_parse_line_and_answers($igt_str, $format, 
							$q1_ans_ptr, 
							$un_ans_cnt_ptr,
							$src_line_ptr,
							$gloss_line_ptr,
							$trans_line_ptr,
							$align_12_map_ptr,
							$align_23_map_ptr, 
							$depend_E_ptr,
							$depend_F_ptr, 
							$parse_line_ptr,
							\@answers);
}



#### return suc
sub str_to_rich_igt_struct_w_parse_line_and_answers {
    my ($igt_str, $format, 
	$q1_ans_ptr, $un_ans_cnt_ptr,
	$src_line_ptr, $gloss_line_ptr, $trans_line_ptr,
	$align_12_map_ptr, $align_23_map_ptr, 
	$depend_E_ptr, $depend_F_ptr, $parse_line_ptr, $answers_ptr) = @_;   

    my $res = 
	str_to_rich_igt_struct_w_answers($igt_str, $format,
					 $q1_ans_ptr, $un_ans_cnt_ptr,
					 $src_line_ptr, $gloss_line_ptr,
					 $trans_line_ptr,
					 $align_12_map_ptr, $align_23_map_ptr, 
					 $depend_E_ptr, $depend_F_ptr,
					 $answers_ptr);

    if(!$res){
	return $res;
    }

    return get_parse_line($igt_str, $parse_line_ptr);
}


### return suc
sub get_parse_line {
    my ($igt_str, $parse_line_ptr) = @_;

    my @lines = split(/\n/, $igt_str);

    my $num = scalar @lines;
    
    my $i = 0;
    my $begin = 0;
    while($i < $num){
	my $line = $lines[$i];
	if($begin){
	    if($line =~ /^\(/){
		$$parse_line_ptr = $line;
		return 1;
	    }
	}else{
	    if($line =~ /Q2\:/){
		$begin = 1;
	    }
	}

	$i++;
    }
    
    return 0;
}

sub str_to_rich_igt_struct {
    my ($igt_str, $format, 
	$q1_ans_ptr, $un_ans_cnt_ptr,
	$src_line_ptr, $gloss_line_ptr, $trans_line_ptr,
	$align_12_map_ptr, $align_23_map_ptr, 
	$depend_E_ptr, $depend_F_ptr) = @_;

    my @answers = ();
    return 
	str_to_rich_igt_struct_w_answers($igt_str, $format, 
					 $q1_ans_ptr, $un_ans_cnt_ptr,
					 $src_line_ptr, $gloss_line_ptr, 
					 $trans_line_ptr,
					 $align_12_map_ptr, $align_23_map_ptr, 
					 $depend_E_ptr, $depend_F_ptr, 
					 \@answers);

}


#### return 1 if succeed (including the case when one is not clean)
#### return 0 o.w.
sub str_to_rich_igt_struct_w_answers {
    my ($igt_str, $format, 
	$q1_ans_ptr, $un_ans_cnt_ptr,
	$src_line_ptr, $gloss_line_ptr, $trans_line_ptr,
	$align_12_map_ptr, $align_23_map_ptr, 
	$depend_E_ptr, $depend_F_ptr, $answers_ptr) = @_;

    my @chunks = ();
    @$answers_ptr = ();  ## answers to the questions

    #### step 1: break into chunks and check whether the IGT is clean
    my $num = break_igt_str_to_chunks($igt_str, $format, 
				      $src_line_ptr, $gloss_line_ptr, 
				      $trans_line_ptr, 
				      $q1_ans_ptr, \@chunks, $answers_ptr, 
				      $un_ans_cnt_ptr);

    if($num < 1){
	print STDOUT "The following IGT is of the wrong format\n\n+$igt_str+\n\n";
	return 0;
    }


    ### gold IGT says that the IGT is not clean.
    if($$q1_ans_ptr == $NO){
	return 1;
    }
    
    if($num != $QUEST_NUM) {
	print STDERR "wrong number of chunks: gold=$num != $QUEST_NUM\n";
	return 0;
    }

    
    ####### step 2: read the word alignment and tgt DS
    my @src_words = split(/\s+/, $$src_line_ptr);
    my $src_leng = scalar @src_words;

    my @gloss_words = split(/\s+/, $$gloss_line_ptr);
    my $gloss_leng = scalar @gloss_words;

    my @trans_words = split(/\s+/, $$trans_line_ptr);
    my $trans_leng = scalar @trans_words;

    my $res = read_word_align_and_DS($src_leng, $gloss_leng, $trans_leng,
				     \@chunks, $answers_ptr,
				     $align_12_map_ptr, $align_23_map_ptr,
				     $depend_E_ptr, $depend_F_ptr);
    return $res;
}


### given two sets of (eng_DS, align_12_map, align_23_map, src_DS)
### calculate a new src_DS
###
sub get_new_dep_tree {
    my ($flag1, $flag2, $flag3, $src_line, $src_headness, $trans_line,
	$align_12_map1, $align_23_map1, $depend_E_ptr1, $depend_F_ptr1,
	$align_12_map2, $align_23_map2, $depend_E_ptr2, $depend_F_ptr2,
	$align_12_map3, $align_23_map3, $depend_E_ptr3, $depend_F_ptr3) = @_;

    if($flag2){
	%$align_12_map3 = %$align_12_map1;
    }else{
	%$align_12_map3 = %$align_12_map2;
    }

    if($flag3){
	%$align_23_map3 = %$align_23_map1;
    }else{
	%$align_23_map3 = %$align_23_map2;
    }

    if($flag1){
	@$depend_E_ptr3 = @$depend_E_ptr1;
    }else{
	@$depend_E_ptr3 = @$depend_E_ptr2;
    }

    ###### get EF and FE alignment
    my %align_13_map = ();
    my %align_31_map = ();

    alignment_composition($align_12_map3, $align_23_map3, \%align_13_map);
    get_reverse_alignment(\%align_13_map, \%align_31_map);


    #### get new src DS
    my @trans_words = split(/\s+/, $trans_line);
    my @src_words = split(/\s+/, $src_line);

    my $res = project_dependency_tree(\@src_words, $src_headness,
				      scalar @trans_words, $depend_E_ptr3,
				      \%align_31_map, \%align_13_map,
				      $depend_F_ptr3);
    return $res;
}


sub print_igt_w_q1_to_q6 {
    my ($fp, $id_line, $src_line, $gloss_line, $trans_line,
	$align_12_map_ptr, $align_23_map_ptr,
	$depend_E_ptr, $depend_F_ptr, $q1_ans, $answers_ptr) = @_;

    my $parse_line = "";

    print_igt_w_q1_to_q6_w_parse_line($fp, $id_line, 
				      $src_line, $gloss_line, $trans_line,
				      $align_12_map_ptr, $align_23_map_ptr,
				      $depend_E_ptr, $depend_F_ptr,
				      $parse_line, $q1_ans, $answers_ptr);
}
    

sub print_igt_w_q1_to_q6_w_parse_line {
    my ($fp, $id_line, $src_line, $gloss_line, $trans_line,
	$align_12_map_ptr, $align_23_map_ptr,
	$depend_E_ptr, $depend_F_ptr, $parse_line, 
	$q1_ans, $answers_ptr) = @_;

    my @src_words = split(/\s+/, $src_line);
    my @gloss_words = split(/\s+/, $gloss_line);
    my @trans_words = split(/\s+/, $trans_line);

    print $fp "\n###########################################################\n";
    print $fp "$id_line\n";
    print $fp "$src_line\n$gloss_line\n$trans_line\n\n";

    my $answer_str = get_answer_str($q1_ans);
    print $fp "######## Q1: IGT is clean? Answer: $answer_str\n\n\n";

    print $fp "############################## Q2: English parse tree \n";
    if($parse_line){
	print $fp "$parse_line\n\n";
	my $indent_str = indentTtree($parse_line);
        print $fp "$indent_str\n";
    }
    
    $answer_str = get_answer_str($answers_ptr->[0]);

    
    print $fp "\n\n###### Q2: English parse tree is correct? Answer: $answer_str\n\n\n\n";


    print $fp "###############################  Q3: English DS \n";
    print $fp "$trans_line\n";

    print_dependency_tree($fp, \@trans_words, $depend_E_ptr);

    $answer_str = get_answer_str($answers_ptr->[1]);
    print $fp "\n\n########  Q3: English DS is correct? Answer: $answer_str\n\n\n\n";


    print $fp "#######################  Q4: src and gloss alignment\n";
    print $fp "$src_line\n$gloss_line\n\n";
    print_word_align($fp, \@src_words, \@gloss_words, $align_12_map_ptr);

    $answer_str = get_answer_str($answers_ptr->[2]);
    print $fp "######## Q4: src and gloss alignment is correct? Answer: $answer_str\n\n\n\n";

    
    print $fp "######################### Q5: gloss and translation alignment\n";
    print $fp "$gloss_line\n$trans_line\n\n";
    print_word_align($fp, \@gloss_words, \@trans_words, $align_23_map_ptr);

    $answer_str = get_answer_str($answers_ptr->[3]);
    print $fp "\n\n######## Q5: gloss and translation alignment is correct? Answer: $answer_str\n\n\n";

    print $fp "############################# Q6: src DS\n";
    print $fp "$src_line\n$gloss_line\n$trans_line\n\n";
    print_dependency_tree($fp, \@src_words, $depend_F_ptr);

    $answer_str = get_answer_str($answers_ptr->[4]);
    print $fp "\n\n####### Q6: src DS is correct? Answer: $answer_str\n\n\n\n\n";
}


sub get_answer_str {
    my ($answer) = @_;

    if($answer eq ""){
	return "";
    }

    if($answer == $NO){
	return "n";
    }

    if($answer == $YES){
	return "y";
    }

}

########################################## sub
### return 1 if succeed, return 0 o.w.
###
###  retrun two alignments and two DSs. 
###   and store both in format 1

sub read_word_align_and_DS {
    my ($src_leng, $gloss_leng, $trans_leng,
	$chunks_ptr, $answers_ptr,
	$align_12_map_ptr, $align_23_map_ptr,
	$depend_E_ptr, $depend_F_ptr) = @_;


    ### step 1: read Q3: English DS
    my $trans_DS_links_wo_null = 0;
    my $trans_DS_links_w_null = 0;
    my %depend_E_hash = ();
    my $res = read_dep_tree($chunks_ptr->[1], $trans_leng, 
			    \$trans_DS_links_wo_null, \$trans_DS_links_w_null,
			    \%depend_E_hash);

    if(!$res){
	return 0;
    }
    
    $res = convert_dep_tree_21($trans_leng, \%depend_E_hash, $depend_E_ptr);
    if(!$res){
	return 0;
    }

    ### step 2: read Q4: src-gloss alignment
    my $align_12_links_wo_null = 0;
    my $align_12_links_w_null = 0;
    my %align_12_map_tmp = ();
    $res = read_word_align($chunks_ptr->[2], $src_leng, $gloss_leng, 
			   \$align_12_links_wo_null, \$align_12_links_w_null,
			   \%align_12_map_tmp);
    if(!$res){
	return 0;
    }

    $res = convert_word_align_21(\%align_12_map_tmp, $align_12_map_ptr);
    if(!$res){
	return 0;
    }

    ### step 3: read Q5: gloss-trans alignment 
    my $align_23_links_wo_null = 0;
    my $align_23_links_w_null = 0;
    my %align_23_map_tmp = ();
    $res = read_word_align($chunks_ptr->[3],
			   $gloss_leng, $trans_leng, 
			   \$align_23_links_wo_null, \$align_23_links_w_null,
			   \%align_23_map_tmp);
    if(!$res){
	return 0;
    }

    $res = convert_word_align_21(\%align_23_map_tmp, $align_23_map_ptr);
    if(!$res){
	return 0;
    }

    #### step 4: read  Q6: Source DS
    my $src_DS_links_wo_null = 0;
    my $src_DS_links_w_null = 0;
    my %depend_F_hash = ();
    $res = read_dep_tree($chunks_ptr->[4], $src_leng, 
			 \$src_DS_links_wo_null, \$src_DS_links_w_null,
			 \%depend_F_hash);

    if(!$res){
	return 0;
    }

    $res = convert_dep_tree_21($src_leng, \%depend_F_hash, $depend_F_ptr);
    if(!$res){
	return 0;
    }

    return 1;
}



##################

## Two word aligns are in format 1
##   e.g., hash1->{src} = "tgt1,tgt2, ..."
## return the number of "src-tgt" pairs (i.e., \# of links)
##
sub mix_two_aligns_in_format1 {
    my ($align1_ptr, $align2_ptr, $new_align_ptr1, $new_align_ptr2) = @_;

    my %tmp_align1 = ();
    my %tmp_align2 = ();
    my %tmp_align3 = ();
    my %tmp_align4 = ();

    convert_word_align_12($align1_ptr, \%tmp_align1);
    convert_word_align_12($align2_ptr, \%tmp_align2);

    my $cnt = mix_two_aligns_in_format2(\%tmp_align1, \%tmp_align2,
					\%tmp_align3, \%tmp_align4);
    convert_word_align_21(\%tmp_align3, $new_align_ptr1);
    convert_word_align_21(\%tmp_align4, $new_align_ptr2);
    return $cnt;
}

## Algorithm for mixing two word alignments:
##   - take the intersection first
##   - for each link in A+B-(A \cup B),
##       (a) take it only if
##           both src and tgt positions have not been covered.
##           => mix1
##       (b) take it only if either src or tgt pos has not been covered yet
##           => mix2
##
sub mix_two_aligns_in_format2 {
    my ($align1_ptr, $align2_ptr, $leng1, $leng2, 
	$new_align_ptr1, $new_align_ptr2) = @_;

    %$new_align_ptr1 = ();
    %$new_align_ptr2 = ();

    my $cnt = 0;
    my %remain_links = ();   # It is A + B -(A \cup B)

    ###### step 1: set the flags
    my @src_covered = (0);   # 1: if the pos is covered already
    my @tgt_covered = (0);  
    for(my $i=0; $i<$leng1; $i++){
	push(@src_covered, 0);
    }

    for(my $i=0; $i<$leng2; $i++){
	push(@tgt_covered, 0);
    }

    ###### step 2: calc the intersection first
    foreach my $key (keys %$align1_ptr){
	my $tmp = $align2_ptr->{$key};
	my ($src, $tgt) = split(/\-/, $key);
	if(defined($tmp) && $tmp){
	    $new_align_ptr1->{$key} = 1;
	    $src_covered[$src] = 1;
	    $tgt_covered[$tgt] = 1;
	    $cnt ++;
	}else{
	    if($src && $tgt){
		$remain_links{$key} = 1;
	    }
	}
    }

    foreach my $key (keys %$align2_ptr){
	my $tmp = $align1_ptr->{$key};
	my ($src, $tgt) = split(/\-/, $key);
	if(!defined($tmp) || !$tmp){
	    ## not in the intersection
	    if($src && $tgt){
		$remain_links{$key} = 1;
	    }
	}
    }


    ##### step 3: check the remaining links, and take ones that 
    ###           both src and tgt are not covered so far.
    foreach my $key (keys %remain_links){
	my ($src, $tgt) = split(/\-/, $key);
	if(!$src_covered[$src] && !$tgt_covered[$tgt]){
	    $new_align_ptr1->{$key} = 1;
	    $src_covered[$src] = 1;
	    $tgt_covered[$tgt] = 1;
	    delete($remain_links{$key});
	    $cnt ++;
	}
    }

    %$new_align_ptr2 = %$new_align_ptr1;

    ##### step 4: check the remaining links, and take ones that 
    ###           either src or tgt has not been covered so far.
    foreach my $key (keys %remain_links){
	my ($src, $tgt) = split(/\-/, $key);
	if(!$src_covered[$src] || !$tgt_covered[$tgt]){
	    $new_align_ptr2->{$key} = 1;
	    $src_covered[$src] = 1;
	    $tgt_covered[$tgt] = 1;
	    delete($remain_links{$key});
	    $cnt ++;
	}    
    }
    
    return $cnt;
}

######################### intersection

### Two word aligns are in format 1
##   e.g., hash1->{src} = "tgt1,tgt2, ..."
## return the number of "src-tgt" pairs (i.e., \# of links)
sub intersect_two_aligns_in_format1 {
    my ($align1_ptr, $align2_ptr, $new_align_ptr) = @_;

    my %tmp_align1 = ();
    my %tmp_align2 = ();
    my %tmp_align = ();

    convert_word_align_12($align1_ptr, \%tmp_align1);
    convert_word_align_12($align2_ptr, \%tmp_align2);

    my $cnt = intersect_two_aligns_in_format2(\%tmp_align1, \%tmp_align2,
					      \%tmp_align);
    convert_word_align_21(\%tmp_align, $new_align_ptr);
    return $cnt;
}


sub intersect_two_aligns_in_format2 {
    my ($align1_ptr, $align2_ptr, $new_align_ptr) = @_;

    %$new_align_ptr = ();

    my $cnt = 0;
    foreach my $key (keys %$align1_ptr){
	my $tmp = $align2_ptr->{$key};
	if(defined($tmp) && $tmp){
	    $new_align_ptr->{$key} = 1;
	    $cnt ++;
	}
    }

    return $cnt;
}


sub union_two_aligns_in_format1 {
    my ($align1_ptr, $align2_ptr, $new_align_ptr) = @_;

    my %tmp_align1 = ();
    my %tmp_align2 = ();
    my %tmp_align = ();

    convert_word_align_12($align1_ptr, \%tmp_align1);
    convert_word_align_12($align2_ptr, \%tmp_align2);

    my $cnt = union_two_aligns_in_format2(\%tmp_align1, \%tmp_align2,
					  \%tmp_align);
    convert_word_align_21(\%tmp_align, $new_align_ptr);
    return $cnt;
}

sub union_two_aligns_in_format2 {
    my ($align1_ptr, $align2_ptr, $new_align_ptr) = @_;

    %$new_align_ptr = ();

    my $cnt = 0;
    foreach my $key (keys %$align1_ptr){
	$new_align_ptr->{$key} = 1;
	$cnt ++;
    }

    foreach my $key (keys %$align2_ptr){
	my $tmp = $new_align_ptr->{$key};
	if(!defined($tmp) || !$tmp){
	    $new_align_ptr->{$key} = 1;
	    $cnt ++;
	}
    }

    return $cnt;
}

sub read_trans_lexicon {
    my ($fp, $hash_ptr) = @_;
    my $cnt = 0;
    while(<$fp>){
	chomp;
	s/^\s+//;
	s/\s+$//;
	next if(/^\s*$/);

	tr/A-Z/a-z/;   # lowercase everything
	my @parts = split(/\s+/, $_);

	my $part_num = scalar @parts;
	if($part_num < 2){
	    print STDERR "wrong format in trans_lexicon: +$_+\n";
	    next;
	}

	my $key = shift(@parts);
	my $val = join("+", @parts);
	my $tmp = $hash_ptr->{$key};
	if(defined($tmp)){
	    $hash_ptr->{$key} .= "+$val";
	}else{
	    $hash_ptr->{$key} = "$val";
	}
	$cnt ++;
    }

    if($debug){
	foreach my $key (keys %$hash_ptr){
	    print STDERR "+$key$+ +$hash_ptr->{$key}+\n";
	}
    }
    return $cnt;

}





################################## misc
sub abs {
    my ($a) = @_;
    if($a < 0){
	return -$a;
    }else{
	return $a;
    }
}

sub remove_space {
    my ($str) = @_;
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    return $str;
}


1;
