#!/usr/bin/env perl

use strict;

use Class::Struct;

my $INTERNAL_NODE = 1;
my $LEAF_NODE = 2;

my $LEFT_DIRECTION = 0;
my $RIGHT_DIRECTION = 1;

################################################### class
#################  Ttree and Tnode
struct( Tnode => {
    type => '$',    # internal, leaf, empty node
    s_tag => '$',   # syntactic tag: POS tag and bracket tag
    f_tag => '@',   # function tags
    indices => '@',  # indices, e.g., "-1" in "NP-1".

    parent => 'Tnode',     # point to parent. The parent of the root is
                           #  not set (since Perl does not have NULL ptr)

    head => 'Tnode',       # point to the head (which is a leaf node)
                           # head is a descendent of the current tnode.

    depends_on => 'Tnode', # this is the parent of the current tnode in the
                           # dependency structure

    ### for internal nodes only
    children => '@',       # pointers to the children
    head_child => 'Tnode', # point to head child
    word_pos_array => '@',  # the list of word positions in the subtree,
                            # starting from 1 (not 0). 
                            # The positions are sorted, but might not be contiguous

    ### for leaf nodes only
    word => '$',           # the spelling of a leaf node
    word_pos => '$'        # the positions in the sent, starting from 1 (not 0)
    }
       );


struct( Ttree => {
    sent_leng => '$',
    root => 'Tnode'
    }
	);


struct(HeadTable => {
    head_table_entries => '%'
}
);

struct( HeadTableEntry => {
    parent_tag => '$',
    direction  => '$',
    children_tag_hash => '%'
}
);



####################################################### sub for Ttree
## return 1 if succeed, return 0 otherwise
## $ttree is init beforehand
sub str_to_ttree {
    my ($ttree_str, $ttree) = @_;

    my $root = new Tnode;

    my $word_cnt = 0;
    my $suc = str_to_ttree_with_tnode($ttree_str, $root, \$word_cnt);

    if(!$suc){
	return 0;
    }

    $ttree->root($root);
    return 1;
}


## return 1 if succeed, return 0 otherwise
## $tnode is 
sub str_to_ttree_with_tnode {
    my ($ttree_str, $tnode, $word_cnt_ptr) = @_;
    
    ## print STDERR "enter: +$ttree_str\n";

    #### (tag word) ...
    if($ttree_str =~ /^\s*\((\S+)\s+(\S+)\s*\)\s*$/){
	my $wd = $2;
	my $tag = $1;
	$tnode->type($LEAF_NODE);
	set_tnode_tags($tag, $tnode);
	## $tnode->s_tag($tag);
	$tnode->word($wd);

	$$word_cnt_ptr ++;
	my $word_pos = $$word_cnt_ptr;
	$tnode->word_pos($word_pos);
	$tnode->children('()');
	
	$tnode->word_pos_array('()');
	my $array_ptr = $tnode->word_pos_array;
	push(@$array_ptr, $word_pos);

	## print STDERR "leaving: leaf: +$ttree_str+\n";
	return 1;
    }

    ### internal node: (S (....) (...))
    if($ttree_str =~ /^\s*\((\S+)\s+(.+)\s*\)\s*$/){
	my $s_tag = $1;
	my $remaining_str = $2;
	
	my @parts = ();

	my $suc = break_into_sibling_strs($remaining_str, \@parts);

	if(!$suc){
	    return 0;
	}

	$tnode->type($INTERNAL_NODE);
	## $tnode->s_tag($s_tag);
	set_tnode_tags($s_tag, $tnode);

	$tnode->children('()');
	my $children_ptr = $tnode->children;

	$tnode->word_pos_array('()');
	my $pos_array_ptr = $tnode->word_pos_array;

	my $children_num = scalar @parts;
	for(my $i=0; $i<$children_num; $i++){
	    my $child = new Tnode;
	    my $suc = str_to_ttree_with_tnode($parts[$i], $child, 
					      $word_cnt_ptr);

	    if(!$suc){
		return 0;
	    }
	    push(@$children_ptr, $child);
	    $child->parent($tnode);

	    push(@$pos_array_ptr, @{$child->word_pos_array});
	}


	## print STDERR "leaving: internal: +$ttree_str+\n\n";
	return 1;
    }

    
    print STDERR "wrong format: +$ttree_str+\n";
    return 0;
}

## e.g., $tag_str = "NP-SBJ-1" => s_tag=NP, f_tag=(SBJ), index=1
sub set_tnode_tags {
    my ($tag_str, $tnode) = @_;

    $tnode->f_tag('()');
    $tnode->indices('()');

    if($tag_str =~ /^(\-[a-z]+\-)/i){
	$tnode->s_tag($1);
	return;
    }

    my @parts = split(/[\-|\=]/, $tag_str);
    $tnode->s_tag($parts[0]);

    my $part_num = scalar @parts;
    if($part_num == 1){
	return;
    }

    my $f_tag_ptr = $tnode->f_tag;
    my $index_ptr = $tnode->indices;

    for(my $i=1; $i<$part_num; $i++){
	my $str = $parts[$i];
	if($str =~ /^[a-z]+$/i){
	    push(@$f_tag_ptr, $str);
	}else{
	    if($str =~ /^\d+$/){
		push(@$index_ptr, $str);
	    }
	}
    }

}


### e.g., str = "(xx) (yy) (zz)"
### => parts = ("(xx)", "(yy)", "(zz)")
###
sub break_into_sibling_strs {
    my ($str, $parts_ptr) = @_;

    ## print STDERR "break: +$str+\n";

    @$parts_ptr = ();

    my $level = 0;

    my @chars = split(//, $str);
    
    my $str = "";
    my $cnt = 0;
    for(my $i=0; $i<scalar @chars; $i++){
	my $char = $chars[$i];
	$str .= $char;
	if($char eq "("){
	    $level ++;
	}elsif($char eq ")"){
	    $level --;
	    if($level == 0){
		if($str !~ /^\s*\(/){
		    return 0;
		}
		push(@$parts_ptr, $str);
		$str = "";
		$cnt ++;
	    }
	}
    }
    
    return $cnt;
}


## return 1 if succeed, return 0 o.w.
sub ttree_to_str {
    my ($ttree, $ttree_str_ptr, $show_head) = @_;

    my $root = $ttree->root;

    if(!defined($root)){
	return 0;
    }

    return ttree_to_str_from_tnode($root, $ttree_str_ptr, $show_head);
}

sub ttree_to_str_from_tnode {
    my ($tnode, $ttree_str_ptr, $show_head) = @_;

    my $str = "";

    if($tnode->type == $LEAF_NODE){
	$str = "(" . $tnode->s_tag . " " . $tnode->word . ") ";
	$$ttree_str_ptr = $str;
	return 1;
    }
    
    my $children_ptr = $tnode->children;

    my $children_num = scalar @$children_ptr;
    for(my $i=0; $i<$children_num; $i++){
	my $child = $children_ptr->[$i];
	my $child_str = "";
	my $suc = ttree_to_str_from_tnode($child, \$child_str, $show_head);

	if(!$suc){
	    return 0;
	}

	$str .= "$child_str ";
    }

    my $str1 = "(" . $tnode->s_tag;
    my $f_tag_ptr = $tnode->f_tag;
    if(scalar @$f_tag_ptr){
	$str1 .= "-" . join("-", @$f_tag_ptr);
    }

    my $index_ptr = $tnode->indices;
    if(scalar @$index_ptr){
	$str1 .= "-" . join("-", @$index_ptr);
    }

    if($show_head){
	my $wd = ($tnode->head)->word;
	$str1 .= "+$wd";
    }

    $str1 .= " $str) ";
    
    $$ttree_str_ptr = $str1;
    return 1;
}



########################################################### sub for head table


### return the number of valid entries.
sub read_head_table {
    my ($fp, $head_table) = @_;

    my $cnt = 0;
    while(<$fp>){
	chomp;
	next if(/^\s*$/);

	my $str = $_;

	my @parts = split(/\s+/, $str);

	my $part_num = scalar @parts;
	if($part_num != 3){
	    print STDERR "wrong format for head table: +$str+\n";
	    next;
	}
	
	my $parent_tag = $parts[0];
	my $direction = $parts[1];
	my @child_tags = split(/\//, $parts[2]);

	my $head_entry = new HeadTableEntry;
	$head_entry->parent_tag($parent_tag);
	if($direction =~ /right/i){
	    $direction = $RIGHT_DIRECTION;
	}else{
	    $direction = $LEFT_DIRECTION;
	}

	$head_entry->direction($direction);
	
	foreach my $child_tag (@child_tags){
	    $head_entry->children_tag_hash->{$child_tag} = 1;
	}

	$head_table->head_table_entries->{$parent_tag} = $head_entry;
	$cnt ++;
    }

    return $cnt;
}


sub print_head_table {
    my ($head_table) = @_;

    my $ptr = $head_table -> head_table_entries;
    foreach my $key (keys %$ptr){
	my $ptr1 = $ptr->{$key};
	my $t1 = $ptr1->parent_tag;
	my $t2 = $ptr1->direction;
	print "$key: $t1, $t2 ";
	my $children_hash = $ptr1->children_tag_hash;
	foreach my $child_tag (keys %$children_hash){
	    print "$child_tag+";
	}
	print "\n";
    }
}


################## find heads
sub find_heads_in_ttree {
    my ($ttree, $head_table, $default_direction) = @_;

    my $root = $ttree->root;
    
    if(defined($root)){
	return find_heads_in_ttree_from_tnode($root, $head_table, 
					      $default_direction);
    }else{
	return 0;
    }
}

#### return 1 if succeed
sub find_heads_in_ttree_from_tnode {
    my ($tnode, $head_table, $default_direction, $depends_on) = @_;

    if($tnode->type == $LEAF_NODE){
	my $wd = $tnode->word;
	$tnode->head($tnode);
	if(defined($depends_on)){
	    $tnode->depends_on($depends_on);
	}
	return 1;
    }

    if($tnode->type == $INTERNAL_NODE){
	my $children_ptr = $tnode -> children;
	my $s_tag = $tnode->s_tag;
	my $head_child;
	my $suc = 0;

	my @children = @$children_ptr;

	$suc = findHeadChild($s_tag, $children_ptr, 
			     $head_table, $default_direction,
			     \$head_child);
	if(!$suc){
	    return 0;
	}

	$suc = find_heads_in_ttree_from_tnode($head_child, $head_table, 
					      $default_direction, $depends_on);
	if(!$suc){
	    return 0;
	}

	for(my $i=0; $i<scalar @children; $i++){
	    my $tmp_child = $children[$i];
	    my $tmp_depends_on = $depends_on;
	    if($tmp_child == $head_child){
		next;
	    }

	    $tmp_depends_on = $head_child->head;
	    $suc = find_heads_in_ttree_from_tnode($children[$i], $head_table,
						  $default_direction, $tmp_depends_on);
	    if(!$suc){
		return 0;
	    }
	}

	my $head = $head_child->head;
	$tnode->head_child($head_child);
	$tnode->head($head);
	return 1;
    }

    return 0;
}


### return success
### Currently, this part is totally specific to English.
## a child is a head child if it has -PRD tag and the parent is VP
sub find_head_child_using_func_tags {
    my ($parent_tag, $children_ptr, $head_child_ptr) = @_;

    if($parent_tag !~ /^VP$/i){
	return 0;
    }

    my @children = @$children_ptr;
    my $children_num = scalar @children;

    for(my $i=0; $i<$children_num; $i++){
	my $child = $children_ptr->[$i];
	if(tnode_has_f_tag($child, "PRD")){
	    $$head_child_ptr = $child;
	    return 1;
	}
    }

    return 0;
}

# return 1 if the tnode has the f_tag. return 0 o.w.
sub tnode_has_f_tag {
    my ($tnode, $f_tag_str) = @_;
    my $f_tag_ptr = $tnode->f_tag;

    my $num = scalar @$f_tag_ptr;
    if($num == 0){
	return 0;
    }

    for(my $i=0; $i<$num; $i++){
	my $cur_tag = $f_tag_ptr->[$i];
	if($cur_tag eq $f_tag_str){
	    return 1;
	}
    }

    return 0;
}

### return success
sub findHeadChild {
    my ($parent_tag, $children_ptr, $head_table, $default_direction,
	$head_child_ptr) = @_;

    ### PTB-specific
    if(find_head_child_using_func_tags($parent_tag, $children_ptr,
				       $head_child_ptr)){
	return 1;
    }

    ### use head percolation table
    my $ptr1 = $head_table->head_table_entries;
    my $ptr = $ptr1->{$parent_tag}; 

    if(!defined($ptr)){
	### no head table entry, use the language default
	if($default_direction == $LEFT_DIRECTION){
	    $$head_child_ptr = $children_ptr->[0];
	}else{
	    my @children = @$children_ptr;
	    $$head_child_ptr = $children_ptr->[scalar @children - 1];
	}
	return 1;
    }

    my $direction = $ptr->direction;

    my $hash_ptr = $ptr->children_tag_hash;

    my @children = @$children_ptr;
    my $children_num = scalar @children;
    if($direction == $LEFT_DIRECTION){
	### search from left to right
	for(my $i=0; $i<$children_num; $i++){
	    my $child = $children_ptr->[$i];
	    my $child_s_tag = $child->s_tag;
	    if(defined($hash_ptr->{$child_s_tag})){
		$$head_child_ptr = $child;
		return 1;
	    }
	}

	$$head_child_ptr = $children_ptr->[0];
    }else{
	for(my $i=$children_num-1; $i>=0; $i--){
	    my $child = $children_ptr->[$i];
	    my $child_s_tag = $child->s_tag;
	    if(defined($hash_ptr->{$child_s_tag})){
		$$head_child_ptr = $child;
		return 1;
	    }
	}

	$$head_child_ptr = $children_ptr->[$children_num-1];
    }	
    return 1;
}


## @$leaf_nodes_ptr stores the list of leaf nodes
## $depend_ptr->[$i]=$j: $i depends on $j, index starts from 1
## in $words_ptr, index starts from 0
sub print_dependency_tree {
    my ($fp, $words_ptr, $depend_ptr) = @_;
    
    for(my $i=0; $i<scalar @$words_ptr; $i++){
	my $wd1 = $words_ptr->[$i];
	my $i1 = $i + 1;

	my $j_str = $depend_ptr->[$i1];

	if(!defined($j_str)){
	    print $fp "$i -2 \# $wd1 NULL\n";
	    next;
	}
	my $wd2_str = "";
	my @js = split(/\,/, $j_str);
	foreach my $j (@js){
	    if($j > 0){
		$wd2_str .= $words_ptr->[$j-1] . ",";
	    }elsif($j == -1){
		$wd2_str .= "*TOP*" . ",";
	    }else{
		$wd2_str .= "NULL" . ",";
	    }
	}
	$wd2_str =~ s/,$//;
	print $fp "$i1 $j_str \# $wd1 $wd2_str\n";    
    }
}


## the depends_on links in ttree are set already.
## index starts at 1
## $leaf_nodes_ptr->[0] and $depend_ptr->[0] are dummy val.
sub get_dependency_tree_from_ttree {
    my ($ttree, $leaf_nodes_ptr, $depend_ptr) = @_;

    @$leaf_nodes_ptr = ();
    @$depend_ptr = ();
    
     
    my $root = $ttree->root;
    get_dependency_tree_from_tnode($root, $leaf_nodes_ptr, $depend_ptr);
}


sub get_dependency_tree_from_tnode {
    my ($tnode, $leaf_nodes_ptr, $depend_ptr) = @_;

    if($tnode->type == $LEAF_NODE){
	my $idx1 = $tnode->word_pos;

	my $depends_on = $tnode->depends_on;
	my $idx2 = -1;

	if(defined($depends_on) && ($depends_on != -1)){
	    $idx2 = $depends_on->word_pos;
	}
	
	if(scalar @$leaf_nodes_ptr == 0){
	    ## elements at index 0 are not used anyway.
	    push(@$leaf_nodes_ptr, $tnode);
	    push(@$depend_ptr, $idx2);
	}

	push(@$leaf_nodes_ptr, $tnode);
	push(@$depend_ptr, $idx2);
	return 1;
    }

    if($tnode->type == $INTERNAL_NODE){
	my $children_ptr = $tnode->children;
	
	my @children = @$children_ptr;

	for(my $i=0; $i<scalar @children; $i++){
	    my $child = $children_ptr->[$i];
	    get_dependency_tree_from_tnode($child, $leaf_nodes_ptr, $depend_ptr);
	}
	return 1;
    }

    return 0;
}


############################### other misc function

## return the number of words
sub get_word_POS_pairs_in_ttree {
    my ($ttree, $words_ptr, $tags_ptr, $keep_EC) = @_;

    @$words_ptr = ();
    @$tags_ptr = ();
    
    get_word_POS_pairs_in_ttree_from_tnode($ttree->root, $words_ptr,
					     $tags_ptr, $keep_EC);
    return scalar @$words_ptr;
}


sub get_word_POS_pairs_in_ttree_from_tnode {
    my ($tnode, $words_ptr, $tags_ptr, $keep_EC) = @_;
    
    if($tnode->type == $LEAF_NODE){
	if(!$keep_EC && ($tnode->s_tag =~ /\-NONE\-/i)){
	    return;
	}
	push(@$words_ptr, $tnode->word);
	push(@$tags_ptr, $tnode->s_tag);
    }else{
	my $children_ptr = $tnode->children;
	for(my $i=0; $i<scalar @$children_ptr; $i++){
	    my $child = $children_ptr->[$i];
	    get_word_POS_pairs_in_ttree_from_tnode($child, $words_ptr,
						   $tags_ptr, $keep_EC);
	}
    }
}


### modified from getSent() in 002/e/ttree2word.pl
## return the number of words
sub get_word_POS_pairs_from_ttree_str {
    my ($ttree_str, $words_ptr, $tags_ptr, $keep_EC) = @_;

    @$words_ptr = ();
    @$tags_ptr = ();

    
    $ttree_str =~ s/\(\s+//g;
    $ttree_str =~ s/\s+\)//g;

    my @parts = split(/\s+/, $ttree_str);

    my $part_num = scalar @parts;

    my $i=0;

    my $word_num = 0;

    while($i < $part_num - 1){
	if($parts[$i] =~ /^\(/ && ($parts[$i] !~ /\)$/) &&
	   $parts[$i+1] !~ /^\(/ && ($parts[$i+1] =~ /\)$/)){
	    
	    
	    my $word = $parts[$i+1];
	    my $tag = $parts[$i];
	    $tag =~ s/^\(+//;
	    $i += 2;

	    if((!$keep_EC) && $tag =~ /\-NONE\-/i){
		next;
	    }

	    $word =~ s/\)+$//;

	    push(@$words_ptr, $word);
	    push(@$tags_ptr, $tag);

	    $word_num ++;

	}else{
	    $i ++;
	}
    }

    return $word_num;
}


########################## misc function
## add -PRD to a NP/PP/ADJP when the verb is a "be" verb.
sub add_func_tags_to_ttree {
    my ($ttree) = @_;

    my $root = $ttree->root;
    add_func_tags_to_ttree_from_tnode($root);
}


sub add_func_tags_to_ttree_from_tnode {
    my ($tnode) = @_;
    
    if($tnode->type == $LEAF_NODE){
	return;
    }

    my $children_ptr = $tnode->children;
    my @children = @$children_ptr;
    my $children_num = @children;
    for(my $i=0; $i<$children_num; $i++){
	my $child = $children[$i];
	add_func_tags_to_ttree_from_tnode($child);
    }


    my $first_child = $children[0];

    ### deal with corpola
    if(($children_num > 1) &&
       ($first_child->type == $LEAF_NODE) &&
       ($tnode->s_tag =~ /^VP$/i) &&
       ($first_child->word =~ /^(be|was|were|is|are|am|being|been|\'s|\'re|\'m)$/i)){
	my $second_child = $children[1];
	if(($second_child->type == $INTERNAL_NODE) &&
	   ($second_child->s_tag =~ /^(NP|PP|ADJP)$/i)){
	    my $new_s_tag = $children[1]->s_tag;
	    $new_s_tag .= "-PRD";
	    $second_child->s_tag($new_s_tag);
	}
    }
}


### Here we assume that the word_pos_array[] is ordered
### return the span: max - min
sub get_tnode_span {
    my ($tnode, $start_ptr, $end_ptr) = @_;

    my $tmp_ptr = $tnode->word_pos_array;
    my $num = scalar @$tmp_ptr;
    
    if($num == 0){
	return 0;
    }
    
    $$start_ptr = $tmp_ptr->[0];
    $$end_ptr = $tmp_ptr->[$num-1];

    return $$end_ptr - $$start_ptr + 1;
}



### for each leaf node, store the tnode pointer in hash_ptr
sub get_leaf_node_ptrs {
    my ($ttree, $hash_ptr) = @_;
    
    get_leaf_node_ptrs_from_tnode($ttree->root, $hash_ptr);
}

sub get_leaf_node_ptrs_from_tnode {
    my ($tnode, $hash_ptr) = @_;

    if($tnode->type == $LEAF_NODE){
	my $pos = $tnode->word_pos;
	$hash_ptr->{$pos} = $tnode;
    }else{
	my $children_ptr = $tnode->children;
	for(my $i=0; $i<scalar @$children_ptr; $i++){
	    my $child = $children_ptr->[$i];
	    get_leaf_node_ptrs_from_tnode($child, $hash_ptr);
	}
    }
}


### return suc
### find the lowest ancestor of tnode1 and tnode2
sub find_common_ancestor_of_two_tnodes {
    my ($ttree, $tnode1, $tnode2, $ancest_ptr) = @_;

    my $root = $ttree->root;
    my $ptr = $tnode1;
    while($ptr != $root){
	if(is_ancestor_in_ttree($ttree, $ptr, $tnode2)){
	    $$ancest_ptr = $ptr;
	    return 1;
	}else{
	    $ptr = $ptr->parent;
	}
    }
    
    $$ancest_ptr = $root;
    return 1;
}


### return 1 if tnode1 is an ancestor of tnode2
sub is_ancestor_in_ttree {
    my ($ttree, $tnode1, $tnode2) = @_;
    
    if($tnode1 == $tnode2){
	return 1;
    }

    if($tnode2 == $ttree->root){
	return 0;
    }

    return is_ancestor_in_ttree($ttree, $tnode1, $tnode2->parent);
}

## tnode2 is an ancestor of tnode1, find the position of the child of tnode2
##  that dominates tnode1
##
## return pos
sub find_child_pos {
    my ($ttree, $tnode1, $tnode2) = @_;
    
    if($tnode2->type == $LEAF_NODE){
	return -1;
    }

    my $children_ptr = $tnode2->children;
    
    for(my $i=0; $i<scalar @$children_ptr; $i++){
	my $child = $children_ptr->[$i];
	if(is_ancestor_in_ttree($ttree, $child, $tnode1)){
	    return $i;
	}
    }

    return -1;
}

## return the number of words
sub get_words_in_ttree {
    my ($ttree, $words_ptr) = @_;
    
    @$words_ptr = ();
    get_words_in_ttree_from_tnode($ttree->root, $words_ptr);

    return scalar @$words_ptr;
}

sub get_words_in_ttree_from_tnode {
    my ($tnode, $words_ptr) = @_;
    
    if($tnode->type == $LEAF_NODE){
	push(@$words_ptr, $tnode->word);
    }else{
	my $children_ptr = $tnode->children;
	for(my $i=0; $i<scalar @$children_ptr; $i++){
	    my $child = $children_ptr->[$i];
	    get_words_in_ttree_from_tnode($child, $words_ptr);
	}
    }
}

### return word str in a subtree
sub subtree_word_str {
    my ($tnode) = @_;
    my @words = ();
    get_words_in_ttree_from_tnode($tnode, \@words);
    return join(" ", @words);
}


## return the sent str
sub sent_str {
    my ($ttree) = @_;
    
    my @words = ();
    get_words_in_ttree($ttree, \@words);
    
    return join(" ", @words);
}


############################### rule extraction
## return the number of rule tokens
## hash2 stores the non-terminal rules, e.g., NP -> N
## hash1 stores the leaf node rules: e.g., N->book
##
## don't clear $rule_hash_ptr
sub extract_CF_rules_from_ttree {
    my ($ttree, $rule_hash1_ptr, $rule_hash2_ptr, $include_f_tag) = @_;

    return extract_CF_rules_from_ttree_from_tnode($ttree->root, 
						  $rule_hash1_ptr,
						  $rule_hash2_ptr,
						  $include_f_tag);
}


## return the number of rule tokens
sub extract_CF_rules_from_ttree_from_tnode {
    my ($tnode, $rule_hash1_ptr, $rule_hash2_ptr, $include_f_tag) = @_;

    if(!defined($tnode)){
	return 0;
    }

    my $s_tag = $tnode->s_tag;
    my @tags = ($s_tag);

    ## leaf node
    if($tnode->type == $LEAF_NODE){
	my $word = $tnode->word;
	push(@tags, $word);
	my $suc = add_CF_rule_to_hash($rule_hash2_ptr, \@tags);
	if(!$suc){
	    print STDERR "adding rule fails for ", join("+", @tags), "\n";
	    return 0;
	}else{
	    return 1;
	}
    }

    #### internal node
    my $cnt = 0;
    my $children_ptr = $tnode->children;
    my @children = @$children_ptr;
    foreach my $child (@children){
	my $tag = $child->s_tag;
	my $f_tag_ptr = $child->f_tag;
	if($include_f_tag && (scalar @$f_tag_ptr > 0)){
	    $tag .= "-" . join("-", @$f_tag_ptr);
	}
	push(@tags, $tag);
	$cnt += extract_CF_rules_from_ttree_from_tnode($child, $rule_hash1_ptr,
						       $rule_hash2_ptr,
						       $include_f_tag);
    }

    my $suc = add_CF_rule_to_hash($rule_hash1_ptr, \@tags);
    if(!$suc){
	print STDERR "adding rule fails for ", join("+", @tags), "\n";
	return 0;
    }else{
	$cnt ++;
    }

    return $cnt;
}


## return suc
sub add_CF_rule_to_hash {
    my ($rule_hash_ptr, $tags_ptr) = @_;

    my $str = join(" ", @$tags_ptr);
    if(defined($rule_hash_ptr->{$str})){
	$rule_hash_ptr->{$str} ++;
    }else{
	$rule_hash_ptr->{$str} = 1;
    }

    return 1;
}


### given "S+NP+VP", return "S -> NP VP"
sub CF_rule_to_str {
    my ($key) = @_;

    my @tags = split(/\s+/, $key);
    
    my $left_tag = shift(@tags);
    my $right_str = join(" ", @tags);

    return "$left_tag -> $right_str";
}

sub print_CF_rules {
    my ($fp, $rule_hash_ptr, $rule_token_ptr, $rule_type_ptr) = @_;

    $$rule_token_ptr = 0;
    $$rule_type_ptr = 0;

    for my $key (sort {$rule_hash_ptr->{$b} <=> $rule_hash_ptr->{$a}} 
		 keys %$rule_hash_ptr) {
	my $str = CF_rule_to_str($key);
	my $val = $rule_hash_ptr->{$key};
	print $fp "$val\t$str\n";
	$$rule_type_ptr ++;
	$$rule_token_ptr += $val;
    }
}


sub print_CF_rules_to_two_files {
    my ($output_file, $rule_hash1_ptr, $rule_hash2_ptr) = @_;

    open(my $output_fp1, ">$output_file.syn") 
    or die "cannot open $output_file.syn\n";

    open(my $output_fp2, ">$output_file.lex") 
    or die "cannot open $output_file.lex\n";

    my $rule_tokens = 0;
    my $rule_types = 0;   ## the number of rule types

    print_CF_rules($output_fp1, $rule_hash1_ptr, \$rule_tokens, \$rule_types);
    print $output_fp1 "syntactic rules: rule_tokens=$rule_tokens, rule_types=$rule_types\n";
    print STDERR "syntactic rules: rule_tokens=$rule_tokens, rule_types=$rule_types\n";

    
    print_CF_rules($output_fp2, $rule_hash2_ptr, \$rule_tokens, \$rule_types);
    print $output_fp2 "lexical rules: rule_tokens=$rule_tokens, rule_types=$rule_types\n";
    print STDERR "lexical rules: rule_tokens=$rule_tokens, rule_types=$rule_types\n";

    close($output_fp1);
    close($output_fp2);
}


### return 1 if tnode is the head child of its parent
### return 0 otherwise
sub is_head_child {
    my ($tnode) = @_;
    
    my $parent = $tnode->parent;
    if(!defined($parent)){
	## the root of the tree
	return 0;
    }

    my $head_child = $parent->head_child;
    if($tnode == $head_child){
	return 1;
    }else{
	return 0;
    }
    
}

1;
