#!/opt/python-3.4/bin/python3.4
# -*- coding: utf-8 -*-

import sys
import math
import operator
import optparse
import statistics
from xigt.codecs import xigtxml

# AGGREGATION Project - Sentential Negation
#
# Anthony Princing
# CLMS Master's Project
# Winter, Spring 2016

DS_ROOT_TAG = 'root'
DS_CONJUNCTION_TAG = 'conj'
POS_VERB_TAG = 'VERB'
POS_CONJUNCTION_TAG = 'CONJ'
PS_QUESTION_TAG = 'SQ'
PS_VERB_TAGS = ['VB', 'VBD', 'VBG', 'VBN', 'VBP', 'VBZ']
MOD_to_modifies = {'S' : 's', 'VP' : 'vp', 'VB' : 'v', 'VBD' : 'v', 'VBG' : 'v', 'VBN' : 'v', 'VBP' : 'v', 'VBZ' : 'v'}



class InferenceIGT(object):

   """ InferenceIGT provides and interface to enriched IGT in Xigt format.
   Each instance of this class represents one IGT and provides methods
   to determine some of the properties of this IGT.
   """
   def __init__(self, XIGT):

      """ Class intialization primarily tests and if possible creates an
      access variable for each tier, setting a truth value on the tier's
      accessibility.

         Args:
            XIGT -- A single IGT in Xigt format
      """
      self.igt = XIGT
      self.igtid = XIGT.id
      self.norm_tier_exists = True
      self.word_tier_exists = True
      self.morph_tier_exists = True
      self.gloss_tier_exists = True
      self.wgloss_tier_exists = True
      self.wpos_tier_exists = True
      self.gwpos_tier_exists = True
      self.wps_tier_exists = True
      self.wds_tier_exists = True
      self.tw_tier_exists = True
      self.twps_tier_exists = True
      self.twds_tier_exists = True
      self.align_tier_exists = True
      
      try:
         self.norm_tier = XIGT['n']
      except (KeyError, ValueError):
         self.norm_tier_exists = False

      try:
         self.word_tier = XIGT['w']
      except (KeyError, ValueError):
         self.word_tier_exists = False

      try:
         self.morph_tier = XIGT['m']
      except (KeyError, ValueError):
         self.morph_tier_exists = False

      try:
         self.gloss_tier = XIGT['g']
      except (KeyError, ValueError):
         self.gloss_tier_exists = False

      try:
         self.wgloss_tier = XIGT['gw']
      except (KeyError, ValueError):
         self.wgloss_tier_exists = False

      try:
         self.wpos_tier = XIGT['w-pos']
      except (KeyError, ValueError):
         self.wpos_tier_exists = False

      try:
         self.gwpos_tier = XIGT['gw-pos']
      except (KeyError, ValueError):
         self.gwpos_tier_exists = False

      try:
         self.wps_tier = XIGT['w-ps']
      except (KeyError, ValueError):
         self.wps_tier_exists = False

      try:
         self.wds_tier = XIGT['w-ds']
      except (KeyError, ValueError):
         self.wds_tier_exists = False

      try:
         self.tw_tier = XIGT['tw']
      except (KeyError, ValueError):
         self.tw_tier_exists = False

      try:
         self.twps_tier = XIGT['tw-ps']
      except (KeyError, ValueError):
         self.twps_tier_exists = False

      try:
         self.twds_tier = XIGT['tw-ds']
      except (KeyError, ValueError):
         self.twds_tier_exists = False

      try:
         self.align_tier = XIGT['a']
      except (KeyError, ValueError):
         self.align_tier_exists = False

      self.is_question = self.get_is_question()
      self.has_conjunction = self.get_has_conjunction()
      self.primary_verb_loc = self.get_primary_verb_loc()

#  end __init__


   def get_w_id(self, gw_id):
      """ Find the w_id associated with the given gw_id.

         Args:
            gw_id -- A word gloss tier id

         Returns:
            string -- w_id
      """
      if self.wgloss_tier[gw_id].alignment:
         return self.wgloss_tier[gw_id].alignment
      else:
         return gw_id.lstrip('g')
#  end get_w_id


   def get_m_id(self, g_id):
      """ Find the m_id (morpheme tier) associated with the give g_id.

         Args:
            g_id -- A gloss tier id

         Returns:
            string -- m_id
      """
      m_id = None
      if g_id:
         if self.gloss_tier[g_id].alignment:
            m_id = self.gloss_tier[g_id].alignment
      return m_id
# end get_m_id


   def get_gw_id(self, w_id):
      """ Find the gw_id (gloss word tier) associated with the given w_id.

         Args:
            w_id -- A word tier id

         Returns:
            string -- gw_id
      """
      gw_id = None
      if w_id:
         gw_id = 'g' + w_id
         if self.wgloss_tier_exists:
            for gw_item in self.wgloss_tier:
               if gw_item.alignment:
                  if gw_item.alignment == w_id:
                     gw_id = gw_item.id
      return gw_id
# end get_gw_id


   def get_is_question(self):
      """ Determine if this IGT represents a question.

         Returns:
            bool -- True if IGT is a question
      """
      ans = False
      if self.wps_tier_exists:
         for wps_item in self.wps_tier:
            if wps_item.value() == PS_QUESTION_TAG:
               ans = True
      return ans
#  end is_question


   def get_has_conjunction(self):
      """ Determine if this IGT contains conjunctions by first checking
      the glossed word POS tier and if that fails the dependency 
      parse tier.

         Returns:
            bool -- True if IGT contains conjunctions
      """
      has_conj = None
      if self.gwpos_tier_exists:
         for gwpos_item in self.gwpos_tier:
            if gwpos_item.value() == POS_CONJUNCTION_TAG:
               has_conj = True
      if (not has_conj) and (self.wds_tier_exists):
         for wds_item in self.wds_tier:
            if wds_item.value() == DS_CONJUNCTION_TAG:
               has_conj = True
      return has_conj
# end has_conjunction


   def get_bound_orthography(self, g_id):
      """ Find the orthography associated with the given g_id.

         Args:
            g_id -- A gloss tier id

        Returns:
           string -- orthography of gloss
      """
      ortho = None
      if g_id:
         if self.morph_tier_exists and self.gloss_tier_exists:
            ortho = self.morph_tier[self.gloss_tier[g_id].alignment].value().lower()
      return ortho
# end get_bound_orthography


   def get_free_orthography(self, word_id):
      """ Find the orthography associated with the given word.

         Args:
            word_id -- A word tier or glossed word tier id

        Returns:
           string -- orthography of word
      """
      ortho = None
      if word_id:
         if word_id.startswith('g'):
            w_id = self.get_w_id(word_id)
         else:
            w_id = word_id
         if w_id:
            if self.word_tier_exists:
               if self.word_tier.get(w_id):
                  ortho = self.word_tier[w_id].value().lower().replace('-', '')
      return ortho
#  end get_free_orthography


   def get_primary_verb_loc(self):
      """ Make several attempts to determine the primary verb in the sentence
      represented by the IGT. First look at the dependency parse tier, then
      the phrase structure parse and POS tiers. If all attempts fail return
      a value of None.

         Returns:
            string -- The word tier (w_id) of the primary verb
      """
      verb_loc = None
      if self.wds_tier_exists:
         for wds_item in self.wds_tier:
            if wds_item.value() == DS_ROOT_TAG:
               if wds_item.get_attribute('dep'):
                  verb_loc = wds_item.get_attribute('dep')
                  break
      if (not verb_loc) and self.wps_tier_exists:
         for wps_item in self.wps_tier:
            if wps_item.value() == 'ROOT':
               child_list = wps_item.get_attribute('children').split()
               if child_list:
                  idx = 0
                  while idx < len(child_list):
                     child_item = self.igt.get_item(child_list[idx])
                     if 'VB' in child_item.value():
                        if child_item.alignment:
                           verb_loc = child_item.alignment
                        break
                     elif child_item.get_attribute('children'):
                        child_list.extend(child_item.get_attribute('children').split())
                     idx += 1
      if (not verb_loc) and self.wps_tier_exists:
         for wps_item in self.wps_tier:
            if wps_item.value() in PS_VERB_TAGS:
               if wps_item.alignment:
                  verb_loc = wps_item.alignment
                  break
      if (not verb_loc) and self.wpos_tier_exists:
         for wpos_item in self.wpos_tier:
            if wpos_item.value() == POS_VERB_TAG:
               if wpos_item.alignment:
                  verb_loc = wpos_item.alignment
                  break
      if (not verb_loc) and self.gwpos_tier_exists:
         for gwpos_item in self.gwpos_tier:
            if gwpos_item.value() == POS_VERB_TAG:
               if gwpos_item.alignment:
                  verb_loc = gwpos_item.alignment.lstrip('g')
                  break
      return verb_loc
# end primary_verb_loc
      

   def all_glosses_in_word(self, word_id):
      """ Generate a list of all glosses for the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            list -- all glosses in word
      """
      glosses = []
      if word_id:
         if word_id.startswith('g'):
            gw_id = word_id
         else:
            gw_id = get_gw_id(word_id)
         if self.gloss_tier_exists:
            for gloss_item in self.gloss_tier:
               if gloss_item.content:
                  gw = gloss_item.content.split('[')[0]
                  if gw == gw_id:
                     glosses.append(gloss_item.value().lower())
      return glosses
# end all_glosses_in_word


   def gram_in_gloss_count(self, gram_list):
      """ Determine how many times any of a list of grams appear in a gloss line.

         Args:
            gram_list -- List of grams to searh for in gloss line

         Returns:
            int -- count of grams
      """
      gram_count = None
      prior_gloss_align = None
      if self.gloss_tier_exists:
         gram_count = 0
#         prior_gloss_align = ''
         for gloss_item in self.gloss_tier:
            if gloss_item.value().lower() in gram_list:
               if gloss_item.alignment:
                  if gloss_item.alignment != prior_gloss_align:
                     gram_count += 1
                     prior_gloss_align = gloss_item.alignment
               else:
                  gram_count += 1
      return gram_count
#  end gram_in_gloss_count


   def is_affixed(self, tier_item):
      """ Determine if the given item displays affixation.

         Args:
            tier_item -- Likely a word tier or glossed word tier id

         Returns:
            bool - True if item shows affixation, False otherwise
      """
      if '-' in tier_item.value():
         return True
      else:
         return False
#  end is_affixed


   def gram_in_tier_item(self, gram_list, tier_item):
      """ Determine if any gram from the gram list appears in the tier item.

         Args:
            gram_list -- The list of grams to search for
            tier_item -- The Xigt object to search

         Returns:
            bool -- True if any gram in item, False otherwise
      """
      if tier_item.value().lower() in gram_list:
         return True
      else:
         return False
# end gram_in_tier_item


   def gram_in_tier(self, gram_list, a_tier):
      """ Determine if a gram from the list appears in the given tier.

         Args:
            gram_list -- The list of grams to search for
            a_tier -- The tier to search

         Returns:
            bool -- True if any gram in tier, False otherwise
      """
      loc_list = []
      for an_item in a_tier:
         for a_gram in gram_list:
            if a_gram in an_item.value().lower():
               if an_item.id not in loc_list:
                  loc_list.append(an_item.id)
      return loc_list
# end gram_in_tier


   def get_pos_tag(self, word_id):
      """ Determine the POS tag of the word provided.
      Check both POS tiers if necessary/possible.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            string -- POS tag of word
      """
      pos = None
      if word_id:
         if word_id.startswith('g'):
            gw_id = word_id
         else:
            gw_id = self.get_gw_id(word_id)
         if self.gwpos_tier_exists:
            for gwpos_item in self.gwpos_tier:
               if gwpos_item.alignment:
                  if gw_id == gwpos_item.alignment:
                     pos = gwpos_item.value()
         if (not pos) and self.wpos_tier_exists:
            for wpos_item in self.wpos_tier:
               if self.wgloss_tier_exists:
                  if self.wgloss_tier[gw_id].alignment == self.wpos_tier.alignment:
                     w_pos = wpos_item.value()
      return pos
#  end get_pos_tag


   def get_adjacent_pos_tags(self, word_id):
      """ Determine the POS tags of the words on either side of the given word.
      Again, check both POS tiers if necessary/possible.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            list -- Item zero is left POS tag, item one is right POS tag
      """
      found = False
      before_id = None
      remember_id = None
      adj_tags = [None, None]
      if word_id:
         if word_id.startswith('g'):
            gw_id = word_id
         else:
            gw_id = self.get_gw_id(word_id)
         if self.gwpos_tier_exists:
            for gwpos_item in self.gwpos_tier:
               if gwpos_item.alignment:
                  if gw_id == gwpos_item.alignment:
                     found = True
                     before_id = remember_id
                  elif found:
                     if before_id:
                        adj_tags[0] = self.igt.get_item(before_id).value()
                     adj_tags[1] = gwpos_item.value()
                     break
               remember_id = gwpos_item.id
         if (not found) and self.wpos_tier_exists:
            for wpos_item in self.wpos_tier:
               if self.wgloss_tier_exists:
                  if self.wgloss_tier[gw_id].alignment == self.wpos_tier.alignment:
                     found = True
                     before_id = remember_id
                  elif found:
                     if before_id:
                        adj_tags[0] = self.igt.get_item(before_id).value()
                     adj_tags[1] = wpos_item.value()
                     break
               remember_id = wpos_item.id
      return adj_tags
# end get_adjacent_pos_tags


   def is_verb_adjacent(self, word_id):
      """ Determine if there is a verb on either side of the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            bool -- True if verb next door, False otherwise
      """
      adj_tags = self.get_adjacent_pos_tags(word_id)
      if (adj_tags[0] in POS_VERB_TAGS) or (adj_tags[1] in POS_VERB_TAGS):
         return True
      else:
         return False
# end is_verb_adjacent


   def get_ps_parent(self, word_id):
      """ Determine the parent phrase structure element for the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            tuple -- Item zero is parent wps_id, and item one is parent phrase structure element name
      """
      parent = None
      child_id = None
      if word_id:
         if word_id.startswith('g'):
            w_id = self.get_w_id(word_id)
         else:
            w_id = word_id
         if self.wps_tier_exists:
            for wps_item in self.wps_tier:
               if wps_item.alignment:
                  if wps_item.alignment == w_id:
                     child_id = wps_item.id
                     break
         if child_id:
            for wps_item in self.wps_tier:
               siblings = wps_item.get_attribute('children')
               if siblings:
                  if child_id in siblings:
                     parent = (wps_item.id, wps_item.value()) 
      return parent
#  end get_ps_parent


   def get_ps_siblings(self, word_id):
      """ Generate a list of the phrase structure sibling, if any, for the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            list -- List of wps_id values for siblings
      """
      siblings = []
      wps_id = None
      if word_id:
         if word_id.startswith('g'):
            w_id = self.get_w_id(word_id)
         else:
            w_id = word_id
         if self.wps_tier_exists:
            for wps_item in self.wps_tier:
               if wps_item.alignment:
                  if wps_item.alignment == w_id:
                     wps_id = wps_item.id
                     break
         if wps_id:
            for wps_item in self.wps_tier:
               children = wps_item.get_attribute('children')
               if children:
                  child_list = children.split()
                  if wps_id in child_list:
                     for a_child in [c for c in child_list if c != wps_id]:
                        if siblings:
                           siblings.append((a_child, self.wps_tier[a_child].value()))
                        else:
                           siblings = [(a_child, self.wps_tier[a_child].value())]
      return siblings
#  end get_ps_siblings


   def get_ps_children(self, word_id):
      """ Generate a list of the phrase structure children, if any, for the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            list -- List of wps_id values for children
      """
      children = []
      wps_id = None
      if word_id:
         if word_id.startswith('g'):
            w_id = self.get_w_id(word_id)
         else:
            w_id = word_id
         if self.wps_tier_exists:
            for wps_item in self.wps_tier:
               if wps_item.alignment:
                  if wps_item.alignment == w_id:
                     wps_id = wps_item.id
                     break
         if wps_id:
            for wps_item in self.wps_tier:
               if wps_item.id == wps_id:
                  children = wps_item.get_attribute('children')
                  if children:
                     child_list = children.split()
                     for a_child in child_list:
                        if child_list:
                           children.append((a_child, self.wps_tier[a_child].value()))
                        else:
                           children = [(a_child, self.wps_tier[a_child].value())]
      return children
#  end get_ps_children


   def get_word_modifies(self, word_id):
      """ Determine what phrase structure element (V, VP, or S) is modified
      by the given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            string -- Phrase structure element value of 'v', 'vp', or 's'
      """
      mod = None
      if word_id:
         if word_id.startswith('g'):
            w_loc = self.get_w_id(word_id)
         else:
            w_loc = word_id
         parent_ps_tag = None
         parent_ps = self.get_ps_parent(w_loc)
         sib_list = self.get_ps_siblings(w_loc)
         mod_id_list = [id for id, tag in sib_list]
         mod_tag_list = [tag for id, tag in sib_list]
         for id in mod_id_list:
            item = self.igt.get_item(id)
            if item:
               sib_children = item.get_attribute('children')
               if sib_children:
                  sib_child_list = sib_children.split()
                  for sib_child in sib_child_list:
                     mod_tag_list.append(self.igt.get_item(sib_child).value())
   
         for ps_tag in mod_tag_list:
            if ps_tag in MOD_to_modifies:
               mod = MOD_to_modifies[ps_tag]
               break
   
         if (not mod) and parent_ps:
            parent_ps_tag = parent_ps[1]
            if parent_ps_tag == 'ROOT':
               mod = 's'
            elif parent_ps_tag == 'S':
               mod = 'vp'
            else:
               mod = 'v'

      return mod
#  end get_word_modifies


   def get_verb_relative_loc(self, word_id):
      """ Determine the verb relative (before or after) location of the 
      given word.

         Args:
            word_id -- A word tier or glossed word tier id

         Returns:
            string -- Representing location, 'before' or 'after', verb
      """
      loc = None
      verb_loc = self.primary_verb_loc
      if not verb_loc:
         return None
      if word_id:
         if word_id.startswith('g'):
            loc = self.get_w_id(word_id)
         else:
            loc = word_id
      if loc and verb_loc:
         if int(loc.lstrip('w')) > int(verb_loc.lstrip('w')):
            return 'after'
         else: # verb_loc > loc:
            return 'before'
      else:
         return None
#  end get_verb_relative_loc

#end class InferenceIGT
