#!/usr/bin/env python3.4

# Alex Burrell
# LING 575 Computational Methods in Endangered Language Documentation
# Spring 2015


import sys
from xigt.codecs import xigtxml


# VerbValenceFromIGT is a utility class useful for identifying verb transitivity
# It has a single public method: infer_valence_from_igt
class VerbValenceFromIGT:
    # Lists of all noun and verb tags seen in "tw-ps" tier
    NOUNS = ['n', 'pro', 'NOUN', 'PRP', 'NN', 'NP', 'PRP$', 'NNS', 'NNP', 'NNPS', 'PRON']
    VERBS = ['vi', 'vt', 'VBG', 'VBP', 'VP', 'VBN', 'VB', 'VBD', 'VBZ', 'VERB']

    # Special cases in English, for auxillary verbs, passive sentences, and wh- phrases
    EXCLUDE = ['do', 'does', 'did', 'has', 'have', 'are']
    PASSIVE = ['is', 'was', 'were', 'be', 'been']
    WH_HOW = ['WP']

    # Public method for identifying verb transitivity
    # Takes the following arguments:
    # - verb    element from "w-pos" tier (Xigt)
    # - igt     containing IGT element (Xigt)
    # - gloss   element from glosses ("g") tier (Xigt)
    # Returns:
    # - string: "intrans" or "trans"
    def infer_valence_from_igt(self, verb, igt, gloss):
        valence = {
            0: 'trans',
            1: 'intrans'
        }
        # Default to transitive
        intransitive = False

        # Keep track of all nouns and verbs encountered in the translation line
        translation_nouns = []
        translation_verbs = []

        # To augment the bilingual alignments tier ("a") try and figure out which
        # translation verb might be a match
        most_likely_alignment = False

        # Identify the sentence as passive or active, and the auxillary if it's passive
        is_passive = False
        auxillary = False

        # Identify the sentence as containing a "wh-" phrase
        is_wh_phrase = False

        # Catch a special case where the gloss contains ".for" (e.g. "look.for", "be.good.for")
        # Through the normal path, these words are marked as intransitive because they are followed
        # by a prepositional phrase, but they should be transitive.
        if gloss.get_content().find('.for') > -1:
            return valence[False]

        # Error catching, in case the "tw-ps" tier is missing in the IGT
        if not igt.get('tw-ps'):
            return valence[intransitive]

        # Only needed in order to get the number of words in the translation line
        tw = igt.get('tw')

        # Get the phrase structure tree of the translation line
        tw_ps = igt.get('tw-ps')

        # Find passive sentences first
        if len(tw) > 1:
            # Get the last word of the translation, cleaned of punctuation
            last_word = igt.get_item(tw_ps[len(tw)-1].alignment).get_content()
            last_word = last_word.replace('.', '').replace(',', '').replace('?', '')

            # Get the second to last word of the translation, cleaned of punctuation
            second_last_word = igt.get_item(tw_ps[len(tw)-2].alignment).get_content()
            second_last_word = second_last_word.replace('.', '').replace(',', '')

            # If the last word POS is a verb and does not end in "ing",
            # and the second to last word is in the list of passive-marking auxillaries
            # then mark as a passive sentence, and track the auxillary alignment
            if tw_ps[len(tw)-1].get_content() in self.VERBS and last_word[-3:] != 'ing' and second_last_word in self.PASSIVE:
                is_passive = True
                auxillary = tw_ps[len(tw)-2].alignment

        # Go through all nodes of the translation phrase structure tree to get a list of the nouns and verbs
        # that will be needed in the next step to find transitivity
        for pos in tw_ps:
            if pos.alignment and pos.get_content() in self.VERBS:
                if igt.get_item(pos.alignment).get_content() in self.EXCLUDE:
                    continue
                if is_passive and pos.alignment == auxillary:
                    continue
                translation_verbs.append(pos.alignment)
            if pos.alignment and pos.get_content() in self.NOUNS:
                translation_nouns.append(pos.alignment)
            if pos.get_content() in self.WH_HOW:
                is_wh_phrase = True

        # This block looks for string matches between the verb's gloss and each translation verb
        # looking for an obvious alignment
        # If the bilingual-alignments tier improves this should be able to be removed
        def matches(target, translation, alignment):
            translation = translation.replace(',', '').replace('.', '').replace('?', '')
            translation = translation.lower()
            # Attempting to compare the root verb didn't actually help much,
            # and was slow - wasn't worth the trade off
            # translation = self.lemmatizer.lemmatize(translation, 'v')
            target = target.lower()
            if target.find(translation) > -1 or translation.find(target) > -1:
                return alignment
            return False
        if len(translation_verbs) > 1:
            target = gloss.get_content()
            for t_verb in translation_verbs:
                translation = igt.get_item(t_verb).get_content()
                # Finds the first match in the sentence
                if not most_likely_alignment:
                    most_likely_alignment = matches(target, translation, t_verb)

        # Creates dictionaries useful for crawling up the phrase structure tree, or down the tree
        # Takes a phrase structure tier element (Xigt)
        # Returns a tuple containing:
        # - dictionary for reading up
        # - dictionary for reading down
        # - list of the verb phrase nodes
        # - reference to the root element of the phrase structure tree
        def create_trees(ps):
            up_tree = {}
            down_tree = {}
            vps = []
            root = False

            for ps in igt.get(ps):
                # These are the nodes for the words themselves
                if ps.alignment:
                    up_tree[ps.alignment] = ps.id
                    down_tree[ps.id] = ps.alignment
                # These are the intermediate nodes
                else:
                    for child in ps.get_attribute('children').split(' '):
                        up_tree[child] = ps.id
                    down_tree[ps.id] = ps.get_attribute('children').split(' ')
                    if ps.get_content() == 'VP':
                        vps.append(ps.id)
                    elif ps.get_content() in self.NOUNS:
                        translation_nouns.append(ps.id)
                    elif ps.get_content() == 'ROOT':
                        root = ps.id

            return (up_tree, down_tree, vps, root)

        def match_english_verbs():
            # Checks to see if this verb's gloss is in bilingual alignments
            # If it is, we're in luck and can only check the transitivity of that one verb
            # Otherwise, at least keep track of the words we know this one is not aligned to
            alignment = False
            not_aligned_to = []
            for bilingual_alignment in igt.get('a'):
                if bilingual_alignment.get_attribute('target') == gloss.id:
                    alignment = bilingual_alignment.get_attribute('source')
                else:
                    not_aligned_to.append(bilingual_alignment.get_attribute('source'))

            # Default to the bilingual alignments tier, but if that wasn't present
            # use the one that was found through string matching
            if not alignment and most_likely_alignment:
                alignment = most_likely_alignment

            up_tree, down_tree, vps, root = create_trees('tw-ps')
            print(up_tree)
            print(down_tree)
            print(vps)
            print(root)
            def has_n_to_right(node, previous_node):
                # Checks to see if the current node has a noun to the right of the given child node
                no_n = True
                if previous_node:
                    for child in down_tree[node][down_tree[node].index(previous_node)+1:]:
                        if child in translation_nouns:
                            no_n = False
                return no_n

            def determine_valence(verb):
                # Starting at the bottom of the tree, work up until either a verb phrase parent
                # or the root node is reached
                # Intransitive means that the container verb phrase does not have a noun child
                # to the right of the verb
                # Transitive means that it has a noun child to the right of the verb
                previous_node = False
                node = up_tree[verb]
                intransitive = True
                while node not in vps and node != root:
                    intransitive = intransitive and has_n_to_right(node, previous_node)
                    previous_node = node
                    node = up_tree[node]

                return intransitive and has_n_to_right(node, previous_node)

            # Default to transitive
            valence = False

            # If there was either a match in the bilingual alignments tier
            # or in the string matching, only that aligned verb can be checked
            if alignment:
                valence = determine_valence(alignment)
                if is_passive and alignment == translation_verbs[-1]:
                    valence = False
            # Otherwise, check all the verbs, and try and make a good guess
            else:
                valences = []
                for t_verb in translation_verbs:
                    if t_verb in not_aligned_to:
                        continue

                    valences.append(determine_valence(t_verb))

                # If it's a passive or wh- statement, set the last verb valence to transitive
                if len(valences) >= 1 and (is_passive or is_wh_phrase):
                    valences[-1] = False

                # print(valences)

                # If all the valences are either true or false, return that value
                if len(valences) >= 1 and (all(valences) or not any(valences)):
                    valence = valences[0]
                # Trying to do something smarter when it can't be sure
                # Return intransitive if that's the majority
                elif len(valences) >= 2:
                    positive = 0
                    negative = 0
                    for v in valences:
                        positive += v
                        negative += not v
                    valence = positive > negative

            return valence

        # Default to transitive for all verbs
        # Good for establishing a baseline
        hypothesis_1 = False

        hypothesis_2 = False
        if len(translation_verbs) > 0:
            hypothesis_2 = match_english_verbs()

        # Choose which hypothesis (or hypotheses) to go with
        intransitive = hypothesis_1

        # print(gloss.get_content(), valence[intransitive])
        return valence[intransitive]


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('Requires at least 1 xigt file')
        exit()

    verbs = {}
    verb_valence_finder = VerbValenceFromIGT()

    for filename in sys.argv[1:]:
        try:
            xigt = xigtxml.load(open(filename, encoding='utf-8'))
        except:
            print('File does not appear to be Xigt compatible')

        # Find all the verbs
        for igt in xigt.igts:
            # print()
            # for sentence in igt.get('n'):
            #     print(sentence.get_content())

            # For each verb in the w-pos tier, find its transitivity, and all other information needed
            # for the choices file lexicon
            pos_tier = igt.get('w-pos')
            if not pos_tier:
                continue
            for pos in pos_tier:
                if pos.get_content()[0].lower() == 'v':
                    word = igt.get_item(pos.alignment)
                    morphemes = [igt.get_item(label) for label in word.referrers()['segmentation']]
                    roots = [morpheme for morpheme in morphemes if morpheme.type == 'Root']
                    morpheme_grams = [morpheme.get_content() for morpheme in roots]
                    gloss_labels = [morpheme.referrers()['alignment'][0] for morpheme in roots]
                    glosses = [igt.get_item(label) for label in gloss_labels]
                    gloss_grams = [gloss.get_content().lower() for gloss in glosses]

                    orth = '-'.join(morpheme_grams)
                    lemma = '-'.join(gloss_grams)

                    valence = verb_valence_finder.infer_valence_from_igt(pos, igt, glosses[0])
                    if valence not in verbs:
                        verbs[valence] = {}
                    if orth not in verbs[valence]:
                        verbs[valence][orth] = {}
                    if lemma not in verbs[valence][orth]:
                        verbs[valence][orth][lemma] = 0

                    verbs[valence][orth][lemma] += 1

    # Print a partial choices file - just the lexicon
    for i, valence in enumerate(verbs.keys(), 1):
        print("verb{}_valence={}".format(i, valence))
        for j, orth in enumerate(verbs[valence].keys(), 1):
            for lemma in verbs[valence][orth].keys():
                print("  verb{}_stem{}_orth={}".format(i, j, orth))
                print("  verb{}_stem{}_pred=_{}_v_rel".format(i, j, lemma))
