﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Text.RegularExpressions;
using System.Xml;

namespace ctn2xigt
{
    class Program
    {
        public const string ODIN = "odin";
        public const string NO_GLOSS = "NO_GLOSS";
        public const string NO_INFO = "NO_INFO";
        public const string SPACE = "__";
        public const string DOT = "_";

        static void Main(string[] args)
        {
            string inputfileLoc = args[0];
            string outputFile = args[1];

            List<SrcIGT> internalIGTS = readChintangData(inputfileLoc);
            List<XigtIGT> xigtList = new List<XigtIGT>();
            int igtCount = 0;
            int skipped = 0;
            foreach (SrcIGT srcphrase in internalIGTS)
            {
                // for small example:
                //if (igtCount == 3)
                //    break;
                //if (!invalidIGTs.Contains(phraseCount))
                //{
                if (srcphrase.words == null || srcphrase.words.Contains(null))
                {
                    skipped++;
                    continue;
                }
                XigtIGT igt = constructXigtIGT(srcphrase);
                if (igt == null)
                {
                    // igt is null when language line or gloss line is not present.
                    skipped++;
                    continue;
                }
                xigtList.Add(igt);
                igtCount++;
                //}
            }
            writeXigtXML(outputFile, xigtList);
            Console.WriteLine("Skipped {0} igts", skipped);

        }

        private static void writeXigtXML(string outPath, List<XigtIGT> igts)
        {
            using (FileStream filestream = new FileStream(outPath, FileMode.Create))
            using (StreamWriter sw = new StreamWriter(filestream))
            using (XmlTextWriter writer = new XmlTextWriter(sw))
            {
                writer.Formatting = Formatting.Indented;
                writer.Indentation = 4;
                writer.WriteStartDocument();
                // ROOT
                writer.WriteStartElement("xigt-corpus");
                // METADATA
                writeMetadataXML(writer);
                // IGTs
                foreach (var igt in igts)
                {
                    //if (igt.id == 711)
                    //    Console.WriteLine("Debug");
                    writer.WriteStartElement("igt");
                    writeAttributes(writer, igt.attributes);
                    // IGT TIERS
                    foreach (var tier in igt.tiers)
                    {
                        writer.WriteStartElement("tier");
                        writeAttributes(writer, tier.attributes);
                        // IGT ITEMS
                        foreach (var item in tier.items)
                        {
                            writer.WriteStartElement("item");
                            writeAttributes(writer, item.attributes);
                            writer.WriteEndElement();
                        }
                        writer.WriteEndElement();
                    }
                    writer.WriteEndElement();
                }
                writer.WriteEndElement();
                writer.WriteEndDocument();
            }
        }

        private static void writeAttributes(XmlTextWriter writer, List<Attribute> attributes)
        {
            foreach (var attr in attributes)
            {
                // Xigt has a content attribute, which is actually content's location, and the content itself.
                // If it is content itself, then there is no "name", just the value.
                if (attr.name != Tier.STRING_CONTENT)
                    writer.WriteAttributeString(attr.name, attr.value);
                else
                    writer.WriteString(attr.value);
            }
        }

        //This is hardcoded for the Chintang file.
        private static void writeMetadataXML(XmlTextWriter writer)
        {
            writer.WriteStartElement("metadata");
            writer.WriteAttributeString("type", "xigt-meta");
            writer.WriteStartElement("meta");
            writer.WriteAttributeString("type", "language");
            writer.WriteAttributeString("tiers", "phrases, glosses, morphemes");
            writer.WriteAttributeString("iso-639-3", "ctn");
            writer.WriteAttributeString("name", "Chintang");
            writer.WriteEndElement();
            writer.WriteStartElement("meta");
            writer.WriteAttributeString("type", "language");
            writer.WriteAttributeString("tiers", "glosses, translations");
            writer.WriteAttributeString("iso-639-3", "eng");
            writer.WriteAttributeString("name", "English");
            writer.WriteEndElement();
            writer.WriteEndElement();
        }

        private static XigtIGT constructXigtIGT(SrcIGT srcphrase)
        {
            //if (srcphrase.id == 282)
            //    Console.WriteLine("Debug");
            RawOdinTier rawOdinTier = new RawOdinTier(srcphrase);
            NormalizedOdinTier normOdinTier = new NormalizedOdinTier(srcphrase);
            //fail case: no language line or no gloss line
            if (String.IsNullOrWhiteSpace(normOdinTier.languageLine) || String.IsNullOrWhiteSpace(normOdinTier.glossLine))
                return null;
            PhrasesTier phrasesTier = new PhrasesTier();
            WordsTier wordsTier = new WordsTier(srcphrase, normOdinTier.languageLine, "words", "w", "p");
            TranslationsTier translationsTier = new TranslationsTier();
            WordsTier translationWordsTier = new WordsTier(srcphrase, normOdinTier.translationLine, "words", "tw", "t");
            MorphemesTier morphemesTier = new MorphemesTier(srcphrase);
            GlossesTier glossesTier = new GlossesTier(srcphrase, morphemesTier, normOdinTier);
            POS_Tier posTier = new POS_Tier(srcphrase);
            Attribute docID = new Attribute("doc-id", "ctn splits");
            //Attribute id = new Attribute("id", srcphrase.id.ToString());
            Attribute id = new Attribute("id", "i" + srcphrase.id);
            XigtIGT igt = new XigtIGT(rawOdinTier, normOdinTier, phrasesTier, wordsTier, morphemesTier, glossesTier, posTier,
                translationsTier, translationWordsTier, docID, id, srcphrase.id);
            return igt;
        }

        public static bool isRoot(SrcMorpheme m)
        {
            return m.morphType == MorphType.Root || m.morphType == MorphType.Stem || m.morphType == MorphType.Bound_Stem
                || m.morphType == MorphType.Bound_Root;
        }
        static List<SrcIGT> readChintangData(string file)
        {
            List<SrcIGT> igts = new List<SrcIGT>();
            int igtCount = 0;
            List<string> lines = File.ReadLines(file, Encoding.UTF8).ToList();
            for (int i = 0; i < lines.Count; i++)
            {
                int firstLineNum = i;
                List<string> igtLines = new List<string>();
                while (!String.IsNullOrWhiteSpace(lines[i]) && i < lines.Count)
                {
                    igtLines.Add(lines[i]);
                    i++;
                }
                SrcIGT igt = new SrcIGT(igtLines, ref firstLineNum, ++igtCount);
                igts.Add(igt);
            }
            return igts;
        }
    }

    class SrcIGT
    {
        public string rf;
        public string elan;
        public string tx;
        public string gw;
        public string eng;
        public string nep;
        public string mph;
        public string normMph;
        public string normGls;
        public string mgl;
        public string lg;
        public string idLine;
        public string ps;
        public string dt;
        public string ed;
        public string rt;
        public List<string> customLines = new List<string>();
        public int id;
        public int lastLineNum;
        public string comment;
        public string sg;
        public List<SrcWord> words;
        public String translation = null;

        public SrcIGT(List<string> lines, ref int lineCount, int id)
        {
            this.id = id;
            lineCount = lineCount + 3;
            this.lastLineNum = lineCount;
            foreach (var ln in lines)
            {
                if (ln.Trim().Contains(' '))
                {
                    string identifier = ln.Substring(ln.IndexOf('\\') + 1, ln.IndexOf(' ')).ToLower().Trim();
                    string content = ln.Substring(ln.IndexOf(' ') + 1).Trim();
                    switch (identifier)
                    {
                        case "ref":
                            this.rf = ln;
                            break;
                        case "elanparticipant":
                            this.elan = content;
                            break;
                        case "tx":
                            this.tx = content;
                            break;
                        case "gw":
                            this.gw = content;
                            break;
                        case "mph":
                            this.mph = content;
                            break;
                        case "mgl":
                            this.mgl = content;
                            break;
                        case "lg":
                            this.lg = content;
                            break;
                        case "id":
                            this.idLine = content;
                            break;
                        case "ps":
                            this.ps = content;
                            break;
                        case "eng":
                            this.eng = content;
                            break;
                        case "nep":
                            this.nep = content;
                            break;
                        case "dt":
                            this.dt = content;
                            break;
                        case "comment":
                            this.comment = content;
                            break;
                        case "ed":
                            this.ed = content;
                            break;
                        case "sg":
                            this.sg = content;
                            break;
                        case "rt":
                            this.rt = content;
                            break;
                        default:
                            this.customLines.Add(identifier + "," + content);
                            break;
                        //throw new Exception("Unexpected line in data file");
                    }
                }
            }
            if (mgl == null || mph == null)
            {
                Console.WriteLine("No gloss line or morpheme line");
                words = null;
            }
            else
                words = buildWords();
            translation = eng;
        }

        List<SrcWord> buildWords()
        {
            List<SrcWord> words = new List<SrcWord>();
            //TODO: also handle the tx line and its alignment to gw
            string[] gwSplit = Regex.Replace(gw, " +", " ").Split(' ');
            string[] posSplit;
            if (this.ps == null)
            {
                posSplit = new string[gwSplit.Length];
                for (int i = 0; i < gwSplit.Length; i++)
                    posSplit[i] = "NoPos";
            }
            else
            {
                posSplit = Regex.Replace(ps, " +", " ").Split(' ');
            }
            string[] glossSplit = Regex.Replace(mgl, " +", " ").Split(' ');
            string[] morphSplit = Regex.Replace(mph, " +", " ").Split(' ');
            if (posSplit.Length != glossSplit.Length)
            {
                Console.WriteLine("POS number not equal to glosses number");
                return null;
            }
            var glst = glossSplit.ToList();
            var mphlst = morphSplit.ToList();
            var poslst = posSplit.ToList();
            glst.RemoveAll(s => s == "-");
            mphlst.RemoveAll(s => s== "-");
            if (glst.Count != mphlst.Count)
            {
                Console.WriteLine("Misaligned gloss and morpheme line");
                return null;
            }
            BNF bnf = new BNF(mph);
            //if (this.id == 711)
            //    Console.WriteLine("Debug");
            List<BNFWord> segmentedWords = bnf.parseSentence();
            if (segmentedWords == null)
                return null;
            //if (segmentedWords.Count != gwSplit.Length)
            //    throw new Exception("wrong number of words");
            int morphIndex = 0;
            for (int i = 0; i < segmentedWords.Count; i++)
            {
                string gword = "PROBLEM_IN_DATA";
                if (segmentedWords.Count == gwSplit.Length)
                    gword = gwSplit[i];
                BNFWord word = segmentedWords[i];
                SrcWord srcword = new SrcWord();
                srcword.txt = gword;
                List<SrcMorpheme> morphemes;
                srcword.norm = buildSegmentedWord(word, out morphemes, glossSplit, ref morphIndex);
                srcword.morphemes = morphemes;
                string pos;
                int rootIdx = word.roots[0].Item2;
                if (rootIdx >= posSplit.Length)
                    pos = "NoPos";
                else
                    pos = posSplit[word.roots[0].Item2].Trim('-');
                POS partOfSpeech;
                bool getPOS = Enum.TryParse(pos, out partOfSpeech);//(typeof(POS), pos);
                if (!getPOS)
                    srcword.pos = POS.NoPOS;
                else
                    srcword.pos = partOfSpeech;
                words.Add(srcword);
            }
            return words;
        }

        string buildSegmentedWord(BNFWord word, out List<SrcMorpheme> morphemes, string[] glosses, ref int currentIdx)
        {
            morphemes = new List<SrcMorpheme>();
            StringBuilder sb = new StringBuilder();
            var infixes = word.infixes.ToList();
            foreach (string prefix in word.prefixes)
            {
                sb.Append(prefix);
                morphemes.Add(new SrcMorpheme(prefix, "prefix", glosses[currentIdx]));
                currentIdx++;
            }
            foreach (var rootIndexPair in word.roots)
            {
                sb.Append(rootIndexPair.Item1);
                //if (this.id == 36)
                //    Console.WriteLine("Debug");
                var gloss = glosses[currentIdx];
                if (gloss == "-")
                    gloss = "-" + glosses[++currentIdx];
                morphemes.Add(new SrcMorpheme(rootIndexPair.Item1, "root", gloss.Trim('-')));
                currentIdx++;
                if (infixes.Count > 0)
                {
                    //assume one infix for now

                    sb.Append(infixes[0]);
                    var infixGloss = glosses[currentIdx];
                    morphemes.Add(new SrcMorpheme(word.infixes[0], "infix", infixGloss));
                    infixes.RemoveAt(0);
                    currentIdx++;
                }
            }
            foreach (string suffix in word.suffixes)
            {
                var gloss = glosses[currentIdx];
                if (gloss == "-")
                    gloss = "-" + glosses[++currentIdx].Trim('-');
                morphemes.Add(new SrcMorpheme(suffix, "suffix", gloss));
                sb.Append(suffix);
                currentIdx++;
            }
            return sb.ToString();
        }

    }

    class BNF
    {
        Scanner sc;
        public BNF(string sentence)
        {
            sc = new Scanner(sentence);
        }

        public List<BNFWord> parseSentence()
        {
            return sentence();
        }

        List<BNFWord> sentence()
        {
            List<BNFWord> words = new List<BNFWord>();
            while (sc.currentToken != null)
            {
                var wd = word();
                if (wd == null)
                    return null;
                words.Add(wd);
            }
            return words;
        }

        BNFWord word()
        {
            BNFWord w = new BNFWord();
            while (sc.currentToken.tok == Token.PREFIX)
            {
                w.prefixes.Add(sc.fetchToken().val);
            }
            if (sc.currentToken.tok != Token.ROOT)
            {
                Console.WriteLine("No root in a word");
                return null;
                //throw new Exception("No root!");
            }
            int rootIdx = sc.idx;
            w.roots.Add(new Tuple<string, int>(sc.fetchToken().val, rootIdx));
            bool hadInfix = false;
            while (sc.currentToken != null && sc.currentToken.tok == Token.INFIX)
            {
                w.infixes.Add(sc.fetchToken().val);
                hadInfix = true;
            }
            if (hadInfix)
            {
                if (sc.currentToken != null && sc.currentToken.tok == Token.ROOT)
                {
                    rootIdx = sc.idx;
                    w.roots.Add(new Tuple<string, int>(sc.fetchToken().val, rootIdx));
                }
                else
                {
                    throw new Exception("Expected root after infix");
                }
            }
            while (sc.currentToken != null && sc.currentToken.tok == Token.HYPHEN)
            {
                sc.fetchToken();
                //sc.idx--;
                if (sc.peekToken().tok == Token.ROOT)
                {
                    rootIdx = sc.idx;
                    w.roots.Add(new Tuple<string, int>('-' + sc.fetchToken().val, rootIdx));
                }
                else // treat as a suffix even if it appears as a prefix
                    w.suffixes.Add('-' + sc.fetchToken().val.Trim('-'));
            }
            while (sc.currentToken != null && sc.currentToken.tok == Token.SUFFIX)
                w.suffixes.Add(sc.fetchToken().val);
            return w;
        }
    }


    class BNFWord
    {
        public List<string> prefixes = new List<string>();
        public List<string> suffixes = new List<string>();
        public List<string> infixes = new List<string>();
        public List<Tuple<string, int>> roots = new List<Tuple<string, int>>();
    }

    enum Token { ROOT, PREFIX, SUFFIX, HYPHEN, INFIX }

    class Scanner
    {
        public KV currentToken;
        string input;
        public int idx;
        public Scanner(string input)
        {
            this.input = Regex.Replace(input, " +", " ").Trim();
            idx = -1;
            fetchToken();
        }

        public KV fetchToken()
        {
            var tmp = currentToken;
            currentToken = getNextToken();
            idx++;
            return tmp;
        }

        public KV peekToken()
        {
            return currentToken;
        }

        private KV getNextToken()
        {
            if (!input.Contains(' '))
            {
                if (input == "")
                    return null;
                else
                {
                    string val = input;
                    input = "";
                    if (val.First() == '-')
                        return new KV { tok = Token.SUFFIX, val = val };
                    return new KV { tok = Token.ROOT, val = val };
                }
            }
            string nextTokContent = input.Substring(0, input.IndexOf(' ')).Trim();
            input = input.Substring(input.IndexOf(' ') + 1);
            if (nextTokContent == "-")
                return new KV { tok = Token.HYPHEN, val = nextTokContent };
            else if (nextTokContent.First() == '-' && nextTokContent.Last() == '-')
                return new KV { tok = Token.INFIX, val = nextTokContent };
            else if (nextTokContent.First() == '-')
                return new KV { tok = Token.SUFFIX, val = nextTokContent };
            else if (nextTokContent.Last() == '-')
                return new KV { tok = Token.PREFIX, val = nextTokContent };
            return new KV { tok = Token.ROOT, val = nextTokContent };
        }
    }

    class KV
    {
        public Token tok;
        public string val;
    }
    class SrcWord
    {
        public List<SrcMorpheme> morphemes;
        public string txt = null;
        public string norm = null;
        public POS pos = POS.Undefined;
        public bool isPunct = false;
        public bool isMulti = false;

        public SrcWord()
        {
            morphemes = new List<SrcMorpheme>();
        }

        public SrcWord(SrcWord copyMe, string txt, string norm, List<SrcMorpheme> morphemes)
        {
            this.pos = copyMe.pos;
            this.isPunct = copyMe.isPunct;
            this.txt = txt;
            this.norm = norm;
            this.morphemes = morphemes;
        }

        public string getGlsRoot()
        {
            foreach (var srcmorph in morphemes)
            {
                if (srcmorph.morphType == MorphType.Root || srcmorph.morphType == MorphType.Stem || srcmorph.morphType == MorphType.Bound_Root
                    || srcmorph.morphType == MorphType.Bound_Stem)
                {
                    return srcmorph.gls;
                }
            }
            return null;
        }
    }

    /*
     * A direct mapping from the source XML.
     * Represents one morpheme which consists of:
     * morph type
     * items types:
     * txt
     * cf
     * hn
     * gls
     * msa
     */
    class SrcMorpheme
    {
        public string txt;
        public string norm;
        public MorphType morphType = MorphType.Undefined;
        public string cf;
        public string gls;
        public string normgls;
        public string msa;
        public string vt; //variantTypes
        public Int16 hn = 153; // arbitrary default; not all entries contain hn, but hn==0 is valid.
        public SrcMorpheme(string txt, string type, string gloss)
        {
            if (type == "prefix")
                morphType = MorphType.Prefix;
            else if (type == "suffix")
                morphType = MorphType.Suffix;
            else if (type == "infix")
                morphType = MorphType.Infix;
            else
                morphType = MorphType.Root;
            this.txt = txt;
            this.gls = gloss;
            this.normgls = gloss;
            norm = txt;
            cf = txt;
        }
    }

    // pos type
    public enum POS
    {
        adv, adj, n, v, dem, det, interj,
        interrog, onom, proform, pro, prt, quant, subordconn,
        Undefined, vi, vt, gm, num, NoPOS
    };

    // morpheme type
    public enum MorphType
    {
        Stem, Root, Bound_Root, Bound_Stem, Enclitic, Suffix, Prefix, Infix, Particle, Phrase, NI, NoType, Undefined
    };

    class XigtIGT
    {
        public int id;
        MorphemesTier morphemesTier;
        RawOdinTier rawOdinTier;
        NormalizedOdinTier normOdinTier;
        GlossesTier glossesTier;
        public Metadata meta = null;
        public List<Attribute> attributes = new List<Attribute>();
        public IEnumerable<Tier> tiers = null;
        private List<Tier> privateTiers = new List<Tier>();
        public XigtIGT(RawOdinTier rawOdinTier, NormalizedOdinTier normOdinTier, PhrasesTier phrasesTier, WordsTier wordsTier,
            MorphemesTier morphemesTier, GlossesTier glossesTier,
            POS_Tier posTier, TranslationsTier translationsTier, WordsTier translationWordsTier,
            Attribute DocID, Attribute id, int igtID)
        {
            this.id = igtID;
            attributes.Add(DocID);
            attributes.Add(id);
            privateTiers.Add((Tier)rawOdinTier);
            privateTiers.Add((Tier)phrasesTier);
            privateTiers.Add((Tier)normOdinTier);
            privateTiers.Add((Tier)wordsTier);
            privateTiers.Add((Tier)morphemesTier);
            privateTiers.Add((Tier)glossesTier);
            privateTiers.Add((Tier)posTier);
            privateTiers.Add((Tier)translationsTier);
            privateTiers.Add((Tier)translationWordsTier);
            tiers = privateTiers;
            this.normOdinTier = normOdinTier;
            this.rawOdinTier = rawOdinTier;
            this.morphemesTier = morphemesTier;
            this.glossesTier = glossesTier;
            validateL();
            //if (!mcb2xigt.invalidIGTs.Contains(this.id))
            validateG();
        }

        private void validateL()
        {
            string line = normOdinTier.languageLine;
            StringBuilder sb = new StringBuilder();
            //if (this.id == 205)
            //    Console.Write(this.id);
            string[] split = line.Split(' ');
            foreach (var item in morphemesTier.items)
            {
                var morphItem = (MorphemesTierItem)item;
                var range = morphItem.attributes.Where(n => n.name == MorphemesTier.SEGMENTATION).ToList()[0].value;
                var match = Regex.Match(range, @"w([0-9]+)\[([0-9]+):([0-9]+)\]");
                int wordIdx = Int32.Parse(match.Groups[1].Value) - 1; // zero-based
                int start = Int32.Parse(match.Groups[2].Value);
                int end = Int32.Parse(match.Groups[3].Value);
                if (end == split[wordIdx].Length)
                    end--;
                string append = split[wordIdx].Substring(start, end - start + 1);
                sb.Append(append);
                if (end == split[wordIdx].Length - 1 && wordIdx != split.Length - 1)
                    sb.Append(' ');
            }
            if (sb.ToString() != line)
                throw new Exception("IGT not validated: morphemes do not add up to language line exactly.");
        }

        private void validateG()
        {
            string line = normOdinTier.glossLine;
            StringBuilder sb = new StringBuilder();
            foreach (var item in glossesTier.items)
            {
                var glossItem = (GlossesTierItem)item;
                var range = glossItem.attributes.Where(n => n.name == MorphemesTier.CONTENT).ToList()[0].value;
                var match = Regex.Match(range, @"\[([0-9]+):([0-9]+)\]");
                int start = Int32.Parse(match.Groups[1].Value);
                int end = Int32.Parse(match.Groups[2].Value);
                if (end == line.Length)
                    end--;
                string append = line.Substring(start, end - start + 1);
                sb.Append(append);
            }
            if (sb.ToString().Trim() != line.Trim())
                throw new Exception("IGT not validated: glosses do not add up to gloss line exactly.");
        }
    }

    class Metadata
    {
        public string id = null;
        public string lang = null;
        public string iso = null;
    }

    public abstract class Tier
    {
        public const string ID = "id";
        public const string ALIGNMENT = "alignment";
        public const string CONTENT = "content";
        public const string STRING_CONTENT = "stringContent";
        public const string SEGMENTATION = "segmentation";
        public const string TYPE = "type";

        public Char[] delimiters = { '-', '.', '=', ':' };

        public IEnumerable<TierItem> items;
        public List<Attribute> attributes = new List<Attribute>();
    }

    class PhrasesTier : Tier
    {
        private List<PhrasesTierItem> itemsList = new List<PhrasesTierItem>();
        public PhrasesTier()
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "phrases"));
            attributes.Add(new Attribute(Tier.ID, "p"));
            attributes.Add(new Attribute(Tier.CONTENT, "n"));
            itemsList.Add(new PhrasesTierItem());
            this.items = itemsList;
        }
    }

    class WordsTier : Tier
    {
        private List<WordTierTtem> itemsList = new List<WordTierTtem>();
        public WordsTier(SrcIGT srcIGT, string line, string type, string id, string segmentation)
        {
            this.attributes.Add(new Attribute(Tier.TYPE, type));
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.SEGMENTATION, segmentation));
            int wordCount = 0;
            for (int i = 0; i < line.Length; i++)
            {
                int start = i;
                while (i < line.Length && line[i] != ' ')
                {
                    i++;
                }
                string word = line.Substring(start, i - start);
                itemsList.Add(new WordTierTtem(id + (++wordCount).ToString(), segmentation + "1[" + start + ":" + i + "]"));
            }
            this.items = itemsList;
        }
    }

    class MorphemesTier : Tier
    {
        private List<MorphemesTierItem> itemsList = new List<MorphemesTierItem>();
        public MorphemesTier(SrcIGT srcIGT)
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "morphemes"));
            this.attributes.Add(new Attribute(Tier.ID, "m"));
            this.attributes.Add(new Attribute(Tier.SEGMENTATION, "w"));
            int currentWord = 1;
            int skippedPunctuationWords = 0;
            foreach (var w in srcIGT.words)
            {
                if (w.isPunct)
                {
                    skippedPunctuationWords++;
                    continue;
                }
                string wordID = "w" + currentWord;
                int morphCount = 1;
                int start = 0;
                foreach (var m in w.morphemes)
                {
                    // need extra "character" for python-style indexig; in affixes is accounted for by delimiters.
                    int end = start + m.norm.Trim('-').Trim('=').Length;
                    string morphID = "m" + currentWord + "." + morphCount;
                    string morphRange = wordID + "[" + start + ":" + end + "]";
                    itemsList.Add(new MorphemesTierItem(morphID, morphRange, m.normgls, m.morphType.ToString()));
                    start = end + 1;
                    morphCount++;
                }
                if (w.morphemes.Count > 0)
                    currentWord++;
            }
            this.items = itemsList;
        }
    }

    class GlossesTier : Tier
    {
        private List<GlossesTierItem> itemsList = new List<GlossesTierItem>();
        public GlossesTier(SrcIGT srcIGT, MorphemesTier morphemesTier, NormalizedOdinTier odinTier)
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "glosses"));
            this.attributes.Add(new Attribute(Tier.ID, "g"));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, "m"));
            this.attributes.Add(new Attribute(Tier.CONTENT, "n"));
            int start = 0;
            int end = 0;
            List<MorphemesTierItem> morphemeItems = (List<MorphemesTierItem>)morphemesTier.items;
            for (int i = 0; i < morphemeItems.Count(); i++)
            {
                MorphemesTierItem m = morphemeItems[i];
                end = m.glossSegment.Trim('-').Trim('=').Length + start;
                itemsList.Add(new GlossesTierItem("g" + m.numericID, "m" + m.numericID, "n2[" + start + ":" + end + "]"));
                start = end + 1;
            }
            //prevalidation check
            if (end != odinTier.glossLine.Length)
            {
                throw new Exception("Gloss line length is not equal to last index used for Xigt Gloss Tier!");
                //mcb2xigt.invalidIGTs.Add(srcIGT.id);
            }
            this.items = itemsList;
        }
    }

    class POS_Tier : Tier
    {
        private List<POS_TierItem> itemsList = new List<POS_TierItem>();
        public POS_Tier(SrcIGT srcIGT)
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "pos"));
            this.attributes.Add(new Attribute(Tier.ID, "w-pos"));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, "w"));
            int wordCount = 1;
            foreach (var w in srcIGT.words)
            {
                if (w.isPunct)
                    continue;
                itemsList.Add(new POS_TierItem("w-pos" + wordCount, "w" + wordCount, w.pos.ToString()));
                wordCount++;
            }
            this.items = itemsList;
        }
    }

    class MorphemeTypeTier : Tier
    {

    }

    class OdinTier : Tier
    {
        public string languageLine;
        public string glossLine;
        public string translationLine;
    }

    class RawOdinTier : OdinTier
    {
        private List<OdinTierItem> itemsList = new List<OdinTierItem>();
        public RawOdinTier(SrcIGT phrase)
        {
            this.attributes.Add(new Attribute(Tier.TYPE, Program.ODIN));
            this.attributes.Add(new Attribute(Tier.ID, "o"));
            this.attributes.Add(new Attribute("state", "raw"));
            constructRawOdinTier(phrase);
            languageLine = Regex.Replace(languageLine, " +", " ");
            glossLine = Regex.Replace(glossLine, " +", " ");
            translationLine = Regex.Replace(translationLine, " +", " ");
            languageLine = languageLine.Trim();
            glossLine = glossLine.Trim();
            translationLine = translationLine.Trim();
        }

        private void constructRawOdinTier(SrcIGT srcphrase)
        {
            StringBuilder odinOrthBuilder = new StringBuilder();
            StringBuilder odinGlossBuilder = new StringBuilder();
            foreach (SrcWord swd in srcphrase.words)
            {
                for (int i = 0; i < swd.morphemes.Count; i++)
                {
                    SrcMorpheme mph = swd.morphemes[i];
                    if (mph.txt != null)
                        odinOrthBuilder.Append(mph.txt);
                    if (mph.gls != null && mph.txt != null)
                    {
                        //prefix
                        if (delimiters.Contains(mph.txt.First()) && mph.gls != null && mph.gls.First() != mph.txt.First())
                            odinGlossBuilder.Append(mph.txt.First());
                        // gloss
                        odinGlossBuilder.Append(mph.gls);
                        // suffix
                        if (delimiters.Contains(mph.txt.Last()) && mph.gls != null && mph.gls.Last() != mph.txt.Last())
                            odinGlossBuilder.Append(mph.txt.Last());
                    }
                }
                odinOrthBuilder.Append(" ");
                odinGlossBuilder.Append(" ");
            }
            // Remove last spaces
            languageLine = odinOrthBuilder.ToString().Trim();
            glossLine = odinGlossBuilder.ToString().Trim();
            translationLine = Regex.Replace(srcphrase.translation.Trim(), " +", " ");
            itemsList.Add(new OdinTierItem("L", srcphrase.lastLineNum - 2, languageLine, "o1"));
            itemsList.Add(new OdinTierItem("G", srcphrase.lastLineNum - 1, glossLine, "o2"));
            itemsList.Add(new OdinTierItem("T", srcphrase.lastLineNum, translationLine, "o3"));
            this.items = itemsList;
        }
    }

    class NormalizedOdinTier : OdinTier
    {
        private List<OdinTierItem> itemsList = new List<OdinTierItem>();
        public NormalizedOdinTier(SrcIGT phrase)
        {
            if (phrase.id == 711)
                Console.WriteLine("Debug");
            this.attributes.Add(new Attribute(Tier.TYPE, Program.ODIN));
            this.attributes.Add(new Attribute(Tier.ID, "n")); // this is the clean normalized odin
            this.attributes.Add(new Attribute("state", "normalized"));
            constructNormOdinTier(phrase);
            languageLine = Regex.Replace(languageLine, " +", " ");
            glossLine = Regex.Replace(glossLine, " +", " ");
            languageLine = Regex.Replace(languageLine, "-=", "=");
            glossLine = Regex.Replace(glossLine, "-=", "=");
            languageLine = Regex.Replace(languageLine, "-+", "-");
            glossLine = Regex.Replace(glossLine, "-+", "-");
            translationLine = Regex.Replace(translationLine, " +", " ");
            languageLine = languageLine.Trim();
            glossLine = glossLine.Trim();
            translationLine = translationLine.Trim();
        }

        private void constructNormOdinTier(SrcIGT srcphrase)
        {
            //if (srcphrase.id == 1122)
            //    Console.WriteLine();
            StringBuilder odinOrthBuilder = new StringBuilder();
            StringBuilder odinGlossBuilder = new StringBuilder();
            foreach (SrcWord swd in srcphrase.words)
            {
                //if (swd.txt == "ovetsikaiganaa")
                //    Console.WriteLine();
                for (int i = 0; i < swd.morphemes.Count; i++)
                {
                    SrcMorpheme mph = swd.morphemes[i];
                    odinOrthBuilder.Append(mph.norm);
                    //prefix
                    if (delimiters.Contains(mph.norm.First()) && mph.normgls.First() != mph.norm.First())
                        odinGlossBuilder.Append(mph.norm.First());
                    // gloss
                    odinGlossBuilder.Append(mph.normgls);
                    // suffix
                    if (delimiters.Contains(mph.norm.Last()) && mph.normgls.Last() != mph.norm.Last())
                        odinGlossBuilder.Append(mph.norm.Last());
                }
                odinOrthBuilder.Append(" ");
                odinGlossBuilder.Append(" ");
            }
            // Remove last spaces
            languageLine = odinOrthBuilder.ToString().Trim();
            glossLine = odinGlossBuilder.ToString().Trim();
            translationLine = Regex.Replace(srcphrase.translation.Trim(), " +", " ");
            itemsList.Add(new OdinTierItem("L", srcphrase.lastLineNum - 2, languageLine, "n1"));
            itemsList.Add(new OdinTierItem("G", srcphrase.lastLineNum - 1, glossLine, "n2"));
            itemsList.Add(new OdinTierItem("T", srcphrase.lastLineNum, translationLine, "n3"));
            this.items = itemsList;
        }
    }

    class TranslationsTier : Tier
    {
        private List<TranslationsTierItem> itemsList = new List<TranslationsTierItem>();
        public TranslationsTier()
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "translations"));
            this.attributes.Add(new Attribute(Tier.ID, "t"));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, "p"));
            this.attributes.Add(new Attribute(Tier.CONTENT, "n"));
            itemsList.Add(new TranslationsTierItem("t1", "p1", "n3"));
            this.items = itemsList;
        }
    }

    public abstract class TierItem
    {
        public List<Attribute> attributes = new List<Attribute>();
    }

    class OdinTierItem : TierItem
    {
        public const string TAG = "tag";
        public const string LINE = "line";
        public OdinTierItem(string tag, int line, string content, string id)
        {
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(OdinTierItem.TAG, tag));
            this.attributes.Add(new Attribute(OdinTierItem.LINE, line.ToString()));
            this.attributes.Add(new Attribute(Tier.STRING_CONTENT, content));
        }

    }

    class WordTierTtem : TierItem
    {
        public WordTierTtem(string id, string segmentation)
        {
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.SEGMENTATION, segmentation));
        }
    }

    class POS_TierItem : TierItem
    {
        public POS_TierItem(string id, string alignment, string content)
        {
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, alignment));
            this.attributes.Add(new Attribute(Tier.STRING_CONTENT, content));
        }
    }

    // need to collapse with POS_TierItem?
    class TranslationsTierItem : TierItem
    {
        public TranslationsTierItem(string id, string alignment, string content)
        {
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, alignment));
            this.attributes.Add(new Attribute(Tier.CONTENT, content));
        }
    }

    class MorphemesTierItem : TierItem
    {
        public Char[] delimiters = { '=', '.', ':', '-', '~' };
        public const string DEFAULT = "ABSENT_GLOSS";
        public string numericID = null;
        public string glossSegment = null;
        public MorphemesTierItem(string id, string segmentation, string gloss, string morphType)
        {
            if (gloss == null)
                gloss = MorphemesTierItem.DEFAULT;
            // for xigt, we need to discard the '=' in clitic glosses, because
            // otherwise we will count it twice in the n2[0:5]-type attributes.
            // Clitics are the only morpheme types in mcb data that have the '=' in both orth and gls field.
            // Other morpheme types only have the delimiter in the orth field.
            else
                if (delimiters.ToList().Contains(gloss.First()))
                    gloss = gloss.Substring(1);
            this.glossSegment = gloss;
            this.numericID = id.Substring(1);
            this.attributes.Add(new Attribute(Tier.TYPE, morphType));
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.SEGMENTATION, segmentation));
        }
    }

    class GlossesTierItem : TierItem
    {
        public GlossesTierItem(string id, string alignment, string content)
        {
            this.attributes.Add(new Attribute(Tier.ID, id));
            this.attributes.Add(new Attribute(Tier.ALIGNMENT, alignment));
            this.attributes.Add(new Attribute(Tier.CONTENT, content));
        }
    }
    class PhrasesTierItem : TierItem
    {
        public PhrasesTierItem()
        {
            this.attributes.Add(new Attribute(Tier.TYPE, "transliteration"));
            this.attributes.Add(new Attribute(Tier.ID, "p1"));
            this.attributes.Add(new Attribute(Tier.CONTENT, "n1"));
        }
    }

    public class MorphemeTypeItem : TierItem
    {

    }

    public class Attribute
    {
        public string name;
        public string value;
        public Attribute(string name, string val)
        {
            this.value = val;
            this.name = name;
        }
    }
}
