#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
__program__ = eval_tam.py
__author__ = MichaelLockwood
__projectclass__ = LING575
__projecttopic__ = AGGREGATION Inferrence
__projectname__ = Evaluate Tense, Aspect, and Mood
__date__ = June2015
__credits__ = None
__collaborators__ = None

This module loads two sets of choices files; a gold standard set and an
auto-generated set. The evaluation procedure compares the gold standard
against the inference set and a baseline set.
"""

import map_gloss
import os

from aggutils import ChoicesReader

class Evaluation:

    agg_path = ''
    choices_path = ''

    def load_data(datasets):
        # Set path to the root of the AGGREGATION folder
        Evaluation.agg_path = os.path.abspath(os.path.join(os.path.abspath(
                            os.path.join(os.getcwd(), os.pardir)), os.pardir))
        # Link to the TAM choices subdirectory for loading auto choices files
        Evaluation.choices_path = os.getcwd() + '/TAM_choices/'
        # Load map_gloss CLSO
        map_gloss.MapGloss.set_restart(False)
        map_gloss.MapGloss.load_combined_structures()
        # Initialize each language and build their choices file evaluations
        for dataset in datasets:
            Language(dataset, datasets[dataset])
        return True

    def set_evaluation(name):
        base, infer = Evaluation.build_summary()
        writer = open(Evaluation.choices_path + str(name) + '.eval', 'w')
        writer.write('Summary for dataset: ' + str(name) + '\n\n')
        writer.write('Baseline Average Scores\n')
        writer.write('\tPrecision: ' + str(base[0]) + '\n')
        writer.write('\tRecall: ' + str(base[1]) + '\n')
        writer.write('\tF-Score: ' + str(base[2]) + '\n\n')
        writer.write('Inference Average Scores\n')
        writer.write('\tPrecision: ' + str(infer[0]) + '\n')
        writer.write('\tRecall: ' + str(infer[1]) + '\n')
        writer.write('\tF-Score: ' + str(infer[2]) + '\n\n')
        for iso in sorted(Language.objects.keys()):
            writer.write(iso + '\n\n')
            writer.write('\tBaseline Scores\n')
            writer.write('\t\tPrecision: ' +
                         str(Language.objects[iso]._base_precision) + '\n')
            writer.write('\t\tRecall: ' +
                         str(Language.objects[iso]._base_recall) + '\n')
            writer.write('\t\tF-Score: ' +
                         str(Language.objects[iso]._base_fscore) + '\n\n')
            writer.write('\tInference Scores\n')
            writer.write('\t\tPrecision: ' +
                         str(Language.objects[iso]._infer_precision) + '\n')
            writer.write('\t\tRecall: ' +
                         str(Language.objects[iso]._infer_recall) + '\n')
            writer.write('\t\tF-Score: ' +
                         str(Language.objects[iso]._infer_fscore) + '\n\n')
            if Language.objects[iso]._infer_confusion_matrix['FNeg']:
                writer.write('\tFalse Negatives\n')
                for gloss in Language.objects[iso]._infer_confusion_matrix[
                    'FNeg']:
                    writer.write('\t\t' + str(gloss) + '\n')
                writer.write('\n')
            if Language.objects[iso]._infer_confusion_matrix['FPos']:
                writer.write('\tFalse Positives\n')
                for gloss in Language.objects[iso]._infer_confusion_matrix[
                    'FPos']:
                    writer.write('\t\t' + str(gloss) + '\n')
                writer.write('\n')
        writer.close()
        return True

    def build_summary():
        # Initialize variables
        base_precision = 0.0
        base_recall = 0.0
        base_fscore = 0.0
        infer_precision = 0.0
        infer_recall = 0.0
        infer_fscore = 0.0
        count = 0
        # Add precision and recall for each language
        for language in Language.objects:
            count += 1
            base_precision += Language.objects[language]._base_precision
            base_recall += Language.objects[language]._base_recall
            infer_precision += Language.objects[language]._infer_precision
            infer_recall += Language.objects[language]._infer_recall
        # Average precision and recall
        base_precision = base_precision / count
        base_recall = base_recall / count
        infer_precision = infer_precision / count
        infer_recall = infer_recall / count
        # Calculate post-average fscores
        base_fscore = 2 * (base_precision * base_recall /
                           (base_precision + base_recall))
        infer_fscore = 2 * (infer_precision * infer_recall /
                            (infer_precision + infer_recall))
        # Return values
        base = (base_precision, base_recall, base_fscore)
        infer = (infer_precision, infer_recall, infer_fscore)
        return base, infer

class Language:

    objects = {}

    def __init__(self, iso, path):
        self._iso = iso
        self._path = path
        self._gold_path = ''
        self._infer_path = ''
        self._infer_confusion_matrix = {}
        self._infer_precision = 0.0
        self._infer_recall = 0.0
        self._infer_fscore = 0.0
        self._base_confusion_matrix = {}
        self._base_precision = 0.0
        self._base_recall = 0.0
        self._base_fscore = 0.0
        Language.objects[iso] = self
        self.set_gold_path()
        self.set_infer_path()
        Choices(iso, 'gold', self._gold_path)
        Choices(iso, 'infer', self._infer_path)
        Baseline(iso)
        self.set_fscores()

    def clear_class_data():
        Language.objects = {}

    def set_gold_path(self):
        if self._path[-1] == '/':
            self._gold_path = (Evaluation.agg_path + self._path +
                               'choices.up')
        else:
            self._gold_path = (Evaluation.agg_path + '/' + self._path +
                               'choices.up')
        return True

    def set_infer_path(self):
        self._infer_path = (Evaluation.choices_path + str(self._iso) +
                            '.choices')
        return True
    
    def set_confusion_matrix(self):
        self._infer_confusion_matrix = {'FNeg': {}, 'FPos': {}, 'TPos': {}}
        self._base_confusion_matrix = {'FNeg': {}, 'FPos': {}, 'TPos': {}}
        self._process_confusion(target='tense')
        self._process_confusion(target='aspect')
        self._process_confusion(target='mood')
        #print(self._iso, self._infer_confusion_matrix)
        return True

    def _process_confusion(self, target='tense'):
        # Tense
        if target == 'tense':
            gold_values = Choices.objects[(self._iso, 'gold')]._tenses
            infer_values = Choices.objects[(self._iso, 'infer')]._tenses
            baseline_values = Baseline.objects[self._iso]._tenses
        # Aspect
        if target == 'aspect':
            gold_values = Choices.objects[(self._iso, 'gold')]._aspects
            infer_values = Choices.objects[(self._iso, 'infer')]._aspects
            baseline_values = Baseline.objects[self._iso]._aspects
        # Mood
        if target == 'mood':
            gold_values = Choices.objects[(self._iso, 'gold')]._moods
            infer_values = Choices.objects[(self._iso, 'infer')]._moods
            baseline_values = Baseline.objects[self._iso]._moods
        # Inference Comparisons
        infer_match = {}
        for i in gold_values:
            match = False
            for j in infer_values:
                for value in infer_values[j]:
                    if i.lower() == value.lower():
                        match = True
                        infer_match[j] = True
            if match == True:
                self._infer_confusion_matrix['TPos'][i] = True
            else:
                self._infer_confusion_matrix['FNeg'][i] = True
        for j in infer_values:
            if j not in infer_match:
                self._infer_confusion_matrix['FPos'][j] = True
        # Baseline Comparisons
        baseline_match = {}
        for i in gold_values:
            match = False
            for j in baseline_values:
                for value in baseline_values[j]:
                    if i.lower() == value.lower():
                        match = True
                        baseline_match[j] = True
            if match == True:
                self._base_confusion_matrix['TPos'][i] = True
            else:
                self._base_confusion_matrix['FNeg'][i] = True
        for j in baseline_values:
            if j not in baseline_match:
                self._base_confusion_matrix['FPos'][j] = True
        return True

    def set_precisions(self):
        if not self._infer_confusion_matrix or not self._base_confusion_matrix:
            self.set_confusion_matrix()
        try:
            self._infer_precision = (
                    float(len(self._infer_confusion_matrix['TPos'])) /
                          (len(self._infer_confusion_matrix['TPos']) +
                           len(self._infer_confusion_matrix['FPos'])))
        except:
           pass
        try:
            self._base_precision = (
                    float(len(self._base_confusion_matrix['TPos'])) /
                          (len(self._base_confusion_matrix['TPos']) +
                           len(self._base_confusion_matrix['FPos'])))
        except:
            pass
        return True

    def set_recalls(self):
        if not self._infer_confusion_matrix or not self._base_confusion_matrix:
            self.set_confusion_matrix()
        try:
            self._infer_recall = (
                    float(len(self._infer_confusion_matrix['TPos'])) /
                          (len(self._infer_confusion_matrix['TPos']) +
                           len(self._infer_confusion_matrix['FNeg'])))
        except:
            pass
        try:
            self._base_recall = (
                    float(len(self._base_confusion_matrix['TPos'])) /
                          (len(self._base_confusion_matrix['TPos']) +
                           len(self._base_confusion_matrix['FNeg'])))
        except:
            pass
        return True

    def set_fscores(self):
        if not self._infer_precision or not self._base_precision:
            self.set_precisions()
        if not self._infer_recall or not self._base_recall:
            self.set_recalls()
        try:
            self._infer_fscore = 2 * (self._infer_precision *
                                      self._infer_recall /
                                  (self._infer_precision + self._infer_recall))
        except:
            pass
        try:
            self._base_fscore = 2 * (self._base_precision * self._base_recall /
                                  (self._base_precision + self._base_recall))
        except:
            pass
        return True
                            
class Choices:

    objects = {}

    def __init__(self, iso, value, path):
        self._iso = iso
        self._value = value
        self._path = path
        self._choices_file = ''
        self._tenses = {}
        self._aspects = {}
        self._moods = {}
        Choices.objects[(iso, value)] = self
        self.load_file()
        self.set_tenses()
        self.set_aspects()
        self.set_moods()

    def load_file(self):
        self._choices_file = ChoicesReader.ChoicesFile(self._path)
        return True

    def set_tenses(self):
        for tense in self._choices_file.tenses():
            if isinstance(tense, list):
                tense = tense[0]
            try:
                self._tenses[tense] = map_gloss.Gloss.objects[(self._iso,
                                                    tense)].get_references()
            except:
                self._tenses[tense] = {tense: True}
        return True

    def set_aspects(self):
        for aspect in self._choices_file.aspects():
            if isinstance(aspect, list):
                aspect = aspect[0]
            try:
                self._aspects[aspect] = map_gloss.Gloss.objects[(self._iso,
                                                    aspect)].get_references()
            except:
                self._aspects[aspect] = {aspect: True}
        return True

    def set_moods(self):
        for mood in self._choices_file.moods():
            if isinstance(mood, list):
                mood = mood[0]
            try:
                self._moods[mood] = map_gloss.Gloss.objects[(self._iso,
                                                    mood)].get_references()
            except:
                self._moods[mood] = {mood: True}
        return True

class Baseline:

    objects = {}
    
    def __init__(self, iso):
        self._iso = iso
        self._tenses = {}
        self._aspects = {}
        self._moods = {}
        Baseline.objects[iso] = self
        self.set_tenses()
        self.set_aspects()
        self.set_moods()

    def set_tenses(self):
        tenses = ['Future', 'Past', 'Present']
        for tense in tenses:
            try:
                self._tenses[tense] = map_gloss.Gloss.objects[(self._iso,
                                                    tense)].get_references()
            except:
                self._tenses[tense] = {tense: True}
        return True

    def set_aspects(self):
        aspects = ['Imperfect', 'Perfect']
        for aspect in aspects:
            try:
                self._aspects[aspect] = map_gloss.Gloss.objects[(self._iso,
                                                    aspect)].get_references()
            except:
                self._aspects[aspect] = {aspect: True}
        return True

    def set_moods(self):
        moods = []
        for mood in moods:
            try:
                self._moods[mood] = map_gloss.Gloss.objects[(self._iso,
                                                    mood)].get_references()
            except:
                self._moods[mood] = {mood: True}
        return True

