#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
extract-lexicon.py [-options] Xigt-input [Xigt-input ...]

This program attempts to extract lexical entries from Interlinear Glossed
Text (IGT) that has been enriched and put into Xigt format (more information
at https://github.com/goodmami/xigt/wiki).

With the proper command line options, it attempts to identify the cases on
nouns, and is able to group lexical entries by case. The lexicon is output in
the standard `choices' file format for usage in the LinGO Grammar Matrix
customization system (http://www.delph-in.net/matrix/customize/matrix.cgi).

Input(s):
Xigt-input - a Xigt-format XML file that contains enriched IGT text
    (If there are multiple input files, they should be the same langauge!)
-options - options for defining case grams, etc.

Output:
choices file parameters for the lexicon (sent to STDOUT)
   (Note that any errors or information are only sent to STDERR)

May 2015, Jeff Heath
  Change to Xigt input, add code to remove case grams from output, add documentation
June 2015, Jeff Heath
  Fix predications from e.g. _n_cat_rel to _cat_n_rel
"""

import sys
import getopt
from xigt.codecs import xigtxml
import re


# parse options and arguments
# see if there's an option for dealing with case values on nouns
try:
  opts, args = getopt.getopt(sys.argv[1:], 'c:s:a:o:N:h')
except getopt.GetoptError as err:
  print(str(err),file=sys.stderr)
  exit(1)

find_case_method = ''
not_marked = ''
sao_case_names = {'s':'', 'a':'', 'o':''}
for opt, arg in opts:
  if opt == '-h':
    print('Usage: extract-lexicon.py [options] Xigt-input [Xigt-input ...]')
    print('  -c grams : automatically detect case from standard grams')
    print('  -c sao   : use case values defined by -s -a -o')
    print('  -s GRAM  : use given gram as subject marker')
    print('  -a GRAM  : use given gram as agent marker')
    print('  -o GRAM  : use given gram as object marker')
    print('  -N GRAM  : unmarked nouns get assigned this case')
    print('  -h       : display this help')
    sys.exit()
  elif opt == '-c':
    if arg in ['grams', 'sao']:
      find_case_method = arg
    else:
      print('Error: Invalid case value extraction method: {} {}.'.format(opt, arg), file=sys.stderr)
      print('       Choose among the following methods: "grams", "sao".', file=sys.stderr)
      sys.exit(1)
  elif opt in ['-s','-a','-o']:
    if arg[0] == "-":
      # it can 'take over' the following option if you don't provide an argument
      print('Error: You must provide an argument for the {} case name.'.format(opt), file=sys.stderr)
      print('       To treat {} case as unmarked, use {} 0.'.format(opt, opt), file=sys.stderr)
      sys.exit(1)
    elif arg == '0':
      sao_case_names[opt[-1]]=''
    else:
      sao_case_names[opt[-1]]=arg
  elif opt == '-N':
    # nouns that aren't marked receive this case name
    not_marked = arg


# an a priori collection of case grams, used in 'grams' method
all_case_grams = ['nom', 'acc', 'abs', 'erg', 'abl', 'all', 
                  'com', 'dat', 'gen', 'ins', 'loc', 'obl']

# if find_case_method is sao, then we only look for specific case
# grams input by the user, all other nouns go in 'othercase'
if find_case_method=='sao':
  all_case_grams = []
  for k, v in sao_case_names.items():
    # only add to list if non-empty
    if v:
      all_case_grams.append(v) 

# storage locations for data we find
nouns={}
verbs={}
dets={}

# treat each argument on the command line as a Xigt file to process
for filename in args:
  # process the Xigt file into a XigtCorpus class so that we can work with it
  corpus = xigtxml.load(open(filename, encoding='utf-8'))
  
  # process all of the IGTs in the corpus
  for igt in corpus.igts:
    # get the tier with the enriched POS tags
    wPOStier = igt.get('w-pos')
    if wPOStier:
      # there is a tier with enriched POS tags, so process the POS items
      for wPOS in wPOStier:
        this_POS = wPOS.get_content()
        word = igt.get_item(wPOS.alignment)
        #orth = word.get_content()
        morpheme_labels = word.referrers()['segmentation']
        morphemes = [igt.get_item(label) for label in morpheme_labels]
        morpheme_grams = [morph.get_content() for morph in morphemes]
        # ??? can there be more than one gloss in the alignment? If so, append all
        gloss_labels = [morph.referrers()['alignment'][0] for morph in morphemes]
        glosses = [igt.get_item(label) for label in gloss_labels]
        gloss_grams = [gloss.get_content().lower() for gloss in glosses]
        
        this_case = not_marked
        if this_POS == 'NN' and find_case_method != '':
          # we found a noun and we are supposed to find case values
          for i, g in enumerate(gloss_grams):
            if g.lower() in all_case_grams:
              this_case = g.lower()
              # remove this case morpheme
              # DISABLED for now
              #del gloss_grams[i]
              #del morpheme_grams[i]
              break
        
        # build orthogrphic form and lemma to store
        orth = '-'.join(morpheme_grams)
        lemma = '-'.join(gloss_grams)

        # got all the info, store it 
        if this_POS == 'NN':
          if not this_case in nouns.keys():
            nouns[this_case]={orth:{lemma:1}}
          elif not orth in nouns[this_case].keys():
            nouns[this_case][orth]={lemma:1}
          else:
            nouns[this_case][orth][lemma] = nouns[this_case][orth].get(lemma, 0) + 1

        elif this_POS[0] == 'V':
          if not orth in verbs.keys():
            verbs[orth]={lemma:1}
          else:
            verbs[orth][lemma] = verbs[orth].get(lemma, 0) + 1

        else: 
          if not orth in dets.keys():
            dets[orth]={lemma:1}
          else:
            dets[orth][lemma] = dets[orth].get(lemma, 0) + 1


# format and print the lexical choices

# nouns
n=0
for case in nouns.keys():
  n+=1
  if find_case_method != '': 
    print("noun{}_feat1_name=case".format(n))
    if case =='':
      print("noun{}_feat1_value=othercase".format(n))
    else:
      print("noun{}_feat1_value={}".format(n,case))
  m=0
  print("noun{}_det=opt".format(n))
  for orth in nouns[case].keys():
    for lemma in nouns[case][orth].keys():
      m+=1
      print("  noun{}_stem{}_orth={}".format(n,m,orth))
      print("  noun{}_stem{}_pred=_{}_n_rel".format(n,m,lemma))

# verbs
# currently outputting all verbs as transitive
n=0
print("verb1_valence=trans")
for orth in verbs.keys():
  for lemma in verbs[orth].keys():
    n+=1
    print("  verb1_stem{}_orth={}".format(n,orth))
    print("  verb1_stem{}_pred=_{}_v_rel".format(n,lemma))

# determiners
n=0
for orth in dets.keys():
  for lemma in dets[orth].keys():
    n+=1
    print("det1_stem{}_orth={}".format(n,orth))
    print("det1_stem{}_pred=_{}_q_rel".format(n,lemma))
