#!/usr/bin/env python3
#
# python class for basic igt collection found in
# igt.all files
# accesses igt by item id, or in sequence
# items support get_meta(), get_id(), get_text, get_gloss(), get_trans()
# get_transt(), get_projt(), get_transtgm(), get_projtgm()
import re
import sys

class IgtAllFile:
    """class for parsing igt.all files into a nice object"""
    iafile = ''
    igts = []
    igt_ids = {}
    def __init__(self, iafile):
        self.iafile = iafile
        self.read_iafile()
        self.build_ids()

    def print_all_igts(self):
        for i in self.igts:
            i.print_self()

    def read_iafile(self):
        """read igt.all file into data structures"""
        self.igts.extend(Igt10(blk) for blk in \
            self.igt_blocks(open(self.iafile,'r')))

    def igt_blocks(self, igt_lines):
        block = []
        for line in igt_lines:
            line = line.strip()
            if line == '':
                if len(block) > 0: # or maybe "if len(block) == 8"?
                    yield block
                # this resets block to [] at any blank line, not only
                # after a successful block
                block = []
            else:
                block += [line]
        yield block



    def build_ids(self):
        """ the meta line of the igt is expected to have an
        id number generated by fei's code, we use that here
        to build a hash to allow access to igts by their ids"""

        ID_RE = re.compile('^Igt_id=(\d+)')
        for igt in self.igts:
            idstr = ''
            m = ID_RE.match(igt.get('meta'))
            if m:
                idstr = m.group(0)
#                print(idstr)
                self.igt_ids[idstr] = igt

class Igt10:
    """ class for individiual igt10s, the following properties are
    made accessible"""

    prop_keys = ['meta','text','gloss','trans', 'sg_align', 'gt_align','transt',
                 'projt', 'transtgm', 'projtgm']
    def __init__(self, propslist):
        if len(propslist) < 10:
            # allow to proceed with a warning
            print("Warning:\n"+\
                  "  IGT example with less than 10 lines found,\n"+\
                  "  the last "+str(10-len(propslist))+" properties\n"+\
                  "  will not appear on this item.", file=sys.stderr)
            for prop in propslist:
                print(prop, file=sys.stderr)
        elif len(propslist) > 10:
            print("Warning:\n"+\
                  "  IGT example with more than 10 lines found,\n"+\
                  "  discarding these extra lines:", file=sys.stderr)
            for i in range(10,len(propslist)):
                print("  "+propslist[i], file=sys.stderr)


        for i in range(0,len(propslist)):
          if propslist[i] == "None": propslist[i]=None

        self.props = dict(zip(self.prop_keys, propslist))
        self.build_align_dicts()

    def build_align_dicts(self):
        """ need to make dictionaries that map between source
            and gloss and between gloss and trans """

        # underlying dicts
        self.sg_dict = {}
        self.gs_dict = {}
        self.gt_dict = {}
        self.tg_dict = {}

        for i in self.props['sg_align'].split('\t'):
            [s, g] = i.split('#')[0].strip().split()
            self.sg_dict[s] = g
            self.gs_dict[g] = s

        for i in self.props['gt_align'].split('\t'):
            [g, t] = i.split('#')[0].strip().split()
            self.gt_dict[g] = t
            self.tg_dict[t] = g

    def get_src_index_from_trans_index(self, index):
        '''Uses 1 based indexing rather than zero, zero represents null mapping'''
        i = ''
        try:
            i = self.tg_dict[index]
        except KeyError:
            ## possible our index is inside a list key rather than a simple one
            for k in self.tg_dict:
                if k.find(',') != -1:
                    l = k.split(',')
                    if index in l:
                        i = self.tg_dict[k]
            if i == '':
                errstr = "-1: key "+index+" not found in tg_dict:\n"
                errstr += str(self.tg_dict)+'\n'
                return errstr

        try:
            return self.gs_dict[i]
        except KeyError:
            errstr = "-2: key "+index+" not found in sg_dict:\n"
            errstr += str(self.sg_dict)+'\n'
            return errstr

    def get_trans_index_from_src_index(self, index):
        '''Uses 1 based indexing rather than zero, zero represents null mapping'''
        try:
            i = self.sg_dict[index]
        except KeyError:
            errstr = '-1: key '+index+' not found in sg_dict:\n'
            errstr += str(self.sg_dict)+'\n'
            return errstr

        try:
            return self.gt_dict[i]
        except KeyError:
            errstr = '-1: key '+index+' not found in gt_dict:\n'
            errstr += str(self.gt_dict)+'\n'
            return errstr

    def get_trans_from_text(self,form):
        wordtoks = self.props['text'].split()
        if form not in wordtoks:
            print(form+" not found in "+self.props['text'])
            return None
        else:
            i = self.get_trans_index_from_src_index(str(wordtoks.index(form)+1))
            if i.startswith('-1'):
                i+="\nno mapping from text word ["+form+"] into gloss line\n"
                return i
            elif i.startswith('-2'):
                i+="\nno mapping from gloss line into text line for given form"
                return i
            else:
                try:
                    # this is a simple word form, we can use the
                    # string like an index
                    return self.props['trans'].split()[int(i)-1]
                except:
                    # we've mapped to a subword token or a multiword token,
                    # split it up and return the relevent portion
                    # or get the series of multiwords and return that
                    if i.find('.') != -1:
                      [wi,mi] = i.split('.')
                      w = self.props['trans'].split()[int(wi)-1]
                      m = w.split('-')[int(mi)-1]
                      return m
                    elif i.find(',') != -1:
                      wilist = i.split(',')
                      out = []
                      for i in [ int(x)-1 for x in wilist ]:
                        try:
                          out.append(self.props['trans'].split()[i-1])
                        except IndexError as e:
                          # big failure:
                          # at least print some info
                          print("get_trans_from_text() has failed with index error", file=sys.stderr)
                          print(self.props['trans'].split(), file=sys.stderr)
                          print(str(i), file=sys.stderr)
                          print("item meta",self.props['meta'],file=sys.stderr)
                          sys.exit(1)
                      return ",".join(out)

    def get_text_from_trans(self, form):
        wordtoks = self.props['trans'].split()
        if form not in wordtoks:
            print(form+" not found in "+self.props['trans'])
            return None
        else:
            i = self.get_src_index_from_trans_index(str(wordtoks.index(form)+1))
            if i.startswith('-1'):
                i+="no mapping from trans word ["+form+"] into gloss line\n"
                i+="transtoks list: "+str(wordtoks)+'\n'
                i+="transtoks.index(form)+1: "+str(wordtoks.index(form)+1)+'\n'
                return i
            elif i.startswith('-2'):
                i+="\nno mapping from gloss line into text line for given form"
                return i
            else:
                try:
                    # this is a simple word form, we can use the
                    # string like an index
                    return self.props['text'].split()[int(i)-1]
                except:
                    # we've mapped to a subword token or a multiword token,
                    # split it up and return the relevent portion
                    # or get the series of multiwords and return that
                    if i.find('.') != -1:
                      [wi,mi] = i.split('.')
                      w = self.props['text'].split()[int(wi)-1]
                      m = w.split('-')[int(mi)-1]
                      return m
                    elif i.find(',') != -1:
                      wilist = i.split(',')
                      out = []
                      for i in [ int(x)-1 for x in wilist ]:
                        out.append(self.props['text'].split()[i])
                      return ",".join(out)


    def get(self, arg):
        if not arg in self.prop_keys:
            print(arg+" is not a key supported by igt10")
            print(" use one of the following")
            for k in self.prop_keys:
                print(k)
        else:
            return self.props[arg]

    def print_self(self, err=False, errf=None):
        for k in self.prop_keys:
            if errf !=None:
                errf.write(self.get(k)+"\n")
            elif err:
                print(self.get(k), file=sys.stderr)
            else:
                print(self.get(k))

leaf_re = re.compile(r'([^\s)]+\))')
def gramtree(igt):
    """Return the projtgm line with the leaves replaced by gloss tokens."""
    try: leaves = leaf_re.findall(igt.get('projtgm'))
    except: return
    pairs = zip(leaves, igt.get('gloss').split())
    tree = str(igt.get('projtgm'))
    for leaf, gloss in pairs:
        tree = tree.replace(leaf, gloss + ')', 1)
    return tree

# if the module is called directly, offer some convenient functionality to
# pull out specific fields
if __name__ == '__main__':
    import sys
    if len(sys.argv) <= 2:
        print('usage: igtall.py INFILE FIELD [FIELD...]\n'
              '  INFILE  : the IgtAll file to load\n'
              '  FIELDs  : meta, text, gloss, trans, sg_align, gt_align,\n'
              '            gt_align, transt, projt, transtgm, projtgm')
        sys.exit()
    ia = IgtAllFile(sys.argv[1])
    fields = sys.argv[2:]
    for igt in ia.igts:
        try:
            igtfields = [gramtree(igt) if field == 'gramtree'
                         else igt.get(field)
                         for field in fields]
            print('\n'.join(igtfields))
        except KeyError as e:
            print('KeyError:', str(e), file=sys.stderr)
            continue
        except TypeError as e:
            print('TypeError:', str(e), file=sys.stderr)
            continue
