#!/opt/python-3.4/bin/python3.4
# -*- coding: utf-8 -*-

import sys
import math
import operator
import optparse
import statistics
from xigt.codecs import xigtxml
from InferenceIGT import InferenceIGT

# AGGREGATION Project - Sentential Negation
#
# Anthony Princing
# CLMS Master's Project
# Winter, Spring 2016

negation_grams = ['neg', 'not', 'neg1', 'neg2']
aux_grams = ['aux']
fxn_grams = ['antip', 'appl', 'ben', 'caus', 'compl', 'decl', 'dur', 'fut', 'imp', 'ind', 'inf', 'intr', 'impv', 'impf', 'irr',
       'pass', 'pfv', 'perf', 'prf', 'prs', 'pres', 'prog', 'proh', 'pst', 'past', 'ptcp', 'purp', 'sbjv', 'tr',
       'hearsay', 'personal', 'resultative', 'continuative', 'able', 'abil']
mod_grams = ['can', 'may', 'might', 'should', 'would', 'could', 'do', 'does', 'did', 'have', 'has', 'had']
cop_grams = ['cop', 'be']
mod_tags = ['adv']
comp_tags = ['prt']
POS_to_free_type = {'ADV' : 'mod', 'PRT' : 'comp'}


def similarity(v1, v2):
   """Cosine similarity function.

   Args:
      v1, v2	-- feature = value pairs (dicts)

   Returns:
      float	-- Similarity measure between 0 and 1 with 1 being exact match
   """
   v_sim = 0
   if v1 and v2:
      dotprod = 0
      for i, element in enumerate(v1):
         if v1[i] and v2[i]:
            dotprod += v1[i]*v2[i]
      sigma = 0
      for e1 in v1:
         if e1:
            sigma += e1**2
      mag1 = math.sqrt(sigma)
      sigma = 0
      for e2 in v2:
         if e2:
            sigma += e2**2
      mag2 = math.sqrt(sigma)
      if mag1 and mag2:
         v_sim = float(dotprod/mag1*mag2)
   return v_sim
# end similarity


class NegationIGT(InferenceIGT):

   """ Class to collect negation related information about a single IGT.
   """

   def __init__(self, NEG_IGT):
      """ Initialize NegationIGT instance by setting all negation related values
      for a single IGT instance.

      Args:
         NEG_IGT	-- An IGT (in Xigt format) that exhibits negation
      """
      InferenceIGT.__init__(self, NEG_IGT)

      self.negators = {}
      self.negation_gram_count = self.gram_in_gloss_count(negation_grams)
      if DEBUG: print("Negation gram count = %d" % self.negation_gram_count)

      if self.gloss_tier_exists:
         self.neg_g_ids = self.gram_in_tier(negation_grams, self.gloss_tier)
      else:
         self.neg_g_ids = None

      if self.wgloss_tier_exists:
         if len(self.neg_g_ids) > 0:
            self.neg_gw_ids = []
            for g_id in self.neg_g_ids:
               self.neg_gw_ids.append(self.igt.get_item(g_id).content.split('[')[0])
         else:
            self.neg_gw_ids = self.gram_in_tier(negation_grams, self.wgloss_tier)
         for i, neg in enumerate(self.neg_gw_ids):
            if DEBUG: print(repr(self.get_adjacent_pos_tags(neg)))
            if len(self.neg_g_ids) > 0:
               g_id = self.neg_g_ids[i]
            else:
               g_id = None
            w_loc = self.get_w_id(neg)
            m_loc = self.get_m_id(g_id)
            morph_type = self.get_neg_morph_type(neg, g_id)
            POS_tag = None
            free_type = None
            order = None
            modifies = None
            if morph_type == 'bound':
               ortho = self.get_bound_orthography(g_id)
            else:
               ortho = self.get_free_orthography(neg)
               POS_tag = self.get_pos_tag(neg)
               free_type = self.get_neg_free_type(neg, g_id)
               if free_type in ('mod', 'comp'):
                  order = self.get_verb_relative_loc(neg)
                  if free_type == 'mod':
                     modifies = self.get_word_modifies(w_loc)
            self.negators['neg'+str(i+1)] = {'gwid'       : neg,
                                             'm_loc'      : m_loc,
                                             'w_loc'      : w_loc,
                                             'POS_tag'    : POS_tag,
                                             'morph_type' : morph_type,
                                             'free_type'  : free_type,
                                             'order'      : order,
                                             'modifies'   : modifies,
                                             'ortho'      : ortho
                                            }
      else:
         self.neg_gw_ids = None


#  end __init__


   def is_neg_aux(self, neg, g_id):
      """ Determine if a negation gram represents negation by auxiliary.

      Args:
         neg	-- A gw_id for an item that was glossed for negation
         g_id	-- The corresponding g_id (gloss tier)

      Returns:
        bool	-- True if auxiliary, False otherwise
      """
      neg_aux = False
      aux_g_ids = self.gram_in_tier(aux_grams + cop_grams, self.gloss_tier)
      if len(aux_g_ids) > 0:
         for aux_gid in aux_g_ids:
            aux_mloc = self.igt.get_item(aux_gid).alignment
            if DEBUG: print("aux_mloc: " + str(aux_mloc))
            if aux_mloc:
               if aux_mloc == self.get_m_id(g_id):
                  if self.igt.get_item(aux_gid).value in cop_grams:
                     if is_verb_adjacent(neg):
                        neg_aux = True
                  else:
                     neg_aux = True
      if not neg_aux:
         gloss_set = set(self.all_glosses_in_word(neg)).difference(set(negation_grams))
         if DEBUG: print(repr(gloss_set))
         fxn_mod_set = set(fxn_grams+mod_grams)
         if DEBUG: print(repr(gloss_set.intersection(fxn_mod_set)))
         if len(gloss_set) > 0:
            if gloss_set == gloss_set.intersection(fxn_mod_set):
               neg_aux = True
      return neg_aux
#end is_neg_aux


   def get_neg_morph_type(self, neg, g_id):
      """ Determine the negation morpheme type: bound or free.

      Args:
         neg	-- A gw_id for an item that was glossed for negation
         g_id	-- The corresponding g_id (gloss tier)

      Returns:
         string -- The values of 'bound' or 'free' as discovered
      """
      if self.is_neg_aux(neg, g_id):
         return 'free'
      elif self.is_affixed(self.wgloss_tier[neg]):
         return 'bound'
      else:
         return 'free'
#  end get_neg_morph_type


   def get_neg_free_type(self, neg, g_id):
      """ Determine the negation free morpheme type: head, mod or comp

      Args:
         neg	-- A gw_id for an item that was glossed for negation
         g_id	-- The corresponding g_id (gloss tier)

      Returns:
         string -- The values of 'head', 'mod' or 'comp' as discovered
      """
      neg_POS = self.get_pos_tag(neg)
      if self.is_neg_aux(neg, g_id):
         if DEBUG: print(repr(self.is_neg_aux(neg, g_id)))
         return 'head'
      elif neg_POS in POS_to_free_type:
         return POS_to_free_type[neg_POS]
      else:
         return None
#  end get_neg_free_type

#end class NegationIGT



class SententialNegation(object):
   """ Class to collect and tally the negation related data about a single language.
   """

   def __init__(self):
      self.exponence = 0
      self.exponence_tally = {}
      self.tallies = {}
      self.ortho_tallies = {}
      self.negationIGTs = {}
      self.negationIGT_count = 0
#  end __init__

   def addIGT(self, negIGT):
      """ Add an IGT that exhibits negation to the list to be tallied.

      Args:
         negIGT	-- An IGT that exhibits negation
      
      Returns:
         bool	-- Success or failure of addition
      """
#      if negIGT.is_question == True or negIGT.has_conjunction == True:
#         return False
      if len(negIGT.negators) == 0:
         return False
      self.negationIGTs[negIGT.igtid] = negIGT
      self.negationIGT_count += 1
      if negIGT.negation_gram_count in self.exponence_tally:
         self.exponence_tally[negIGT.negation_gram_count] += 1
      else:
         self.exponence_tally[negIGT.negation_gram_count] = 1

      return True
#  end addIGT

   @property
   def neg_exp(self):
      """ Determine the negation exponence by finding max exponence tally.
      """
      exp = None
      if DEBUG: print(repr(self.exponence_tally))
      (exp_count, exp) = max((self.exponence_tally[keyval], keyval) for keyval in self.exponence_tally)
      self.exponence = exp
      return exp
#  end neg_exp


   @property
   def is_Jespersen(self):
      """ Determine if the language could be in Jespersen cycle.
      """
      J_val = False
      if len(self.exponence_tally) > 1:
         exp_list = [(k, t) for k, t in self.exponence_tally.items()]
         exp_list.sort(key=operator.itemgetter(1), reverse=True)
         if (exp_list[0][0] in [1, 2]) and (exp_list[1][0] in [1, 2]):
            if not self.ignore_min(exp_list[0][1], exp_list[1][1]):
               J_val = True
      return J_val
# end is_Jespersen


   def init_tallies(self, tally_o, neg_key):
      """ Initialize a negation related values structure to record tallies
      for a single negation item.

      Args:
         tally_o	-- The tally structure to be initialized
         neg_key	-- The negation item for which tallies will be recorded

      """
      tally_o[neg_key] = {'count'  : 0,
                          'bound'  : 0,
                          'free'   : 0,
                          'head'   : 0,
                          'comp'   : 0,
                          'mod'    : 0,
                          'v'      : 0,
                          'vp'     : 0,
                          's'      : 0,
                          'after'  : 0,
                          'before' : 0 }
# end init_tallies


   def position_tally(self, negators):
      """ Update tallies of negation related values for all postional negation items for a single IGT.

      Args:
         negators	-- A dict containing negation values of all positional negation items for a single IGT
      """
      for neg in negators:
         if neg not in self.tallies:
            self.init_tallies(self.tallies, neg)
         self.tallies[neg]['count'] += 1
         if negators[neg]['morph_type']:
            if negators[neg]['morph_type'] == 'bound':
               self.tallies[neg][negators[neg]['morph_type']] += 1
            elif negators[neg]['free_type']:
               self.tallies[neg][negators[neg]['morph_type']] += 1
         if negators[neg]['free_type']:
            self.tallies[neg][negators[neg]['free_type']] += 1
         if negators[neg]['modifies']:
            self.tallies[neg][negators[neg]['modifies']] += 1
         if negators[neg]['order']:
            self.tallies[neg][negators[neg]['order']] += 1
# end position_tallies


   def ortho_tally(self, negators):
      """ Update tallies of negation related values for all orthography negation items for a single IGT.

      Args:
         negators	-- A dict containing negation values of all orthography negation items for a single IGT
      """
      for neg in negators:
         ortho_key = negators[neg]['ortho']
         if ortho_key:
            if ortho_key not in self.ortho_tallies:
               self.init_tallies(self.ortho_tallies, ortho_key)
            self.ortho_tallies[ortho_key]['count'] += 1
            if negators[neg]['morph_type']:
               self.ortho_tallies[ortho_key][negators[neg]['morph_type']] += 1
            if negators[neg]['free_type']:
               self.ortho_tallies[ortho_key][negators[neg]['free_type']] += 1
            if negators[neg]['modifies']:
               self.ortho_tallies[ortho_key][negators[neg]['modifies']] += 1
            if negators[neg]['order']:
               self.ortho_tallies[ortho_key][negators[neg]['order']] += 1
# end ortho_tallies


   def ignore_min(self, num1, num2, DEVMUL=2):
      """ Determine if the smaller of two values is significant or noise.

      Args:
         num1, num2	-- integer values
         DEVMUL		-- multiplier

      Returns:
         bool -- True if min values in not significant, True otherwise
      """
      if DEVMUL*min([num1, num2]) < statistics.stdev([num1, num2]):
         return True
      else:
         return False
# end ignore_min

   def get_morph_type(self, neg_key):
      """ Examine frequency tallies to determine negation morphological type, bound or free for language.

      Args:
         neg	-- identifier of negation item

      Returns:
         string -- 'bound', 'free', or 'variable'
      """
      if neg_key in self.tallies:
         bound_tally = self.tallies[neg_key]['bound']
         free_tally = self.tallies[neg_key]['free']
      elif neg_key in self.ortho_tallies:
         bound_tally = self.ortho_tallies[neg_key]['bound']
         free_tally = self.ortho_tallies[neg_key]['free']
      else:
         return None
      if (bound_tally > 0) and (free_tally > 0) and (not self.ignore_min(bound_tally, free_tally)):
         return 'variable'
      elif bound_tally > free_tally:
         return 'bound'
      elif free_tally > bound_tally:
         return 'free'
      else:
         return None
# end get_morph_type


   def get_free_type(self, neg_key):
      """ Examine frequency tallies to determine negation free morphological type, head, mod or comp for language.

      Args:
         neg	-- identifier of negation item

      Returns:
         string	-- 'head', 'mod', or 'comp'
      """
      f_type = None
      if neg_key in self.tallies:
         (ft_count, f_type) = max((self.tallies[neg_key][keyval], keyval) for keyval in self.tallies[neg_key] if keyval in ['head', 'comp', 'mod'])
      elif neg_key in self.ortho_tallies:
         (ft_count, f_type) = max((self.ortho_tallies[neg_key][keyval], keyval) for keyval in self.ortho_tallies[neg_key] if keyval in ['head', 'comp', 'mod'])
      if ft_count > 0:
         return f_type
      else:
         return None
# end get_free_type


   def get_modifies(self, neg_key):
      """ Examine frequency tallies to determine what negation modifier modifies for the language.

      Args:
         neg	-- identifier of negation item

      Returns:
         string	-- 'v', 'vp', or 's'
      """
      m_type = None
      if neg_key in self.tallies:
         (m_count, m_type) = max((self.tallies[neg_key][keyval], keyval) for keyval in self.tallies[neg_key] if keyval in ['v', 'vp', 's'])
      elif neg_key in self.ortho_tallies:
         (m_count, m_type) = max((self.ortho_tallies[neg_key][keyval], keyval) for keyval in self.ortho_tallies[neg_key] if keyval in ['v', 'vp', 's'])
      return m_type
# end get_modifies


   def get_order(self, neg_key):
      """ Examine frequency tallies to determine verb relative order of negation modifier or complement for the language.

      Args:
         neg	-- identifier of negation item

      Returns:
         string	-- 'before', 'after' or 'either'
      """
      if neg_key in self.tallies:
         before_tally = self.tallies[neg_key]['before']
         after_tally = self.tallies[neg_key]['after']
      elif neg_key in self.ortho_tallies:
         before_tally = self.ortho_tallies[neg_key]['before']
         after_tally = self.ortho_tallies[neg_key]['after']
      else:
         return None
      if (before_tally > 0) and (after_tally > 0) and (not self.ignore_min(before_tally, after_tally, 0.33)):
         return 'either'
      elif before_tally > after_tally:
         return 'before'
      elif after_tally > before_tally:
         return 'after'
      else:
         return None
# end get_order


   def get_matching_orthography(self, neg_key):
      """ For a positional negation item, find its matching orthography.

      Args:
         neg	-- identifier of negation item

      Returns:
         string	-- the best matching orthography item
      """
      if neg_key in self.tallies:
         orth_list = []
         neg_vec = (self.tallies[neg_key]['bound'], self.tallies[neg_key]['free'],\
                    self.tallies[neg_key]['before'], self.tallies[neg_key]['after'])
         for orth_key in self.ortho_tallies:
            orth_vec = (self.ortho_tallies[orth_key]['bound'], self.ortho_tallies[orth_key]['free'],\
                        self.ortho_tallies[orth_key]['before'], self.ortho_tallies[orth_key]['after'])
            orth_list.append((orth_key, similarity(neg_vec, orth_vec)))
         orth_list.sort(key=operator.itemgetter(1), reverse=True)
         if len(orth_list) > 0:
            return orth_list[0][0]
         else:
            return None
      else:
         return None
# end get_matching_orthography

#end class SententialNegation


def output_choices(choices, SN, lang_exponence, Jespersen):

   """ Use the tally results and write the appropriate choces to the output file specified.

   Args:
      choices		-- the output file descriptor
      SN		-- the SententialNegation class instance used for tallies
      lang_exponence	-- the exponence value calculated
      Jespersen		-- the Jespersen setting calculated
   """
   if Jespersen: choices.write("# Possible Jespersen Cycle!\n")
   choices.write("section=sentential-negation\n")
   if lang_exponence > 3:
      choices.write("neg-exp=more\n")
   else:
      choices.write("neg-exp=%d\n" % lang_exponence)
   if lang_exponence == 0:
      choices.write("vpc-0-neg=create\n")
   elif lang_exponence == 1:
      lang_simple_morph_type = SN.get_morph_type('neg1')
      if (lang_simple_morph_type == 'bound') or (lang_simple_morph_type == 'variable'):
         choices.write("infl-neg=on\n")
      if (lang_simple_morph_type == 'free') or (lang_simple_morph_type == 'variable'):
         lang_simple_free_type = SN.get_free_type('neg1')
         if lang_simple_free_type == 'head':
            choices.write("neg-aux=on\n")
            choices.write("neg-aux-index=1\n")
            choices.write("\nsection=word-order\n")
            choices.write("has-aux=yes\n")
            choices.write("\nsection=lexicon\n")
            choices.write("  aux1_name=neg\n")
            choices.write("  aux1_sem=add-pred\n")
            choices.write("    aux1_stem1_pred=neg_rel\n")

         else:
            order1 = SN.get_order('neg1')
            ortho1 = SN.get_matching_orthography('neg1')
            if lang_simple_free_type == 'mod':
               modifies1 = SN.get_modifies('neg1')
               choices.write("adv-neg=on\n")
               if modifies1:
                  choices.write("neg-mod=%s\n" % modifies1)
               if order1:
                  choices.write("neg-order=%s\n" % order1)
               if ortho1:
                  choices.write("neg-adv-orth=%s\n" % ortho1)
            elif lang_simple_free_type == 'comp':
               choices.write("comp-neg=on\n")
               if order1:
                  choices.write("comp-neg-order=%s\n" % order1)
               if ortho1:
                  choices.write("comp-neg-orth=%s\n" % ortho1)
            else:
               pass
   #         
   elif lang_exponence == 2:
      orth1 = SN.get_matching_orthography('neg1')
      orth2 = SN.get_matching_orthography('neg2')
      morph1 = SN.get_morph_type(orth1)
      morph2 = SN.get_morph_type(orth2)
      free1 = None
      free2 = None
      order1 = None
      order2 = None
      modifies1 = None
      modifies2 = None
   #   selected1 = None
   #   selected2 = None
      if morph1 != 'bound':
         free1 = SN.get_free_type(orth1)
         if free1 != 'head':
            order1 = SN.get_order(orth1)
            if free1 == 'mod':
               modifies1 = SN.get_modifies(orth1)
      if morph2 != 'bound':
         free2 = SN.get_free_type(orth2)
         if free2 != 'head':
            order2 = SN.get_order(orth2)
            if free2 == 'mod':
               modifies2 = SN.get_modifies(orth2)

      if (morph1 == 'bound') and (morph2 == 'bound'): 
         bi_type = (morph1, morph2)
      elif morph1 == 'bound':
         bi_type = (morph1, free2)
      elif morph2 == 'bound':
         bi_type = (free1, morph2)
      else:
         bi_type = (free1, free2)

      if bi_type in [('bound', 'bound'), ('bound', 'head'), ('head', 'bound'), ('head', 'head')]:
         choices.write("infl-neg=on\n")
         choices.write("bineg-type=infl-infl\n")
         choices.write("neg1-type=b\n")
         choices.write("neg2-type=b\n")

      elif bi_type in [('bound', 'mod'), ('mod', 'bound'), ('head', 'mod'), ('mod', 'head')]:
         choices.write("infl-neg=on\n")
         choices.write("adv-neg=on\n")
         choices.write("bineg-type=infl-mod\n")
         if bi_type in [('bound', 'mod'), ('head', 'mod')]:
            choices.write("neg1-type=b\n")
            choices.write("neg2-type=fm\n")
            if modifies2:
               choices.write("neg-mod-infl-mod-neg=%s\n" % modifies2)
            if order2:
               choices.write("neg-mod-order-infl-mod-neg=%s\n" % order2)
            if orth2:
               choices.write("neg-mod-orth=%s\n" % orth2)
         else:
            choices.write("neg1-type=fm\n")
            choices.write("neg2-type=b\n")
            if modifies1:
               choices.write("neg-mod-infl-mod-neg=%s\n" % modifies1)
            if order1:
               choices.write("neg-mod-order-infl-mod-neg=%s\n" % order1)
            if orth1:
               choices.write("neg-mod-orth=%s\n" % orth1)

      elif bi_type in [('bound', 'comp'), ('comp', 'bound'), ('head', 'comp'), ('comp', 'head')]:
         choices.write("infl-neg=on\n")
         choices.write("comp-neg=on\n")
   #      choices.write("adv-neg=on\n")
         choices.write("bineg-type=infl-comp\n")
         if bi_type in [('bound', 'comp'), ('head', 'comp')]:
            choices.write("neg1-type=b\n")
            choices.write("neg2-type=fc\n")
            if orth2:
               choices.write("comp-neg-orth=%s\n" % orth2)
            if order2:
               choices.write("comp-neg-order-infl-head=%s\n" % order2)
   #         if selected2:
   #            choices.write("comp-neg-head-infl-comp-neg=%s\n" % selected2)
         else:
            choices.write("neg1-type=fc\n")
            choices.write("neg2-type=b\n")
            if orth1:
               choices.write("comp-neg-orth=%s\n" % orth1)
            if order1:
               choices.write("comp-neg-order-infl-head=%s\n" % order1)
   #         if selected1:
   #            choices.write("comp-neg-head-infl-comp-neg=%s\n" % selected1)

      else: # Attested logic only allows comp_comp if both are free.
         choices.write("comp-neg=on\n")
         choices.write("adv-neg=on\n")
         choices.write("bineg-type=comp-comp\n")
         choices.write("neg1-type=fc\n")
         choices.write("neg2-type=fc\n")
   #      if selected1:
   #         choices.write("comp-neg1-head=%s\n" % selected1)
         if order1:
            choices.write("comp-neg1-order=%s\n" % order1)
         if orth1:
            choices.write("comp-neg1-orth=%s\n" % orth1)
   #      if selected2:
   #         choices.write("comp-neg2-head=%s\n" % selected2)
         if order2:
            choices.write("comp-neg2-order=%s\n" % order2)
         if orth2:
            choices.write("comp-neg2-orth=%s\n" % orth2)

# end output_choices

if __name__ == '__main__':
   """ Determine the sentential negation settings for the language represented in the 
   enriched IGT file provided as input parameter one.

   Args:
      xigt_input	-- the enriched IGT input file
      choices_file	-- the output file for sentential negation choices
   """
   cmd = optparse.OptionParser()
   cmd.add_option("-d", action="store_true", dest="debug")
   cmd.add_option("--debug", action="store_true", dest="debug")
   cmd.set_defaults(debug=False)
   opts, args = cmd.parse_args()
   DEBUG = opts.debug

   numargs = len(args)

   if numargs != 2:
      sys.stderr.write("Usage: %s [-d | --debug] XIGT_FILE CHOICES_FILE" % sys.argv[0])
      raise SystemExit
   else:
      xigt_input = args[0]
      choices_output = args[1]

   xigt_file = open(xigt_input, encoding='utf-8')
   xc = xigtxml.load(xigt_file)
   xigt_file.close()

   sent_neg = SententialNegation()

   for i,an_igt in enumerate(xc):

      igt_inst = InferenceIGT(an_igt)
      if igt_inst.is_question:
         if DEBUG: print("IGT " + igt_inst.igtid + " is a question.")
#      continue
      if igt_inst.has_conjunction:
         if DEBUG: print("IGT " + igt_inst.igtid + " is conjunctive.")
#      continue
      if igt_inst.gram_in_gloss_count(negation_grams):
         neg_inst = NegationIGT(an_igt)
         sent_neg.addIGT(neg_inst)
         if DEBUG: print(repr(neg_inst.igtid))
         if DEBUG: print("Primary verb location: %s" % str(neg_inst.primary_verb_loc))
         if DEBUG: print(repr(neg_inst.neg_g_ids))
         if DEBUG: print(repr(neg_inst.neg_gw_ids))
         if DEBUG: print(repr(neg_inst.negators))
      else:
         pass

   lang_exponence = sent_neg.neg_exp
   Jespersen = sent_neg.is_Jespersen

   for negIGT in sent_neg.negationIGTs:
      if Jespersen:
         sent_neg.position_tally(sent_neg.negationIGTs[negIGT].negators)
         sent_neg.ortho_tally(sent_neg.negationIGTs[negIGT].negators)
      elif sent_neg.negationIGTs[negIGT].negation_gram_count == lang_exponence:
         sent_neg.position_tally(sent_neg.negationIGTs[negIGT].negators)
         sent_neg.ortho_tally(sent_neg.negationIGTs[negIGT].negators)
   if DEBUG: print(repr(sent_neg.tallies))
   if DEBUG: print(repr(sent_neg.ortho_tallies))

   choices_fd = open(choices_output, mode="w", encoding="utf-8", newline="\n")
   output_choices(choices_fd, sent_neg, lang_exponence, Jespersen)
   choices_fd.close()
