#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
__program__ = infer_tam.py
__author__ = MichaelLockwood
__projectclass__ = LING575
__projecttopic__ = AGGREGATION Inferrence
__projectname__ = Infer Tense, Aspect, and Mood
__date__ = May2015
__credits__ = None
__collaborators__ = None

This module loads enriched XIGT data and automatically builds TAM choices
files. This additionally automatically loads eval_tam.py to evaluate the
precision, recall, and f-score of the automated choices file output against
the gold standard.
"""

import eval_tam
import map_gloss
import os
import sys
import xigt.ref

from xigt.codecs import xigtxml

class Inference:

    agg_path = ''
    choices_path = ''

    def load_data(datasets):
        # Set path to the root of the AGGREGATION folder
        Inference.agg_path = os.path.abspath(os.path.join(os.path.abspath(
                            os.path.join(os.getcwd(), os.pardir)), os.pardir))
        # Create a TAM choices subdirectory for outputs
        Inference.choices_path = os.getcwd() + '/TAM_choices/'
        if not os.path.exists(Inference.choices_path):
            os.makedirs(Inference.choices_path)
        # Initialize and process objects with map_gloss.py
        data = []
        for dataset in datasets:
            Language.objects[dataset] = Language(dataset, datasets[dataset])
            data.append((dataset, Language.objects[dataset]._xigt_path))
        # Process XIGT files with map_gloss.py
        map_gloss.MapGloss.main_xigt(data)
        # Load TAM hierarchy
        Inference.load_hierarchy()
        return True

    def load_hierarchy():
        Inference.hierarchy = {'Tense':
                                   {'Future': {},
                                    'Past':
                                        {'Perfect': True,
                                         'Recent': True,
                                         'Removed': True},
                                    'Present': {}
                                   },
                               'Aspect':
                                   {'Continuous': {},
                                    'Distributive': {},
                                    'Durative': {},
                                    'Habitual': {},
                                    'Indicative': {},
                                    'Imperfective': {},
                                    'Perfective': {},
                                    'Progressive': {},
                                    'Resultative': {},
                                   },
                               'Mood':
                                   {'Irrealis': {},
                                    'Personal': {},
                                    'Prohibtive': {},
                                    'Purposive': {},
                                    'Subjunctive': {}
                                   }
                               }
        return True

    def set_choices():
        for iso in Language.objects:
            Language.objects[iso].set_choices()
        return True

    def run_evaluation(dataset, name):
        eval_tam.Evaluation.load_data(dataset)
        eval_tam.Evaluation.set_evaluation(name)
        Language.clear_class_data()
        eval_tam.Language.clear_class_data()

class Language:

    objects = {}
    hierarchy = {}
    mapping = {}
    datasets = {}

    def __init__(self, iso, path):
        self._iso = iso
        self._path = path
        self._xigt_path = ''
        self._choices_infile = ''
        self._choices_text = ''
        self._choices_outfile = ''
        self._mapping = {}
        Language.objects[iso] = path
        self.set_xigt_path()
        self.set_choices_infile()
        self.set_choices_outfile()

    def clear_class_data():
        Language.objects = {}
        Language.hierarchy = {}
        Language.mapping = {}
        Language.datasets = {}

    def set_mapping(self):
        for gloss in map_gloss.Language.objects[self._iso]._glosses:
            current = map_gloss.Gloss.objects[(self._iso, gloss)].get_gold()
            for value in Inference.hierarchy:
                if current in Inference.hierarchy[value]:
                    if current not in self._mapping:
                        self._mapping[current] = []
                    break
                for subvalue in Inference.hierarchy[value]:
                    if current in Inference.hierarchy[value][subvalue]:
                        if subvalue not in self._mapping:
                            self._mapping[subvalue] = []
                        if current not in self._mapping[subvalue]:
                            self._mapping[subvalue] = self._mapping.get(
                                                subvalue, 0) + [current]
                        break
        return True

    def set_choices(self):
        self.set_mapping()
        self.add_choices('section=tense-aspect-mood\n' +
                         'tense-definition=choose\n')
        for tense in Inference.hierarchy['Tense']:
            self.set_tense(tense)
        a = 1
        for aspect in Inference.hierarchy['Aspect']:
            a = self.set_aspect(aspect, a)
        m = 1
        for mood in Inference.hierarchy['Mood']:
            m = self.set_mood(mood, m)
        self.export_choices()
        return True
            
    def set_xigt_path(self):
        if self._path[-1] == '/':
            self._xigt_path = (Inference.agg_path + self._path +
                               'testsuite-enriched.xml')
        else:
            self._xigt_path = (Inference.agg_path + '/' + self._path +
                               '/testsuite-enriched.xml')
        return True

    def set_choices_infile(self):
        if self._path[-1] == '/':
            self._choices_infile = (Inference.agg_path + self._path +
                                    'choices.up')
        else:
            self._choices_infile = (Inference.agg_path + '/' + self._path +
                                    '/choices.up')
        return True

    def set_choices_outfile(self):
        self._choices_outfile = (Inference.choices_path + str(self._iso) +
                                 '.choices')
        return True

    def export_choices(self):
        writer = open(self._choices_outfile, 'w')
        writer.write(self._choices_text)

    def add_choices(self, value):
        self._choices_text += value

    def set_tense(self, tense):
        if tense in self._mapping:
            self.add_choices(str(tense).lower() + '=on\n')
            t = 1
            for subvalue in self._mapping[tense]:
                author = self.get_best_author(subvalue)
                self.add_choices(' ' + str(tense).lower() + '-subtype' +
                                 str(t) + '_name=' + str(author) + '\n')
                t += 1
        return True

    def set_aspect(self, aspect, a):
        if aspect in self._mapping:
            author = self.get_best_author(aspect)
            self.add_choices(' aspect' + str(a) + '_name=' + str(author) +
                             '\n')
            self.add_choices('  aspect' + str(a) + '_supertype1_name=aspect\n')
            a += 1
        return a

    def set_mood(self, mood, m):
        if mood in self._mapping:
            author = self.get_best_author(mood)
            self.add_choices(' mood' + str(m) + '_name=' + str(author) + '\n')
            self.add_choices('  mood' + str(m) + '_supertype1_name=mood\n')
            m += 1
        return m

    def get_best_author(self, gold):
        try:
            options = map_gloss.Language.objects[self._iso
                                                 ]._observed[gold].keys()
            # Select the most frequent equivalent gloss
            frequent = (0, '')
            for opt in options:
                if (map_gloss.Gloss.objects[(self._iso, opt)]._observed >
                    frequent[0]):
                    frequent = (map_gloss.Gloss.objects[(self._iso, opt)
                                                        ]._observed, opt)
            self._author = frequent[1]
            return self._author
        except:
            return gold
        
"""
User Interface Functions------------------------------------------------
"""
dev1 = {# Dev 1
        'jpn': '/data/567/dev1/jpn/',
        'jup': '/data/567/dev1/jup/',
        'pbt': '/data/567/dev1/pbt/',
        'rus': '/data/567/dev1/rus/',
        'sna': '/data/567/dev1/sna/'
        }
dev2 = {# Dev 2
        'bre': '/data/567/dev2/bre/',
        'fra': '/data/567/dev2/fra/',
        'lut': '/data/567/dev2/lut/',
        'ojg': '/data/567/dev2/ojg/',
        'sci': '/data/567/dev2/sci/',
        'tam': '/data/567/dev2/tam/'
        }
test = {# Test
        'ain': '/data/567/test/ain/',
        'hbs': '/data/567/test/hbs/',
        'hix': '/data/567/test/hix/',
        'inh': '/data/567/test/inh/',
        'jaa': '/data/567/test/jaa/',
        'kat': '/data/567/test/kat/',
        'tha': '/data/567/test/tha/',
        'vie': '/data/567/test/vie/',
        }

Inference.load_data(dev1)
Inference.set_choices()
Inference.run_evaluation(dev1, 'dev1')

Inference.load_data(dev2)
Inference.set_choices()
Inference.run_evaluation(dev2, 'dev2')

Inference.load_data(test)
Inference.set_choices()
Inference.run_evaluation(test, 'test')
