#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
__program__ = map_gloss.py
__author__ = MichaelLockwood
__projectclass__ = LING575
__projecttopic__ = AGGREGATION Inferrence
__projectname__ = Mapping Input Glosses to Standard Outputs
__date__ = May2015
__credits__ = None
__collaborators__ = None

This module allows a user to map a set of input glosses from an interlinear
gloss text source to a condensed output that attempts to correct mistakes and
merge redundant glosses. The main glossing function calls the correct MapGloss
class functions for the user and manages that process. For more control a user
may consult the Map_Gloss_Tutorial.pdf which explains each function and how to
command greater control of the system. Another optional interface is to have
map_gloss.py directly load IGT using XIGT, the AGGREGATION project's standard
for storing IGT. This function is called xigt_glossing and takes a list of XIGT
files as input. The user must make sure their PYTHONPATH is directed to the
proper XIGT directory in order for the function to work properly.
"""

import inspect
import re
import sys

class CLSO:
    """Data format created by Michael Lockwood that stores files of
    classes and objects (hence the name CLS - class and O - ojbect).

    Note the following code option was tried in previous versions but
    was replaced:

    import ast
    ast.literal_eval() # Convert str to dict, tuple, list
    """

    objects = {}

    def __init__(self, file, option='r'):
        self._file = open(str(file)+'.clso', option)
        CLSO.objects[file] = self

    def load(self):
        clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
        cls = ''
        values = {}
        for line in self._file:
            line = line.rstrip()
            # If blank, continue
            if not line:
                continue
            # If ID line set class to the value between the angle brackets
            if re.search('<[\w]*>', line):
                cls = line[1:-1]
            # If class variable process immediately
            elif line[0] == '$':
                exec(line[1:])
            # If value delimiter process the object and reset the value DS
            elif line == '@':
                for member in clsmembers:
                    if cls == member[0]:
                        # Add __init__ arguments
                        init_values = inspect.getargspec(eval(cls+'.__init__'))
                        init_code = cls + '('
                        for value in init_values.args[1:]:
                            if value in values:
                                if re.search('\'', values[value]):
                                    init_code += '\"' + values[value] + '\", '
                                else:
                                    init_code += '\'' + values[value] + '\', '
                                del values[value]
                            elif ('_'+value) in values:
                                if re.search('\'', values['_'+value]):
                                    init_code += ('\"' + values['_'+value] +
                                                  '\", ')
                                else:
                                    init_code += ('\'' + values['_'+value] +
                                                  '\', ')
                                del values['_'+value]
                            else:
                                raise ValueError('Missing argument in CLSO' +
                                                 'object from ' + str(cls))
                        init_code += ')'
                        obj = eval(init_code)
                        # Add all other values
                        for value in values:
                            try:
                                exec('obj.'+str(value)+'=' +
                                     str(values[value]))
                            except:
                                try:
                                    exec('obj.'+str(value) + '=\''+
                                         str(values[value] + '\''))
                                except:
                                    exec('obj.'+str(value) + '=\"'+
                                         str(values[value] + '\"'))
                values = {}
            # If close line process the object and reset the cls DS
            elif re.search('</[\w]*\>', line):
                cls = ''
            # If it is a value, process as value
            else:
                line = line.split('=')
                values[line[0]] = line[1]
        return True

    def _dump_helper(self, obj, cls_variables):
        methods = {}
        temp = inspect.getmembers(obj, inspect.ismethod)
        for method in temp:
            methods[method[0]] = True
        for value in dir(obj):
            if not re.search('__[\w]*__', value):
                if value not in methods and value not in cls_variables:
                    self._file.write(value + '=' + str(eval('obj.' + value))
                                     + '\n')
        self._file.write('@\n')
        return True

    def dump(self, obj):
        cls = str(obj.__class__)[17:-2]
        if cls[0] == '.':
            cls = cls[1:]
        cls_variables = []
        for value in dir(eval(cls)):
            if not re.search('__[\w]*__', value):
                cls_variables.append(value)
        self._file.write('<' + cls + '>\n')
        self._dump_helper(obj, cls_variables)
        self._file.write('</' + cls + '>\n\n')
        return True

    def bulk_dump(self, cls):
        self._file.write('<' + cls + '>\n')
        cls_variables = []
        for value in dir(eval(cls)):
            if not re.search('__[\w]*__', value):
                cls_variables.append(value)
        cls_functions = {}
        for func in inspect.getmembers(eval(cls), inspect.isfunction):
            cls_functions[func[0]] = True
        for var in cls_variables:
            if var != 'objects' and var not in cls_functions:
                self._file.write('$' + cls + '.' + str(var) + '=' + str(
                                 eval(cls +'.' + str(var))) + '\n')
        for obj in eval(str(cls) + '.objects'):
            if isinstance(obj, tuple):
                self._dump_helper(eval(str(cls) + '.objects[' + str(obj) +
                                       ']'), cls_variables)
            elif isinstance(obj, str):
                self._dump_helper(eval(str(cls) + '.objects[\'' + str(obj) +
                                       '\']'), cls_variables)
            else:
                raise TypeError('Object must be string or tuple.')
        self._file.write('</' + cls + '>\n\n')

    def close(self):
        self._file.close()

class MapGloss:
    """The MapGloss class manages the process of mapping input glosses
    to intermediate and/or final compliant glosses. Objects are
    individual input glosses that must have a language iso value. The
    compliant gloss output could be one of three user preferences:

    0 -- the leipzig gloss
    1 -- the gold gloss
    2 -- the best author gloss

    What 2 does is read in the entire set and if it discovers that two
    tags are likely synonymous it will merge them to form the most
    common version of the tag. Whereas 0 and 1 take input tags and make
    them either leipzig or gold, the author tag mapping tries to fix
    author made mistakes of inconsistently tagging the same gloss.

    The two current methods are delphi and levenshtein.

    The delphi method (not to be confused with the DELPH-IN consortium)
    allows a collective group of users to define likely corrections. For
    example: suppose that IMPF is found as input quite often. This was
    probably confused for IMPV which is the leipzig version of the
    imperfective aspect. The delphi method would change all instances of
    IMPF to IMPV if leipzig were selected or Imperfect if gold were
    selected.

    The levenshtein is currently set as the back-off to the delphi
    method. That means if the gloss in the input is not leipzig, gold,
    or in delphi, the levenshtein algorithm will try to find the closest
    matching leipzig or gold tag. A user input option allows the user to
    decide between ambiguities as they arise. The default is False so
    the system will prefer the first match but calling the wrapper
    function glossing with user_input=True will allow the user to
    decide how to resolve the ambiguity.

    Note that pickle files are used to store values so that the training
    of the class dictionaries does not need to occur each time. However
    if the user desires for them to be rebuilt simply call the glossing
    function with restart=True.

    Another note about architecture: the levenshtein glosses and lookup
    glosses for output are stored by iso so as to separate preferences
    between languages. At current the only impact this would have is
    that the user could decide that ambiguous levenshtein outcomes
    should be different between languages. It will also impact the
    author tag results if one occurs in one language more than the
    other. Although truly this functionality would not improve the
    system much unless more learning methods are applied.
    """

    max_distance = 0
    user_input = False
    user_preference = 1
    learning = ['delphi', 'levenshtein']
    restart = False

    def main(data, user_input=False, user_preference=2, restart=True):
        """Function for the user to call to process a set(s) of glosses.

        Keyword arguments:
          data -- the list of gloss sets organized into tuples as
                  following: (iso_code, [tag1, tag2, ... , tagn])
          user_input -- allow the user to give feedback on ambiguities
                        if True, otherwise the system will decide
          user_preference -- allow the user to decide what output system
                             the tags should be; 0) Leipzig, 1) GOLD,
                             2) author
          restart -- if True will restart files, otherwise will load
                     CLSO files for each DS
        """
        # Check data formatting
        if not isinstance(data, list):
            data = [data]
        for item in data:
            if len(item) != 2:
                raise ValueError('Each set must be a tuple with (iso, [tags])')
            if not isinstance(item[0], str):
                raise ValueError('Each set must be a tuple with (iso, [tags])')
            if not isinstance(item[1], list):
                raise ValueError('Each set must be a tuple with (iso, [tags])')
        # Set user variables
        MapGloss.set_user_input(user_input)
        MapGloss.set_user_preference(user_preference)
        MapGloss.set_restart(restart)
        # Load data structures
        MapGloss.load_combined_structures()
        # Process and convert data
        out_data = []
        for item in data:
            # Train each gloss
            for gloss in item[1]:
                MapGloss.training(item[0], gloss)
            # Set the author values and reference dictionary after training
            MapGloss.prepare_decoding()
            # Decode all glosses
            out_item = []
            for gloss in item[1]:
                out_item.append(MapGloss.decoding(item[0], gloss))
            out_data.append(out_item)
        # Export data structures for future use
        MapGloss.export_combined_structures()
        return out_data

    def main_xigt(data, user_input=False, user_preference=2, restart=True):
        """Function for the user to call to process a set(s) of glosses.
        Ensure that /src/xigt is in the PYTHONPATH.

        Keyword arguments:
          data -- the list of gloss sets organized into tuples as
                  following: (iso_code, xigt_file)
          user_input -- allow the user to give feedback on ambiguities
                        if True, otherwise the system will decide
          user_preference -- allow the user to decide what output system
                             the tags should be; 0) Leipzig, 1) GOLD,
                             2) author
          restart -- if True will restart files, otherwise will load
                     CLSO files for each DS
        """
        # Import xigtxml from xigt.codecs to allow reading of xigt files
        from xigt.codecs import xigtxml
        # Check data formatting
        if not isinstance(data, list):
            data = [data]
        for item in data:
            if len(item) != 2:
                raise ValueError('Each set must be a tuple with ' +
                                 '(iso, xigt_file)')
        # Set user variables
        MapGloss.set_user_input(user_input)
        MapGloss.set_user_preference(user_preference)
        MapGloss.set_restart(restart)
        # Load data structures
        MapGloss.load_combined_structures()
        # Process and convert data
        out_data = []
        for item in data:
            # Open the current xigt file
            file = open(item[1])
            xc = xigtxml.load(file)
            file.close()
            # Train each gloss
            for igt in xc:
                # Capture the morphemes
                morphemes = {}
                for morpheme in igt.get('m'):
                    morphemes[morpheme.id] = morpheme.value()
                # Process by gloss
                for gloss in igt.get('g'):
                    try:
                        MapGloss.training(item[0], gloss.value(),
                                          morphemes[gloss.alignment])
                    except:
                        MapGloss.training(item[0], gloss.value())
            # Set the author values and reference dictionary after training
            MapGloss.prepare_decoding()
            # Decode all glosses
            out_item = []
            for igt in xc:
                for gloss in igt.get('g'):
                    out_item.append(MapGloss.decoding(item[0], gloss.value()))
            out_data.append(out_item)
        # Export data structures for future use
        MapGloss.export_combined_structures()
        return out_data

    def training(iso, gloss, morpheme=''):
        if iso not in Language.objects:
            Language(iso)
        if (iso, gloss) not in Gloss.objects:
            Gloss(iso, gloss)
            Gloss.objects[(iso, gloss)].set_values()
        Gloss.objects[(iso, gloss)].add_observation()
        if morpheme != '':
            Gloss.objects[(iso, gloss)].set_morpheme(morpheme)
        return True

    def prepare_decoding():
        for iso in Language.objects:
            Language.objects[iso].set_references()
            for gloss in Language.objects[iso]._glosses:
                Gloss.objects[(iso, gloss)].set_author()

    def decoding(iso, gloss):
        return Gloss.objects[(iso, gloss)].get_output_gloss()

    def clear_class_data():
        """Clear the class DS to ensure that previous values are not
        saved.
        """
        MapGloss.max_distance = 1
        MapGloss.user_input = False
        MapGloss.user_preference = 1
        MapGloss.learning = ['delphi', 'levenshtein']
        MapGloss.restart = False
        return True

    def load_individual_structures():
        """Load individual CLSO files for all DS.
        """
        Leipzig.load()
        Gold.load()
        Delphi.load()
        Language.load()
        Gloss.load()
        Segment.load()
        Levenshtein.load()
        Lexicon.load()
        Morpheme.load()
        return True

    def export_individual_structures():
        """Export individual CLSO files for all DS.
        """
        Leipzig.export()
        Gold.export()
        Delphi.export()
        Language.export()
        Gloss.export()
        Segment.export()
        Levenshtein.export()
        Lexicon.export()
        Morpheme.export()
        return True

    def load_combined_structures():
        """Load a CLSO file containing all DS.
        """
        if MapGloss.restart == False:
            try:
                reader = CLSO('map_gloss', option='r')
                reader.load()
                return True
            except:
               pass
        MapGloss.load_individual_structures()

    def export_combined_structures():
        """Export a CLSO file containing all DS.
        """
        writer = CLSO('map_gloss', option='w')
        writer.bulk_dump('Leipzig')
        writer.bulk_dump('Gold')
        writer.bulk_dump('Delphi')
        writer.bulk_dump('Language')
        writer.bulk_dump('Gloss')
        writer.bulk_dump('Segment')
        writer.bulk_dump('Levenshtein')
        writer.bulk_dump('Lexicon')
        writer.bulk_dump('Morpheme')
        return True

    def set_max_distance(value):
        """Set the max distance for Levenshtein value acceptance.

        Keyword arguments
          value -- any integer
        """
        if isinstance(value, int):
            MapGloss.max_distance = value
        else:
            raise ValueError('Value must be integer.')
    
    def set_user_input(value):
        """Set the user input between True and False

        Keyword arguments
          value -- True) allow input, False) do not allow input
        """
        if isinstance(value, bool):
            MapGloss.user_input = value
        else:
            raise ValueError('Value must True or False.')

    def set_user_preference(value):
        """Set the user preference between 0, 1, and 2.

        Keyword arguments
          value -- 0) leipzig, 1) gold, 2) author
        """
        if value == 0 or value == 1 or value == 2:
            MapGloss.user_preference = value
        else:
            raise ValueError('Value must be equal to 0, 1, or 2.')

    def set_restart(value):
        """Set the restart between True and False

        Keyword arguments
          value -- True) restart data structure, False) do not restart
        """
        if isinstance(value, bool):
            MapGloss.restart = value
        else:
            raise ValueError('Value must True or False.')

    def user_addition(gloss):
        """Allow the user to add values for glosses that have either an
        unknown leipzig or GOLD value. If neither values are known, then
        a delphi entry is added to map an incorrect value to the correct
        leipzig and GOLD values.
        """
        # Determine type of situation (is gloss leipzig, GOLD, or delphi?)
        value = input('\nIs the gloss ' + str(gloss) +
                      ':\n\n\t0) Leipzig\n\t1) GOLD\n\t2) Neither\n' +
                      '\nPlease select an option (int only): ')
        # Leipzig selection
        if int(value) == 0:
            gold = input('\nWhat would you say is the GOLD ' +
                         'equivalent of ' + str(gloss) +
                         ': ')
            Leipzig(gloss, gold, pseudo=True)
            Gold(gold, gloss, pseudo=True)
            return gloss
        # GOLD selection
        elif int(value) == 1:
            leipzig = input('\nWhat would you say is the ' +
                            'Leipzig equivalent of ' +
                            str(gloss) + ': ')
            Gold(gloss, leipzig, pseudo=True)
            Leipzig(leipzig, gloss, pseudo=True)
            return gloss
        # Delphi selection
        else:
            leipzig = input('\nWhat would you say is the ' +
                            'Leipzig equivalent of ' +
                            str(gloss) + ': ')
            gold = input('\nWhat would you say is the GOLD ' +
                         'equivalent of ' + str(gloss) +
                         ': ')
            Delphi(gloss, leipzig, gold)
            return leipzig

    def get_example():
        print([('yak', ['hungry', 'IMPF', 'PST', '1SG', 'IMPF', 'IMFP']),
                ('rus', ['Ivan', 'buy', '3SG', 'PST', 'dog', 'ACC', 'SG'])])
        return [('yak', ['hungry', 'IMPF', 'PST', '1SG', 'IMPF', 'IMFP']),
                ('rus', ['Ivan', 'buy', '3SG', 'PST', 'dog', 'ACC', 'SG'])]

class Leipzig:

    objects = {}

    def __init__(self, leipzig, gold, pseudo=False):
        self._leipzig = leipzig
        self._gold = gold
        self._pseudo = pseudo
        self._iso_codes = {}
        Leipzig.objects[leipzig] = self
        if gold not in Gold.objects:
            Gold(gold, leipzig, pseudo)

    def load(restart=False):
        """Load the leipzig CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('leipzig', option='r')
                reader.load()
                return True
            except:
                pass
        leipzig_map = {# Person and Number
                       '1': ['1', 'FirstPerson'],
                       '2': ['2', 'SecondPerson'],
                       '3': ['3', 'ThirdPerson'],
                       'SG': ['SG', 'SingularNumber'],
                       'PL': ['PL', 'PluralNumber'],
                       '1SG': ['1SG', 'FirstPerson-SingularNumber'],
                       '2SG': ['2SG', 'SecondPerson-SingularNumber'],
                       '3SG': ['3SG', 'ThirdPerson-SingularNumber'],
                       '1PL': ['1PL', 'FirstPerson-PluralNumber'],
                       '2PL': ['2PL', 'SecondPerson-PluralNumber'],
                       '3PL': ['3PL', 'ThirdPerson-PluralNumber'],
                       'SBJ': ['SBJ', 'Subject'],
                       'OBJ': ['OBJ', 'Object'],
                       # Case
                       'ABL': ['ABL', 'Ablative'],
                       'ABS': ['ABS', 'Absolutive'],
                       'ACC': ['ACC', 'Accusative'],
                       'ALL': ['ALL', 'Allative'],
                       'BEN': ['BEN', 'Benefactive'],
                       'DAT': ['DAT', 'Dative'],
                       'ERG': ['ERG', 'Ergative'],
                       'GEN': ['GEN', 'Genitive'],
                       'INS': ['INS', 'Instrumental'],
                       'LOC': ['LOC', 'Locative'],
                       'OBL': ['OBL', 'Oblique'],
                       'VOC': ['VOC', 'Vocative'],
                       # Gender
                       'F': ['F', 'Feminine'],
                       'M': ['M', 'Masculine'],
                       'N': ['N', 'Neuter'],
                       # Tense
                       'FUT': ['FUT', 'Future'],
                       'NFUR': ['NPST', 'NonFuture', 'Psuedo'],
                       'NPRS': ['NPST', 'NonPresent', 'Psuedo'],
                       'NPST': ['NPST', 'NonPast', 'Psuedo'],
                       'PRF': ['PRF', 'Perfect'],
                       'PRS': ['PRS', 'Present'],
                       'PST': ['PST', 'Past'],
                       'REC': ['REC', 'Recent', 'Pseudo'],
                       'REM': ['REM', 'Removed', 'Pseudo'],
                       # Aspect
                       'CONT': ['CONT', 'Continuous', 'Pseudo'],
                       'DISTR': ['DISTR', 'Distributive'],
                       'DUR': ['DUR', 'Durative'],
                       'HAB': ['HAB', 'Habitual'],
                       'IND': ['IND', 'Indicative'],
                       'IPFV': ['IPFV', 'Imperfective'],
                       'PFV': ['PFV', 'Perfective'],
                       'PROG': ['PROG', 'Progressive'],
                       # Mood
                       'IRR': ['IRR', 'Irrealis'],
                       #'PERS': ['PERS', 'Personal'],
                       'PROH': ['PROH', 'Prohibitive'],
                       'PURP': ['PURP', 'Purposive'],
                       'RES': ['RES', 'Resultative'],
                       'SBJV': ['SBJV', 'Subjunctive'],
                       # Other Verb
                       'IMP': ['IMP', 'Imperative'],
                       'INF': ['INF', 'Infinitive'],
                       'INTR': ['INTR', 'Intransitive'],
                       'TR': ['TR', 'Transitive'],
                       # Coordination
                       'CONJ': ['CONJ', 'Conjunction', 'Pseudo'],
                       # Other
                       'DEM': ['DEM', 'Demonstrative']
                      }
        # From map create objects for each value
        for gloss in leipzig_map:
            # If pseudo, update map and ensure pseudo=True for the object
            if len(leipzig_map[gloss]) == 3:
                leipzig_obj = Leipzig(gloss, leipzig_map[gloss][1],
                                      pseudo=True)
            else:
                leipzig_obj = Leipzig(gloss, leipzig_map[gloss][1])
            # Add the object to the dictionary of leipzig objects
            Leipzig.objects[gloss] = leipzig_obj 
        return True

    def export():
        """Export the leipzig CLSO file.
        """
        writer = CLSO('leipzig', option='w')
        writer.bulk_dump('Leipzig')
        return True

    def delete_leipzig(leipzig_value):
        """Delete a leipzig value and its corresponding gold value.

        Keyword arguments:
          leipzig_value -- the value of the leipzig gloss
        """
        gold_value = Leipzig.options[leipzig_value].get_gold()
        del Leipzig.objects[leipzig_value]
        del Gold.objects[gold_value]
        return True

    def get_leipzig(self):
        """Return the leipzig value of the object.
        """
        return self._leipzig

    def get_gold(self):
        """Return the gold value of the object.
        """
        return self._gold

    def get_pseudo(self):
        """Return the pseudo value of the object.
        """
        return self._pseudo

    def get_iso_codes(self):
        """Return the iso code values of the object.
        """
        return self._iso_codes

    def get_iso_list(self):
        """Return the iso code values of the object in a list.
        """
        return self._iso_codes.keys()

class Gold:

    objects = {}

    def __init__(self, gold, leipzig, pseudo=False):
        self._gold = gold
        self._leipzig = leipzig
        self._pseudo = pseudo
        self._iso_codes = {}
        Gold.objects[gold] = self
        if leipzig not in Leipzig.objects:
            Leipzig(leipzig, gold, pseudo)

    def load(restart=False):
        """Load the gold CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('gold', option='r')
                reader.load()
                return True
            except:
                pass
        gold_map = {# Person and Number
                    'FirstPerson': ['1', 'FirstPerson'],
                    'SecondPerson': ['2', 'SecondPerson'],
                    'ThirdPerson': ['3', 'ThirdPerson'],
                    'SingularNumber': ['SG', 'SingularNumber'],
                    'PluralNumber': ['PL', 'PluralNumber'],
                    'FirstPerson-SingularNumber': ['1SG',
                                                'FirstPerson-SingularNumber'],
                    'SecondPerson-SingularNumber': ['2SG',
                                                'SecondPerson-SingularNumber'],
                    'ThirdPerson-SingularNumber': ['3SG',
                                                'ThirdPerson-SingularNumber'],
                    'FirstPerson-PluralNumber': ['1PL',
                                                 'FirstPerson-PluralNumber'],
                    'SecondPerson-PluralNumber': ['2PL',
                                                  'SecondPerson-PluralNumber'],
                    'ThirdPerson-PluralNumber': ['3PL',
                                                 'ThirdPerson-PluralNumber'],
                    'Subject': ['SBJ', 'Subject'],
                    'Object': ['OBJ', 'Object'],
                    # Case
                    'Ablative': ['ABL', 'Ablative'],
                    'Absolutive': ['ABS', 'Absolutive'],
                    'Accusative': ['ACC', 'Accusative'],
                    'Allative': ['ALL', 'Allative'],
                    'Benefactive': ['BEN', 'Benefactive'],
                    'Dative': ['DAT', 'Dative'],
                    'Ergative': ['ERG', 'Ergative'],
                    'Genitive': ['GEN', 'Genitive'],
                    'Instrumental': ['INS', 'Instrumental'],
                    'Locative': ['LOC', 'Locative'],
                    'Oblique': ['OBL', 'Oblique'],
                    'Vocative': ['VOC', 'Vocative'],
                    # Gender
                    'Feminine': ['F', 'Feminine'],
                    'Masculine': ['M', 'Masculine'],
                    'Neuter': ['N', 'Neuter'],
                    # Tense
                    'Future': ['FUT', 'Future'],
                    'NonFuture': ['NPST', 'NonFuture', 'Psuedo'],
                    'NonPresent': ['NPST', 'NonPresent', 'Psuedo'],
                    'NonPast': ['NPST', 'NonPast'],
                    'Perfect': ['PRF', 'Perfect'],
                    'Present': ['PRS', 'Present'], 
                    'Past': ['PST', 'Past'],
                    'Recent': ['REC', 'Recent', 'Pseudo'],
                    'Removed': ['REM', 'Removed', 'Pseudo'],
                    # Aspect
                    'Continuous': ['CONT', 'Continuous'],
                    'Distributive': ['DISTR', 'Distributive'],
                    'Durative': ['DUR', 'Durative'],
                    'Habitual': ['HAB', 'Habitual'],
                    'Indicative': ['IND', 'Indicative'],
                    'Imperfective': ['IPFV', 'Imperfective'],
                    'Perfective': ['PFV', 'Perfective'],
                    'Progressive': ['PROG', 'Progressive'],
                    # Mood
                    'Irrealis': ['IRR', 'Irrealis'],
                    #'Personal': ['PERS', 'Personal'],
                    'Prohibtive': ['PROH', 'Prohibitive'],
                    'Purposive': ['PURP', 'Purposive'],
                    'Resultative': ['RES', 'Resultative'],
                    'Subjunctive': ['SBJV', 'Subjunctive'],
                    # Other Verb
                    'Imperative': ['IMP', 'Imperative'],
                    'Infinitive': ['INF', 'Infinitive'],
                    'Intransitive': ['INTR', 'Intransitive'],
                    'Transitive': ['TR', 'Transitive'],
                    # Coordination
                    'Conjunction': ['CONJ', 'Conjunction', 'Pseudo'],
                    # Other
                    'DEM': ['DEM', 'Demonstrative']
                   }
        # From map create objects for each value
        for gloss in gold_map:
            # If pseudo, update map and ensure pseudo=True for the object
            if len(gold_map[gloss]) == 3:
                gold_obj = Gold(gloss, gold_map[gloss][0], pseudo=True)
            else:
                gold_obj = Gold(gloss, gold_map[gloss][0])
            # Add the object to the dictionary of gold objects
            Gold.objects[gloss] = gold_obj 
        return True

    def export():
        """Export the gold CLSO file.
        """
        writer = CLSO('gold', option='w')
        writer.bulk_dump('Gold')
        return True

    def delete_gold(gold_value):
        """Delete a gold value and its corresponding leipzig value.

        Keyword arguments:
          gold_value -- the value of the gold gloss
        """
        leipzig_value = Gold.objects[gold_value].get_leipzig()
        del Leipzig.objects[leipzig_value]
        del Gold.objects[gold_value]
        return True

    def get_gold(self):
        """Return the gold value of the object.
        """
        return self._gold

    def get_leipzig(self):
        """Return the leipzig value of the object.
        """
        return self._leipzig

    def get_pseudo(self):
        """Return the pseudo value of the object.
        """
        return self._pseudo

    def get_iso_codes(self):
        """Return the iso code values of the object.
        """
        return self._iso_codes

    def get_iso_list(self):
        """Return the iso code values of the object in a list.
        """
        return self._iso_codes.keys()

class Delphi:

    objects = {}

    def __init__(self, delphi, leipzig, gold):
        self._delphi = delphi
        self._leipzig = leipzig
        self._gold = gold
        self._iso_codes = {}
        Delphi.objects[delphi] = self

    def load(restart=False):
        """Load the delphi CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('delphi', option='r')
                reader.load()
                return True
            except:
                pass
        delphi_map = {# Case
                      'INST': ['INS', 'Instrumental', 'INST'],
                      # Gender
                      'FEM': ['F', 'Feminine', 'FEM'],
                      'MASC': ['M', 'Masculine', 'MASC'],
                      'NEUT': ['N', 'Neuter', 'NEUT'],
                      # Tense
                      'PRES': ['PRS', 'Present', 'PRES'],
                      # Aspect
                      'Continuative': ['CONT', 'Continuous', 'Continuative'],
                      'IMP': ['IPVF', 'Imperfective', 'IMP'],
                      'IMPF': ['IPVF', 'Imperfective', 'IMPF'],
                      'IMPV': ['IPVF', 'Imperfective', 'IMPV'],
                      'PER': ['PFV', 'Perfective', 'PER'],
                      'PERF': ['PFV', 'Perfective', 'PERF'],
                      # Other Verb
                      'ITR': ['INTR', 'Intransitive'],
                      'INTRANS': ['INTR', 'Intransitive'],
                      'TRANS': ['TR', 'Transitive'],
                      # Ignore Section
                      'AND': ['CONJ', 'Conjunction', 'AND'],
                      'And': ['CONJ', 'Conjunction', 'And'],
                      'and': ['CONJ', 'Conjunction', 'and'],
                      'BUT': ['CONJ', 'Conjunction', 'BUT'],
                      'But': ['CONJ', 'Conjunction', 'But'],
                      'but': ['CONJ', 'Conjunction', 'but'],
                      'FV': ['FV', 'FV', 'FV'],
                      'in': ['in', 'in', 'in'],
                      'INT': ['INT', 'INT', 'INT'],
                      'PRE': ['PRE', 'PRE', 'PRE'],
                      'red': ['red', 'red', 'red']
                     }
        # From map create objects for each value
        for gloss in delphi_map:
            delphi_obj = Delphi(gloss, delphi_map[gloss][0],
                                delphi_map[gloss][1])
            # Add the object to the dictionary of delphi objects
            Delphi.objects[gloss] = delphi_obj 
        return True

    def export():
        """Export the delphi CLSO file.
        """
        writer = CLSO('delphi', option='w')
        writer.bulk_dump('Delphi')
        return True

    def delete_delphi(delphi_value):
        """Delete a delphi value.

        Keyword arguments:
          delphi_value -- the value of the delphi gloss
        """
        del Delphi.objects[delphi_value]
        return True

    def get_delphi(self):
        """Return the delphi value of the object.
        """
        return self._delphi

    def get_leipzig(self):
        """Return the leipzig value of the object.
        """
        return self._leipzig

    def get_gold(self):
        """Return the gold value of the object.
        """
        return self._gold

    def get_iso_codes(self):
        """Return the iso code values of the object.
        """
        return self._iso_codes

    def get_iso_list(self):
        """Return the iso code values of the object in a list.
        """
        return self._iso_codes.keys()

class Language:

    objects = {}

    def __init__(self, iso):
        self._iso = iso
        self._glosses = {}
        self._segments = {}
        self._morphemes = {}
        self._observed = {}
        self._references = {}
        Language.objects[iso] = self

    def load(restart=False):
        """Load the languages CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('language', option='r')
                reader.load()
                return True
            except:
                pass
        Language.objects = {}
        return True

    def export():
        """Export the languages CLSO file.
        """
        writer = CLSO('language', option='w')
        writer.bulk_dump('Language')
        return True

    def delete_language(iso_value):
        """Delete a language value.

        Keyword arguments:
          iso_value -- the iso value of the language
        """
        del Language.objects[iso_value]
        return True

    def get_iso(self):
        """Return the iso value of the object.
        """
        return self._iso

    def get_glosses(self):
        """Return the glosses value of the object.
        """
        return self._glosses

    def get_segments(self):
        """Return the segments value of the object.
        """
        return self._segments

    def get_morphemes(self):
        """Return the morphemes value of the object.
        """
        return self._morphemes

    def get_observed(self):
        """Return the observed value of the object.
        """
        return self._observed

    def set_references(self):
        """Set the reference data structure by building indexed sets of
        the values in the gloss_set. This essentially builds concepts or
        rather groups of equivalent values whether they are leipzig,
        GOLD, or authored. This may be important for the work in the
        inferrence procedures when comparing choices files for the
        purpose of evaluation.
        """
        # For each gloss for the language
        for gloss in self._glosses:
            values = [gloss,
                      Gloss.objects[(self._iso, gloss)].get_leipzig(),
                      Gloss.objects[(self._iso, gloss)].get_gold(),
                      Gloss.objects[(self._iso, gloss)].get_author()]
            # For each equivalent value of that gloss
            for value in values:
                # If entry does not exist, add it to references
                if value not in self._references:
                    self._references[value] = {}
                # Add all equivalents to current
                for entry in values:
                    self._references[value][entry] = True
            # Rerun each equivalent value to add the known equivalents of each
            # equivalent to all
            for value in values:
                for entry in values:
                    for item in self._references[entry]:
                        self._references[value][item] = True
            # Add the current gloss and author to previous
            for entry in self._references[values[0]]:
                for item in self._references[entry]:
                    if item != entry:
                        self._references[item][values[0]] = True
            for entry in self._references[values[3]]:
                    for item in self._references[entry]:
                        self._references[item][values[3]] = True
        # For each gloss, apply references
        for gloss in self._glosses:
            Gloss.objects[(self._iso, gloss)].set_references()
        return True

    def get_references(self):
        """Return the references value of the object.
        """
        return self._references

    def print_values(self):
        for gloss in self._glosses:
            Gloss.objects[(self._iso, gloss)].print_values()

class Gloss:

    objects = {} #(iso, gloss)

    def __init__(self, iso, gloss):
        self._iso = iso
        self._gloss = gloss
        self._leipzig = ''
        self._gold = ''
        self._author = ''
        self._delphi = False
        self._levenshtein = False
        self._segments = {}
        self._morphemes = {}
        self._observed = 0
        self._references = {}
        Language.objects[iso]._glosses[gloss] = True
        Gloss.objects[(iso, gloss)] = self

    def load(restart=False):
        """Load the glosses CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('gloss', option='r')
                reader.load()
                return True
            except:
                pass
        Gloss.objects = {}
        return True

    def export():
        """Export the glosses CLSO file.
        """
        writer = CLSO('gloss', option='w')
        writer.bulk_dump('Gloss')
        return True

    def delete_gloss(iso_value, gloss_value):
        """Delete a gloss value.

        Keyword arguments:
          iso_value -- the iso value of the gloss
          gloss_value -- the gloss value
        """
        del Gloss.objects[(iso_value, gloss_value)]
        return True

    def get_iso(self):
        """Return the iso name for the language from the input.
        """
        return self._iso

    def get_input_gloss(self):
        """Return the gloss input.
        """
        return self._gloss

    def get_output_gloss(self):
        """Return the gloss output.
        """
        if MapGloss.user_preference == 0:
            output = self.get_leipzig()
        elif MapGloss.user_preference == 1:
            output = self.get_gold()
        elif MapGloss.user_preference == 2:
            output = self.get_author()
        return output

    def get_leipzig(self):
        """Return the leipzig version of the gloss.
        """
        if self._leipzig == '':
            self.set_values()
        return self._leipzig

    def get_gold(self):
        """Return the gold version of the gloss.
        """
        if self._gold == '':
            self.set_values()
        return self._gold

    def set_author(self):
        """Set the author version of the gloss.
        """
        if (self._gloss not in Leipzig.objects and
            self._gloss not in Gold.objects):
            try:
                options = Language.objects[self._iso]._observed[self._gold].keys()
                # Select the most frequent equivalent gloss
                frequent = (0, '')
                for opt in options:
                    if Gloss.objects[(self._iso, opt)]._observed > frequent[0]:
                        frequent = (Gloss.objects[(self._iso, opt)]._observed, opt)
                self._author = frequent[1]
            except:
                self._author = self._gloss
        return True

    def get_author(self):
        """Return the author version of the gloss.
        """
        if self._author == '':
            self.set_values()
        return self._author

    def get_delphi(self):
        """Return the delphi match of the gloss iff it exists.
        """
        return self._delphi

    def get_levenshtein(self):
        """Return the levenshtein match of the gloss iff it exists.
        """
        return self._levenshtein

    def set_values(self):
        """Add an entry for the object's gloss to the gloss lookup if it
        does not yet exist.
        """
        # Process segmentation
        self.set_segmentation()
        # Reset author value
        self._author = ''
        # If gloss is Leipzig set as Leipzig value
        if self._gloss in Leipzig.objects:
            self._leipzig = Leipzig.objects[self._gloss].get_leipzig()
            self._gold = Leipzig.objects[self._gloss].get_gold()
            self._author = Leipzig.objects[self._gloss].get_leipzig()
        # If gloss is GOLD set as GOLD value
        elif self._gloss in Gold.objects:
            self._leipzig = Gold.objects[self._gloss].get_leipzig()
            self._gold = Gold.objects[self._gloss].get_gold()
            self._author = Gold.objects[self._gloss].get_gold()
        # Otherwise proceed to ranked learning methods
        else:
            for method in MapGloss.learning:
                # If the method is delphi, lookup possible delphi values
                if method == 'delphi':
                    if self._author == '':
                        if self._gloss in Delphi.objects:
                            self._leipzig = Delphi.objects[
                                                    self._gloss].get_leipzig()
                            self._gold = Delphi.objects[
                                                    self._gloss].get_gold()
                            self._author = Delphi.objects[
                                                    self._gloss].get_delphi()
                            self._delphi = True
                # If the method is levenshtein, process levenshtein value
                elif method == 'levenshtein':
                    if self._author == '':
                        Levenshtein(self._iso, self._gloss)
                        Levenshtein.objects[(self._iso, self._gloss)
                                            ].set_levenshtein()
                        values = Levenshtein.objects[(self._iso, self._gloss)
                                            ].get_levenshtein()
                        self._leipzig = values[0]
                        self._gold = values[1]
                        self._author = values[2]
                        self._levenshtein = True
        # Use gold gloss to set benchmark for author gloss disambiguation
        if self._gold not in Language.objects[self._iso]._observed:
            Language.objects[self._iso]._observed[self._gold] = {}
        Language.objects[self._iso]._observed[self._gold][self._gloss] = True
        return True

    def add_segment(self, segment):
        self._segments[segment] = self._segments.get(segment, 0) + 1
        return True

    def set_segmentation(self):
        """Locate all segments within the gloss, if the gloss is Leipzig
        or Gold it will match itself.
        """
        # Try each Leipzig value
        for leipzig in Leipzig.objects:
            # If there is a match
            if re.search(leipzig, self._gloss):
                # Add the segment to the segments list
                self.add_segment(leipzig)
                match = re.search(leipzig, self._gloss)
                # Create then populate segment object
                if (self._iso, leipzig) not in Segment.objects:
                    Segment(self._iso, leipzig)
                Segment.objects[(self._iso, leipzig)].set_combination(
                                                                self._gloss)
                # Process values to the left and right of the segment
                self._set_segmentation(self._gloss[:match.span()[0]])
                self._set_segmentation(self._gloss[match.span()[1]:])
        # Try each Gold value
        for gold in Gold.objects:
            # If there is a match
            if re.search(gold, self._gloss):
                # Add the segment to the segments list
                self.add_segment(gold)
                match = re.search(gold, self._gloss)
                # Create then populate segment object
                if (self._iso, gold) not in Segment.objects:
                    Segment(self._iso, gold)
                Segment.objects[(self._iso, gold)].set_combination(self._gloss)
                # Process values to the left and right of the segment
                self._set_segmentation(self._gloss[:match.span()[0]])
                self._set_segmentation(self._gloss[match.span()[1]:])
        return True

    def _set_segmentation(self, gloss):
        """Locate all segments within the gloss, if the gloss is Leipzig
        or Gold it will match itself. This is for inner recursion of the
        remainder.
        """
        if not gloss:
            return False
        setter = True
        # Try each Leipzig value
        for leipzig in Leipzig.objects:
            # If there is a match
            if re.search(leipzig, gloss):
                setter = False
                match = re.search(leipzig, gloss)
                # Process values to the left and right of the segment
                self._set_segmentation(gloss[:match.span()[0]])
                self._set_segmentation(gloss[match.span()[1]:])
        # Try each Gold value
        for gold in Gold.objects:
            # If there is a match
            if re.search(gold, gloss):
                setter = False
                match = re.search(gold, gloss)
                # Process values to the left and right of the segment
                self._set_segmentation(gloss[:match.span()[0]])
                self._set_segmentation(gloss[match.span()[1]:])
        # If no match has been found
        if setter == True:
            # Add the segment to the segments list
            self.add_segment(gloss)
            # Create then populate segment object
            if (self._iso, gloss) not in Segment.objects:
                Segment(self._iso, gloss)
            Segment.objects[(self._iso, gloss)].set_combination(gloss)
        return True

    def get_segments(self):
        """Return the segment value of the object.
        """
        return self._segments

    def set_morpheme(self, morpheme):
        """Increment the count of the morpheme for the gloss.

        Keyword arguments:
          morpheme -- the morpheme value associated with the gloss
        """
        self._morphemes[morpheme] = self._morphemes.get(morpheme, 0) + 1
        # Increment the gloss count for the matching morpheme object, create an
        # object first if one does not exist
        if (self._iso, morpheme) not in Morpheme.objects:
            Morpheme(self._iso, morpheme)
        Morpheme.objects[(self._iso, morpheme)].set_gloss(self._gloss)

    def get_morphemes(self):
        """Return the morphemes value of the object.
        """
        return self._morphemes

    def add_observation(self):
        """Add an entry to the observation set for the iso_code and
        object gloss which will be used to determine the author gloss.
        """
        # Increment occurrence count for the gloss
        self._observed += 1
        return True

    def get_observed(self):
        """Return the observed value for the gloss.
        """
        return self._observed

    def set_references(self):
        """Set the reference data structure by building indexed sets of
        the values in the gloss_set. This essentially builds concepts or
        rather groups of equivalent values whether they are leipzig,
        GOLD, or authored. This may be important for the work in the
        inferrence procedures when comparing choices files for the
        purpose of evaluation.
        """
        # Utilize result from Language.set_references() to set gloss references
        self._references = Language.objects[self._iso]._references[self._gloss]
        return True

    def get_references(self):
        """Return the references value for the gloss.
        """
        return self._references

    def print_values(self):
        print_segments = ''
        for segment in self._segments:
            print_segments += ('\n\t\t' + str(segment) + ': ' +
                                str(self._segments[segment]))
        print_morphemes = ''
        for morpheme in self._morphemes:
            print_morphemes += ('\n\t\t' + str(morpheme) + ': ' +
                                str(self._morphemes[morpheme]))
        print_references = ''
        for reference in self._references:
            print_references += ('\n\t\t' + str(reference))
        print(str(self._gloss) +
              '\n\tLeipzig: ' + str(self._leipzig) +
              '\n\tGOLD: ' + str(self._gold) +
              '\n\tAuthor: ' + str(self._author) +
              '\n\tDelphi: ' + str(self._delphi) +
              '\n\tLevenshtein: ' + str(self._levenshtein) +
              '\n\tSegments:' + print_segments +
              '\n\tMorphemes:' + print_morphemes +
              '\n\tObservations: ' + str(self._observed) +
              '\n\tReferences:' + print_references)
        return True              

class Segment:

    objects = {} #(iso, gloss)

    def __init__(self, iso, segment):
        self._iso = iso
        self._segment = segment
        self._combinations = {}
        self._observed = 0
        Language.objects[iso]._segments[segment] = True
        Segment.objects[(iso, segment)] = self

    def load(restart=False):
        """Load the segment CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('segment', option='r')
                reader.load()
                return True
            except:
                pass
        Segment.objects = {}
        return True

    def export():
        """Export the segment CLSO file.
        """
        writer = CLSO('segment', option='w')
        writer.bulk_dump('Segment')
        return True

    def delete_segment(iso_value, segment_value):
        """Delete a morpheme value.

        Keyword arguments:
          iso_value -- the iso value of the gloss
          segment_value -- the segment value
        """
        del Segment.objects[(iso_value, segment_value)]
        return True

    def get_iso(self):
        """Return the iso name for the language from the input.
        """
        return self._iso

    def get_segment(self):
        """Return the segment input.
        """
        return self._segment

    def set_combination(self, gloss):
        """Increment the count of the gloss for the segment.

        Keyword arguments:
          gloss -- the gloss value associated with the segment
        """
        self.add_observation()
        self._combinations[gloss] = self._combinations.get(gloss, 0) + 1

    def get_combinations(self):
        """Return the combinations value of the object.
        """
        return self._combinations

    def add_observation(self):
        """Add an entry to the observation set for the iso_code and
        object segment.
        """
        # Increment occurrence count for the segment
        self._observed += 1
        return True

    def get_observed(self):
        """Return the observed value for the segment.
        """
        return self._observed

class Levenshtein:

    objects = {} #(iso, gloss)

    def __init__(self, iso, gloss):
        self._iso = iso
        self._gloss = gloss
        self._levenshtein = ''
        Levenshtein.objects[(iso, gloss)] = self

    def load(restart=False):
        """Load the levenshtein CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('levenshtein', option='r')
                reader.load()
                return True
            except:
                pass
        Levenshtein.objects = {}
        return True

    def export():
        """Export the levenshtein CLSO file.
        """
        writer = CLSO('levenshtein', option='w')
        writer.bulk_dump('Levenshtein')
        return True

    def delete_levenshtein(iso, gloss_value):
        """Delete a gloss value.

        Keyword arguments:
          iso -- the iso value of the language
          gloss_value -- the value of the levenshtein gloss
        """
        del Levenshtein.objects[(iso, gloss_value)]
        return True

    def get_iso(self):
        """Return the iso name for the language from the input.
        """
        return self._iso

    def get_gloss(self):
        """Return the gloss input.
        """
        return self._gloss

    def set_levenshtein(self):
        """Set the levenshtein determined best leipzig and gold gloss for
        the object's gloss.
        """
        # Initialize to an impossibly high distance
        lookup = 1000000
        lookup_tag = ['']
        if (self._iso, self._gloss) in Lexicon.objects:
            self._levenshtein = [self._gloss, self._gloss, self._gloss]
            return True
        else:
            # Process all leipzig and gold values for most common matches
            for gloss in Leipzig.objects:
                lookup, lookup_tag = self._handle_levenshtein(gloss, lookup,
                                                          lookup_tag)
            for gloss in Gold.objects:
                lookup, lookup_tag = self._handle_levenshtein(gloss, lookup,
                                                          lookup_tag)
            # If the score is greater than the maximum set as self._gloss
            if lookup > MapGloss.max_distance:
                lookup_tag = self._gloss
                Lexicon(self._iso, self._gloss)
            # If user input allow the user to decide
            elif MapGloss.user_input == True:
                # Build input question to ask the user their preference
                options = ('\t0) ' + str(self._gloss) + ' is not a gloss but' +
                           ' a word\n')
                o = 1
                for tag in lookup_tag:
                    options += '\t' + str(o) + ') ' + str(tag) + '\n'
                    o += 1
                options += ('\t' + str(o) + ') ' + str(self._gloss) +
                            ' is a gloss itself\n')
                value = input('\nFor the input gloss ' + self._gloss +
                              ' there were ' + str(len(lookup_tag)) +
                              ' values of equal weight:\n\n' + options +
                              '\nPlease select the number of the gloss you' +
                              ' prefer (int only): ')
                # If 0 selected, set value to gloss itself
                if int(value) == 0:
                    Lexicon(self._gloss)
                    lookup_tag = self._gloss
                # If final option selected, set value to gloss itself, and add
                # pseudo GOLD and Leipzig values
                elif int(value) == o:
                    lookup_tag = MapGloss.user_addition()
                # Otherwise set to the value the user selected
                else:
                    lookup_tag = lookup_tag[int(value) - 1]
            # Otherwise select the first lowest scoring
            else:
                lookup_tag = lookup_tag[0]
            # Set the best value based on whether it is leipzig or gold
            if lookup_tag in Leipzig.objects:
                self._levenshtein = (Leipzig.objects[lookup_tag].get_leipzig(),
                                     Leipzig.objects[lookup_tag].get_gold(),
                                     self._gloss)
            elif lookup_tag in Gold.objects:
                self._levenshtein = (Gold.objects[lookup_tag].get_leipzig(),
                                     Gold.objects[lookup_tag].get_gold(),
                                     self._gloss)
            else:
                self._levenshtein = (self._gloss, self._gloss, self._gloss)
            return True
        
    def _handle_levenshtein(self, gloss, lookup, lookup_tag):
        """Intermediate levenshtein function that handles the
        calculations and assignments of lookup values.
        """
        distance = self._calculate_levenshtein(self._gloss.lower(),
                                               gloss.lower())
        # Rerank possible glosses based on minimum distance threshold
        if distance < lookup:
            lookup = distance
            lookup_tag = [gloss]
        elif distance == lookup:
            lookup_tag.append(gloss)
        return lookup, lookup_tag
                
    def _calculate_levenshtein(self, a, b):
        """This function actually applies the levenshtein algorithm
        between an the object's gloss and a leipzig or gold tag.
        """
        # Reset to calculate whereas b has greater length
        if len(a) < len(b):
            return self._calculate_levenshtein(b, a)
        # If b has length of 0 return length of a
        if len(b) == 0:
            return len(a)
        previous_row = range(len(b) + 1)
        for i, c1 in enumerate(a):
            current_row = [i + 1]
            for j, c2 in enumerate(b):
                insertions = previous_row[j + 1] + 1
                deletions = current_row[j] + 1
                substitutions = previous_row[j] + (c1 != c2)
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row
        return previous_row[-1]

    def get_levenshtein(self):
        """Return the levenshtein value of the object.
        """
        return self._levenshtein

class Lexicon:

    objects = {} #(iso, gloss)

    def __init__(self, iso, gloss):
        self._iso = iso
        self._gloss = gloss
        Lexicon.objects[(iso, gloss)] = self

    def load(restart=False):
        """Load the lexicon CLSO file.
        """
        if restart == False:
            try:
                reader = CSLO('lexicon', option='r')
                reader.load()
                return True
            except:
                pass
        Lexicon.objects = {}
        return True

    def export():
        """Export the lexicon CLSO file.
        """
        writer = CLSO('lexicon', option='w')
        writer.bulk_dump('Lexicon')
        return True

    def delete_lexicon(iso, lexicon_value):
        """Delete a lexicon value.

        Keyword arguments:
          iso -- the iso value of the language
          lexicon_value -- the value of the lexicon gloss
        """
        del Lexicon.objects[(iso, lexicon_value)]
        return True

    def get_iso(self):
        """Return the iso name for the language from the input.
        """
        return self._iso

    def get_gloss(self):
        """Return the gloss input.
        """
        return self._gloss

class Morpheme:

    objects = {} #(iso, morpheme)

    def __init__(self, iso, morpheme):
        self._iso = iso
        self._morpheme = morpheme
        self._glosses = {}
        self._observed = 0
        Language.objects[iso]._morphemes[morpheme] = True
        Morpheme.objects[(iso, morpheme)] = self

    def load(restart=False):
        """Load the morpheme CLSO file.
        """
        if restart == False:
            try:
                reader = CLSO('morpheme', option='r')
                reader.load()
                return True
            except:
                pass
        Morpheme.objects = {}
        return True

    def export():
        """Export the morpheme CLSO file.
        """
        writer = CSLO('morpheme', option='w')
        writer.bulk_dump('Morpheme')
        return True

    def delete_morpheme(iso_value, morpheme_value):
        """Delete a morpheme value.

        Keyword arguments:
          iso_value -- the iso value of the gloss
          morpheme_value -- the morpheme value
        """
        del Morpheme.objects[(iso_value, morpheme_value)]
        return True

    def get_iso(self):
        """Return the iso name for the language from the input.
        """
        return self._iso

    def get_morpheme(self):
        """Return the morpheme input.
        """
        return self._morpheme

    def set_gloss(self, gloss):
        """Increment the count of the gloss for the morpheme.

        Keyword arguments:
          gloss -- the gloss value associated with the morpheme
        """
        self.add_observation()
        self._glosses[gloss] = self._glosses.get(gloss, 0) + 1

    def get_glosses(self):
        """Return the glosses value of the object.
        """
        return self._glosses

    def add_observation(self):
        """Add an entry to the observation set for the iso_code and
        object morpheme.
        """
        # Increment occurrence count for the gloss
        self._observed += 1
        return True

    def get_observed(self):
        """Return the observed value for the gloss.
        """
        return self._observed

class Baseline:
    
    objects = {} #(iso, gloss)
    languages = {'Aggregate': {}}

    def __init__(self, iso, gloss):
        self._iso = iso
        self._gloss = gloss
        self._observed = 1
        self._standard = ''
        self._error_type = ''
        Baseline.objects[(iso, gloss)] = self

    def main(data, name, restart=False):
        """Function for the user to call to assess baseline IGT data.
        Ensure that /src/xigt is in the PYTHONPATH.

        Keyword arguments:
          data -- the list of gloss sets organized into tuples as following:
                  (iso_code, xigt_file)
          name -- the name of the dataset
          restart -- if True will restart files, otherwise will load CLSO
                     files for each DS
        """
        # Import xigtxml from xigt.codecs to allow reading of xigt files
        from xigt.codecs import xigtxml
        # Check data formatting
        if not isinstance(data, list):
            data = [data]
        # Load data
        if restart == False:
            Baseline.load(name)
            if Baseline.objects:
                Baseline.report(name)
                Baseline.objects = {}
                return True
        # Process and convert data
        for item in data:
            # Open the current xigt file
            file = open(item[1])
            xc = xigtxml.load(file)
            file.close()
            # Load each gloss
            for igt in xc:
                # Process by gloss
                for gloss in igt.get('g'):
                    gls = str(gloss.value()).lower()
                    if (item[0], gls) not in Baseline.objects:
                        Baseline(item[0], gls)
                    else:
                        Baseline.objects[(item[0], gls)].add_observation()
        # Process and write baseline report
        Baseline.seek_input()
        Baseline.report(name)
        Baseline.export(name)
        Baseline.objects = {}
        return True

    def load(name, restart=False):
        """Load the baseline CLSO file.
        """
        if restart == False:
            try:
                reader = CLSO('baseline_' + str(name), option='r')
                reader.load()
                return True
            except:
                pass
        Baseline.objects = {}
        return True

    def export(name):
        """Export the baseline CLSO file.
        """
        writer = CLSO('baseline_' + str(name), option='w')
        writer.bulk_dump('Baseline')
        return True

    def seek_input():
        """Seek user input on baseline values.
        """
        print('For each gloss please first enter:\n' +
                  '\t0) If it is correct\n' +
                  '\t1) If it is misspelled\n' +
                  '\t2) If it is confused with another gloss\n' +
                  '\t3) If it is a concept that needs to be glossed\n' +
                  '\t4) If it is a combination that should be divided\n' +
                  '\t5) If it is ambiguous\n' +
                  '\t6) If it is identified only by the user\n' +
                  '\t7) If it is an unrecoverable data entry error\n' +
                  '\t8) If it is a part-of-speech tag and not a word\n' +
                  '\t9) If it is a word and not a gloss\n')
        iso = ''
        for gloss in sorted(Baseline.objects.keys()):
            if iso != gloss[0]:
                print('Language: ' + str(gloss[0]))
                iso = gloss[0]
            valid = False
            while not valid:
                error = input('\t' + gloss[1] + ': ')
                valid = True
                if str(error) == '0':
                    Baseline.objects[gloss]._error_type = 'correct'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                elif str(error) == '1':
                    Baseline.objects[gloss]._error_type = 'misspelled'
                    Baseline.objects[gloss]._standard = Baseline.seek_standard(
                                                        gloss[1])
                elif str(error) == '2':
                    Baseline.objects[gloss]._error_type = 'confused'
                    Baseline.objects[gloss]._standard = Baseline.seek_standard(
                                                        gloss[1])
                elif str(error) == '3':
                    Baseline.objects[gloss]._error_type = 'incomplete'
                    Baseline.objects[gloss]._standard = Baseline.seek_standard(
                                                        gloss[1], repeat=True)
                elif str(error) == '4':
                    Baseline.objects[gloss]._error_type = 'combined'
                    Baseline.objects[gloss]._standard = Baseline.seek_standard(
                                                        gloss[1], repeat=True)
                elif str(error) == '5':
                    Baseline.objects[gloss]._error_type = 'ambiguous'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                elif str(error) == '6':
                    Baseline.objects[gloss]._error_type = 'user-identified'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                elif str(error) == '7':
                    Baseline.objects[gloss]._error_type = 'unrecovered'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                elif str(error) == '8':
                    Baseline.objects[gloss]._error_type = 'part-of-speech'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                elif str(error) == '9':
                    Baseline.objects[gloss]._error_type = 'word'
                    Baseline.objects[gloss]._standard = gloss[1]
                    print()
                else:
                    print('\t\tYou have entered an unrecognized value, try' +
                          ' again.')
                    valid = False
                if valid == True and str(error) != '9' and str(error) != '8':
                    Baseline.report_error(gloss[0], error)
        return True

    def seek_standard(gloss, repeat=False):
        """Ask the user to enter the standard value for the gloss.
        """
        if repeat == True:
            s_list = []
            standard = ''
            while str(standard) != '0':
                standard = input('\tWhat is one of the glosses for ' +
                                 str(gloss) + ' (use 0 if no more values): ')
                if str(standard) != '0' and standard != '':
                    s_list.append(standard)
            standard = s_list
        else:
            standard = input('\tWhat is the correct value for ' + str(gloss) +
                             ': ')
        print()
        return standard

    def _run_statistics():
        """Report errors to the langauges DS for reporting.
        """
        Baseline.languages = {'Aggregate': {}} 
        for key in Baseline.objects: 
            if key[0] not in Baseline.languages:
                Baseline.languages[key[0]] = {}
            if Baseline.objects[key]._error_type == 'correct':
                code = '0'
            elif Baseline.objects[key]._error_type == 'misspelled':
                code = '1'
            elif Baseline.objects[key]._error_type == 'confused':
                code = '2'
            elif Baseline.objects[key]._error_type == 'incomplete':
                code = '3'
            elif Baseline.objects[key]._error_type == 'combined':
                code = '4'
            elif Baseline.objects[key]._error_type == 'ambiguous':
                code = '5'
            elif Baseline.objects[key]._error_type == 'user-identified':
                code = '6'
            elif Baseline.objects[key]._error_type == 'unrecovered':
                code = '7'
            else:
                continue
            Baseline.languages[key[0]][code] = Baseline.languages[key[0]].get(
                                                                code, 0) + 1
            Baseline.languages['Aggregate'][code] = Baseline.languages[
                                                'Aggregate'].get(code, 0) + 1
        return True

    def report(name):
        """Report all of the Baseline statistics to file.
        """
        Baseline._run_statistics()
        file = open('baseline_' + str(name), 'w')
        file.write('Baseline Statistics for: ' + str(name) + '\n\n')
        iso_map = {}
        for iso in Baseline.languages:
            iso_map[iso] = True
        del iso_map['Aggregate']
        iso_list = ['Aggregate'] + sorted(iso_map.keys())
        for iso in iso_list:
            total = 0.0
            for error in Baseline.languages[iso]:
                total += Baseline.languages[iso][error]
            err = {}
            for error in Baseline.languages[iso]:
                err[error] = Baseline.languages[iso][error] / total
            file.write(str(iso) + ' Statistics:\n')
            if '0' in Baseline.languages[iso]:
                file.write('\t0 - Correct Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['0'], err['0']))
            if '1' in Baseline.languages[iso]:
                file.write('\t1 - Misspelled Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['1'], err['1']))
            if '2' in Baseline.languages[iso]:
                file.write('\t2 - Confused Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['2'], err['2']))
            if '3' in Baseline.languages[iso]:
                file.write('\t3 - Incomplete Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['3'], err['3']))
            if '4' in Baseline.languages[iso]:
                file.write('\t4 - Combined Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['4'], err['4']))
            if '5' in Baseline.languages[iso]:
                file.write('\t5 - Ambiguous Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['5'], err['5']))
            if '6' in Baseline.languages[iso]:
                file.write('\t6 - Authored Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['6'], err['6']))
            if '7' in Baseline.languages[iso]:
                file.write('\t7 - Unrecovered Gloss: {0} | {1}%\n'.format(
                    Baseline.languages[iso]['7'], err['7']))
            file.write('\n')
        file.close()
        return True

    def delete_baseline(iso, gloss):
        """Delete a baseline value.

        Keyword arguments:
          iso -- the iso value of the language
          gloss -- the value of the baseline gloss
        """
        del Baseline.objects[(iso, gloss)]
        return True

    def add_observation(self):
        """Add an entry to the observation set for the iso_code and
        gloss.
        """
        # Increment occurrence count for the gloss
        self._observed += 1
        return True

import random
import string

class Joint:

    objects = {}

    def __init__(self):
        self.glosses = {}
        self.joints = {}
        while True:
            ID = ''.join(random.SystemRandom().choice(string.ascii_uppercase +
                                            string.digits) for _ in range(8))
            if ID not in Joint.objects:
                self._ID = ID
                Joint.objects[ID] = self
                break

    def set_joint(self, gloss1, gloss2, cond1=0, cond2=0, joint=0):
        glosses = [gloss1, gloss2]
        glosses_sort = sorted(glosses)
        if glosses != glosses_sort:
            temp = cond1
            cond1 = cond2
            cond2 = temp
        joint_gloss = JointGloss(self._ID, glosses_sort[0], glosses_sort[1])
        self.glosses[(glosses_sort[0], glosses_sort[1])] = joint_gloss
        self.joints[joint_gloss] = True
        if cond1:
            joint_gloss.set_conditional_1(cond1)
        if cond2:
            joint_gloss.set_conditional_2(cond2)
        if joint:
            joint_gloss.set_joint(joint)
        if cond1 and cond2 and joint:
            joint_gloss.set_probability()
        return joint_gloss

    def get_probability(self, gloss1, gloss2):
        glosses = [gloss1, gloss2]
        glosses = sorted(glosses)
        return JointGloss.objects[self._ID, glosses[0], glosses[1]].probability

class JointGloss:

    objects = {}

    def __init__(self, ID, gloss1, gloss2):
        self._ID = ID
        self._gloss1 = gloss1
        self._gloss2 = gloss2
        JointGloss.objects[ID, gloss1, gloss2] = self

    def set_conditional_1(self, prob):
        self.conditional_1 = prob
        return True

    def set_conditional_2(self, prob):
        self.conditional_2 = prob
        return True

    def set_joint(self, prob):
        self.joint = prob
        return True

    def set_probability(self):
        try:
            #self.probability = (self.conditional_1 * self.conditional_2 *
            #                   self.joint)
            self.probability = self.joint
        except ValueError:
            print("Cannot set probability for joint glosses because " +
                  "conditional and/or joint probabilities are unknown")
        return True

class Training:
    
    objects = {} #(iso, gloss)

    languages = 8
    trained_glosses = {"IMPF": {"yak": True, "rus": True},
                       "PRF": {"yak": True, "rus": True, "sna": True},
                       "IMP": {},
                       "1": {"yak": True, "rus": True, "sna": True,
                             "eng": True, "pbt": True, "ang": True,
                             "heb": True, "ara": True},
                       "2": {"yak": True, "rus": True, "sna": True,
                             "eng": True, "pbt": True, "ang": True,
                             "heb": True, "ara": True},
                       "3": {"yak": True, "rus": True, "sna": True,
                             "eng": True, "pbt": True, "ang": True,
                             "heb": True, "ara": True}}

    distances = {"IMPF": {1:1, 2:1},
                 "PRF": {0:2, 1:1},
                 "IMP": {},
                 "1": {0:8},
                 "2": {0:8},
                 "3": {0:8}}

    def set_individual(trained_glosses, distances, languages):
        M = {}
        for gloss in distances:
            for dist in distances[gloss]:
                distances[gloss][dist] = (distances[gloss].get(dist, 0) /
                                          (len(trained_glosses[gloss])+1.0))
            distances[gloss]['<default>'] = 1 / (len(trained_glosses[gloss])+1.0)
        for gloss in trained_glosses:
            M[gloss] = ((len(trained_glosses)+1)/(languages+1.0), distances[gloss])
        return M

    def set_joint(trained_glosses, languages):
        J = Joint()
        for G1 in trained_glosses:
            for G2 in trained_glosses:
                if G1 == G2:
                    continue
                G1_and_G2 = 1.0 # Laplace Smoothing
                for iso in trained_glosses[G1]:
                    if iso in trained_glosses[G2]:
                        G1_and_G2 += 1
                J.set_joint(G1, G2,
                            cond1=(G1_and_G2/(len(trained_glosses[G2])+1)),
                            cond2=(G1_and_G2/(len(trained_glosses[G1])+1)),
                            joint=(G1_and_G2/(languages+1)))
        return J
    """
    M = set_individual(trained_glosses, distances, languages)
    for gloss in M:
        print(gloss, M[gloss])

    J = set_joint(trained_glosses, languages)
    for gloss in J.joints:
        print(gloss._gloss1, '\t', gloss._gloss2, '\t', gloss.probability)

    print('\n\n\n')

    N = ("IMP", "PRF", "2")
    """

class BeamSearch:

    objects = {}
    memoized = {}

    def __init__(self, N, M, joint):
        self._N = N
        self._M = M
        self._joint = joint
        while True:
            ID = ''.join(random.SystemRandom().choice(string.ascii_uppercase +
                                            string.digits) for _ in range(8))
            if ID not in BeamSearch.objects:
                self._ID = ID
                BeamSearch.objects[ID] = self
                BeamSearch.memoized[ID] = ((), 0)
                break
        self.search(self._N, self._M, self._joint, (), 1)
        print(BeamSearch.memoized[self._ID])

    def search(self, N, M, joint, path, prev):
        for m in M:
            if m not in path:
                prob = (M[m][0] * levenshtein_probability(N[0], m, M[m][1]) *
                        prev)
                #for gloss in path:
                #    prob = prob * joint.get_probability(m, gloss)
                if prob > BeamSearch.memoized[self._ID][1]:
                    if len(N) > 1:
                        self.search(N[1:], M, joint, path + (m,), prob)
                    else:
                        BeamSearch.memoized[self._ID] = (path + (m,), prob)
        return BeamSearch.memoized[self._ID]

    #obj = BeamSearch(N, M, J)
