#!/opt/python-2.7/bin/python2.7
# -*- coding: utf-8 -*-
'''
@author David Wax - waxd@u.washington.edu
'''
import sys, os, re, tarfile, random, argparse, copy, shutil;
import igt;
import pickle;
import time #debugging


#############
# Arguments
#############
parser = argparse.ArgumentParser(description="Loads a settings file and run a set of tests.");

parser.add_argument("settings_file_name", help="The name of the configuration file.");

args = parser.parse_args();

#############
# Classes
#############

class Setting_Interface:
    '''
    @author David Wax
    
    This Interface allows for safe(ish) manipulation of settings.
    '''
    
    def get(self, key):
        '''
        @param key (String) - The key value of the setting
        @return List - The list of values matching the key
        
        This returns the list of values associated with the key
        @throw AssertionError if the value isn't a list.
        '''
        assert isinstance(self.settings[key], list);
        return self.settings[key];

    def set(self, key, value, warn=True):
        '''
        @param key (String) - One of the settings to be changed
        @param value (List or object) - The value the setting will be changed to
        @return True if the value was correctly added
                False if the value was not added.
   Ã¢     
        This sets the settings for key to the list provided or a list created from
        the single object provided.
        '''
        if key not in self.settings:
            if warn:
                sys.stderr.write("Warning: key \""+str(key)+"\"value \""+str(value)+"\" pair was not added to the Settings.\n"); 
            return False;
        if not isinstance(value, list):
            self.settings[key] = [value];
        else:
            self.settings[key] = value;
        return True;

    def add(self, key, value, strict=True):
        if key not in self.settings:
            if strict:
                raise Exception("key \""+str(key)+"\" is not valid.");
            return False;
        if value == "None":
            value = None;
        elif value == "True":
            value = True;
        elif value == "False":
            value = False;
        self.get(key).append(value);
        return True;

class Action(Setting_Interface):
    def run_test(self, igt_library, actions_list, iso, results, fout, tarball):
        if len(self.get_states()) > 1:
            print ("Saving "+str(self.get_states(iso)));
            saved_state = igt_library.pickle();
            print ("Saved State as:"+saved_state);
        else:
            saved_state = None;
        #loop though actions
        states = self.get_states(iso)
        for state in self.get_states(iso):
            if saved_state:
                print ("Loading state:"+saved_state);
                start = time.time()
                igt_library.load(saved_state);
                end = time.time()
            print ("Running Settings:"+str(state));
            start = time.time()
            result = results+self.execute(igt_library, state, tarball);
            end = time.time()
            if actions_list:
                start = time.time()
                actions_list[0].run_test(igt_library, actions_list[1:], iso, result, fout, tarball);
                end = time.time()
            else:
                start = time.time()
                fout.write(result);
                fout.write("\n");
                end = time.time()
        if saved_state and os.path.exists(saved_state):
            start = time.time()
            os.remove(saved_state);
            end = time.time()

    def get_states(self, iso="iso", keys=None):
        '''
        This function returns an array of dictionaries which represents the possible
        states of an action.
        '''
        if keys is None:
            keys = list(self.settings.keys());
            if "source" in keys:
                keys.remove("target");
            return self.get_states(iso, keys);
        elif len(keys) == 0:
            return [{}];
        else:
            key = keys[0];
            states = [];
            for data in self.get(key):
                for state in self.get_states(iso, keys[1:]):
                    if key == "source":
                        index = self.get(key).index(data);
                        state = dict({"target":self.get("target")[index]}.items() |state.items());
                    if isinstance(data, str):
                        states.append(dict({key:re.sub("\$iso",iso,data)}.items() | state.items()));
                    else:
                        states.append(dict({key:data}.items() | state.items()));
            return states;


class Pos_Tag(Action):
    def __init__(self, tag):
        tags = tag.split(',')
        self.settings = {"tags": tags,
                         "iso": ["$iso"],
                         }

    #def verify(self):
    #    if self.get("tag")[0] != "V":
    #        sys.stderr.write("Only verbs supported for POS tagging currently.\n");
    #        exit(1);

    def print_header(self, fout=sys.stdout):
        fout.write(",POS Tagging");

    def print_settings_header(self, fout=sys.stdout):
        fout.write(",POS");

    def execute(self, igt_library, state, tarball): 
        #tags = self.settings['tags']
        tag = state['tags']
        igt_library.mark_pos(state["iso"], tag);
        return ","+state["tags"];
       
class Output(Action):
    def __init__(self, name):
        self.settings = {"name": [name],
                         "compression": [],
                         "wordlist": [],
                         "output iso": ["$iso"],
                         "groups":["training"],
                         "iso":["$iso"],
                         "language":[],
                         };
        self.output_files_count = 0

    def verify(self):
        '''
        This function converts items in the settings which need to be changed to
        integers (or throws an error), and fills in defaults.
        '''
        compression = [];
        for value in self.get("compression"):
            if value is not None:
                try:
                    compression.append(float(value));
                except ValueError as e:
                    sys.stderr.write(e);
                    exit(1);
            else:
                compression.append(value);
        self.set("compression",compression);
        if not self.get("language"):
            self.set("language", None);
 
    def print_header(self, fout=sys.stdout):
        fout.write(","+self.get("name")[0]+",");

    def print_settings_header(self, fout=sys.stdout):
        fout.write(",compression,choices name");

    def execute(self, igt_library, state, tarball):
        start = time.time()
        choices_file = igt_library.make_choices(state["output iso"], state["groups"]);
        #choices_file.lexicon.remove_grandchild_inputs();
        if state["compression"]:
            choices_file.lexicon.compact_lexicon(state["compression"]);
        choices_file_name = igt_library.print_choices(
                state["output iso"], choices_file, state["language"], 
                state["groups"], None);
        wordlist_file_name = "wordlist"
        wordlist_file = open(wordlist_file_name, "w");
        igt_library.print_words(state["iso"],flag=igt.LANGUAGE, tag="verb", groups=state["groups"], file=wordlist_file);
        wordlist_file.close();
        output_filename = "choices" + str(self.output_files_count)
        self.output_files_count = self.output_files_count + 1
        shutil.copyfile(choices_file_name, output_filename)
        #print "TAR:", os.path.split(wordlist_file_name)[1]
        #print os.path.split(choices_file_name)[1]
        #print output_filename
        #tarball.add(os.path.split(wordlist_file_name)[1]);
        #tarball.add(os.path.split(choices_file_name)[1]);
        tarball.add(output_filename)
        os.remove(wordlist_file_name);
        os.remove(choices_file_name);
        os.remove(output_filename)
        end = time.time()
        print ("Done! Output class execute method took: " + str(end - start))
        return ","+str(state["compression"])+","+choices_file_name.split("/")[-1];

class Test(Setting_Interface):
    '''
    @author David Wax

    This is a dictionary which only stores a specific set of keys
    used to run a test.
    '''
    actions = ["pos tags", "output"]
    def __init__(self, filename):
        '''
        @param String filename - The name of the configuration file to load.
        @return Test
        '''
        start = time.time() #debugging 
        self.settings = {"iso": [],
                         "data directory": [],
                         "file extension": [],
                         "output id": [],
                         "output dir": [],
                         "hyphens": [],
                         "actions": [],                
                         }

        file = open(filename, "r");
        for line in file:
            line = line.split("#",1)[0].strip();
            if line:
                ''' allow colons in lines, such as regular expressions'''
                (setting, values) = (line.split(":",1)[0].strip(), line.split(":",1)[1].strip())
                setting = setting.strip();
                if setting in self.settings:
                    for value in values.split(","):
                        value = value.strip();
                        if not self.add(setting, value):
                            raise Exception("Setting:"+setting+" value:"+value+" pair invalid.");
                elif setting in self.actions:
                    if setting == "pos tags":
                         self.add("actions", Pos_Tag(values));
                    elif setting == "output":
                        self.add("actions", Output(values));
                else:
                    for value in values.split(","):
                        value = value.strip();
                        if not self.last_action().add(setting, value):
                            raise Exception("Setting:"+setting+" value:"+value+" pair invalid.");
        if not self.settings["iso"]:
            raise Exception("Configuration file missing iso declariation.");
        if not self.settings["data directory"]:
            raise Exception("Configuration file missing data directory declariation.");
        if not self.get("file extension"):
            self.set("file extension", ".check");
        if len(self.get("output id")) == 0:
            self.set("output id", None);
        if len(self.get("output dir")) == 0:
            self.set("output dir", None);
        elif len(self.get("output id")) > 1:
            sys.stderr.write("Multiple value for output id not allowed\n");
            exit(1);
        #for action in self.get("actions"):
        #    action.verify();
        end = time.time()

    def last_action(self):
        return self.get("actions")[-1];

    def print_header(self, fout=sys.stdout):
        fout.write("General Settings,");
        for action in self.get("actions"):
            action.print_header(fout);
        fout.write("\n");

    def print_settings_header(self, fout=sys.stdout):
        fout.write("iso,percent");
        for action in self.get("actions"):
            action.print_settings_header(fout);
        fout.write("\n");

    def run_test(self, igt_library, fout, tarball):
        self.print_header(fout);
        self.print_settings_header(fout);
        start = time.time()
        igt_library.process();
        end = time.time()
        if len(self.get("iso")) > 1:
            print ("Saving Initial State");
            start = time.time() 
            saved_state = igt_library.pickle();
            end = time.time()
            print ("Saved State as:",saved_state);
        else:
            saved_state = None;
        for iso in self.get("iso"):
            if saved_state:
                print ("Loading Initial State");
                start = time.time()
                igt_library.load(saved_state);
                end = time.time()
            results = iso
            start = time.time()
            self.get("actions")[0].run_test(igt_library, self.get("actions")[1:], iso, results, fout, tarball);
            end = time.time()
        if saved_state and os.path.exists(saved_state):
            start = time.time()
            os.remove(saved_state);
            end = time.time()

#############
# Main
#############

def main(args):
    '''
    This method loads the settings file in the args, and runs the tests
    defined by that file.
    '''
    print ("Loading Configuration File",args.settings_file_name);
    settings = Test(args.settings_file_name);
    fout_dir = settings.get("output dir")[0];
    fout_id = settings.get("output id")[0];
    if fout_dir is None:
        fout_dir = "./";
    elif str(fout_dir)[len(str(fout_dir))-1] is not '//':
        fout_dir = str(fout_dir) + '/';
    filescount = 0
    if fout_id is None:
        fout_id = filescount;
    tar_name = str(fout_dir)+"choices"+str(fout_id)+str(filescount)+".tar";
    while os.path.exists(tar_name):
        filescount += 1;
        tar_name = str(fout_dir)+"choices"+str(fout_id)+str(filescount)+".tar";
    fout_name = "results";
    results_file_name = str(fout_dir)+fout_name+str(fout_id)+".csv";
    if os.path.exists(results_file_name):
        bonus_id = 0;
        results_file_name = str(fout_dir)+fout_name+str(fout_id)+"-"+str(bonus_id)+".csv";
        while os.path.exists(results_file_name):
            bonus_id += 1;
            results_file_name = str(fout_dir)+fout_name+str(fout_id)+"-"+str(bonus_id)+".csv";
    hyphens = settings.get("hyphens")[0]
    igt_library = igt.IGT_library(hyphens = hyphens);
    igt_library.read_files(settings.get("data directory")[0], settings.get("file extension")[0]);
    fout = open(results_file_name, "w");
    tarball = tarfile.open(tar_name, "a");
    settings.run_test(igt_library, fout, tarball);
    fout.close();
    tarball.add(results_file_name, arcname=os.path.split(results_file_name)[1])
    os.remove(results_file_name);
    tarball.close();

#This is the only call at the top level of the script
#and should be left as a means of running any main functions
main(args);
