#!/usr/bin/perl
#########################################
## texttoplain.pl
## 
## this script takes tex files with igt 
## in the 'covington' format (on STDIN)
## and outputs just the IGT in a plain 
## format
##
## step one:
##  * remove TIPA codes and replace with utf-8
##  * so far this is just \textbeltl -> ɬ
## step two:
##  * convert covington format to plain-text
##    triplets
##  
## the 'covington' format looks like this:
## \gll expects two lines separated by CR
## with the same number of tokens, tokens
## can be grouped by { }, eg:
##
## \gll foreign words by whitespace
## translated words by {white space}
##
## then the 'free translation' line is marked
## by \glt
## \glend closes the environment
use warnings; 
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
sub replace_tipa($);
my $total = 0;
my $warnings = 0;
while($line = <>){
  # first step is to ignore lines not in a \gll 
  if ($line=~m/\\gll/){
    $total++;
    # print STDERR "found a gll group:\n";
    # found one, feed the lines into an array until 
    # we find \glend
    my @lines;
    push(@lines, replace_tipa($line));
    $next = <>;
    while ($next!~m/\\glend/){
      push(@lines, $next);
      $next = <>;
    }
    # okay, we've got a group of lines which is a \gll
    # and TIPA is already removed
    # next step is to 
    #   ** validate the lines, 
    #   ** complain if they aren't valid
    @lines = validate_igt(\@lines);
    # now we should be able to just print a 
    # plain text igt to STDOUT
    for (@lines){
      print "@$_\n";
    }
    print "\n";
  } else {
    next;
  }
}
print STDERR "---\n";
print STDERR "total number of IGT instances in this file: $total\n";
print STDERR "$warnings generated warnings.\n";

sub replace_tipa($) {
  if ($_[0]=~/\\textipa{(\\\w+)?}/){ 
#    print STDERR "texttipa found on line $.: char=$1\n";
    my $replace = '';
    if ($1 eq '\textbeltl'){
      $replace = "\x{026c}"; 
    }
#    print STDERR "  replacement is $replace\n"; 
    $_[0]=~s/\\textipa{[\w\\]+?}/$replace/g;
  }
  return $_[0];
}

sub validate_igt{
 # here we just want to make sure that 
 # there are three lines and that 
 # we can properly tokenize them
 my (@lines) = @{$_[0]};
 if (scalar(@lines) != 3){
   print STDERR "warning: igt group doesn't contain 3 lines:\n@lines\n";
   $warnings++;
 }

 # language line
 @l1 = split (/\s+/, $lines[0]);
 $dis = shift @l1;
 if ($dis ne '\gll'){
   print STDERR "warning: first token of language line should be \\gll\n, found $dis\n";
   $warnings++;
 }
 @l1 = &parans_for_dots(\@l1);

 # gloss line
 ## has to be split with respect to { }
 ## we can join the words in { } with .
 @l2 = split (/\s+/, $lines[1]);
 @l2 = &parans_for_dots(\@l2);
 # print "@l2";
 if (scalar(@l2) != scalar(@l1)){
   print STDERR "warning: length of language and gloss lines unequal!\n  l1 [".scalar(@l1)."]: @l1\n  l2 [".scalar(@l2)."]: @l2\n";
   $warnings++;
 } 

 # free translation line
 ## just remove the \glt and done
 @l3 = split (/\s+/, $lines[2]);
 $dis = shift @l3;
 if ($dis ne '\glt'){
   print STDERR "warning: first token of free translation line should be \\glt\n, found $dis\n";
 }

 return (\@l1,\@l2,\@l3);
}

sub parans_for_dots {
  my(@line) = @{$_[0]}; 
  my @to_return;
  my $i;
  for ($i=0;$i<=$#line;$i++){
    if ($line[$i]=~m/{/){
      # remove opening bracket (sometimes brackets are doubled)
      my $n = 0;

      # transform $line[$i] to a char array
      my @chars = split("", $line[$i]);

      my $len = $#chars + 1;
#      print STDERR length($line[$i]);
#      print STDERR "line[i] still exists: $line[$i]\n";
      while (($n < $len) && ($chars[$n] eq '{')){
        $n++;
      }
      $line[$i] = substr $line[$i], $n;  


#      print STDERR "altered line[$i] still exists: $line[$i]\n";



      # variable to_join is array of enclosed items
      my @to_join;
      while ($line[$i]!~m/}/){
        #print STDERR "this is not the last to token, keep joining\n";
        push (@to_join, $line[$i]);
        #print STDERR "joining $line[$i] with @to_join\n";
        $i+=1;
      }

      # remove closing bracket (sometimes they're doubled)
      $n = -1;

      # transform $line[$i] to a char array
      @chars = split("",$line[$i]);
      $len = (-1 * $#chars); # inverted length!
      while (($n > $len) && ($chars[$n] eq '}')){
        $n--;
      }
      $line[$i] = substr $line[$i],0, $n+1; 


      # put the final token in
      push (@to_join, $line[$i]);

      # put the joined words into the to_return array
      push (@to_return, join(".",@to_join));

    } else {
      push (@to_return, $line[$i]);
    }
  }
  return @to_return;
}
