#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Authors: Kristen Howell, Olga Zamaraeva

# This scripts reads in a xigt corpus and a param file, which
# specifies which options to use for various methods like find_case, and returns a choices file.

# -*- coding: utf-8 -*-
import sys,os
from mom import MatrixChoices, Choices, MainClasses
from XigtifiedToolbox import XigtReader, MorphTypeGuesser, FeatDict
from xigt.codecs import xigtxml
from inference_scripts import infer_arg_opt, infer_case_frame_from_ds, infer_case_from_counts_xigt, \
    infer_case_from_grams_xigt, infer_lex_items, infer_person, infer_sent_neg, infer_wo_from_ds, infer_coordination, infer_adnominal_possession, \
    filter_lex_rules
from aggutils.deffile import MatrixDefFile
from aggutils import reenrich

parent_dir = os.path.join(os.getcwd(),os.pardir)
sys.path.append(parent_dir)
import re
import operator

def get_valid_lines(lines):
  """
  Return all lines that are not empty. Also, strip out byte order marks
  from the first line if it has one.
  """
  if len(lines) > 0:
    lines[0] = lines[0].strip()
  lines = [l.strip() for l in lines if l.strip() != '']
  return lines

def build_param_dict(lines, all_params):
    """
    Check each line for a valid parameter. Add all paramaters to
    param dict
    """
    params = {}
    for line in lines:
        (param, value) = line.split(':',1)
        if param in all_params:
            params[param] = value.strip()
    return params

def add_basic_choices(choices, config_file):
    choices.__setitem__('version', '32')
    choices.__setitem__('language',  MainClasses.Settings(config_file).language)
    choices.__setitem__('iso-code',  MainClasses.Settings(config_file).iso)
    return choices

def extract_case_system(xigt_corpus, method):
    """
    call infer_case method according to a given method: sao or gram
    """
    if method == 'gram':
        if __name__ == '__main__':
            case_system = infer_case_from_grams_xigt.infer_case(xigt_corpus)
        # KPH map split-erg to erg-abs, because we have no heuristic to detirmine what the split is on
        if case_system == 'split-erg':
            case_system = 'erg-abs'
        if case_system.lower() == 'nom-acc':
            case_grams = {'s': 'nom', 'a': 'nom', 'o': 'acc'}
        elif case_system.lower() == 'erg-abs':
            case_grams = {'s': 'abs', 'a': 'erg', 'o': 'abs'}
        elif case_system.lower() == 'none':
            case_grams = {'s': '', 'a': '', 'o': ''}
        elif case_system.lower() == 'split-erg':
            case_grams = {'s': 'nom', 'a': 'erg', 'o': 'acc'}
    elif method == 'sao':
        if __name__ == '__main__':
            sao_output = infer_case_from_counts_xigt.infer_case(xigt_corpus)
        case_grams = {'s': sao_output['s'], 'a': sao_output['a'], 'o': sao_output['o']}
        case_system = sao_output['case']
    return(case_system, case_grams)

def my_extract_case_frame(case_grams, arg_opt_choices, agr_morphemes, igt2items, xc, config_file):
    pos_tags = MainClasses.Settings(config_file).verb_tags
    options = {}
    options['corpus'] = corpus_file
    options['case_names'] = case_grams
    obj_drop = False
    if 'obj-drop' in arg_opt_choices:
        obj_drop = True
    options['obj_drop'] = obj_drop
    verbs = infer_case_frame_from_ds.extract_case_frame(options, agr_morphemes, igt2items, pos_tags, xc)

    return verbs

#def my_extract_gender(xigt_corpus):
#    genders = infer_gender.infer_gender(xigt_corpus)
#    return genders

def add_genders_to_choices(genders, choices):
    n = 1
    for gender in genders:
        #print(gender)
        choices.__setitem__(('gender' + str(n) + '_name'), gender)
        n += 1
    return choices

def my_infer_arg_opt(xigt_corpus, config_file):
    verb_tags = MainClasses.Settings(config_file).verb_tags
    arg_opt_choices, agr_morphemes = infer_arg_opt.infer_arg_opt(xigt_corpus, verb_tags=verb_tags)
    # KPH since we don't identify which verbs are lexically lisenced for pro-drop, map lexically lisenced to all
    if arg_opt_choices['obj-drop'] == 'obj-drop-lex':
        arg_opt_choices['obj-drop'] = 'obj-drop-all'
    return arg_opt_choices, agr_morphemes

def add_arg_opt_to_choices(arg_opt_choices, choices):
    for choice in arg_opt_choices:
        choices[choice] = arg_opt_choices[choice]
    choices = infer_arg_opt.add_argument_marking_constraints(choices)
    return choices

def my_infer_sent_neg(xigt_corpus, auxiliaries, igt2items):
    sent_neg_choices, negation_auxiliaries = infer_sent_neg.infer_sent_neg(xigt_corpus, auxiliaries, igt2items)
    return sent_neg_choices, negation_auxiliaries

def add_sent_neg_to_choices(sent_neg_choices, choices):
    for choice in sent_neg_choices:
        choices[choice] = sent_neg_choices[choice]
    return choices

def my_infer_word_order(xigt_corpus, verbs, case_system, igt2items, config_file, known_glosses):
    verb_tags = MainClasses.Settings(config_file).verb_tags
    wo_choices, auxiliaries = infer_wo_from_ds.infer_word_order(xigt_corpus, verbs, case_system, igt2items, verb_tags, known_glosses)
    return wo_choices, auxiliaries

def add_word_order_to_choices(wo_choices, choices):
    for choice in wo_choices:
        choices[choice] = wo_choices[choice]
    return choices

def my_infer_coord(xc, config_file, subj_drop_choice, word_order, known_glosses):
    verb_tags = MainClasses.Settings(config_file).verb_tags
    noun_tags = MainClasses.Settings(config_file).noun_tags
    if subj_drop_choice == 'all':
        subj_drop = True
    else:
        subj_drop = False
    coord_choices = infer_coordination.infer_coordination(xc, verb_tags, noun_tags, subj_drop, word_order, known_glosses)
    return coord_choices

def add_coord_to_choices(coord_choices, choices):
    for choice in coord_choices:
        choices[choice] = coord_choices[choice]
    return choices
 
def my_infer_adnom_poss(xc, feats, mainchoices):
    poss_choices = infer_adnominal_possession.infer_adnom_poss(xc, feats, mainchoices)
    return poss_choices

def add_adnom_poss_to_choices(poss_choices, choices):
    for choice in poss_choices:
        choices[choice] = poss_choices[choice]
    return choices


def read_in_xigt(settings):
    xigt_reader = XigtReader.Xigt_Reader(settings, verbose=False)
    xigt_reader.process_igts(settings.data_file)
    xigt_reader.print_skipped(settings.out_dir + '/skipped_items.txt')
    return xigt_reader

def run_mom(settings, xigt_reader, agr_morphemes=None):
    filescount = 0
    fout_name, tar_name, arc_name = MainClasses.set_filenames(filescount, settings)
    mom = MainClasses.MOM(name='AGG-MOM', output_dir=settings.out_dir,
                              tag_sets=settings.pos_tags, lexitem_classes=settings.lexitem_classes,
                              precluster_only=settings.precluster_only)
    choices = mom.execute(pos_items=xigt_reader.items[settings.pos],settings=settings,
                            patterns=settings.patterns,tarball=None,arc_name=arc_name,
                            ignore_affix_types=settings.ignore_affixes,verbose=False,pos_tag=settings.pos,
                            escape_special_characters=settings.escape_special_characters,agr_morphemes=agr_morphemes)
    return choices

def get_features(choices):
    affix_features = collect_affix_features(choices)
    lexical_features = collect_lexical_features(choices)
    features = affix_features.copy()
    for feat in lexical_features:
        if feat in features:
            features[feat].update(lexical_features[feat])
        else:
            features[feat] = lexical_features[feat]
    #features.update(lexical_features)
    return(features)

def collect_affix_features(choices):
    affix_features = {}
    pos_list = ['verb-pc', 'noun-pc']
    for pos in pos_list:
        for pc in choices[pos]:
            for lrt in pc['lrt']:
                for feat in lrt['feat']:
                    values = feat['value'].split(', ')
                    if feat['name'] not in affix_features:
                        affix_features[feat['name']] = set()
                    for value in values:
                        if value not in affix_features[feat['name']]:
                            affix_features[feat['name']].add(value)
                    for lri in lrt['lri']:
                        for feat in lri['feat']:
                            values = feat['value'].split(', ')
                            if feat['name'] not in affix_features:
                                affix_features[feat['name']] = set()
                            for value in values:
                                affix_features[feat['name']].add(value)
    return affix_features

def collect_lexical_features(choices):
    lexical_features = {}
    pos_list = ['verb', 'noun', 'aux', 'det']
    for pos in pos_list:
        if pos in choices:
            for word in choices[pos]:
                for feat in word['feat']:
                    values = feat['value'].split(', ')
                    if feat['name'] not in lexical_features:
                        lexical_features[feat['name']] = set()
                    for value in values:
                        lexical_features[feat['name']].add(value)
    return(lexical_features)


def set_case_choices(choices, case_system, features):
    # we really only handle nom-acc and erg-abs. split-erg is mapped to erg-abs
    if 'case' in features:
        case_features = features['case']
    else:
        case_features = set()
    for adp in choices['adp']:
        for feat in adp['feat']:
            if feat['name'] == 'case':
                if feat['value'] not in case_features:
                    case_features.add(feat['value'])
    # KPH: I want to make sure this is done after aux inference, because we need to know when the case system is none
    if case_system == 'none':# and len(case_features) != 0:
        case_system = 'nom-acc'
        print('case grams found despite no case system. positing nom-acc case-system so the matrix can support these cases. case-frame has not been added to verbs')
    choices['case-marking'] = case_system
    if case_system == 'erg-abs':
        choices['erg-abs-erg-case-name'] = FeatDict.ERGATIVE
        choices['erg-abs-abs-case-name'] = FeatDict.ABSOLUTIVE
        case_features.discard(FeatDict.ERGATIVE)
        case_features.discard(FeatDict.ABSOLUTIVE)
    if case_system == 'nom-acc':
        choices['nom-acc-nom-case-name'] = FeatDict.NOMINATIVE
        choices['nom-acc-acc-case-name'] = FeatDict.ACCUSATIVE
        case_features.discard(FeatDict.NOMINATIVE)
        case_features.discard(FeatDict.ACCUSATIVE)
    if case_system != 'none':
        for i, f in enumerate(list(case_features)):
            if f:
                choices['case' + str(i + 1) + '_name'] = f

def set_png_tam_choices(choices, features):
    # Person
    # Person
    choices['person'] = infer_person.person(features)
    if '1' in choices['person']:
        first_per, num = infer_person.first_person(features)
        choices['first-person'] = first_per
        if first_per == 'incl-excl':
            choices['incl-excl-number'] = num

    # Number
    if 'number' in features:
        i = 1
        for number in features['number']:
            choices['number'+str(i)+'_name'] = number
            i += 1

    # Gender
    if 'gender' in features:
        i = 1
        for gender in features['gender']:
            choices['gender'+str(i)+'_name'] = gender
            i += 1

    # TAM
    TAM = ['tense', 'aspect', 'mood']
    if 'tense' in features:
        choices['tense-definition'] = 'build'
    for type in TAM:
        if type in features:
            i = 1
            for feat in features[type]:
                choices[type + str(i) + '_name'] = feat
                choices[type + str(i) + '_supertype' + str(i) + '_name'] = type
                i += 1

    # FORM
    if 'finite' in features or 'nonfinite' in features:
        choices['form-fin-nf'] = 'on'

    if 'pernum' in features:
        choices = infer_person.update_pernum_feats(choices)

    return choices

def extract_pronouns(choices):
    choices = infer_lex_items.infer_pronouns(choices)
    return choices

def extract_auxiliaries(choices, auxiliaries, negation_auxiliaries):
    choices = infer_lex_items.add_auxiliaries(choices, auxiliaries, negation_auxiliaries)
    return choices

def extract_adpositions(xc, choices, config_file, wo_choices):
    adp_tags = MainClasses.Settings(config_file).adp_tags
    noun_tags = MainClasses.Settings(config_file).noun_tags
    choices = infer_lex_items.extract_adpositions(xc, choices, adp_tags, noun_tags, wo_choices)
    return choices

def add_additional_noun_choices(choices):
    # add 3sg to nouns that aren't pronouns
    # add det optional to nouns that aren't pronouns if dets are allowed
    choices = infer_lex_items.add_common_noun_features(choices)
    return choices

def get_punctuation(choices):
    punct_chars = {'!', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?',
                   '@', '[', ']', '^', '_', '`', '{', '|', '}', '~', '\\'}
    punct = set()
    pos_list = ['verb', 'noun', 'det', 'adp', 'aux']
    for pos in pos_list:
        for word in choices[pos]:
            for stem in word['stem']:
                for char in stem['orth']:
                    if char in punct_chars:
                        punct.add(char)
            for char in word['orth']: # some lexical types (eg. adp) do not put the orth inside a stem
                if char in punct_chars:
                    punct.add(char)
        for pc in choices[pos+'-pc']:
            for lrt in pc['lrt']:
                for lri in lrt['lri']:
                    if 'orth' in lri:
                        for char in lri['orth']:
                            if char in punct_chars:
                                if char == '.':
                                    print()
                                punct.add(char)
    if punct != set():
        choices['punctuation-chars'] = 'keep-list'
        choices['punctuation-chars-list'] = ''.join(punct)
    return choices





# Main routine
all_params = ['infer case', 'data file']
param_file = open(sys.argv[1]).readlines()
param_dict = build_param_dict(get_valid_lines(param_file), all_params)
corpus_file = param_dict['data file']
settings = MainClasses.Settings(sys.argv[1])
# Xigt Reader contains a special data structure of items that were created based
# on the xigtified data. In particular, the items correspond to the word level
# of the sentences (they correspond to words) and they have all the information
# that MOM later is going to need (POS, affixes, gloss, translation, which IGT they came from, etc.)
xigt_reader = read_in_xigt(settings)
choices = MatrixChoices.ChoiceDict()
corpus = open(corpus_file, encoding='utf-8')
xc = xigtxml.load(corpus)
corpus.close()
config_file = sys.argv[1]
known_glosses = MorphTypeGuesser.MorphTypeGuesser(settings.known_glosses_file)

# Call various inference scripts and add to choices file
arg_opt_choices, agr_morphemes = my_infer_arg_opt(xc, config_file)
case_system, case_grams = extract_case_system(xigt_reader.full_xc, param_dict['infer case'])

# Uncomment the below two lines to test on a small file:
#case_system = 'erg-abs'
#case_grams = {'s': 'abs', 'a': 'erg', 'o': 'abs'}
# Here, case inference and transitivity info will be set as attributes on the Xigt Reader objects,
# and that is what MOM will be looking at.
verbs = my_extract_case_frame(case_grams, arg_opt_choices, agr_morphemes, xigt_reader.igt2items, xigt_reader.xc,config_file)
wo_choices, auxiliaries = my_infer_word_order(xc, verbs, case_system, xigt_reader.igt2items, config_file, known_glosses.known_glosses)
coord_choices = my_infer_coord(xc, config_file, arg_opt_choices['subj-drop'], wo_choices['word_order'], known_glosses.known_glosses)
sent_neg_choices, negation_auxiliaries = my_infer_sent_neg(xc, auxiliaries, xigt_reader.igt2items)
#poss_choices = my_infer_adnom_poss()

#genders = my_extract_gender(xigt_reader.xc)
# Now, MOM runs on the Xigt Reader objects, with the transitivity and case frame
# info set, to infer morphotactics. MOM should only merge items with the same or
# underspecified transitivity and case frame. E.g. a transitive item
# should never be assigned an intransitive class.
settings.pos = 'verb'
mom_verb_choices = run_mom(settings,xigt_reader,agr_morphemes=agr_morphemes)
settings.pos = 'noun'
mom_noun_choices = run_mom(settings,xigt_reader)
choices = Choices.build_matrix_choices(mom_verb_choices.lexicon,mom_noun_choices.lexicon, settings)
# This is kind of weird. visualize is a function of lexicon, but it doesn't actually use
# the lexicon, it uses the matrix choices (passed in as the second argument). So it shouldn't
# matter that I'm using mom_verb_choices to call both here

choices = filter_lex_rules.filter_reduntant_zero_rules(choices)
#choices = add_genders_to_choices(genders, choices)
choices = add_arg_opt_to_choices(arg_opt_choices, choices)
choices = add_sent_neg_to_choices(sent_neg_choices, choices)
choices = add_word_order_to_choices(wo_choices, choices)
choices = add_coord_to_choices(coord_choices, choices)
choices = extract_pronouns(choices)
choices = add_additional_noun_choices(choices)
choices = extract_adpositions(xc, choices, config_file, wo_choices)
choices = extract_auxiliaries(choices, auxiliaries, negation_auxiliaries)
features = get_features(choices)
set_case_choices(choices, case_system, features)
choices = set_png_tam_choices(choices, features)
choices = get_punctuation(choices)
adnom_poss_choices, choices = my_infer_adnom_poss(xc, features, choices)
choices = add_adnom_poss_to_choices(adnom_poss_choices, choices)

choices = add_basic_choices(choices, config_file)

# write choices file
print('OUTPUT DIR: {0}'.format(settings.out_dir))
matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
choices_file = settings.out_dir + 'latest_choices'
formatted_choices = MatrixDefFile('./src/aggutils/matrixdef')

formatted_choices.save_choices(matrix_choices,choices_file)
#with open(settings.out_dir + 'latest_choices', 'w', encoding="UTF-8") as f:
    #mom_choices.print_to(file=f)
    # The below will simply print out the lexicon
    # and position classes choices, without sections or version or language, unless those things are
    # actually in the choices.
    #matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
    #f.write(str(matrix_choices.choices))
