### $Id: deffile.py,v 1.16 2008-09-30 23:50:02 lpoulson Exp $

######################################################################
# This module is currently a bit of a hybrid.  Most of the code is
# part of the MatrixDefFile class, which is used both to parse
# ./matrixdef and to emit HTML.  However, a couple of the methods
# on that class, while they do output HTML, don't have anything to
# do with the matrixdef file.  The HTML-generation code should
# probably be split out into a separate module/class.  Later.
#   - sfd 3/5/2008

######################################################################
# imports

import os
import cgitb
import glob
import re
import tarfile
import gzip
import zipfile
import cgi

from mom.MatrixChoices import ChoicesFile
from aggutils.matrix_utils import tokenize_def, get_name

from collections import defaultdict

######################################################################
# HTML blocks, used to create web pages


def merge_quoted_strings(line):
    i = 0
    while i < len(line):
        j = 0
        in_quotes = False
        while j < len(line[i]):
            if line[i][j] == '"' and (j == 0 or line[i][j-1] != '\\'):
                in_quotes = not in_quotes
            j += 1

        # if we reach the end of a line inside a quoted string, merge with
        # the next line and reprocess the newly-merged line
        if in_quotes:
            line[i] += line[i+1] # crash here implies an unbalanced '"'
            del line[i+1]
        else:
            i += 1

    return line

######################################################################
# Archive helper functions
#   make_tgz(dir) and make_zip(dir) create an archive called
#   dir.(tar.gz|zip) that contains the contents of dir

def make_tgz(dir):

    # ERB First get rid of existing file because gzip won't
    # overwrite existing .tgz meaning you can only customize
    # grammar once per session.

    if os.path.exists('matrix.tar.gz'):
        os.remove('matrix.tar.gz')

    archive = dir + '.tar'
    t = tarfile.open(archive, 'w')
    t.add(dir)
    t.close()

    g = gzip.open(archive + '.gz', 'wb')
    f = open(archive, 'rb')
    g.write(f.read())
    f.close()
    g.close()

    os.remove(archive)


def add_zip_files(z, dir):
    files = os.listdir(dir)
    for f in files:
        cur = os.path.join(dir, f)
        if os.path.isdir(cur):
            add_zip_files(z, cur)
        else:
            z.write(cur, cur)


def make_zip(dir):
    z = zipfile.ZipFile(dir + '.zip', 'w')
    add_zip_files(z, dir)
    z.close()


# Replace variables of the form {name} in word using the dict iter_vars
def replace_vars(word, iter_vars):
    for k in iter_vars.keys():
        word = re.sub('\\{' + k + '\\}', str(iter_vars[k]), word)
    return word


# From a list of triples of strings [string1, string2, ...], return
# a string containing a JavaScript-formatted list of strings of the
# form 'string1:string2'.
def js_array(list):
    val = ''
    for l in list:
        val += '\'' + l[0] + ':' + l[1] + '\',\n'
    val = val[:-2]  # trim off the last ,\n
    return val

# From a list of triples of strings [string1, string2, ...], return
# a string containing a JavaScript-formatted list of strings of the
# form 'string1:string2:string3'. This is used to convey features,
# values and category (category of feature).
def js_array3(list):
    val = ''
    for l in list:
        val += '\'' + l[0] + ':' + l[1] + ':' + l[3] + '\',\n'
    val = val[:-2]  # trim off the last ,\n
    return val

# From a list of triples of strings [string1, string2, ...], return
# a string containing a JavaScript-formatted list of strings of the
# form 'string1:string2:string3:string4'. This is used to convey features,
# values, category (category of feature), a flag feature 'customized'.
def js_array4(list):
    val = ''
    for l in list:
        val += '\'' + l[0] + ':' + l[1] + ':' + l[3] + ':' + l[4] + '\',\n'
    val = val[:-2]  # trim off the last ,\n
    return val

######################################################################
# MatrixDefFile class
# This class and its methods are used to parse Matrix definition
# formatted files (currently just the file ./matrixdef), and based
# on the contents, to produce HTML pages and save choices files.

class MatrixDefFile:
    # links between names and friendly names for
    # use in links on html navigation menu
    sections = { 'general':'General Information',
                 'word-order':'Word Order', 'number':'Number',
                 'person':'Person', 'gender':'Gender', 'case':'Case', 
                 'adnom-poss':'Adnominal Possession', 'direct-inverse':'Direct-inverse', 
                 'tense-aspect-mood':'Tense, Aspect and Mood', 'evidentials': 'Evidentials',
                 'other-features':'Other Features', 'sentential-negation':'Sentential Negation',
                 'coordination':'Coordination', 'matrix-yes-no':'Matrix Yes/No Questions',
                 'info-str':'Information Structure',
                 'arg-opt':'Argument Optionality',
                 'clausal-comp':'Clausal Complements',
                 'nominalclause':'Nominalized Clauses',
                 'clausalmods':'Clausal Modifiers', 'lexicon':'Lexicon',
                 'morphology':'Morphology','toolbox-import':'Toolbox Import',
                 'test-sentences':'Test Sentences','gen-options':'TbG Options',
                 'ToolboxLexicon':'Toolbox Lexicon'}
    # used to link section names to their documentation
    # page name in the delph-in wiki
    doclinks = { 'general':'GeneralInfo',
                 'word-order':'WordOrder', 'number':'Number',
                 'person':'Person', 'gender':'Gender', 'case':'Case',
                 'adnom-poss':'AdnominalPossession',
                 'direct-inverse':'DirectInverse', 'tense-aspect-mood':'TenseAspectMood', 
                 'evidentials': 'Evidentials',
                 'other-features':'OtherFeatures', 'sentential-negation':'SententialNegation',
                 'coordination':'Coordination', 'matrix-yes-no':'YesNoQ',
                 'info-str':'InformationStructure',
                 'arg-opt':'ArgumentOptionality',
                 'clausal-comp':'ClausalComplements',
                 'nominalclause':'Nominalized Clauses',
                 'clausalmods':'Clausal Modifiers', 'lexicon':'Lexicon',
                 'morphology':'Morphology','toolbox-import':'ImportToolboxLexicon',
                 'test-sentences':'TestSentences','gen-options':'TestByGeneration',
                 'ToolboxLexicon':'ImportToolboxLexicon'}
    def_file = ''
    v2f = {}
    f2v = {}

    def __init__(self, def_file):
        self.def_file = def_file
        f = open(self.def_file,encoding='utf-8')
        self.def_lines = merge_quoted_strings(f.readlines())
        f.close()
        self.make_name_map()

    ######################################################################
    # Variable/friendly name mapping

    # initialize the v2f and f2v dicts
    def make_name_map(self):
        for l in self.def_lines:
            l = l.strip()
            if len(l):
                w = tokenize_def(l)
                if len(w) >= 3:
                    ty = w[0]
                    vn = w[1]
                    fn = w[2]
                    if ty in ('Text', 'TextArea', 'Check', 'Radio',
                              'Select', 'MultiSelect', '.'):
                        self.v2f[vn] = fn
                        self.f2v[fn] = vn

    # return the friendly name for a variable, or the variable if none
    # is defined
    def f(self, v):
        if v in self.v2f:
            return self.v2f[v]
        else:
            return v

    # return the variable for a friendly name, or the friendly name if
    # none is defined
    def v(self, f):
        if f in self.f2v:
            return self.f2v[f]
        else:
            return f


    ######################################################################
    # Utility methods

    # TJT 2014-08-28: Adding method to find if some choice exists
    # in order to enable skipping UI elements based on this choice
    def check_choice_switch(self, switch, choices):
        """
        This method checks to see if a given string or regex (the switch)
        is in a given choices file. This method is used to see if a given
        UI element should be skipped. As such, the boolean return value
        is a little bit backwards. The function returns True if the switch
        is not found, and false if it is found.
        """
        if not switch:
            return True
        # Switch can have multiple variable names
        switches = switch.strip('"').split('|')
        # Switch contains choice name and value
        # split on the rightmost "=" just in case...
        switches = [switch.rsplit('=',1) if "=" in switch else switch
                    for switch in switches]
        # Default to true
        #skip_it = {switch[0] if isinstance(switch, list) else switch: True
        #           for switch in switches}
        # TJT: 12-19-14 changing this to loop instead of comprehension
        skip_it = {}
        for switch in switches:
            if isinstance(switch, list):
                skip_it[switch[0]] = True
            else:
                skip_it[switch] = True
        for instance in switches:
            if isinstance(instance, list):
                switch = instance[0].strip()
                values = instance[1].strip().strip("()").split(',') # Get multiple values
            else:
                switch = instance
                values = False # set default
            results = choices.get_regex(switch)
            if results:
                # Found a match
                if not values:
                    # If no value, then switching on a key,
                    # which was found, so don't skip it!
                    skip_it[switch] = False
                else:
                    for item in results:
                        if item[1] in values:
                            skip_it[switch] = False
                            break
        # if all false, don't skip it
        # else, skip it
        return any(skip_it.values())


    def save_choices_section(self, lines, f, choices,
                             iter_level = 0, prefix = ''):
        already_saved = {}  # don't save a variable more than once
        i = 0
        while i < len(lines):
            word = tokenize_def(lines[i])
            if len(word) == 0:
                pass
            # TJT 2014-5-27: changing this from list to tuple
            elif word[0] in ('Check', 'Text', 'TextArea',
                             'Radio', 'Select', 'MultiSelect', 'File','Hidden'):
                vn = word[1]
                if prefix + vn not in already_saved:
                    already_saved[prefix + vn] = True
                    val = ''
                    if choices.get(prefix + vn):
                        val = choices.get(prefix + vn)
                        if word[0] == 'TextArea':
                            val = '\\n'.join(val.splitlines())
                    if vn and val:
                        f.write('  '*iter_level) # TJT 2014-09-01: Changing this to one write from loop
                        f.write(prefix + vn + '=' + val + '\n')
            elif word[0] == 'BeginIter':
                (iter_name, iter_var) = word[1].replace('}', '').split('{', 1)
                i += 1
                beg = i
                while True:
                    word = tokenize_def(lines[i])
                    if len(word) == 0:
                        pass
                    elif word[0] == 'EndIter' and word[1] == iter_name:
                        break
                    i += 1
                end = i

                for var in choices.get(prefix + iter_name):
                    self.save_choices_section(lines[beg:end], f, choices,
                                              iter_level = iter_level + 1,
                                              prefix =
                                              prefix + iter_name + \
                                              str(var.iter_num()) + '_')

            i += 1


    # Read the choices_file, stripping out the section associated with
    # the 'section' member of form_data, and replacing it with all the
    # values in form_data.  Use self.def_file to keep the choices file
    # in order.
    def save_choices(self, choices_object, choices_file):
        choices = ChoicesFile(choice_dict=choices_object)

        # Now pass through the def file, writing out either the old choices
        # for each section or, for the section we're saving, the new choices
        f = open(choices_file, 'w', encoding='utf-8')
        f.write('\n') # blank line in case an editor inserts a BOM
        f.write('version=' + str(choices.current_version()) + '\n\n')

        cur_sec = ''
        cur_sec_begin = 0
        i = 0
        while i < len(self.def_lines):
            word = tokenize_def(self.def_lines[i])
            if len(word) == 0:
                pass
            elif word[0] == 'Section':
                if cur_sec:
                    self.save_choices_section(self.def_lines[cur_sec_begin:i], f, choices)
                    f.write('\n')
                cur_sec = word[1]
                cur_sec_begin = i + 1
                f.write('section=' + cur_sec + '\n')
            i += 1
        # Make sure to save the last section
        if cur_sec_begin:
            self.save_choices_section(self.def_lines[cur_sec_begin:i], f, choices)

        f.close()


    def create_neg_aux_choices(self, choices,form_data):
        '''this is a side effect of the existence of neg-aux
        in the form data, it puts some lines pertaining to a neg-aux
        lexical item into the choices file object unless they are
        already there. returns a ChoicesFile instance, and an int
        which is the index number of the created neg-aux'''

        # get the next aux number
        next_n = choices['aux'].next_iter_num() if 'aux' in choices else 1

        # create a new aux with that number
        choices['aux%d_name' % next_n] = 'neg'

        # copy that to nli for adding further information
        nli = choices['aux'].get_last()
        nli['sem']='add-pred'
        nli['stem1_pred'] = 'neg_rel'

        if 'bineg-type' in form_data.keys() and \
                        form_data['bineg-type'].value =='infl-head':
            nli['compfeature1_name']='form'
            nli['compfeature1_value']='negform'

        # if auxiliaries are off, turn them on
        choices['has-aux'] = 'yes'
        return choices, next_n

    def create_infl_neg_choices(self, old_choices, new_choices):
        vpc = new_choices['vpc-0-neg']
        lrt = ''
        if vpc == 'create':
            next_n = old_choices['verb-pc'].next_iter_num() if old_choices['verb-pc'] else 1
            old_choices['verb-pc%d_name' % next_n] = 'negpc'
            vpc = old_choices['verb-pc'].get_last()
            vpc['lrt1_name'] = 'neg'
            lrt = old_choices['verb-pc'].get_last()['lrt'].get_last()
            new_choices['vpc-0-neg'] = 'verb-pc'+str(next_n)

        else:
            next_n = old_choices[vpc]['lrt'].next_iter_num() if old_choices[vpc]['lrt'] else 1
            # create new lrt in this position class
            old_choices[vpc]['lrt%d_name' % next_n] = 'neg'
            # add some features for negation and empty PHON
            lrt = old_choices[vpc]['lrt'].get_last()

        lrt['feat1_name']= 'negation'
        lrt['feat1_value'] = 'plus'
        lrt['feat1_head'] = 'verb'
        lrt['lri1_inflecting'] = 'no'
        return old_choices, new_choices

