#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Authors: Kristen Howell

# This scripts reads in a xigt corpus and a param file, which
# specifies which options to use for MOM as well as which baseline system to run.
# The output is a choices file

# -*- coding: utf-8 -*-
import sys,os
from mom import MainClasses, MatrixChoices, Choices, FeatDict, MorphTypeGuesser
from XigtifiedToolbox import XigtReader
from xigt.codecs import xigtxml
from aggutils.deffile import MatrixDefFile

parent_dir = os.path.join(os.getcwd(),os.pardir)
sys.path.append(parent_dir)
import random

def get_valid_lines(lines):
  """
  Return all lines that are not empty. Also, strip out byte order marks
  from the first line if it has one.
  """
  if len(lines) > 0:
    lines[0] = lines[0].strip()
  lines = [l.strip() for l in lines if l.strip() != '']
  return lines

def build_param_dict(lines, all_params):
    """
    Check each line for a valid parameter. Add all paramaters to
    param dict
    """
    params = {}
    for line in lines:
        print(line)
        (param, value) = line.split(':',1)
        if param in all_params:
            params[param] = value.strip()
    return params

def add_basic_choices(choices, config_file):
    choices.__setitem__('version', '32')
    choices.__setitem__('language',  MainClasses.Settings(config_file).language)
    choices.__setitem__('iso-code',  MainClasses.Settings(config_file).iso)
    return choices

def read_in_xigt(settings):
    xigt_reader = XigtReader.Xigt_Reader(settings, verbose=False)
    xigt_reader.process_igts(settings.data_file)
    xigt_reader.print_skipped(settings.out_dir + '/skipped_items.txt')
    return xigt_reader

def run_mom(settings, xigt_reader, agr_morphemes=None, ignore_features=False):
    filescount = 0
    fout_name, tar_name, arc_name = MainClasses.set_filenames(filescount, settings)
    mom = MainClasses.MOM(name='AGG-MOM', output_dir=settings.out_dir,
                              tag_sets=settings.pos_tags, lexitem_classes=settings.lexitem_classes,
                              precluster_only=settings.precluster_only)
    choices = mom.execute(pos_items=xigt_reader.items[settings.pos],settings=settings,
                            patterns=settings.patterns,tarball=None,arc_name=arc_name,
                            ignore_affix_types=settings.ignore_affixes,verbose=False,pos_tag=settings.pos,
                            escape_special_characters=settings.escape_special_characters,
                            agr_morphemes=agr_morphemes,ignore_features=ignore_features)
    return choices

def get_punctuation(choices):
    punct_chars = {'!', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?',
                   '@', '[', ']', '^', '_', '`', '{', '|', '}', '~', '\\'}
    punct = set()
    pos_list = ['verb', 'noun', 'det', 'adp', 'aux']
    for pos in pos_list:
        for word in choices[pos]:
            for stem in word['stem']:
                for char in stem['orth']:
                    if char in punct_chars:
                        punct.add(char)
            for char in word['orth']: # some lexical types (eg. adp) do not put the orth inside a stem
                if char in punct_chars:
                    punct.add(char)
        for pc in choices[pos+'-pc']:
            for lrt in pc['lrt']:
                for lri in lrt['lri']:
                    if 'orth' in lri:
                        for char in lri['orth']:
                            if char in punct_chars:
                                if char == '.':
                                    print()
                                punct.add(char)
    if punct != set():
        choices['punctuation-chars'] = 'keep-list'
        choices['punctuation-chars-list'] = ''.join(punct)
    return choices

def toss_coin():
    toss = random.randint(0,1)
    if toss == 0:
        coin = 'heads'
    else:
        coin = 'tails'
    return coin

def set_verb_valence(igt2items, baseline):
    for i in igt2items:
        for w in igt2items[i]:
            if baseline == 'broad':
                igt2items[i][w].transitivity = 'trans'
                igt2items[i][w].case_frame = {'subj':'','obj':''}
            else:
                if toss_coin() == 'heads':
                    igt2items[i][w].transitivity = 'trans'
                    igt2items[i][w].case_frame = {'subj': '', 'obj':''}
                else:
                    igt2items[i][w].transitivity = 'intrans'
                    igt2items[i][w].case_frame = {'subj': ''}

def add_wo_choices(choices, baseline):
    #word-order
    if baseline == 'broad':
        choices.__setitem__('word-order', 'free')
    elif baseline == 'typ':
        choices.__setitem__('word-order', 'sov') #wals
    else:
        word_orders = ['sov','svo','vso','osv','ovs','vos','v-final','v-initial','free','v2']
        rc = random.randint(0,len(word_orders)-1)
        choices.__setitem__('word-order', word_orders[rc])

    #dets
    if baseline == 'broad':
        choices.__setitem__('has-dets', 'yes')
        if toss_coin() == 'heads':
            choices.__setitem__('noun-det-order', 'det-noun')
        else:
            choices.__setitem__('noun-det-order', 'noun-det')
    if baseline == 'typ':
        choices.__setitem__('has-dets', 'yes') #wals 37a
        choices.__setitem__('noun-det-order', 'noun-det') #wals 88a
    else:
        if toss_coin() == 'heads':
            choices.__setitem__('has-dets', 'yes')
            if toss_coin() == 'heads':
                choices.__setitem__('noun-det-order', 'det-noun')
            else:
                choices.__setitem__('noun-det-order', 'noun-det')
        else:
            choices.__setitem__('has-dets', 'no')

    #aux
    # we cannot identify the actually choices without inference, so there is not point in saying the langauge has them
    choices.__setitem__('has-aux', 'no')
    return choices

def add_required_feature_choices(choices, baseline):
    # baseline can't have features without inference
    # these are required choices for a valid choices file
    choices.__setitem__('person', 'none')
    choices.__setitem__('case-marking', 'none')
    return choices

def add_coord_choices(choices, baseline):
    if baseline == 'broad':
        choices.__setitem__('cs1_n', 'on')
        choices.__setitem__('cs1_np', 'on')
        choices.__setitem__('cs1_vp', 'on')
        choices.__setitem__('cs1_s', 'on')
        choices.__setitem__('cs1_pat', 'a')
    if baseline == 'random':
        strategy = False
        if toss_coin() == 'heads':
            choices.__setitem__('cs1_n', 'on')
            strategy = True
        if toss_coin() == 'heads':
            choices.__setitem__('cs1_np', 'on')
            strategy = True
        if toss_coin() == 'heads':
            choices.__setitem__('cs1_vp', 'on')
            strategy = True
        if toss_coin() == 'heads':
            choices.__setitem__('cs1_s', 'on')
            strategy = True
        if strategy:
            choices.__setitem__('cs1_pat', 'a')
    # typologically most common is monosyndedic. This is useless without the orthography
    return choices

def add_arg_opt_choices(choices, baseline):
    if baseline == 'broad':
        choices.__setitem__('subj-drop', 'subj-drop-all')
        choices.__setitem__('subj-mark-drop', 'subj-mark-drop-opt')
        choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-opt')
        choices.__setitem__('obj-drop', 'obj-drop-all')
        choices.__setitem__('obj-mark-drop', 'obj-mark-drop-opt')
        choices.__setitem__('obj-mark-no-drop', 'obj-mark-no-drop-opt')
    # we are just choosing between all and nothing. lex has a downstream dependency, so we exclude it
    # this is analagous to the choice to exclude mono and moni coordination
    if baseline == 'random':
        if toss_coin() == 'heads':
            choices.__setitem__('subj-drop', 'subj-drop-all')
            choices.__setitem__('subj-mark-drop', 'subj-mark-drop-opt')
            choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-opt')
        # else do nothing-- the language doesn't have subj drop
        if toss_coin() == 'heads':
            choices.__setitem__('obj-drop', 'obj-drop-all')
            choices.__setitem__('obj-mark-drop', 'obj-mark-drop-opt')
            choices.__setitem__('obj-mark-no-drop', 'obj-mark-no-drop-opt')
       # else to nothing
    if baseline == 'typ': # wals 101A for subj. no info for obj-drop
        choices.__setitem__('subj-drop', 'subj-drop-all')
        choices.__setitem__('subj-mark-drop', 'subj-mark-drop-opt')
        choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-opt')
    return choices

def extract_dets(choices, xc, baseline):
    det_tags = ['det', 'dem', 'quant', 'num', 'def']
    determiners = set()
    for igt in xc:
        pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos')
        if not pos_tier:
            pos_tier = igt.get('m-pos')
        if not pos_tier:
            pos_tier = igt.get('gw-pos')
        if not pos_tier:
            pos_tier = igt.get('x')
        if not pos_tier:
            continue
        for pos in pos_tier:
            if pos.value():
                if pos.value().lower() in det_tags:
                    w_id = pos.attributes['alignment']
                    if 'm' in w_id:
                        w_id = igt.get('m').get(w_id).attributes['segmentation']
                    elif 'gw' in w_id:
                        w_id = igt.get('gw').get(w_id).attributes['alignment']
                    elif 'g' in w_id:
                         w_id = igt.get('g').get(w_id).attributes['alignment']
                    word = igt.get('w').get(w_id)
                    orth = word.value().strip('\"')
                    determiners.add(orth)
    det_count = 1
    for det in determiners:
        choices.__setitem__('det' + str(det_count) + '_name', 'det' + str(det_count))
        choices.__setitem__('det' + str(det_count) + '_stem1_orth', det)
        pred = '_exist_q_rel'
        choices.__setitem__('det' + str(det_count) + '_stem1_pred', pred)
        det_count += 1
    return choices

def add_det_choice_to_nouns(choices, baseline):
    det_choices = ['obl','opt','imp']
    for noun_type in choices['noun']:
        if baseline == 'broad':
            choices.__setitem__(noun_type.full_key + '_det', 'opt')
        else:
            rc = random.randint(0, len(det_choices) - 1)
            choices.__setitem__(noun_type.full_key + '_det', det_choices[rc])
    return choices

# Main routine
all_params = ['baseline', 'data file']
param_file = open(sys.argv[1]).readlines()
param_dict = build_param_dict(get_valid_lines(param_file), all_params)
corpus_file = param_dict['data file']
settings = MainClasses.Settings(sys.argv[1])
# Xigt Reader contains a special data structure of items that were created based
# on the xigtified data. In particular, the items correspond to the word level
# of the sentences (they correspond to words) and they have all the information
# that MOM later is going to need (POS, affixes, gloss, translation, which IGT they came from, etc.)
xigt_reader = read_in_xigt(settings)
choices = MatrixChoices.ChoiceDict()
corpus = open(corpus_file, encoding='utf-8')
xc = xigtxml.load(corpus)
corpus.close()
config_file = sys.argv[1]
known_glosses = MorphTypeGuesser.MorphTypeGuesser(settings.known_glosses_file)

baseline = param_dict['baseline']
set_verb_valence(xigt_reader.igt2items, baseline)

# Now, MOM runs on the Xigt Reader objects, with the transitivity and case frame
# info set, to infer morphotactics. MOM should only merge items with the same or
# underspecified transitivity and case frame. E.g. a transitive item
# should never be assigned an intransitive class.
settings.pos = 'verb'
mom_verb_choices = run_mom(settings,xigt_reader,ignore_features=True)
settings.pos = 'noun'
mom_noun_choices = run_mom(settings,xigt_reader,ignore_features=True)
choices = Choices.build_matrix_choices(mom_verb_choices.lexicon,mom_noun_choices.lexicon, settings)
choices = add_basic_choices(choices, config_file)
choices = get_punctuation(choices)
choices = add_wo_choices(choices, baseline)
choices = add_required_feature_choices(choices, baseline)
choices = add_coord_choices(choices, baseline)
choices = add_arg_opt_choices(choices, baseline)
if choices['has-dets'] == 'yes':
    choices = extract_dets(choices, xc, baseline)
choices = add_det_choice_to_nouns(choices, baseline)
# write choices file
print('OUTPUT DIR: {0}'.format(settings.out_dir))
matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
choices_file = settings.out_dir + 'latest_choices'
formatted_choices = MatrixDefFile('./src/aggutils/matrixdef')

formatted_choices.save_choices(matrix_choices,choices_file)

