#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kristen Howell
# this script has been tested on a handful of languages from ODIN 2.1.
# Bugs may be found when testing it on other datasets

from mom import MatrixChoices
from xigt.codecs import xigtxml
import sys
from aggutils.find_root import find_root
import re

# a set of verbs collected from the ERG that allow object dropping.
OBJ_DROP_VERBS = open('./src/aggutils/obj_drop_verbs').readlines()

i = 0
while i < len(OBJ_DROP_VERBS):
    OBJ_DROP_VERBS[i] = OBJ_DROP_VERBS[i].strip()
    i +=1

def infer_arg_opt(xc, filter_3sg=True, verb_tags=['VERB']):
    count_dependencies(xc)
    transitive_verbs = {}
    intransitive_verbs = {}
    trans_verb_tokens = {}
    intrans_verb_tokens = {}
    arguments = {'overt-subj': 0, 'dropped-subj': 0, 'overt-obj': 0, 'dropped-obj': 0, 'non-eng-drop': 0}
    agr_markers = {'overt-subj-mark': 0, 'dropped-subj-mark': 0, 'overt-obj-mark': 0, 'dropped-obj-mark': 0}
    unglossed = 0
    twolines = 0
    non_empty_igt = 0
    for igt in xc:
        #skip if igt doesn't have three lines
        if not igt.get('w') or not igt.get('gw') or not igt.get('tw'):
            twolines += 1
            continue
        else:
            non_empty_igt += 1

        verb_sets = get_verbs(igt, verb_tags)
        for verb_set in verb_sets:
            #skip igt with no english subject, eg. imperatives
            if not english_subj(igt, verb_set):
                continue
            #skip subordinate verbs
            if subordinate_verb(igt, verb_set):
                continue
            #skip passive verbs
            if is_passive(igt, verb_set):
                continue
            subj, obj = get_dependents(igt, verb_set)
            threesg_subj, threesg_obj = is_threesg(igt, subj, obj, verb_set)
            agreement_markers = get_agreement_markers(igt, verb_set)
            if agreement_markers == (None, None):
                unglossed += 1
                continue
            if subj:
                dropped_subj = False
            else:
                dropped_subj = True
            if obj:
                dropped_obj = False
                transitive = True
                non_eng_drop = False
            else:
                if check_for_pronouns(igt) != '':
                    # english has a pronoun, so the object truely dropped
                    non_eng_drop = True
                    dropped_obj = True
                    transitive = True
                else:
                    non_eng_drop = False
                    if check_for_english_pro_drop(igt, verb_set) == True:
                        # english also drops in this case
                        dropped_obj = True
                        transitive = True
                    else:
                        # the verb is intransitive
                        transitive = False
            verb = verb_set[0]
            token = ''
            for verb in verb_set:
                lemma = igt.get('w').get(verb).value()
                token += lemma+' '
            if transitive:
                if igt.id in transitive_verbs:
                    transitive_verbs[igt.id][verb] = {'dropped-subj': dropped_subj, 'dropped-obj': dropped_obj,
                                                      'agreement_markers': agreement_markers,
                                                      'non-eng-drop': non_eng_drop, 'threesg_subj': threesg_subj,
                                                      'threesg_obj': threesg_obj}
                else:
                    transitive_verbs[igt.id] = {}
                    transitive_verbs[igt.id][verb] = {'dropped-subj': dropped_subj, 'dropped-obj': dropped_obj,
                                                      'agreement_markers': agreement_markers,
                                                      'non-eng-drop': non_eng_drop, 'threesg_subj': threesg_subj,
                                                      'threesg_obj': threesg_obj}
                if token in trans_verb_tokens:
                    trans_verb_tokens[token] += 1
                else:
                    trans_verb_tokens[token] = 1
            else:
                if igt.id in intransitive_verbs:
                    intransitive_verbs[igt.id][verb] = {'dropped-subj': dropped_subj,
                                                        'agreement_markers': agreement_markers,
                                                        'threesg_subj': threesg_subj}
                else:
                    intransitive_verbs[igt.id] = {}
                    intransitive_verbs[igt.id][verb] = {'dropped-subj': dropped_subj,
                                                        'agreement_markers': agreement_markers,
                                                        'threesg_subj': threesg_subj}
                if token in intrans_verb_tokens:
                    intrans_verb_tokens[token] += 1
                else:
                    intrans_verb_tokens[token] = 1


    subj_markers = set()
    obj_markers = set()
    portmanteau_markers =set()
    for igt in intransitive_verbs:
        for verb in intransitive_verbs[igt]:
            agreement_markers = intransitive_verbs[igt][verb]['agreement_markers']
            if filter_3sg == False or intransitive_verbs[igt][verb]['threesg_subj'] == False:
                if intransitive_verbs[igt][verb]['dropped-subj'] == True:
                    arguments['dropped-subj'] += 1
                    if agreement_markers != []:
                        agr_markers['dropped-subj-mark'] += 1
                else:
                    arguments['overt-subj'] += 1
                    if agreement_markers != []:
                        agr_markers['overt-subj-mark'] += 1
                if len(agreement_markers) > 0:
                    for gloss, orth in agreement_markers:
                        if orth:
                            subj_markers.add(orth)
                    # gloss_token = ''
                    # for gloss in agreement_glosses:
                    #     gloss_token += gloss+'.'
                    # if 'subj_gloss' in verb_tokens[verb]:
                    #     if gloss_token in verb_tokens[verb]['subj_gloss']:
                    #         verb_tokens[verb]['subj_gloss'][gloss_token] += 1
                    #     else:
                    #         verb_tokens[verb]['subj_gloss'][gloss_token] = 1
                    # else:
                    #     verb_tokens[verb]['subj_gloss'] = {gloss_token: 1}

    for igt in transitive_verbs:
        for verb in transitive_verbs[igt]:
            agreement_markers = transitive_verbs[igt][verb]['agreement_markers']
            agreement_glosses = [x[0] for x in agreement_markers]
            marks_separate_args = separate_args(agreement_glosses)
            subj_gloss = False
            for gloss, orth in agreement_markers:
                if orth:
                    if marks_separate_args:
                        portmanteau_markers.add(orth)
                        subj_markers.discard(orth)
                    if orth in subj_markers:
                        subj_gloss = True
                    elif marks_separate_args:
                        portmanteau_markers.add(orth)
                    else:
                        obj_markers.add(orth)
            if (filter_3sg == False) or (transitive_verbs[igt][verb]['threesg_subj'] == False):
                if transitive_verbs[igt][verb]['dropped-subj'] == True:
                    arguments['dropped-subj'] += 1
                    if subj_gloss or marks_separate_args:
                        agr_markers['dropped-subj-mark'] += 1
                else:
                    arguments['overt-subj'] += 1
                    if subj_gloss or marks_separate_args:
                        agr_markers['overt-subj-mark'] += 1

            if filter_3sg == False or transitive_verbs[igt][verb]['threesg_obj'] == False:
                if transitive_verbs[igt][verb]['dropped-obj'] == True:
                    arguments['dropped-obj'] += 1
                    if (not subj_gloss and len(agreement_glosses) > 0) or marks_separate_args:
                        agr_markers['dropped-obj-mark'] += 1
                else:
                    arguments['overt-obj'] += 1
                    if (not subj_gloss and len(agreement_glosses) > 0) or marks_separate_args:
                        agr_markers['overt-obj-mark'] += 1
                if transitive_verbs[igt][verb]['non-eng-drop'] == True:
                    arguments['non-eng-drop'] += 1

    """ print('non-empty-igt: ' + str(non_empty_igt))
    print('overt subjects: ' + str(arguments['overt-subj']))
    print('dropped subjects: ' + str(arguments['dropped-subj']))
    print('overt objects: ' + str(arguments['overt-obj']))
    print('dropped objects: ' + str(arguments['dropped-obj']))
    print('marked overt subjects: ' + str(agr_markers['overt-subj-mark']))
    print('marked dropped subjects: ' + str(agr_markers['dropped-subj-mark']))
    print('marked overt objects: ' + str(agr_markers['overt-obj-mark']))
    print('marked dropped objects: ' + str(agr_markers['dropped-obj-mark'])) """
    # print('percent overt subjects with marking: ' + str(float(agr_markers['overt-subj-mark']/arguments['overt-subj'])))
    # print('percent dropped subjects with marking: ' + str(float(agr_markers['dropped-subj-mark'] / arguments['dropped-subj'])))
    # print('percent overt objects with marking: ' + str(float(agr_markers['overt-obj-mark'] / arguments['overt-obj'])))
    # print('percent dropped objects with marking: ' + str(float(agr_markers['dropped-obj-mark'] / arguments['dropped-obj'])))
    # print('two lines: ' + str(twolines))
    # print('unglossed: ' + str(unglossed))
    intrans_tokens = 0
    unique_intrans_tokens = 0
    for verb in intrans_verb_tokens:
        unique_intrans_tokens += 1
        intrans_tokens += intrans_verb_tokens[verb]
    #print('total intrans verbs: ' + str(intrans_tokens))
    #print('unique intrans verbs: ' + str(unique_intrans_tokens))
    trans_tokens = 0
    unique_trans_tokens = 0
    for verb in trans_verb_tokens:
        unique_trans_tokens += 1
        trans_tokens += trans_verb_tokens[verb]
    # print('total trans verbs: ' + str(trans_tokens))
    # print('unique trans verbs: ' + str(unique_trans_tokens))
    # print('subject glosses: ' + str(len(subj_agr_glosses)))
    # print('intransitive verbs:')
    # for verb in intrans_verb_tokens:
    #     print(verb + ': ' + str(intrans_verb_tokens[verb]))
    # print('transitive verbs:')
    # for verb in trans_verb_tokens:
    #     print(verb + ': ' + str(trans_verb_tokens[verb]))
    # print('subject glosses:')
    # for gloss in subj_agr_glosses:
    #     print(gloss)

    choices = set_choices(arguments, agr_markers)
    return choices, {'subj_morphemes': subj_markers, 'obj_morphemes': obj_markers, 'portmanteau_morphemes': portmanteau_markers}

def separate_args(agreement_glosses):
    # returns true if a set of agreement markers correspond to two or more separate arguments
    # checks for agr marks that belong to the same category, eg. two number markers.
    separate_args = False
    person_glosses = ['1','2','3',]
    number_glosses = ['sg','pl']
    gender_glosses = ['f','m','n']
    valence_glosses = ['o', 's(?!g)', 'subj','obj']
    pronoun_glosses = ['i', 'you', 'he', 'she', 'it', 'they', 'me', 'him', 'her', 'them']
    mutually_excl_glosses = person_glosses + number_glosses + gender_glosses + valence_glosses
    gloss_string = ''
    for gloss in agreement_glosses:
        if gloss:
            gloss_string += gloss
    if len(agreement_glosses) >= 1:
        if agreement_glosses[0]:
            pronouns = 0
            for gloss in agreement_glosses:
                if gloss in pronoun_glosses:
                    pronouns += 1
            if pronouns > 1:
                separate_args = True
            elif pronouns == 1:
                for gloss in mutually_excl_glosses:
                    if re.search(gloss, gloss_string):
                        separate_args = True
            else:
                per_count = 0
                for gloss in person_glosses:
                    per_count += len(re.findall(gloss,gloss_string))
                if per_count > 1:
                    separate_args = True
                num_count = 0
                for gloss in number_glosses:
                    num_count += len(re.findall(gloss, gloss_string))
                if num_count > 1:
                    separate_args = True
                gen_count = 0
                for gloss in gender_glosses:
                    gen_count += len(re.findall(gloss, gloss_string))
                if gen_count > 1:
                    separate_args = True
                val_count = 0
                for gloss in valence_glosses:
                    val_count += len(re.findall(gloss, gloss_string))
                if val_count > 1:
                    separate_args = True
    return separate_args

def get_verbs(igt, verb_tags):
    # collects the verbs in an igt and groups verbs that are in dependency relationships with each other.
    # returns a list of verb sets, which contain verbs that are dependents of each other
    all_verbs = []
    verb_sets = []
    pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
    if pos_tier:
        for pos in pos_tier:
            if pos.value() in verb_tags:
                verb = pos.attributes['alignment']
                if 'gw' in verb:
                    if 'alignment' in igt.get('gw').get(verb).attributes:
                        verb = igt.get('gw').get(verb).attributes['alignment']
                        # weed out capitalized "verbs". eg. the proper noun "Can" in Turkish
                if is_capitalized(igt, verb):
                    continue
                if 'gw' not in verb:
                    all_verbs.append(verb)
                else:
                    print(igt.id + ': debugging get_verbs. alignment betwee gw and w missing')
    w_ds = igt.get('w-ds')
    if w_ds:
        while all_verbs != []:
            verb_set = [all_verbs[0]]
            for item in w_ds:
                if 'head' in item.attributes:
                    if (not item.value()) or item.value() == 'aux' or item.value() == 'dep':
                        if all_verbs[0] == item.attributes['head'] and item.attributes['dep'] in all_verbs:
                            verb_set.append(item.attributes['dep'])
                            all_verbs.remove(item.attributes['dep'])
                        elif all_verbs[0] == item.attributes['dep'] and item.attributes['head'] in all_verbs:
                            verb_set.append(item.attributes['head'])
                            all_verbs.remove(item.attributes['head'])
            all_verbs.remove(all_verbs[0])
            verb_sets.append(verb_set)
    return verb_sets

def get_agreement_markers(igt, verbs):
    # collects all agreement glosses for a list of verb ids
    all_agreement_glosses = ['1','2','3','sg','du','pl', '1sg', '2sg', '3sg', '1du','2du','3du','1pl','2pl','3pl','1s','2s','3s','1p','2p','3p','a','s','o','p','t','excl','f','m','n','nm','fam','form','inanim','incl']
    all_agreement_glosses += ['1fsg','1sgf','1nmsg','1sgnm','1msg','1sgm','1nsg','1sgn','1fpl','1plf','1nmpl','1plnm','1mpl','1plm','1npl','1pln','1f','1nm','1m','1n','1o','1s','1fo','1nmo','1fs','1nms','1mo','1ms','1no','1ns','1a']
    all_agreement_glosses += ['2fsg','2sgf','2nmsg','2sgnm','2msg','2sgm','2nsg','2sgn','2fpl','2plf','2nmpl','2plnm','2mpl','2plm','2npl','2pln','2f','2nm','2m','2n','2o','2s','2fo','2fs','2nmo','2nms','2mo','2ms','2no','2ns','2a']
    all_agreement_glosses += ['3fsg','3sgf','3nmsg','3sgnm','3msg','3sgm','3nsg','3sgn','3fpl','3plf','3nmpl','3plnm','3mpl','3plm','3npl','3pln','3f','3nm','3m','3n','3o','3s','3fo','3fs','3nmo','3nms','3mo','3ms','3no','3ns','3a']
    all_agreement_glosses += ['s3sg','s1sg','s2sg','s3pl','s1pl','s2pl','o3sg','o1sg','o2sg','o3pl','o1pl','o2pl','fem','masc','neut','as']
    case_marked = []
    for gloss in all_agreement_glosses:
        case_marked.append(gloss + 'n')
        case_marked.append(gloss + 'd')
        case_marked.append(gloss + 'e')
        case_marked.append(gloss + 'g')
        case_marked.append(gloss + 'a')
    pers_marked = []
    for gloss in all_agreement_glosses:
        pers_marked.append('pers'+gloss)
    all_agreement_glosses += case_marked
    all_agreement_glosses += pers_marked
    all_agreement_glosses += ['1nom','2nom','3nom','1acc','2acc','3acc','1erg','2erg','3erg','1abs','2abs','3abs']
    all_agreement_glosses += ['mo','ms','fo','fs','no','ns','nmo','nms']
    all_agreement_glosses += ['subj', 'obj', 't','dt']#t is redundant here for transitive
    all_agreement_glosses += ['i', 'you', 'he', 'she', 'it', 'they', 'me', 'him', 'her', 'them', 'cl', 'scl', 'ocl']
    agreement_markers = []
    gloss_words = igt.get('gw')
    # assume that the sentence isn't glossed, until we see one that is
    no_glossing = True
    if gloss_words:
        for gw in gloss_words:
            if 'alignment' in gw.attributes:
                if gw.attributes['alignment'] in verbs:
                    #gloss_word = gw.value()
                    #gloss_word = re.sub('(\(|\)|_)', '.', gloss_word)
                    # this hard to read list comrehension returns the gloss and respective orthography for each morpheme in the word
                    #glosses = [(x.value(), igt.get('m').get(x.attributes['alignment']).value() if igt.get('m').get(x.attributes['alignment']) else None) for x in igt.get('g') if 'segmentation' in x.attributes and gw.id in x.attributes['segmentation']]
                    glosses = []
                    for x in igt.get('g'):
                        if 'segmentation' in x.attributes and gw.id in x.attributes['segmentation']:
                            gloss = x.value()
                            orth = None
                            if 'alignment' in x.attributes:
                                if igt.get('m').get(x.attributes['alignment']):
                                    orth = igt.get('m').get(x.attributes['alignment']).value()
                            glosses.append((gloss, orth))
                    for gloss, orth in glosses:
                        grams = gloss.lower().split('.')
                        if len(glosses) > 1 or len(grams) > 1:
                            no_glossing = False
                        agr_gloss = ''
                        for gram in grams:
                            if gram in all_agreement_glosses:
                                if gram == 't':
                                    gram = 'so'
                                if gram == 'dt':
                                    gram = 'sog'
                                agr_gloss += gram
                        if agr_gloss != '':
                            agreement_markers.append((gloss, orth))
    if no_glossing == True:
        agreement_markers = [(None, None)]
    return agreement_markers

def get_dependents(igt, verbs):
    # returns the id of the subject and object that correspond to a set of verbs
    subj = None
    obj = None
    subj_types = ['nsubj', 'csubj', 'nsubjpass', 'csubjpass']
    w_ds = igt.get('w-ds')
    if w_ds:
        for item in w_ds:
            if 'head' in item.attributes:
                if item.attributes['head'] in verbs:
                    if item.value() in subj_types:
                        subj = item.attributes['dep']
                    elif item.value() == 'dobj':
                        obj = item.attributes['dep']
    return subj, obj

def is_threesg(igt, subj, obj, verb_set):
    # returns a tuple of true/false for each subj and obj, based on whether they are 3rd person singular
    # If 3sg isn't in the gloss we look for a 3sg pronoun. if "it" is in the translation, both the subj and obj will be marked 3sg
    # first check the language tier, but if subj or obj isn't overt, check the translation
    threesg_subj = False
    threesg_obj = False
    THREESG_SUBJ_PRONOUNS = ['he','she','it']
    THREESG_OBJ_PRONOUNS = ['him', 'her', 'it']
    w_ps = igt.get('w-ps')
    tw_ds = igt.get('tw-ds')
    tw = igt.get('tw')
    tw_ps = igt.get('tw-ps')
    if subj:
        if w_ps:
            for item in w_ps:
                if 'alignment' in item.attributes:
                    if item.attributes['alignment'] == subj:
                        if item.value() == 'NNP' or item.value() == 'NN':
                            threesg_subj = True
                        elif item.value() == 'PRP':
                            gw_tier = igt.get('gw')
                            # is this even necessary if I'm going to check the trans line anyway?
                            for gw in gw_tier:
                                if 'alignment' in gw.attributes:
                                    if gw.attributes['alignment'] == subj:
                                        if '3sg' in gw.value().lower():
                                            threesg_subj = True
                                        else:
                                            for pron in THREESG_SUBJ_PRONOUNS:
                                                if pron in gw.value().lower():
                                                    threesg_subj = True

    else:
        for verb in verb_set:
            eng_verb = get_english_wid(igt, verb)
            for item in tw_ds:
                if item.value() == 'nsubj' or item.value() == 'nsubjpass':
                    if item.attributes['head'] == eng_verb:
                        subj = item.attributes['dep']
                        if tw_ps:
                            for item in tw_ps:
                                if 'alignment' in item.attributes:
                                    if item.attributes['alignment'] == subj:
                                        if item.value() == 'NNP' or item.value() == 'NN':
                                            threesg_subj = True
                                        elif item.value() == 'PRP':
                                            for word in tw:
                                                if word.id == subj:
                                                    eng_subj = word.value()
                                                    if eng_subj.lower() in THREESG_SUBJ_PRONOUNS:
                                                        threesg_subj = True

    if obj:
        if w_ps:
            for item in w_ps:
                if 'alignment' in item.attributes:
                    if item.attributes['alignment'] == obj:
                        if item.value() == 'NNP' or item.value() == 'NN':
                            threesg_obj = True
                        elif item.value() == 'PRP':
                            gw_tier = igt.get('gw')
                            for gw in gw_tier:
                                if 'alignment' in gw.attributes:
                                    if gw.attributes['alignment'] == obj:
                                        if '3sg' in gw.value().lower():
                                            threesg_obj = True
                                        else:
                                            for pron in THREESG_OBJ_PRONOUNS:
                                                if pron in gw.value().lower():
                                                    threesg_obj = True
    else:
        for verb in verb_set:
            eng_verb = get_english_wid(igt, verb)
            for item in tw_ds:
                if item.value() == 'dobj':
                    if item.attributes['head'] == eng_verb:
                        obj = item.attributes['dep']
                        if tw_ps:
                            for item in tw_ps:
                                if 'alignment' in item.attributes:
                                    if item.attributes['alignment'] == obj:
                                        if item.value() == 'NNP' or item.value() == 'NN':
                                            threesg_obj = True
                                        elif item.value() == 'PRP':
                                            for word in tw:
                                                if word.id == obj:
                                                    eng_obj = word.value()
                                                    if eng_obj.lower() in THREESG_OBJ_PRONOUNS:
                                                        threesg_obj = True

    return threesg_subj, threesg_obj

def check_for_english_pro_drop(igt, verb_set):
    # returns true if the verb is in a set of English verbs that allow object dropping
    pro_drop = False
    for verb in verb_set:
        for word in igt.get('w'):
            if word.id == verb:
                verb_item = word
        if verb_item:
            verb_lemma, eng_root = find_root(igt, verb_item)
            if eng_root in OBJ_DROP_VERBS:
                pro_drop = True
    return pro_drop

def check_for_pronouns(igt):
    """
    From and IGT, determine if the direct object in the English translation
    is a pronoun. If so, return the item corresponding with that pronoun.
    Otherwise return an empty string. """
    pronoun = ''
    eng_acc_pronouns = ['me', 'us', 'you', 'him', 'her', 'them', 'it', 'someone', 'something', 'somewhere']
    #for tier in xp.findall(igt, 'tier[@type="dependencies"][@dep="tw"]'):
    tw_ds = igt.get('tw-ds')
    if tw_ds:
        for item in tw_ds:
            if item.value() == 'dobj':
                obj = igt.get_item(item.attributes['dep']).value()
                if obj.lower() in eng_acc_pronouns:
                    pronoun = obj
    return pronoun

def get_english_wid(igt, word_id):
    # return the id of the translation verb corresponding with a givin language verb
    gw_tier = igt.get('gw')
    gloss_word = ''
    for gw in gw_tier:
        if 'alignment' in gw.attributes:
            if gw.attributes['alignment'] == word_id:
                gloss_word = gw.id
    g_tier = igt.get('g')
    glosses = []
    if g_tier:
        for g in g_tier:
            if gloss_word == g.content or gloss_word == g.segmentation:
                glosses.append(g.id)
    trans_word = ''
    for tier in igt:
        if tier.type == 'bilingual-alignments':
            for a in tier:
                if 'target' in a.attributes:
                    if a.attributes['target'] in glosses:
                        trans_word = a.attributes['source']
    return trans_word

def subordinate_verb(igt, verb_set):
    # returns true if a verb is subordinate
    # since we are using this to weed out non-finite forms,
    # this excludes ccomp, because we assume those to be finite more often
    is_subord = False
    w_ds = igt.get('w-ds')
    SUBORDINATE_DEPS = ['xcomp', 'rcmod', 'vmod', 'prepc', 'pcomp', 'advcl']
    if w_ds:
        for item in w_ds:
            if item.value() in SUBORDINATE_DEPS:
                if item.attributes['dep'] in verb_set:
                    is_subord = True
    return is_subord

def is_passive(igt, verb_set):
    # returns true if a verb is passive
    is_pass = False
    w_ds = igt.get('w-ds')
    for verb in verb_set:
        eng_verb = get_english_wid(igt, verb)
        tw_ds = igt.get('tw-ds')
        for item in tw_ds:
            if item.value() == 'auxpass':
                if item.attributes['head'] == eng_verb:
                    is_pass = True
    return is_pass

def english_subj(igt, verb_set):
    # returns true if any of the verbs' Englsih translations have a subject
    subj_types = ['nsubj', 'csubj', 'nsubjpass', 'csubjpass']
    has_subj = False
    for verb in verb_set:
        eng_verb = get_english_wid(igt, verb)
        if eng_verb != '':
            tw_ds = igt.get('tw-ds')
            for item in tw_ds:
                if item.value() in subj_types:
                    if item.attributes['head'] == eng_verb:
                        has_subj = True

    return has_subj

def is_capitalized(igt, verb):
    # returns true if the verb is capitalized and is not the first word of the sentence
    is_cap = False
    trans_word = get_english_wid(igt, verb)
    tw_tier = igt.get('tw')
    word_string = ''
    if tw_tier:
        for tw in tw_tier:
            if tw.id == trans_word:
                word_string = tw.value()
    if re.search('[A-Z]', word_string) and trans_word != 'tw1':
        is_cap = True
    return is_cap

def set_choices(arguments, agr_markers):
    # creates the choices object and sets arg-opt choices
    choices = MatrixChoices.ChoiceDict()
    # compare ratios of each argument and argument marker to predict req, opt, not
    # if we didn't see any dropped arguments, return an empty choices object: this basically means no argument optionality
    if arguments['dropped-subj'] == 0 and arguments['dropped-obj'] == 0:
        return choices
    # we predict all because that is most common
    if arguments['dropped-subj'] != 0:
        choices.__setitem__('subj-drop', 'subj-drop-all')
        #print('subj-drop-all')
    #else:
    #    choices.__setitem__('subj-drop', 'subj-drop-lex')
    # if there is at least one dropped obj on a verb that isn't an english dropping verb, we say all
    # this isn't perfect because find root isn't perfect. maybe I should do some math here
    if arguments['non-eng-drop'] != 0:
        choices.__setitem__('obj-drop', 'obj-drop-all')
        #print('obj-drop-all')
    elif arguments['dropped-obj'] != 0:
        choices.__setitem__('obj-drop', 'obj-drop-lex')
        #print('obj-drop-lex')
    # for whether marking is opt, req or not, we use a 5% margin of error at the top and 10% at the bottom
    if 'subj-drop' in choices:
        # for subject context we just say always (we're not doing inference on this yet)
        # excluding subj con for evaluation
        #choices.__setitem__('subj-con', 'subj-con-always')
        if arguments['overt-subj'] != 0:
            if agr_markers['overt-subj-mark'] / arguments['overt-subj'] > 0.75:
                choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-req')
                #print('subj-mark-no-drop-req')
            elif agr_markers['overt-subj-mark'] / arguments['overt-subj'] > 0.1:
                choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-opt')
                #print('subj-mark-no-drop-opt')
            else:
                choices.__setitem__('subj-mark-no-drop', 'subj-mark-no-drop-not')
                #print('subj-mark-no-drop-not')
        if arguments['dropped-subj'] != 0:
            if agr_markers['dropped-subj-mark'] / arguments['dropped-subj'] > 0.75:
                choices.__setitem__('subj-mark-drop', 'subj-mark-drop-req')
                #print('subj-mark-drop-req')
            elif agr_markers['dropped-subj-mark'] / arguments['dropped-subj'] > 0.1:
                choices.__setitem__('subj-mark-drop', 'subj-mark-drop-opt')
                #print('subj-mark-drop-opt')
            else:
                choices.__setitem__('subj-mark-drop', 'subj-mark-drop-not')
                #print('subj-mark-drop-not')
    if 'obj-drop' in choices:
        if arguments['overt-obj'] != 0:
            if agr_markers['overt-obj-mark'] / arguments['overt-obj'] > 0.75:
                choices.__setitem__('obj-mark-no-drop', 'obj-mark-no-drop-req')
                #print('obj-mark-no-drop-req')
            elif agr_markers['overt-obj-mark'] / arguments['overt-obj'] > 0.1:
                choices.__setitem__('obj-mark-no-drop', 'obj-mark-no-drop-opt')
                #print('obj-mark-no-drop-opt')
            else:
                choices.__setitem__('obj-mark-no-drop', 'obj-mark-no-drop-not')
                #print('obj-mark-no-drop-not')
        if arguments['dropped-obj'] != 0:
            if agr_markers['dropped-obj-mark'] / arguments['dropped-obj'] > 0.75:
                choices.__setitem__('obj-mark-drop', 'obj-mark-drop-req')
                #print('obj-mark-drop-req')
            elif agr_markers['dropped-obj-mark'] / arguments['dropped-obj'] > 0.1:
                choices.__setitem__('obj-mark-drop', 'obj-mark-drop-opt')
                #print('obj-mark-drop-opt')
            else:
                choices.__setitem__('obj-mark-drop', 'obj-mark-drop-not')
                #print('obj-mark-drop-not')
    return choices

def count_dependencies(xc):
    # counts the number of dependency tiers in the xigt corpus to check if there are enough for inference
    w_ds_count = 0
    tw_ds_count = 0
    w_ds_root_count = 0
    tw_ds_root_count = 0
    w_pos_count = 0
    for igt in xc:
        non_empty_wds = False
        non_empty_twds = False
        w_ds = igt.get('w-ds')
        if w_ds:
            w_ds_count += 1
            for item in w_ds:
                #if item.value() == 'root':
                # w_ds_root_count += 1
                if item.value():
                    non_empty_wds = True
            if non_empty_wds:
                w_ds_root_count += 1
        tw_ds = igt.get('tw-ds')
        if tw_ds:
            tw_ds_count += 1
            for item in tw_ds:
                #if item.value(): == 'root':
                #    tw_ds_root_count += 1
                if item.value():
                    non_empty_twds = True
            if non_empty_twds:
                tw_ds_root_count += 1
        w_pos = igt.get('w-pos')
        if w_pos:
            w_pos_count += 1

    #print('w-ds count: ' + str(w_ds_count))
    #print('non-empty w-ds count: ' + str(w_ds_root_count))
    #print('tw-ds count: ' + str(tw_ds_count))
    #print('tw-ds root count: ' + str(tw_ds_root_count))
    #print('w-pos count: ' + str(w_pos_count))
    #if w_ds_root_count < 500:
    #    print('WARNING: corpus does not have enough verb dependencies for inference')

def add_argument_marking_constraints(choices):
    # KPH this method is for adding the argument dropping constraints to the morphological rules. It assumes that the
    # choices fiel has verb pcs and is called from main.py
    #KPH we don't need to do anything if there isn't pro drop or if there isn't argument marking
    if 'subj-drop' not in choices and 'obj-drop' not in choices:
        return choices
    if choices['subj-mark-drop'] == 'subj-mark-drop-not' and choices['subj-mark-no-drop'] == 'subj-mark-no-drop-not':
        return choices
    for pc in choices['verb-pc']:
        subj_mark_pc = False
        obj_mark_pc = False
        for lrt in pc['lrt']:
            subj_mark_lrt = False
            obj_mark_lrt = False
            for feat in lrt['feat']:
                if feat['name'] == 'person' or feat['name'] == 'number' or feat['name'] == 'gender':
                    agr_pc = True
                    if feat['head'] == 'subj':
                        subj_mark_lrt = True
                        subj_mark_pc = True
                    if feat['head'] == 'obj':
                        obj_mark_lrt = True
                        obj_mark_pc = True
                    if feat['head'] == 'both':
                        subj_mark_lrt = True
                        obj_mark_lrt = True
                        subj_mark_pc = True
                        obj_mark_pc = True
            if subj_mark_lrt:
                id = lrt['name'] + '_feat' + str(len(lrt['feat'])+1)
                if choices['subj-mark-drop'] == 'subj-mark-drop-not':
                    if choices['subj-mark-no-drop'] == 'subj-mark-no-drop-opt' or choices['subj-mark-no-drop'] == \
                                                                                        'subj-mark-no-drop-req':
                        choices[id + '_name'] = 'dropped-arg'
                        choices[id + '_value'] = 'not-permitted'
                        choices[id + '_head'] = 'subj'
                elif choices['subj-mark-no-drop'] == 'subj-mark-no-drop-not':
                    if choices['subj-mark-drop'] == 'subj-mark-drop-opt' or choices['subj-mark-drop'] == 'subj-mark-drop-req':
                        choices[id + '_name'] = 'overt-arg'
                        choices[id + '_value'] = 'not-permitted'
                        choices[id + '_head'] = 'subj'
                elif choices['subj-mark-no-drop'] == 'subj-mark-no-drop-req':
                    if choices['subj-mark-drop'] == 'subj-mark-drop-opt':
                        choices[id + '_name'] = 'dropped-arg'
                        choices[id + '_value'] = 'permitted'
                        choices[id + '_head'] = 'subj'
                elif choices['subj-mark-no-drop'] == 'subj-mark-no-drop-opt':
                    if choices['subj-mark-drop'] == 'subj-mark-drop-req':
                        choices[id + '_name'] = 'overt-arg'
                        choices[id + '_value'] = 'permitted'
                        choices[id + '_head'] = 'subj'
            if obj_mark_lrt:
                if subj_mark_lrt:
                    id = lrt['name'] + '_feat' + str(len(lrt['feat'])+2)
                else:
                    id = lrt['name'] + '_feat' + str(len(lrt['feat'])+1)
                if choices['obj-mark-drop'] == 'obj-mark-drop-not':
                    if choices['obj-mark-no-drop'] == 'obj-mark-no-drop-opt' or choices['obj-mark-no-drop'] == \
                            'obj-mark-no-drop-req':
                        choices[id + '_name'] = 'dropped-arg'
                        choices[id + '_value'] = 'not-permitted'
                        choices[id + '_head'] = 'obj'
                elif choices['obj-mark-no-drop'] == 'obj-mark-no-drop-not':
                    if choices['obj-mark-drop'] == 'obj-mark-drop-opt' or choices[
                        'obj-mark-drop'] == 'obj-mark-drop-req':
                        choices[id + '_name'] = 'overt-arg'
                        choices[id + '_value'] = 'not-permitted'
                        choices[id + '_head'] = 'obj'
                elif choices['obj-mark-no-drop'] == 'obj-mark-no-drop-req':
                    if choices['obj-mark-drop'] == 'obj-mark-drop-opt':
                        choices[id + '_name'] = 'dropped-arg'
                        choices[id + '_value'] = 'permitted'
                        choices[id + '_head'] = 'obj'
                elif choices['obj-mark-no-drop'] == 'obj-mark-no-drop-opt':
                    if choices['obj-mark-drop'] == 'obj-mark-drop-req':
                        choices[id + '_name'] = 'overt-arg'
                        choices[id + '_value'] = 'permitted'
                        choices[id + '_head'] = 'obj'
        if subj_mark_pc and obj_mark_pc:
            subj_feat_name = ''
            if choices['subj-mark-drop'] == 'subj-mark-drop-not' and choices['subj-mark-no-drop'] == 'subj-mark-no-drop-req':
                subj_feat_name = 'overt-arg'
            elif choices['subj-mark-drop'] == 'subj-mark-drop-req' and choices['subj-mark-no-drop'] == 'subj-mark-no-drop-not':
                subj_feat_name = 'dropped-arg'
            if subj_feat_name != '':
                obj_feat_name = ''
                if choices['obj-mark-drop'] == 'obj-mark-drop-not' and choices[
                    'obj-mark-no-drop'] == 'obj-mark-no-drop-req':
                    obj_feat_name = 'overt-arg'
                elif choices['obj-mark-drop'] == 'obj-mark-drop-req' and choices[
                    'obj-mark-no-drop'] == 'obj-mark-no-drop-not':
                    obj_feat_name = 'dropped-arg'
                if obj_feat_name != '':
                    choices[pc['name'] + '_obligatory'] = 'on'
                    zero_rule = False
                    # KPH check to see if there is already a zero inflecting rule in this pc
                    for lrt in pc['lrt']:
                        for lri in lrt['lri']:
                            if lri['inflecting'] == 'no':
                                zero_rule = True
                    # KPH if not, add one
                    if not zero_rule:
                        id = pc['name'] + '_lrt' + str(len(pc['lrt']) + 1)
                        choices[id + '_name'] = id
                        choices[id + '_feat1_name'] = subj_feat_name
                        choices[id + '_feat1_value'] = 'not-permitted'
                        choices[id + '_feat1_head'] = 'subj'
                        choices[id + '_feat1_name'] = obj_feat_name
                        choices[id + '_feat1_value'] = 'not-permitted'
                        choices[id + '_feat1_head'] = 'obj'
                        choices[id + '_lri1_inflecting'] = 'no'
        elif subj_mark_pc:
            feat_name = ''
            if choices['subj-mark-drop'] == 'subj-mark-drop-not' and choices['subj-mark-no-drop'] == 'subj-mark-no-drop-req':
                feat_name = 'overt-arg'
            elif choices['subj-mark-drop'] == 'subj-mark-drop-req' and choices['subj-mark-no-drop'] == 'subj-mark-no-drop-not':
                feat_name = 'dropped-arg'
            if feat_name != '':
                choices[pc['name']+'_obligatory'] = 'on'
                zero_rule = False
                # KPH check to see if there is already a zero inflecting rule in this pc
                for lrt in pc['lrt']:
                    for lri in lrt['lri']:
                        if lri['inflecting'] == 'no':
                            zero_rule = True
                #KPH if not, add one
                if not zero_rule:
                    id = pc['name'] + '_lrt' + str(len(pc['lrt']) + 1)
                    choices[id + '_name'] = id
                    choices[id + '_feat1_name'] = feat_name
                    choices[id + '_feat1_value'] = 'not-permitted'
                    choices[id + '_feat1_head'] = 'subj'
                    choices[id + '_lri1_inflecting'] = 'no'
        elif obj_mark_pc:
            feat_name = ''
            if choices['obj-mark-drop'] == 'obj-mark-drop-not' and choices['obj-mark-no-drop'] == 'obj-mark-no-drop-req':
                feat_name = 'overt-arg'
            elif choices['obj-mark-drop'] == 'obj-mark-drop-req' and choices['obj-mark-no-drop'] == 'obj-mark-no-drop-not':
                feat_name = 'dropped-arg'
            if feat_name != '':
                choices[pc['name']+'_obligatory'] = 'on'
                zero_rule = False
                # KPH check to see if there is already a zero inflecting rule in this pc
                for lrt in pc['lrt']:
                    for lri in lrt['lri']:
                        if lri['inflecting'] == 'no':
                            zero_rule = True
                #KPH if not, add one
                if not zero_rule:
                    id = pc['name'] + '_lrt' + str(len(pc['lrt']) + 1)
                    choices[id + '_name'] = id
                    choices[id + '_feat1_name'] = feat_name
                    choices[id + '_feat1_value'] = 'not-permitted'
                    choices[id + '_feat1_head'] = 'obj'
                    choices[id + '_lri1_inflecting'] = 'no'
    return choices

def write_choices_file(choices, file):
    # writes the chocies file
    with open(file, 'w', encoding="UTF-8") as f:
        matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
        f.write(str(matrix_choices.choices))

if __name__ == '__main__':
    corpus = open(sys.argv[1], encoding='utf-8')
    xc = xigtxml.load(corpus)
    arg_opt_choices = infer_arg_opt(xc)
    write_choices_file(arg_opt_choices, 'C:/Users/Kristen/Documents/code/aggregation/choices_'+sys.argv[2])