#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kristen Howell
# this script infers the answers to the word order page and extracts lexical entries for detirminers and nouns
# with the appropriate choices

from XigtifiedToolbox import FeatDict
from mom import MatrixChoices
from xigt.codecs import xigtxml
import re

FXN = ['ANTIP', 'APPL', 'BEN', 'CAUS', 'COMPL', 'DECL', 'DUR', 'FUT', 'IMP', 'IND', 'INF', 'INTR', 'IMPV', 'IMPF',
       'IRR',
       'PASS', 'PFV', 'PERF', 'PRF', 'PRS', 'PRES', 'PROG', 'PROH', 'PST', 'PAST', 'PTCP', 'PURP', 'SBJV', 'TR',
       'HEARSAY', 'PERSONAL', 'RESULTATIVE', 'CONTINUATIVE', 'ABLE', 'ABIL']
MOD = ['can', 'may', 'might', 'should', 'would', 'could', 'do', 'does', 'did']
BE = ['be', 'is', 'am', 'are', 'was', 'were']
HAVE = ['have', 'has', 'had']
KNOWN_FEATS = FeatDict.FeatureDictionary().all_keys
feat_dict = FeatDict.FeatureDictionary().dictionaries

def infer_aux(xc, verbs, case_system, igt2items, known_glosses, wo_choices, verb_tags=['verb', 'v', 'VERB']):
    aux_dict = {}
    second_pos_count = 0
    igt_count = 0
    for igt in xc:
        if verby_phrase(igt):
            igt_count += 1  # in principle this should count all igt, but in practice we miss too many, so we'll only check where we found an aux
        auxiliaries = find_auxiliaries(igt, known_glosses, verb_tags)
        if auxiliaries[0] != []:
            multiple_aux = has_multiple_aux(auxiliaries[1], igt)
            for gloss, aux in zip(auxiliaries[0], auxiliaries[1]):
                if aux == 'w2':
                    second_pos_count += 1
                if igt.id in igt2items and aux in igt2items[igt.id]:
                    orth = str(igt2items[igt.id][aux].root)
                else:
                    orth = get_orth(gloss, igt).lower()
                comp = find_verbal_complement(aux, igt, verb_tags)
                position = get_aux_comp_position(aux, comp)
                comp_type = get_complement_type(aux, comp, igt)
                subj_case = get_subj_case(igt, comp, verbs, igt2items)
                if subj_case == 'verb' and case_system == 'none':
                    subj_case = 'none'
                comp_features = get_complement_features(comp, igt)
                aux_features = get_aux_features(gloss, igt)
                if orth in aux_dict:
                    if multiple_aux:
                        aux_dict[orth]['multiple_aux'] += 1
                    if subj_case == 'none':
                        aux_dict[orth]['subj_case'] = subj_case # this is probably unnecessaary because it would already be none
                    elif subj_case != aux_dict[orth]['subj_case']:
                        aux_dict[orth]['subj_case'] = 'verb' # KPH if the subj case is different, it probably is coming from the verb
                else:
                    aux_dict[orth] = {'multiple_aux': 0, 'position': {'before': 0, 'after': 0}, 'comp_type': {'s': 0, 'vp': 0, 'v': 0},
                                         'subj_case': subj_case, 'comp_features': {}, 'aux_features': {}}
                if position != '':
                    aux_dict[orth]['position'][position] += 1
                if comp_type != '':
                    aux_dict[orth]['comp_type'][comp_type] += 1
                if multiple_aux:
                    aux_dict[orth]['multiple_aux'] += 1
                for feat in comp_features:
                    if feat in aux_dict[orth]['comp_features']:
                        aux_dict[orth]['comp_features'][feat] += 1
                    else:
                        aux_dict[orth]['comp_features'][feat] = 1
                for feat in aux_features:
                    if feat in aux_dict[orth]['aux_features']:
                        aux_dict[orth]['aux_features'][feat] += 1
                    else:
                        aux_dict[orth]['aux_features'][feat] = 1

    choices = set_aux_choices(aux_dict, wo_choices)
    # todo tune this threshold after updating known feats
    if igt_count == 0:
        second_pos = False
    elif float(second_pos_count/igt_count) > .75:
        second_pos = True
    else:
        second_pos = False
    return choices, aux_dict, second_pos

def find_auxiliaries(igt, known_glosses, verb_tags):
    aux_glosses = []
    auxes = []
    # KPH: get the low hanging fruit of auxiliaries that are glossed as such
    if igt.get('g'):
        for g in igt.get('g'):
            grams = re.split('\.|-', g.value())
            if 'AUX' in grams:
                aux_glosses.append(g.id)
                auxes.append(get_word_from_gloss(g, igt))

    # KPH: then look for words with common auxiliary glosses that don't also contain roots
    if igt.get('gw'):
        for gw in igt.get('gw'):
            potential_aux = True
            gw_glosses = [g for g in igt.get('g') if 'segmentation' in g.attributes and g.attributes['segmentation'] == gw.id]
            possible_roots = set()
            for g in gw_glosses:
                glosses = re.split('-|=', g.value())
                for gloss in glosses:
                    for gram in re.split('\.', gloss):
                        if gram.upper() not in FXN and gram.lower() not in KNOWN_FEATS and gram.lower() not in HAVE+BE+MOD and gram not in known_glosses:
                            potential_aux = False
                        else:
                            if gram in FXN:
                                possible_roots.add(g.id)
                            if gram in HAVE and not transitive_have_in_translation(igt):
                                possible_roots.add(g.id)
                            if gram in BE:
                                possible_roots.add(g.id)
            if potential_aux:
                if is_verby(gw):#KPH this is meant to distinguish between auxiliaries and other function words like dets and prons
                    if not is_copula(gw, igt, verb_tags):
                        if 'alignment' in gw.attributes:
                            if possible_roots != set():
                                aux_glosses.append(max(possible_roots)) # there should only be one, but if their are more, take the rightmost
                            else:
                                aux_glosses.append(max([g.id for g in gw_glosses])) #KPH I'm grabbing the rightmost gloss as the root, because I think mom takes the rightmost morpheme if all glosses are known glosses
                            if gw.attributes['alignment'] not in auxes:
                                auxes.append(gw.attributes['alignment'])
    return aux_glosses, auxes

def verby_phrase(igt):
    verby = False
    if igt.get('tw-ds'):
        root = ''
        for ds in igt.get('tw-ds'):
            if ds.value() == 'root':
                if 'dep' in ds.attributes:
                    root = ds.attributes['dep']
        if root != '':
            if igt.get('tw-pos'):
                for pos in igt.get('tw-pos'):
                    if pos.attributes['alignment'] == root:
                        if pos.value() == 'VERB':
                            if igt.get('tw-w-aln'):
                                for alignment in igt.get('tw-w-aln'):
                                    if alignment.attributes['source'] == root: #check that this actually aligns to something (i added this because wambaya has a null copula)
                                        verby = True
    return verby

def get_word_from_gloss(g, igt):
    m = g.attributes['alignment'].split(',')[0]  # KPH sometimes the gloss is aligned to multiple morphs- just take the first
    morph_item = igt.get('m').get(m)
    if 'segmentation' in morph_item.attributes:
        w = morph_item.attributes['segmentation']
    return w

def transitive_have_in_translation(igt):
    transitive_have = False
    if not igt.get('tw'):
        return transitive_have
    for tw in igt.get('tw'):
        if tw.value() in HAVE:
            if igt.get('tw-ds'):
                for dep in igt.get('tw-ds'):
                    if 'head' in dep.attributes:
                        if dep.attributes['head'] == tw.id:
                            if dep.value() == 'dobj':
                                transitive_have = True
    return transitive_have

def is_verby(gw):
    verby_features = {}
    verby_features.update(feat_dict['tenseFeatures'].copy())
    verby_features.update(feat_dict['aspectFeatures'].copy())
    verby_features.update(feat_dict['moodFeatures'].copy())
    verby_features.update(feat_dict['negFeatures'].copy())
    verby_features.update(feat_dict['tenseAspectFeatures'].copy())
    verby_features.update(feat_dict['voiceFeatures'].copy())
    verby_features.update(feat_dict['formFeatures'].copy())
    verby_features.update(feat_dict['aktionsartFeatures'].copy())
    verby_features.update(feat_dict['evidFeatures'].copy())
    verby_features.update(feat_dict['interFeatures'].copy())
    verby_features.update(feat_dict['politeFeatures'].copy())
    verby = False
    grams = re.split('\.|-', gw.value())
    for gram in grams:
        if gram.lower() in verby_features:
            verby = True
    return verby

def get_orth(g, igt):
    m_id = igt.get('g').get(g).attributes['alignment']
    orth = igt.get('m').get(m_id).value()
    return orth

def find_verbal_complement(aux, igt, verb_tags):
    comp = get_complement_from_ds(aux, igt)
    if comp == '':
        comp = guess_complement_from_proximity(aux, igt, verb_tags)
    return comp

def get_complement_from_ds(aux, igt):
    comp = ''
    if igt.get('w-ds'):
        for dep in igt.get('w-ds'):
            if dep.value() == 'aux':
                if dep.attributes['dep'] == aux and 'head' in dep.attributes:
                    comp = dep.attributes['head']
    return comp

def guess_complement_from_proximity(aux, igt, verb_tags):
    comp = ''
    verbs = []
    pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
    if pos_tier:
        for pos in pos_tier:
            if pos.value() in verb_tags:
                verb = pos.attributes['alignment']
                if 'gw' in verb:
                    if 'alignment' in igt.get('gw').get(verb).attributes:
                        verb = igt.get('gw').get(verb).attributes['alignment']
                    else:
                        continue
                verbs.append(verb)
    for verb in verbs:
        if verb == aux:
            continue # this shouldn't be happen because this is the same as the copula check
        else:
            verb_id = int(verb[1:])
            aux_id = int(aux[1:])
            if abs(verb_id-aux_id) == 1:
                comp = verb
    return comp

def is_copula(gw, igt, verb_tags):
    copula = False
    # is be in the gloss
    for gloss in re.split('\.|-|=', gw.value()):
        if gloss in BE:
            copula = True
    # is is the main verb of the sentence
    if copula:
        verbs = []
        pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
        if pos_tier:
            for pos in pos_tier:
                if pos.value() in verb_tags:
                    verb = pos.attributes['alignment']
                    if 'gw' in verb:
                        if 'alignment' in igt.get('gw').get(verb).attributes:
                            verb = igt.get('gw').get(verb).attributes['alignment']
                    verbs.append(verb)
        # if len(verbs) == 1: # KPH this will only work for simple sentences
        #     if verbs[0] == gw.attributes['alignment']:
        #         copula = True
        if len(verbs) > 1:
            for verb in verbs:
                if igt.get('w-ds'):
                    for dep in igt.get('w-ds'):
                        if 'dep' in dep.attributes and 'head' in dep.attributes:
                            for verb2 in verbs:
                                if (dep.attributes['dep'] == verb and dep.attributes['head'] == verb2):
                                    copula = False
    return copula

def get_aux_comp_position(aux_str, comp_str):
    if aux_str == '':
        aux = 0
    else:
        aux = int(aux_str[1:])
    if comp_str == '':
        comp = 0
    else:
        comp = int(comp_str[1:])
    if comp == 0:
        position = ''
    elif comp < aux:
        position = 'before'
    elif comp > aux:
        position = 'after'
    return position

def get_complement_type(aux_id, comp_id, igt):
    # does a subject or object intervene between aux and v?
    if comp_id == '':
        return ''
    s = 0
    o = 0
    comp = int(comp_id[1:])
    aux = int(aux_id[1:])
    type = ''
    subj_deps = ['nsubj', 'nsubjpass', 'csubj', 'csubjpass']
    if igt.get('w-ds'):
        for dep in igt.get('w-ds'):
            if 'head' in dep.attributes:
                if dep.value() in subj_deps and dep.attributes['head'] == comp_id:
                    s = int(dep.attributes['dep'][1:])
                if dep.value() == 'dobj' and dep.attributes['head'] == comp_id:
                    o = int(dep.attributes['dep'][1:])
    if o != 0:
        if o < aux < comp:
            type = 'v'
        if comp < aux < o:
            type = 'v'
    if s != 0:
        if s < aux < comp:
            type = 'vp'
        if comp < aux < s:
            type = 'vp'
        if aux < s < comp:
            type = 's'
        if comp < s < aux:
            type = 's'
    return type

def has_multiple_aux(aux_list, igt):
    multiple_aux = False
    if len(aux_list) > 1:
        # because projection often isn't possible for auxes, we aren't going to look for a dependency between teh two
        # instead, we'll just look for two adjacent auxes
        ids = [int(id[1:]) for id in aux_list]
        i = 1
        while i < len(ids):
            if ids[i]-ids[i-1] == 1:
                multiple_aux = True
            i += 1
    return multiple_aux

def get_subj_case(igt, comp, verbs, igt2items):
    # get the case of the verb in the grammar
    if comp == '':
        return ''
    if igt.id not in igt2items:
        return ''
    if comp not in igt2items[igt.id]:
        return ''
    verb_lemma = str(igt2items[igt.id][comp].root)
    if verb_lemma in verbs['intrans']:
        verb_subj_case = verbs['intrans'][verb_lemma]['case_frame']['subj']
    elif verb_lemma in verbs['trans']:
        verb_subj_case = verbs['trans'][verb_lemma]['case_frame']['subj']
    else:
        verb_subj_case = ''

    igt_subj_case = ''
    subj_deps = ['nsubj', 'nsubjpass', 'csubj', 'csubjpass']
    if igt.get('w-ds'):
        for dep in igt.get('w-ds'):
            if 'head' in dep.attributes:
                if dep.value() in subj_deps and dep.attributes['head'] == comp:
                    subj = dep.attributes['dep']
                    for gw in igt.get('gw'):
                        if 'alignment' in gw.attributes:
                            if gw.attributes['alignment'] == subj:
                                glosses = re.split('\.|-', gw.value())
                                for gloss in glosses:
                                    if gloss.lower() in feat_dict['caseFeatures']:
                                        igt_subj_case = gloss

    if igt_subj_case.lower() == verb_subj_case.lower():
        aux_subj_case = 'verb'
    else:
        aux_subj_case = igt_subj_case.lower()
    return aux_subj_case

def get_complement_features(comp, igt):
    comp_features = set()
    for gw in igt.get('gw'):
        if 'alignment' in gw.attributes:
            if gw.attributes['alignment'] == comp:
                glosses = re.split('\.|-',gw.value())
                for gloss in glosses:
                    if gloss.lower() in KNOWN_FEATS:
                        comp_features.add(gloss.lower())
    return comp_features

def get_aux_features(g, igt):
    features = set()
    value = igt.get('g').get(g).value()
    glosses = re.split('\.', value)
    for gloss in glosses:
        if gloss.lower() in KNOWN_FEATS:
            features.add(gloss.lower())
    return features

def set_aux_choices(aux_dict, wo_choices):
    choices = MatrixChoices.ChoiceDict()
    before_count = 0
    after_count = 0
    v_count = 0
    vp_count = 0
    s_count = 0
    multiple_count = 0
    for aux in aux_dict:
        before_count += aux_dict[aux]['position']['before']
        after_count += aux_dict[aux]['position']['after']
        v_count += aux_dict[aux]['comp_type']['v']
        vp_count += aux_dict[aux]['comp_type']['vp']
        s_count += aux_dict[aux]['comp_type']['s']
        multiple_count += aux_dict[aux]['multiple_aux']
    if aux_dict == {}:
        choices.__setitem__('has-aux', 'no')
    else:
        choices.__setitem__('has-aux', 'yes')
        choices.__setitem__('form-fin-nf', 'on')
        if multiple_count > 0:
            choices.__setitem__('multiple-aux', 'yes')
        else:
            choices.__setitem__('multiple-aux', 'no')
        if before_count > after_count:
            choices.__setitem__('aux-comp-order', 'before')
        else:
            choices.__setitem__('aux-comp-order', 'after')
        if v_count > 0:
            choices.__setitem__('aux-comp', 'v')
        elif vp_count > 0:
            if choices['aux-comp-order'] == 'before' and wo_choices['word-order'] == 'osv':
                choices.__setitem__('aux-comp', 'v')
            elif choices['aux-comp-order'] == 'after' and wo_choices['word-order'] == 'vso':
                choices.__setitem__('aux-comp', 'v')
            else:
                choices.__setitem__('aux-comp', 'vp')
        else:
            if choices['aux-comp-order'] == 'before' and wo_choices['word-order'] == 'osv':
                choices.__setitem__('aux-comp', 'v')
            elif choices['aux-comp-order'] == 'after' and wo_choices['word-order'] == 'vso':
                choices.__setitem__('aux-comp', 'v')
            elif s_count > 0:
                choices.__setitem__('aux-comp', 's') #kph we're trying to disfavor s because it's uncommon and doesn't work with subj features. so we only infer it if there is evidence for it and nothing else
            else:
                choices.__setitem__('aux-comp', 'vp') #kph otherwise favor vp because v is expensive (because of argument composition)

    # does tha language have auxes?
    # do they go before or after their complement?
    # are their complements s, vp or v?
    # can their be multiple auxes per clause?
    return choices
