#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kriten Howell
# this script has been tested on a handful of languages from ODIN 2.1.
# Bugs may be found when testing it on more datasets
# Parts of this code come from extract-lexicon-xigt.py by Jeff Heath

import sys
import re
import getopt
from xigt import xigtpath as xp
from xigt import ref
from XigtifiedToolbox import FeatDict

feat_dict = FeatDict.FeatureDictionary()

def get_options():
    """
    Read in the user input- return options, a dictionary with the corpus, method
    and case names, if specified
    """
    options = {}
    user_case_names = {'s': '', 'a': '', 'o': '', 'n': ''}
    try:
        opts, args = getopt.getopt(sys.argv[1:], 's:a:o:c:h')
    except getopt.GetoptError as err:
        print(str(err), file=sys.stderr)
        exit(1)
    options['corpus'] = args[0]
    for opt, arg in opts:
        print('opt: ', opt)
        print('arg: ', arg)
        if opt == '-h':
            print('Usage: extract-case-frame.py [options] Xigt-corpus')
            print('  -c gram  : automatically detect case using the gram methodology')
            print('  -c sao   : automatically detect case using the sao methodology')
            print('  -N GRAM  : use given gram for unmarked nouns')
            print('  -h       : display this help')
            print(' to specify case grams yourself, enter value for s, a, and o:')
            print('  -s GRAM  : use given gram as subject marker')
            print('  -a GRAM  : use given gram as agent marker')
            print('  -o GRAM  : use given gram as object marker')
            sys.exit()
        elif opt == '-c':
            if arg in ['gram', 'sao']:
                options['method'] = arg
            else:
                print('Error: Invalid case value extraction method: {} {}.'.format(opt, arg), file=sys.stderr)
                print('       Choose among the following methods: "gram", "sao".', file=sys.stderr)
                sys.exit(1)
        elif opt in ['-s', '-a', '-o']:
            if arg[0] == "-":
                # it can 'take over' the following option if you don't provide an argument
                print('Error: You must provide an argument for the {} case name.'.format(opt), file=sys.stderr)
                print('       To treat {} case as unmarked, use {} 0.'.format(opt, opt), file=sys.stderr)
                sys.exit(1)
            elif arg == '0':
                user_case_names[opt[-1]] = ''
            else:
                user_case_names[opt[-1]] = arg
        elif opt == '-N':
            # nouns that aren't marked receive this case name
            user_case_names['n'] = arg
        options['case_names'] = user_case_names
    return options

def get_count_grams(corpus):
    """
    Return case grams from infer_case_from_counts
    """
    import infer_case_from_counts_xigt
    if __name__ == '__main__':
        sao_output = infer_case_from_counts_xigt.infer_case(corpus)
    count_case_grams = {'s': sao_output['s'], 'a': sao_output['a'], 'o': sao_output['o']}
    return count_case_grams

def get_gram_grams(corpus):
    """
    Return case grams from infer_case_from_grams
    """
    import infer_case_from_grams_xigt
    if __name__ == '__main__':
        gram_case_system = infer_case_from_grams_xigt.infer_case(corpus)
        print('case system= ', gram_case_system)
    if gram_case_system.lower() == 'nom-acc':
        gram_case_grams = {'s': 'nom', 'a': 'nom', 'o': 'acc'}
    elif gram_case_system.lower() == 'erg-abs':
        gram_case_grams = {'s': 'abs', 'a': 'erg', 'o': 'abs'}
    elif gram_case_system.lower() == 'none':
        gram_case_grams = {'s': '', 'a': '', 'o': ''}
    elif gram_case_system.lower() == 'split-erg':
        gram_case_grams = {'s': 'nom', 'a': 'erg', 'o': 'acc'}
    return gram_case_grams

def get_dependents_case(igt, dep, all_case_grams, sao_case_grams, gloss_ids):
    """
    From and igt and a noun (called dep as this will be an item corresponding to
    a dependent of the verb), return the case marking on that noun
    """
    dep_case = ''
    item = igt.get_item(dep.attributes['dep'])
    dependent = igt.get_item(dep.attributes['dep'])
    referrers = ref.referrers(igt, item.id, ['alignment'])['alignment']
    glosses = [r for r in referrers if r in gloss_ids]
    for each in glosses:
        dep_gloss_word = igt.get_item(each).value().lower()
        dep_grams = re.split('-|\.|\(\)|:|\/', dep_gloss_word)
        for each in dep_grams:
            if each in all_case_grams:
                dep_case = each
        if dep_case == '':
            if 'n' in sao_case_grams:
                dep_case = sao_case_grams['n']
    return dep_case

def check_for_eng_obj(igt, verb_id):
    """
    From and IGT, determine if the direct object in the English translation
    is a pronoun or if the translation verb has a direct object. If so return True,
    otherwise, return False. """
    obj = False
    eng_acc_pronouns = ['me', 'us', 'you', 'him', 'her', 'them', 'it', 'someone', 'something', 'somewhere']
    for tier in xp.findall(igt, 'tier[@type="dependencies"][@dep="tw"]'):
        for item in tier:
            if item.value() == 'dobj':
                obj = igt.get_item(item.attributes['dep']).value()
                if obj.lower() in eng_acc_pronouns:
                    obj = True
    if not obj:
        eng_verb = get_english_wid(igt, verb_id)
        if eng_verb != '':
            if igt.get('tw-ds'):
                for dep in igt.get('tw-ds'):
                    if dep.value() == 'dobj' and 'head' in dep.attributes:
                        if dep.attributes['head'] == verb_id:
                            obj = True
    return obj

def get_english_wid(igt, word_id):
    # return the id of the translation verb corresponding with a givin language verb
    gw_tier = igt.get('gw')
    gloss_word = ''
    for gw in gw_tier:
        if 'alignment' in gw.attributes:
            if gw.attributes['alignment'] == word_id:
                gloss_word = gw.id
    g_tier = igt.get('g')
    glosses = []
    if g_tier:
        for g in g_tier:
            if gloss_word == g.content or gloss_word == g.segmentation:
                glosses.append(g.id)
    trans_word = ''
    for tier in igt:
        if tier.type == 'bilingual-alignments':
            for a in tier:
                if 'target' in a.attributes:
                    if a.attributes['target'] in glosses:
                        trans_word = a.attributes['source']
    return trans_word

def check_for_obj_agr(igt, verb_id, agr_morphemes):
    obj = False
    morphemes = re.split('-|=', igt.get('w').get(verb_id).value())
    if igt.get('gw').get('g'+verb_id):
        glosses = re.split('-|=', igt.get('gw').get('g'+verb_id).value())
    else:
        glosses = []
    for morph in morphemes:
        if morph.lower() in agr_morphemes['obj_morphemes'] or morph.lower() in agr_morphemes['portmanteau_morphemes']:
            obj = True
    for gloss in glosses:
        for dict in feat_dict.dictionaries:
            if 'head' in dict.lower():
                if gloss.lower() in feat_dict.dictionaries[dict]:
                    if 'obj' in feat_dict.dictionaries[dict][gloss.lower()]:
                        obj = True
    return obj

def get_verb_case_frame(igt, all_case_grams, sao_case_grams, igt2items, pos_tags, agr_morphemes):
    """
    From an IGT and a set of case grams, create a list of verbs in that IGT,
    each of which is represented by a dictionary containing the verbs transitivity,
    orthography of its root, translation, and case frame"""
    if igt.get('gw') is None: #or igt.get('g') is None: # Should we be relying on 'gw' tier though?
        return {}
    gloss_ids = set(i.id for i in igt.get('gw'))
    verbs = []
    #for tier in xp.findall(igt, '//tier[@type="pos"][@alignment="w"]'):
    #  this is very slow; perhaps we could assume w-pos or pos? That's what MOM does.
    # KPH: Typically gw-pos is annotated with 'VERB' etc, while w-pos and pos may use different
    # tags (eg from the stanford parser). If the more detailed tier is missing, check the next tier up.
    pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos')
    if not pos_tier:
        pos_tier = igt.get('gw-pos')
    verb = {}
    if pos_tier:
        for item in pos_tier:
            #if item is not None and item.value().upper() == 'VERB':
            if item.text:
                if 'v.' in item.text.lower() or item.text in pos_tags:
                    verb_item = igt.get_item(item.attributes['alignment'])
                    if verb_item is not None:
                        # if we are looking at gloss words, we need to traverse up to the word
                        verb_id = ''
                        if 'gw' in verb_item.id:
                            if 'alignment' in verb_item.attributes:
                                word_item = igt.get_item(verb_item.attributes['alignment'])
                                verb_id = word_item.id
                            elif 'segmentation' in verb_item.attributes:
                                word_item = igt.get_item(verb_item.attributes['segmentation'])
                                verb_id = word_item.id
                        elif 'm' in verb_item.id:
                            if 'segmentation' in verb_item.attributes and igt.get_item(verb_item.attributes['segmentation']):
                                word_item = igt.get_item(verb_item.attributes['segmentation'])
                                verb_id = word_item.id
                            elif 'alignment' in verb_item.attributes:
                                word_item = igt.get_item(verb_item.attributes['alignment'])
                                verb_id = word_item.id
                        else:
                            verb_id = verb_item.id
                        if not (igt.id in igt2items and verb_id in igt2items[igt.id]):
                            return {}
                        mom_item = igt2items[igt.id][verb_id]
                        verb_lemma = mom_item.root
                        translation = '.'.join(mom_item.root.glosses)
                        #verb_lemma, translation = find_root.find_root(igt, verb_item)
                        if verb_lemma == '':
                            return {}
                        subj_case = ''
                        obj_case = ''
                        nsubj = ''
                        dobj = ''
                        clausal_complement = False
                        ditransitive = False
                        #for tier in xp.findall(igt, 'tier[@type="dependencies"][@dep="w"]'): # Can we just assume w-ds here, for speed?
                        ds_tier = igt.get('w-ds')
                        if ds_tier:
                            for item in ds_tier:
                                if item is not None:
                                    if 'head' in item.attributes:
                                        if item.value() == 'nsubj' and verb_id in item.attributes['head']:
                                            nsubj = 'overt'
                                            subj_case = get_dependents_case(igt, item, all_case_grams, sao_case_grams, gloss_ids)
                                        if item.value() == 'dobj' and verb_id in item.attributes['head']:
                                            dobj = 'overt'
                                            obj_case = get_dependents_case(igt, item, all_case_grams, sao_case_grams, gloss_ids)
                                        if item.value() == 'xcomp' and verb_id in item.attributes['head']:
                                            clausal_complement = True
                                        if item.value() == 'iobj'and verb_id in item.attributes['head']:
                                            ditransitive = True
                        if clausal_complement == True or ditransitive == True:
                            continue
                        elif nsubj == 'overt' and dobj == 'overt':
                            # both overt
                            # transitive
                            transitivity = 'trans'
                            case_frame = {'subj': subj_case, 'obj': obj_case}
                        elif nsubj == 'overt':
                            #  nsubj overt
                            #  intransitive or dropped object
                            object = check_for_eng_obj(igt, verb_id)
                            if not object:
                                object = check_for_obj_agr(igt, verb_id, agr_morphemes)
                            if not object:
                                transitivity = 'intrans'
                                case_frame = {'subj': subj_case}
                            else:
                                transitivity = 'trans'
                                case_frame = {'subj': subj_case, 'obj': obj_case}#, 'obj': sao_case_grams['o']}
                        elif dobj == 'overt':
                            # dobj overt
                            # transitive verb, dropped subject
                            transitivity = 'trans'
                            case_frame = {'subj': subj_case, 'obj': obj_case}#{'subj': sao_case_grams['a'], 'obj': obj_case}
                        else:
                            # nothing overt
                            object = check_for_eng_obj(igt, verb_id)
                            if not object:
                                object = check_for_obj_agr(igt, verb_id, agr_morphemes)
                            if not object:
                                transitivity = 'intrans'
                                case_frame = {'subj': subj_case}#{'subj': sao_case_grams['a']}
                            else:
                                transitivity = 'trans'
                                case_frame = {'subj': subj_case, 'obj': obj_case}#{'subj': sao_case_grams['a'], 'obj': sao_case_grams['o']}
                        verb = {transitivity: {'orth': str(verb_lemma), 'eng': translation, 'case_frame': case_frame, 'igt': igt.id, 'verb_id': verb_id}}
                        # KPH 5/20/19 Instead of updating the mom item below, I am going to update the mom items after all
                        # transitivity has been found for all verbs, so that we don't get multiple entries (eg. trans and
                        # intrans for a single verb in a pro-drop language
                        #mom_item.transitivity = transitivity
                        #mom_item.case_frame = case_frame
                        verbs.append(verb)
    return verbs

def add_to_dict(verbs, verb, obj_drop, sao_case_grams):
    """
    Add each verb in a list of verb (from one IGT) to the dictionary of verbs for the langauge
    """
    if 'trans' in verb:
        if verb['trans']['orth'] in verbs['trans']:
            if verb['trans']['case_frame']['subj']:
                verbs['trans'][verb['trans']['orth']]['case_frame']['subj'] = verb['trans']['case_frame']['subj']
            if verb['trans']['case_frame']['obj']:
                verbs['trans'][verb['trans']['orth']]['case_frame']['obj'] = verb['trans']['case_frame']['obj']
            verbs['trans'][verb['trans']['orth']]['id'].append((verb['trans']['igt'], verb['trans']['verb_id']))
        else:
            verbs['trans'][verb['trans']['orth']] = {}
            verbs['trans'][verb['trans']['orth']]['case_frame'] = {}
            if verb['trans']['case_frame']['subj']:
                verbs['trans'][verb['trans']['orth']]['case_frame']['subj'] = verb['trans']['case_frame']['subj']
            else:
                verbs['trans'][verb['trans']['orth']]['case_frame']['subj'] = sao_case_grams['s']
            if verb['trans']['case_frame']['obj']:
                verbs['trans'][verb['trans']['orth']]['case_frame']['obj'] = verb['trans']['case_frame']['obj']
            else:
                verbs['trans'][verb['trans']['orth']]['case_frame']['obj'] = sao_case_grams['o']
            verbs['trans'][verb['trans']['orth']]['eng'] = verb['trans']['eng']
            verbs['trans'][verb['trans']['orth']]['id'] = [(verb['trans']['igt'], verb['trans']['verb_id'])]
        if obj_drop and (verb['trans']['orth'] in verbs['intrans']):
            del verbs['intrans'][verb['trans']['orth']]
    elif 'intrans' in verb:
        if obj_drop and (verb['intrans']['orth'] in verbs['trans']):
            if verb['intrans']['orth'] in verbs['intrans']:
                del verbs['intrans'][verb['intrans']['orth']]
        else:
            if verb['intrans']['orth'] in verbs['intrans']:
                if verb['intrans']['case_frame']['subj']:
                    verbs['intrans'][verb['intrans']['orth']]['case_frame']['subj'] = verb['intrans']['case_frame']['subj']
                verbs['intrans'][verb['intrans']['orth']]['id'].append((verb['intrans']['igt'], verb['intrans']['verb_id']))
            else:
                verbs['intrans'][verb['intrans']['orth']] = {}
                if verb['intrans']['case_frame']['subj']:
                    verbs['intrans'][verb['intrans']['orth']]['case_frame'] = verb['intrans']['case_frame']
                else:
                    verbs['intrans'][verb['intrans']['orth']]['case_frame'] = {'subj': sao_case_grams['a']}
                verbs['intrans'][verb['intrans']['orth']]['eng'] = verb['intrans']['eng']
                verbs['intrans'][verb['intrans']['orth']]['id'] = [(verb['intrans']['igt'], verb['intrans']['verb_id'])]
    return verbs

def extract_case_frame(options, agr_morphemes, igt2items, pos_tags, xc):
    """
    The Main method- extract the case frame of all verbs in a corpus
    """
    #xigt_path = os.path.abspath(".")
    #sys.path.append(xigt_path)
    #file = open(options['corpus'])
    #xc = xigtxml.load(file)
    if 'method' in options:
        if options['method'] == 'sao':
            sao_case_grams = get_count_grams(options['corpus'])
        if options['method'] == 'gram':
            sao_case_grams = get_gram_grams(options['corpus'])
    else:
        sao_case_grams = (options['case_names'])
    all_case_grams = ['nom', 'acc', 'abs', 'erg', 'abl', 'all',
                      'com', 'dat', 'gen', 'ins', 'loc', 'obl']
    for gram in sao_case_grams:
        if gram not in all_case_grams:
            all_case_grams.append(sao_case_grams[gram])

    verbs = {'trans': {}, 'intrans': {}}
    for igt in xc:
        verbs_in_igt = get_verb_case_frame(igt, all_case_grams, sao_case_grams, igt2items, pos_tags, agr_morphemes)
        for verb in verbs_in_igt:
            for verb in verbs_in_igt:
                verbs = add_to_dict(verbs, verb, options['obj_drop'], sao_case_grams)
    types = ['intrans', 'trans']
    for type in types:
        for verb in verbs[type]:
            for igt_id, verb_id in verbs[type][verb]['id']:
                mom_item = igt2items[igt_id][verb_id]
                mom_item.transitivity = type
                mom_item.case_frame = verbs[type][verb]['case_frame']
    return verbs

def print_verbs(verbs):
    # KPH I've changed verbs from a dict to a list- this method will no longer work
    print('intransitive verbs')
    for verb in verbs['intrans']:
        print(verb, '\t', verbs['intrans'][verb]['eng'], '\tsubject: ', verbs['intrans'][verb]['case_frame']['subj'])
    print()
    print('transitive verbs')
    for verb in verbs['trans']:
        print(verb, '\t', verbs['trans'][verb]['eng'],  '\tsubject: ', verbs['trans'][verb]['case_frame']['subj'], '\tobject: ', verbs['trans'][verb]['case_frame']['obj'])

if __name__ == '__main__':
    options = get_options()
    verbs = extract_case_frame(options)
    print_verbs(verbs)
