#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kriten Howell
# this script has been tested on a handful of languages from ODIN 2.1.
# Bugs may be found when testing it on more datasets
# Parts of this code come from extract-lexicon-xigt.py by Jeff Heath and some code written by Alex Burrel

import sys
import os
import re
import getopt
from collections import Counter
from xigt import xigtpath as xp
from xigt import ref
from xigt.codecs import xigtxml
from src.aggutils import find_root
from XigtifiedToolbox import XigtReader

# KPH: Some global variables
# these sets for nouns and verbs should probably be replaced with the mom variables. Just using them short-term while adapting the VerbValenceFromIGT script
NOUNS = ['n', 'pro', 'NOUN', 'PRP', 'NN', 'NP', 'NNS', 'NNP', 'NNPS', 'PRON', 'WP'] # KPH: removed PRP$-- this is for posessive pronouns, added WP (wh-pronouns)
VERBS = ['vi', 'vt', 'VBG', 'VBP', 'VP', 'VBN', 'VB', 'VBD', 'VBZ', 'VERB']

def get_options():
    """
    Read in the user input- return options, a dictionary with the corpus, method
    and case names, if specified
    """
    options = {}
    user_case_names = {'s': '', 'a': '', 'o': '', 'n': ''}
    try:
        opts, args = getopt.getopt(sys.argv[1:], 's:a:o:c:h')
    except getopt.GetoptError as err:
        print(str(err), file=sys.stderr)
        exit(1)
    options['corpus'] = args[0]
    for opt, arg in opts:
        print('opt: ', opt)
        print('arg: ', arg)
        if opt == '-h':
            print('Usage: extract-case-frame.py [options] Xigt-corpus')
            print('  -c gram  : automatically detect case using the gram methodology')
            print('  -c sao   : automatically detect case using the sao methodology')
            print('  -N GRAM  : use given gram for unmarked nouns')
            print('  -h       : display this help')
            print(' to specify case grams yourself, enter value for s, a, and o:')
            print('  -s GRAM  : use given gram as subject marker')
            print('  -a GRAM  : use given gram as agent marker')
            print('  -o GRAM  : use given gram as object marker')
            sys.exit()
        elif opt == '-c':
            if arg in ['gram', 'sao']:
                options['method'] = arg
            else:
                print('Error: Invalid case value extraction method: {} {}.'.format(opt, arg), file=sys.stderr)
                print('       Choose among the following methods: "gram", "sao".', file=sys.stderr)
                sys.exit(1)
        elif opt in ['-s', '-a', '-o']:
            if arg[0] == "-":
                # it can 'take over' the following option if you don't provide an argument
                print('Error: You must provide an argument for the {} case name.'.format(opt), file=sys.stderr)
                print('       To treat {} case as unmarked, use {} 0.'.format(opt, opt), file=sys.stderr)
                sys.exit(1)
            elif arg == '0':
                user_case_names[opt[-1]] = ''
            else:
                user_case_names[opt[-1]] = arg
        elif opt == '-N':
            # nouns that aren't marked receive this case name
            user_case_names['n'] = arg
        options['case_names'] = user_case_names
    return options

def get_count_grams(corpus):
    """
    Return case grams from infer_case_from_counts
    """
    import src.infer_case_from_counts_xigt
    if __name__ == '__main__':
        sao_output = src.infer_case_from_counts_xigt.infer_case(corpus)
    count_case_grams = {'s': sao_output['s'], 'a': sao_output['a'], 'o': sao_output['o']}
    return count_case_grams

def get_gram_grams(corpus):
    """
    Return case grams from infer_case_from_grams
    """
    import src.infer_case_from_grams_xigt
    if __name__ == '__main__':
        gram_case_system = src.infer_case_from_grams_xigt.infer_case(corpus)
        print('case system= ', gram_case_system)

    if gram_case_system.lower() == 'nom-acc':
        gram_case_grams = {'s': 'nom', 'a': 'nom', 'o': 'acc'}
    elif gram_case_system.lower() == 'erg-abs':
        gram_case_grams = {'s': 'abs', 'a': 'erg', 'o': 'abs'}
    elif gram_case_system.lower() == 'none':
        gram_case_grams = {'s': '', 'a': '', 'o': ''}
    elif gram_case_system.lower() == 'split-erg':
        gram_case_grams = {'s': 'nom', 'a': 'erg', 'o': 'acc'}
    return gram_case_grams

def get_translation_verbs(igt):
    # Go through all nodes of the translation phrase structure tree to get a list of the nouns and verbs
    # that will be needed in the next step to find transitivity
    tw_ps_tier = igt.get('tw-ps')
    translation_verbs = []
    if not tw_ps_tier:
        return translation_verbs
    for pos in tw_ps_tier:
        if pos.alignment and pos.get_content() in VERBS:
            translation_verbs.append(pos.alignment)
    return translation_verbs

def get_verb_tw(igt, verb_id, translation):
    # The following is adapted from VerbValenceFromIGT.py
    # first check for an actually alignment
    alignment = False
    not_aligned_to = []
    for bilingual_alignment in igt.get('a'):
        if bilingual_alignment.get_attribute('target') == verb_id: #target can be gloss tier
            alignment = bilingual_alignment.get_attribute('source')
        else:
            not_aligned_to.append(bilingual_alignment.get_attribute('source'))

    if not alignment:
        translation = translation.replace(',', '').replace('.', '').replace('?', '')
        translation = translation.lower()
        translation_verbs = get_translation_verbs(igt)
        if len(translation_verbs) > 1:
            #target = gloss.get_content()
            for t_verb_id in translation_verbs:
                t_verb = igt.get_item(t_verb_id).get_content()
                # Finds the first match in the sentence
                if t_verb.find(translation) > -1 or translation.find(t_verb) > -1:
                    return t_verb_id
        elif len(translation_verbs) == 1:
            return translation_verbs[0]
    else:
        return alignment

def get_english_verb(verb_tw, igt):
    tw_tier = igt.get('tw')
    for tw in tw_tier:
        if tw.id == verb_tw:
            english_verb = tw.value()
    return english_verb

def get_translation_valence(verb_tw, igt):
    translation_nouns = get_translation_nouns(igt)
    up_tree, down_tree, vps, root = create_trees(igt, 'tw-ps', translation_nouns)
    # Starting at the bottom of the tree, work up until either a verb phrase parent
    # or the root node is reached
    # Intransitive means that the container verb phrase does not have a noun child
    # to the right of the verb
    # Transitive means that it has a noun child to the right of the verb
    previous_node = False
    node = up_tree[verb_tw]
    intransitive = True
    while node not in vps and node != root:
        intransitive = intransitive and has_n_to_right(node, previous_node, down_tree, translation_nouns)
        previous_node = node
        node = up_tree[node]
    intransitive = intransitive and has_n_to_right(node, previous_node, down_tree, translation_nouns)
    if intransitive:
        return 'intrans'
    else:
        return 'trans'

def get_translation_nouns(igt):
    tw_ps_tier = igt.get('tw-ps')
    translation_nouns = []
    for pos in tw_ps_tier:
        if pos.alignment and pos.get_content() in NOUNS:
            translation_nouns.append(pos.alignment)
    return translation_nouns

def create_trees(igt, ps, translation_nouns):
    # From VerbValenceFromIGT.py
    # Creates dictionaries useful for crawling up the phrase structure tree, or down the tree
    # Takes a phrase structure tier element (Xigt)
    # Returns a tuple containing:
    # - dictionary for reading up
    # - dictionary for reading down
    # - list of the verb phrase nodes
    # - reference to the root element of the phrase structure tree
    up_tree = {}
    down_tree = {}
    vps = []
    root = False

    for ps in igt.get(ps):
        # These are the nodes for the words themselves
        if ps.alignment:
            up_tree[ps.alignment] = ps.id
            down_tree[ps.id] = ps.alignment
        # These are the intermediate nodes
        else:
            for child in ps.get_attribute('children').split(' '):
                up_tree[child] = ps.id
            down_tree[ps.id] = ps.get_attribute('children').split(' ')
            if ps.get_content() == 'VP':
                vps.append(ps.id)
            elif ps.get_content() in NOUNS:
                translation_nouns.append(ps.id)
            elif ps.get_content() == 'ROOT':
                root = ps.id

    return (up_tree, down_tree, vps, root)


def has_n_to_right(node, previous_node, down_tree, translation_nouns):
    # Checks to see if the current node has a noun to the right of the given child node
    no_n = True
    if previous_node:
        for child in down_tree[node][down_tree[node].index(previous_node) + 1:]:
            if child in translation_nouns:
                no_n = False
    return no_n

def get_verb_case_frame(igt, all_case_grams, sao_case_grams, igt2items, pos_tags):
    """
    From an IGT and a set of case grams, create a list of verbs in that IGT,
    each of which is represented by a dictionary containing the verbs transitivity,
    orthography of its root, translation, and case frame"""
    if igt.get('gw') is None: #or igt.get('g') is None: # Should we be relying on 'gw' tier though?
        return {}
    gloss_ids = set(i.id for i in igt.get('gw'))
    verbs = []
    #for tier in xp.findall(igt, '//tier[@type="pos"][@alignment="w"]'):
    #  this is very slow; perhaps we could assume w-pos or pos? That's what MOM does.
    # KPH: not sure how old the above comment is. we miss a lot of verbs if we don't check gw-pos.
    # the if statements should help with the speed
    # KPH: Typically gw-pos is annotated with 'VERB' etc, while w-pos and pos may use different
    # tags (eg from the stanford parser). If the more detailed tier is missing, check the next tier up.
    pos_tier = igt.get('pos')
    if not pos_tier:
        pos_tier = igt.get('w-pos')
    if not pos_tier:
        pos_tier = igt.get('gw-pos')
    verb = {}
    if pos_tier:
        for item in pos_tier:
            if 'v.' in item.value().lower() or item.value() in pos_tags:
                verb_item = igt.get_item(item.attributes['alignment'])
                if verb_item is not None:
                    # if we are looking at gloss words, we need to traverse up to the word
                    # KPH: the next chunk is looking for the verb.id. This shouldn't need to change
                    if 'gw' in verb_item.id:
                        if 'alignment' in verb_item.attributes:
                            word_item = igt.get_item(verb_item.attributes['alignment'])
                            verb_id = word_item.id
                        elif 'segmentation' in verb_item.attributes:
                            word_item = igt.get_item(verb_item.attributes['segmentation'])
                            verb_id = word_item.id
                    elif 'm' in verb_item.id:
                        if 'segmentation' in verb_item.attributes:
                            word_item = igt.get_item(verb_item.attributes['segmentation'])
                            verb_id = word_item.id
                        elif 'alignment' in verb_item.attributes:
                            word_item = igt.get_item(verb_item.attributes['alignment'])
                            verb_id = word_item.id
                    else:
                        verb_id = verb_item.id
                    if not (igt.id in igt2items and verb_id in igt2items[igt.id]):
                        return {}
                    mom_item = igt2items[igt.id][verb_id]
                    verb_lemma = mom_item.root
                    translation = '.'.join(mom_item.root.glosses)
                    verb_tw = get_verb_tw(igt, verb_id, translation)
                    if verb_tw:
                        # english_verb = get_english_verb(verb_tw, igt)
                        transitivity = get_translation_valence(verb_tw, igt)
                    else:
                        return {}
                    # make sure eng verb item is not false
                    #verb_lemma, translation = find_root.find_root(igt, verb_item)
                    if verb_lemma == '':
                        return {}
                    clausal_complement = False
                    ditransitive = False
                    if clausal_complement == True or ditransitive == True:
                        return {}
                    elif transitivity == 'trans':
                        case_frame = {'subj': sao_case_grams['a'], 'obj': sao_case_grams['o']}
                    elif transitivity == 'intrans':
                        case_frame = {'subj': sao_case_grams['s']}
                    verb = {transitivity: {'orth': verb_lemma, 'eng': translation, 'case_frame': case_frame}}
                    # Here, the information about transitivity and case frame is set for MOM to use
                    # in the future.
                    mom_item.transitivity = transitivity
                    mom_item.case_frame = case_frame
                verbs.append(verb) #Is this indentation correct? Should we be adding emtpy "verbs"?
    return verbs

def add_to_dict(verbs, verb):
    """
    Add each verb in a list of verb (from one IGT) to the dictionary of verbs for the langauge
    """
    if 'trans' in verb:
        if verb['trans']['orth'] in verbs['trans']:
            if verb['trans']['case_frame']['subj'] not in verbs['trans'][verb['trans']['orth']]['case_frame']['subj'] or verb['trans']['case_frame']['obj'] not in verbs['trans'][verb['trans']['orth']]['case_frame']['obj']:
                count = 2
                while True:
                    if verb['trans']['orth']+str(count) in verbs['trans']:
                        if verb['trans']['case_frame']['subj'] not in verbs['trans'][verb['trans']['orth']+str(count)]['case_frame']['subj'] or verb['trans']['case_frame']['obj'] not in verbs['trans'][verb['trans']['orth']+str(count)]['case_frame']['obj']:
                            count += 1
                        else:
                            verbs['trans'][verb['trans']['orth'] + str(count)] = {}
                            verbs['trans'][verb['trans']['orth'] + str(count)]['case_frame'] = verb['trans'][
                                'case_frame']
                            verbs['trans'][verb['trans']['orth'] + str(count)]['eng'] = verb['trans']['eng']
                            break
                    else:
                        verbs['trans'][verb['trans']['orth'] + str(count)] = {}
                        verbs['trans'][verb['trans']['orth']+str(count)]['case_frame'] = verb['trans']['case_frame']
                        verbs['trans'][verb['trans']['orth'] + str(count)]['eng'] = verb['trans']['eng']
                        break
        else:
            verbs['trans'][verb['trans']['orth']] = {}
            verbs['trans'][verb['trans']['orth']]['case_frame'] = verb['trans']['case_frame']
            verbs['trans'][verb['trans']['orth']]['eng'] = verb['trans']['eng']
    elif 'intrans' in verb:
        if verb['intrans']['orth'] in verbs['intrans']:
            if verb['intrans']['case_frame']['subj'] not in verbs['intrans'][verb['intrans']['orth']]['case_frame']['subj']:
                count = 2
                while True:
                    if verb['intrans']['orth'] + str(count) in verbs['intrans']:
                        if verb['intrans']['case_frame']['subj'] not in verbs['intrans'][verb['intrans']['orth']+str(count)]['case_frame']['subj']:
                            count += 1
                        else:
                            verbs['intrans'][verb['intrans']['orth'] + str(count)] = {}
                            verbs['intrans'][verb['intrans']['orth'] + str(count)]['case_frame'] = verb['intrans'][
                                'case_frame']
                            verbs['intrans'][verb['intrans']['orth'] + str(count)]['eng'] = verb['intrans']['eng']
                            break
                    else:
                        verbs['intrans'][verb['intrans']['orth'] + str(count)] = {}
                        verbs['intrans'][verb['intrans']['orth']+str(count)]['case_frame'] = verb['intrans']['case_frame']
                        verbs['intrans'][verb['intrans']['orth'] + str(count)]['eng'] = verb['intrans']['eng']
                        break
        else:
            verbs['intrans'][verb['intrans']['orth']] = {}
            verbs['intrans'][verb['intrans']['orth']]['case_frame'] = verb['intrans']['case_frame']
            verbs['intrans'][verb['intrans']['orth']]['eng'] = verb['intrans']['case_frame']
    return verbs

def extract_case_frame(options, igt2items, pos_tags, xc):
    """
    The Main method- extract the case frame of all verbs in a corpus
    """
    #xigt_path = os.path.abspath(".")
    #sys.path.append(xigt_path)
    #file = open(options['corpus'])
    #xc = xigtxml.load(file)
    # Note: sao_case_grams means subject, agent, object case grams. Not to be confused with 'sao' another name used for the 'count' method
    if 'method' in options:
        if options['method'] == 'sao':
            sao_case_grams = get_count_grams(options['corpus'])
        if options['method'] == 'gram':
            sao_case_grams = get_gram_grams(options['corpus'])
    else:
        sao_case_grams = (options['case_names'])
    all_case_grams = ['nom', 'acc', 'abs', 'erg', 'abl', 'all',
                      'com', 'dat', 'gen', 'ins', 'loc', 'obl']
    for gram in sao_case_grams:
        if gram not in sao_case_grams:
            all_case_grams.append(sao_case_grams[gram])

    verbs = {'trans': {}, 'intrans': {}}
    for igt in xc:
        # first let's skip any igt with passives
        # KPH: I'm checking the whole sentence because PASS might be marked on an auxiliary.
        # this may skip serial verbs unnecessarily
        glosses = igt.get('g')
        is_passive = False
        for gloss in glosses:
            if 'PASS' in gloss.get_content():
                is_passive = True
        if is_passive == False:
            verbs_in_igt = get_verb_case_frame(igt, all_case_grams, sao_case_grams, igt2items, pos_tags)
            for verb in verbs_in_igt:
                verbs = add_to_dict(verbs, verb)
    return verbs

def print_verbs(verbs):
    print('intransitive verbs')
    for verb in verbs['intrans']:
        print(verb, '\t', verbs['intrans'][verb]['eng'], '\tsubject: ', verbs['intrans'][verb]['case_frame']['subj'])
    print()
    print('transitive verbs')
    for verb in verbs['trans']:
        print(verb, '\t', verbs['trans'][verb]['eng'],  '\tsubject: ', verbs['trans'][verb]['case_frame']['subj'], '\tobject: ', verbs['trans'][verb]['case_frame']['obj'])

if __name__ == '__main__':
    options = get_options()
    verbs = extract_case_frame(options)
    print_verbs(verbs)