from mom import MatrixChoices
from XigtifiedToolbox import FeatDict
from xigt.codecs import xigtxml
import sys
import re
from operator import itemgetter
from collections import Counter

KNOWN_FEATS = FeatDict.FeatureDictionary().all_keys
COORD_GLOSSES = ['coord', 'conj', 'and', 'cconj']

def infer_coordination(xc, verb_tags, noun_tags, subj_drop, word_order, known_glosses):
    coord_strategies = {}
    for igt in xc:
        english_coords = english_coordination(igt)
        for eng_coord in english_coords:
            eng_coordinands = get_coordinands_from_dependencies(igt, [eng_coord], 'tw-ds')
            eng_phrase_type = get_eng_phrase_type(igt, eng_coordinands, eng_coord, verb_tags, noun_tags)
            #coordinators_from_alignment = get_coordinators_from_alignment(igt, eng_coord)
            coordinators = get_coordinators_from_gloss(igt, verb_tags, noun_tags)
            coordinands = get_coordinands_from_dependencies(igt, coordinators, 'w-ds')
            if any(compound_word(igt, c, known_glosses) for c in coordinators): #KPH this is intended to weed out words like "one.and.a.half" or "land.and.sea"
                continue

            if coordinators == {}:
                mark = 'NONE'
            else:
                possible_types = [coordinators[x]['type'] for x in coordinators]
                mark = Counter(possible_types).most_common()[0][0]

            if eng_phrase_type == 'np':
                if mark == 'affix':
                    phrase_type = 'n'
                else:
                    phrase_type = 'np'
            else:
                if mark == 'affix': #affix marked coord is not supported for verbal constituents in the matrix
                    continue
                elif eng_phrase_type == 'vp' and not subj_drop:
                    phrase_type = 'vp'
                else:
                    phrase_type = 's'

            if len(coordinators) == 0:
                pattern = 'a'
            elif len(coordinators) == 1:
                pattern = 'mono'
            elif len(coordinators) == len(eng_coordinands): # KPH I don't expect the coordinand count to be accurate so we'll use the english coordinand count
                pattern = 'omni'
            else:
                #pattern = 'poly' # KPH since mono entails poly and we don't want both, we'll default to mono
                pattern = 'mono'

            if coordinators == {}:
                orth = 'NONE'
            else:
                possible_orths = [coordinators[x]['orth'] for x in coordinators]
                orth = Counter(possible_orths).most_common()[0][0]

            if mark == 'affix':
                position = prefix_or_suffix(igt, coordinators)
            else:
                position = ''
                if len(coordinands) == len(eng_coordinands):
                    position = before_or_after(coordinators, coordinands)
                else:
                    if 'w1' in coordinators:
                        position = 'before'
                    elif 'w'+str(len(vars(igt.get('w')))+1) in coordinators:
                        position = 'after'

            if phrase_type in coord_strategies:
                if mark in coord_strategies[phrase_type]:
                    if orth in coord_strategies[phrase_type][mark]:
                        if pattern in coord_strategies[phrase_type][mark][orth]:
                            if position != '':
                                coord_strategies[phrase_type][mark][orth][pattern].append(position)
                        else:
                            coord_strategies[phrase_type][mark][orth][pattern] = [position]
                    else:
                        coord_strategies[phrase_type][mark][orth] = {pattern: [position]}
                else:
                    coord_strategies[phrase_type][mark] = {orth: {pattern: [position]}}
            else:
                coord_strategies[phrase_type] = {mark: {orth: {pattern: [position]}}}
    choices = set_choices(coord_strategies, word_order)

    return choices

def english_coordination(igt):
    coords = []
    if igt.get('tw-pos'):
        for tw in igt.get('tw-pos'):
            if tw.value() == 'CCONJ':
                if igt.get('tw').get(tw.attributes['alignment']).value() == 'and':
                    coords.append(tw.attributes['alignment'])
    return coords

def get_coordinators_from_alignment(igt, eng_coord):
    coordinators = {}
    for tier in igt:
        if tier.type == 'bilingual-alignments' and tier.attributes['target'] == 'w':
            for a in tier:
                if 'source' in a.attributes:
                    if a.attributes['source'] == eng_coord:
                        coordinator = a.attributes['target']
                        coordinators[coordinator] = {}
                        coordinators[coordinator]['type'] = get_coordinator_type(igt, coordinator)
                        #coordinators[coordinator]['orth'] = igt.get('w').get(coordinator)
                        orth = igt.get('w').get(coordinator).value()
                        orth = orth.rstrip('\"')
                        orth = orth.lstrip('\"')
                        coordinators['orth'] = orth
    return coordinators

def get_coordinators_from_gloss(igt, verb_tags, noun_tags):

    coordinators = {}
    if not igt.get('g'):
        return coordinators
    for gloss in igt.get('g'):
        for g in re.split('\.', gloss.value()):
            if g.lower() in COORD_GLOSSES:
                gw = gloss.attributes['segmentation']
                if 'alignment' in igt.get('gw').get(gw).attributes:
                    coordinator = igt.get('gw').get(gw).attributes['alignment']
                    coord = {}
                    coord['type'] = get_coordinator_type(igt, coordinator, verb_tags, noun_tags)
                    if coord['type'] == 'affix':
                        coord['orth'] = igt.get('m').get(gloss.attributes['alignment']).value()
                    else:
                        orth = igt.get('w').get(coordinator).value()
                        orth = orth.rstrip('\"')
                        orth = orth.lstrip('\"')
                        coord['orth'] = orth
                    coordinators[coordinator] = coord
    return coordinators

def get_coordinator_type(igt, coordinator, verb_tags, noun_tags):
    type = 'word'
    pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
    if pos_tier:
        for pos in pos_tier:
            if pos.attributes['alignment'] == coordinator or pos.attributes['alignment'] == 'g'+coordinator:
                if pos.value() in verb_tags or pos.value() in noun_tags:
                    type = 'affix'
    return type

def get_coordinands_from_dependencies(igt, coordinators, tier):
    coordinands = []
    if igt.get(tier):
        for dep in igt.get(tier):
            if dep.value() == 'cc' and 'head' in dep.attributes and 'dep' in dep.attributes:
                if dep.attributes['dep'] in coordinators:
                    coordinands.append(dep.attributes['head'])
                    for cdep in igt.get(tier):
                        if cdep.value() == 'conj' and 'head' in cdep.attributes and 'dep' in cdep.attributes:
                            if cdep.attributes['head'] in coordinands and cdep.attributes['dep'] not in coordinators:
                                coordinands.append(cdep.attributes['dep'])
                            elif cdep.attributes['dep'] in coordinands and cdep.attributes['head'] not in coordinators:
                                coordinands.append(cdep.attributes['head'])
    return coordinands

def get_coordinands_from_position(igt, coordinators, verb_tags, noun_tags):
    coordinands = {}
    coords = sorted([x for x in coordinators])
    coords = ['w0']+coords+['w'+str(max([int(x.id[1:]) for x in igt.get('w')])+1)]
    adjacent_words = []
    i = 0
    while i < len(coords):
        adjacent_words.append(['w'+str(x) for x in range(int(coords[i][1:])+1, int(coords[i+1][1:]))])
        i += 1

    # next you'll need to look at intervening words

    return coordinands

def get_eng_phrase_type(igt, coordinands, coordinator, verb_tags, noun_tags):
    phrase_type = ''
    if 'VERB' not in verb_tags:
        verb_tags.append('VERB')
    if 'NOUN' not in noun_tags:
        noun_tags.append('NOUN')
    poses = {'verb': 0, 'noun': 0}
    for pos in igt.get('tw-pos'):
        if 'alignment' in pos.attributes:
            if pos.attributes['alignment'] in coordinands:
                if pos.value() in verb_tags:
                    poses['verb'] += 1
                if pos.value() in noun_tags:
                    poses['noun'] += 1
    if poses['noun'] > poses['verb']:
        phrase_type = 'np'
    else:
        phrase_type = 'vp'
        if igt.get('w-ds'):
            for dep in igt.get('w-ds'):
                if dep.value() == 'nsubj' and 'head' in dep.attributes:
                    if dep.attributes['head'] in coordinands:
                        if dep.attributes['dep'] > coordinator:
                            phrase_type = 's'
    return phrase_type

def prefix_or_suffix(igt, coordinators):
    possible_position = []
    position = ''
    for coord in coordinators:
        if coordinators[coord]['orth'][0] == '-' or coordinators[coord]['orth'][0] == '=':
            position = 'after'
        elif coordinators[coord]['orth'][-1] == '-' or coordinators[coord]['orth'][-1] == '=':
            position == 'before'
    if position == '':
        for gw in igt.get('gw'):
            if gw.attributes['alignment'] in coordinators:
                glosses = re.split('\.|-|=',gw.value())
                for gloss in glosses:
                    if gloss.lower() in KNOWN_FEATS and gloss.lower() not in COORD_GLOSSES:
                        glosses.remove(gloss)
                if glosses[0].lower() in COORD_GLOSSES:
                    possible_position.append('before')
                elif glosses[len(glosses)-1].lower() in COORD_GLOSSES:
                    possible_position.append('after')
            if possible_position != []:
                position = Counter(possible_position).most_common()[0][0]
    return position

def before_or_after(coordinators, coordinands):
    position = ''
    if coordinators != [] and coordinands != []:
        min_coordinator = min([int(x[1:]) for x in coordinators])
        max_coordinator = max([int(x[1:]) for x in coordinators])
        min_coordinand = min([int(x[1:]) for x in coordinands])
        max_coordinand = max([int(x[1:]) for x in coordinands])
        if min_coordinator < min_coordinand:
            position = 'before'
        if max_coordinator > max_coordinand:
            position = 'after'
    return position

def compound_word(igt, coord, known_glosses):
    compound = False
    gw = igt.get('gw').get('g'+coord).value()
    glosses = re.split('\.|-', gw)
    i = 0
    while i < len(glosses)-1:
        if glosses[i] in COORD_GLOSSES:
            if glosses[i-1] not in known_glosses and glosses[i+1] not in known_glosses:
                compound = True
        i += 1
    return compound

def set_choices(coord_strategies, word_order):
    choices = MatrixChoices.ChoiceDict()
    strategy_count = 1
    for phrase_type in coord_strategies:
        for coord_type in coord_strategies[phrase_type]:
            for orth in coord_strategies[phrase_type][coord_type]:
                for synd_type in coord_strategies[phrase_type][coord_type][orth]:
                    if synd_type == 'a':
                        choices.__setitem__('cs' + str(strategy_count) + '_' + phrase_type, 'on')
                        choices.__setitem__('cs' + str(strategy_count) + '_' + 'pat', synd_type)
                    else:
                        position = Counter(coord_strategies[phrase_type][coord_type][orth][synd_type]).most_common()[0][0]
                        if position == '':
                            if word_order in ['sov', 'osv', 'v-final']:
                                position = 'after'
                            elif coord_type == 'affix':
                                position = 'after'
                            else:
                                position = 'before'
                        choices.__setitem__('cs' + str(strategy_count) + '_' + phrase_type, 'on')
                        choices.__setitem__('cs' + str(strategy_count) + '_' + 'pat', synd_type)
                        choices.__setitem__('cs' + str(strategy_count) + '_' + 'mark', coord_type)
                        choices.__setitem__('cs' + str(strategy_count) + '_' + 'order', position)
                        if coord_type == 'affix':
                            if position == 'before':
                                if orth[-1] == '-' or orth[-1] == '=':
                                    choices.__setitem__('cs' + str(strategy_count) + '_' + 'orth', orth)
                                else:
                                    choices.__setitem__('cs' + str(strategy_count) + '_' + 'orth', orth + '-')
                            else:
                                if orth[0] == '-' or orth[0] == '=':
                                    choices.__setitem__('cs' + str(strategy_count) + '_' + 'orth', orth)
                                else:
                                    choices.__setitem__('cs' + str(strategy_count) + '_' + 'orth', '-' + orth)
                        else:
                            choices.__setitem__('cs' + str(strategy_count) + '_' + 'orth', orth)
                    strategy_count += 1
    return choices
