#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
infer-gender

This program attempts to infer the gender system (including class system)
from Interlinear Glossed Text (IGT) that has been enriched and put in Xigt
format (https://github.com/goodmami/xigt/wiki). The gender system discovered,
as well as properly classified nouns, will be output to a `choices' file for
usage in the LinGO Grammar Matrix customization system
(http://www.delph-in.net/matrix/customize/matrix.cgi).

Input(s):
Xigt-input - a Xigt-format XML file that contains enriched IGT text
    (If there are multiple input files, they should be the same langauge!)
-options - options for manually defining gender grams, etc.

Output:
choices file parameters for gender and the lexicon (sent to STDOUT)
   (Note that any errors or extra information are only sent to STDERR)

June 2015, Jeff Heath
  Initial development version
"""

import sys
import getopt
from xigt.codecs import xigtxml
import xigt.ref
import re
import string
#from aggutils.xigt_nltk_tree import *
#import nltk


def usage():
    """Print command line usage information."""
    print('Usage: infer_gender.py [options] Xigt-input [Xigt-input ...]')
    print('  -g g1,g2,... : manually define genders to use')
    print('  -h           : display this help')


'''
Trace the gw tier (gloss word, aligned to word) item corresponding to a g tier (gloss, aligned to morpheme) item
'''
def get_corr_gword(igt, g_tier, g_id):
    refs = g_tier[g_id].referents()
    for r in refs:
        if refs[r]:
            if refs[r][0].startswith('gw'):
                return refs[r][0]
            elif refs[r][0].startswith('m'):
                m_refs = igt.get_item(refs[r][0]).referents()
                for m_ref in m_refs:
                    if m_refs[m_ref]:
                        if m_refs[m_ref][0].startswith('w'):
                            # Assume the w_id corresponds directly to gw_id (is this true? one gloss word per word?)
                            id = 'gw' + m_refs[m_ref][0][1:]
                            # If it exists:
                            if igt.get_item(id):
                                return id


def get_corr_word(igt,gw_id):
    if igt.get_item(gw_id).alignment:
        if igt.get_item(gw_id).alignment.startswith('w'):
            return igt.get_item(gw_id).alignment
    # Assume the w_id corresponds directly to gw_id (is this true? one gloss word per word?)
    id = 'w' + gw_id[2:]
    # If it exists:
    if igt.get_item(id):
        return id


def merge_nouns(nouns, src_gen, dest_gen):
    """Merge the nouns from src_gen into those of dest_gen"""
    # first make sure both exist in the dictionary
    if src_gen not in nouns:
        nouns[src_gen] = {}
    if dest_gen not in nouns:
        nouns[dest_gen] = {}
    # process all lexemes in the source
    for lexeme in nouns[src_gen]:
        if lexeme in nouns[dest_gen]:
            # common lexeme to merge
            for gloss in nouns[src_gen][lexeme]:
                # create new gloss, or add to what is already there
                nouns[dest_gen][lexeme][gloss] = nouns[dest_gen][lexeme].get(gloss, 0) + \
                                                 nouns[src_gen][lexeme][gloss]
        else:
            # lexeme not there yet, so just copy
            nouns[dest_gen][lexeme] = nouns[src_gen][lexeme].copy()
    del nouns[src_gen]


def infer_gender(xigt_corpus):
    # default collection of gender grams (in lowercase)
    # including animate, inanimate, human, rational, zoic (animals), vegetable
    # and including Arabic numeral numbered classes up to 20, and some Roman numerals

    # define these separately and add them in later
    # because we need them in the code for some special processing
    arabic_numeral_classes = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
                              '11', '12', '13', '14', '15', '16', '17', '18', '19', '20']
    roman_numeral_classes = ['i', 'v', 'ii', 'iii', 'iv', 'vi', 'vii', 'viii', 'ix', 'x']

    all_gender_grams = ['m', 'masc', 'f', 'fem', 'n', 'neut', 'anim', 'inan', 'inanim',
                        'hum', 'nhum', 'nonhum', 'r', 'rat', 'nr', 'nonrat', 'zo', 'veg',
                        'cl1', 'cl2', 'cl3', 'cl4', 'cl5', 'cl6', 'cl7', 'cl8', 'cl9', 'cl10',
                        'cl11', 'cl12', 'cl13', 'cl14', 'cl15', 'cl16', 'cl17', 'cl18', 'cl19', 'cl20',
                        'g1', 'g2', 'g3', 'g4', 'g5', 'g6', 'g7', 'g8', 'g9', 'g10',
                        'g11', 'g12', 'g13', 'g14', 'g15', 'g16', 'g17', 'g18', 'g19', 'g20']
    all_gender_grams.extend(arabic_numeral_classes)
    all_gender_grams.extend(roman_numeral_classes)

    # sets of genders which tend to go together
    gender_sets = [{'m', 'f', 'n'}, {'masc', 'fem', 'neut'}, {'anim', 'inan', 'inanim'},
                   {'cl1', 'cl2', 'cl3', 'cl4', 'cl5', 'cl6', 'cl7', 'cl8', 'cl9', 'cl10',
                    'cl11', 'cl12', 'cl13', 'cl14', 'cl15', 'cl16', 'cl17', 'cl18', 'cl19', 'cl20'},
                   {'g1', 'g2', 'g3', 'g4', 'g5', 'g6', 'g7', 'g8', 'g9', 'g10',
                    'g11', 'g12', 'g13', 'g14', 'g15', 'g16', 'g17', 'g18', 'g19', 'g20'}]
    gender_sets.append(set(arabic_numeral_classes))
    gender_sets.append(set(roman_numeral_classes))


    # manual_genders = False
    # for opt, arg in opts:
    # 	if opt == '-h':
    # 		usage()
    # 		sys.exit()
    # 	elif opt == '-g':
    # 		# manually define the genders?
    # 		all_gender_grams = arg.split(',')
    # 		manual_genders = True

    # create regex pattern for matching gender names
    # (can have optional person number in front, but it can't be followed by a digit)
    all_gender_patt = re.compile(r'^([1-4](?!\d))?('+'|'.join(all_gender_grams)+')$')


    # storage location for noun data we find
    # it is stored: nouns[gender][lexeme][gloss] = count
    # so for each gender there is a dictionary of lexemes
    # for each lemexe there is a dictionary of glosses
    # and the value stored is a count of how many times it occurs
    nouns = {}

    # total counts for IGTS and genders
    IGT_count = 0
    gender_counts = {}

    # process all of the IGTs in the corpus
    IGT_count += len(xigt_corpus)
    for igt in xigt_corpus:
        process_igt(all_gender_patt, gender_counts, igt, nouns)

    # See if there are any genders to output for the choices file

    # gender_counts provides a dictionary of all genders (with their counts)
    # both ones where a noun was stored (in nouns[]) and where it was not stored

    # attempt to merge m/masc and f/fem
    # KPH: when we input this to MOM, we will probably want to retain alternate glosses like masc/m
    if 'm' in gender_counts and 'masc' in gender_counts:
        gender_counts['masc'] += gender_counts['m']
        del gender_counts['m']
        merge_nouns(nouns, 'm', 'masc')
    if 'f' in gender_counts and 'fem' in gender_counts:
        gender_counts['fem'] += gender_counts['f']
        del gender_counts['f']
        merge_nouns(nouns, 'f', 'fem')

    # remove genders 1/2/3 if there aren't any other Arabic numeral class members
    genders = set(gender_counts.keys())
    if genders.intersection(arabic_numeral_classes[:3]):
        # there are one or more genders 1/2/3
        if len(genders.intersection(arabic_numeral_classes[3:])) < 2:
            # but there are at most one class number higher than those
            for gender in genders.intersection(arabic_numeral_classes[:3]):
                # get rid of this gender
                del gender_counts[gender]
                if gender in nouns:
                    del nouns[gender]

    # remove gender i, v if there aren't any other Roman numeral class members
    genders = set(gender_counts.keys())
    if genders.intersection(roman_numeral_classes[:2]):
        # there are one or more genders i, v
        if len(genders.intersection(roman_numeral_classes[2:])) < 2:
            # but there are at most one class number other than those
            for gender in genders.intersection(roman_numeral_classes[:2]):
                # get rid of this gender
                del gender_counts[gender]
                if gender in nouns:
                    del nouns[gender]
    # remove very infrequent genders
    count_limit = int(IGT_count * 0.05)
    genders_to_keep = remove_infrequent(count_limit, gender_counts, gender_sets)
    # check for upper/lower case equivalents, and remove the upper case version
    harmonize_case(nouns)
    # check for singular/plural gloss equivalents in same lexeme
    harmonize_number(nouns)
    output_choices(genders_to_keep, nouns)
    return genders_to_keep


def harmonize_number(nouns):
    for gender in nouns:
        for lexeme in list(nouns[gender].keys()):
            # and there is a lower-case version of it in noun dict
            to_delete = []
            for wordgloss in nouns[gender][lexeme]:
                # check if lower case version has same word gloss
                if wordgloss + 's' in nouns[gender][lexeme]:
                    # transfer the wordgloss count of the plural to the singular version
                    nouns[gender][lexeme][wordgloss] += nouns[gender][lexeme][wordgloss + 's']
                    to_delete.append(wordgloss + 's')
            # delete any duplicate plurals found
            for item in to_delete:
                del nouns[gender][lexeme][item]


def harmonize_case(nouns):
    for gender in nouns:
        for lexeme in list(nouns[gender].keys()):
            if lexeme != lexeme.lower():
                # this is a lexeme that has some upper-case in it
                lexeme_lower = lexeme.lower()
                if lexeme_lower in nouns[gender].keys():
                    # and there is a lower-case version of it in noun dict
                    to_delete = []
                    for wordgloss in nouns[gender][lexeme]:
                        # check if lower case version has same word gloss
                        if wordgloss in nouns[gender][lexeme_lower]:
                            # transfer the wordgloss count to the lower case version
                            nouns[gender][lexeme_lower][wordgloss] += nouns[gender][lexeme][wordgloss]
                            to_delete.append(wordgloss)
                        # if that didn't work, try a lower case version of wordgloss
                        elif wordgloss.lower() in nouns[gender][lexeme_lower]:
                            # transfer the wordgloss count to the lower case version
                            nouns[gender][lexeme_lower][wordgloss.lower()] += nouns[gender][lexeme][wordgloss]
                            to_delete.append(wordgloss)
                    # delete any upper case version that got transferred
                    for item in to_delete:
                        del nouns[gender][lexeme][item]
                    # check if there are no more glosses
                    if len(nouns[gender][lexeme]) == 0:
                        # no more glosses for this lexeme, so delete it
                        del nouns[gender][lexeme]


def remove_infrequent(count_limit, gender_counts, gender_sets):
    genders_to_keep = set()
    genders_to_toss = set()
    for gender in sorted(gender_counts):
        # print('gender {} count = {}'.format(gender, gender_counts[gender]))
        if gender_counts[gender] >= count_limit:
            genders_to_keep.add(gender)
        else:
            genders_to_toss.add(gender)
    # reconsider each of the genders to toss
    for gender in genders_to_toss:
        for gender_set in gender_sets:
            if gender in gender_set and gender_set.intersection(genders_to_keep):
                # we are keeping a related gender, so don't toss this one
                genders_to_keep.add(gender)

    return genders_to_keep


def output_choices(genders_to_keep, nouns):
    # check to see if we have any genders to report
    if len(genders_to_keep) > 0:
        # output genders found
        print("section=gender")
        n = 0
        for gender in sorted(genders_to_keep):
            n += 1
            print("  gender{}_name={}".format(n, gender))

    # format and print the lexicon (nouns) for the choices file
    if len(nouns) > 0 and len(genders_to_keep) > 0:
        # output nouns with their genders
        print("section=lexicon")
        n = 0
        for gender in sorted(genders_to_keep):
            n += 1
            # output the header for this noun type
            print("  noun{}_name=common-{}".format(n, gender))
            print("    noun{}_feat1_name=gender".format(n))
            print("    noun{}_feat1_value={}".format(n, gender))

            if gender not in nouns:
                # we don't actually have any nouns to output for this gender
                continue

            m = 0
            for lexeme in sorted(nouns[gender].keys()):
                for gloss in sorted(nouns[gender][lexeme].keys()):
                    m += 1
                    print("    noun{}_stem{}_orth={}".format(n, m, lexeme))
                    print("    noun{}_stem{}_pred=_{}_n_rel".format(n, m, gloss))


def process_igt(all_gender_patt, gender_counts, igt, nouns):
    #print(igt.id) # for debugging
    # get the tier with the gloss tags (alignment to morphemes)
    gTier = igt.get('g')
    # get the tier with the connection between gloss words and words
    gwTier = igt.get('gw')
    # get the translation word-level phrase-structure tier
    tw_psTier = igt.get('tw-ps')
    # get the bilingual-alignments tier
    aTier = igt.get('a')
    # get the word phrase-structure tier
    w_psTier = igt.get('w-ps')
    if gTier:
        # can't look for gender grams if we don't have this tier
        glosswordID_NNs = []
        wordID_NNs = []
        if tw_psTier:
            # first find all of the trans word IDs for the NN's (or NNP's) in the translation line
            twID_NNs = [x.alignment for x in tw_psTier if x.value()[:2] == 'NN']
            if aTier:
                # get corresponding gloss IDs from the bilingual-alignments tier
                glossID_NNs = [x.get_attribute('target') for x in aTier if x.get_attribute('source') in twID_NNs]
                # get corresponding glosses from the glosses tier
                gloss_NNs = [gTier[x].value() for x in glossID_NNs]
                # get corresponding gloss word IDs from the glosses tier
                glosswordID_NNs = []
                for g_id in glossID_NNs:
                    gw_id = get_corr_gword(igt, gTier, g_id)
                    if gw_id:
                        glosswordID_NNs.append(gw_id)
                if gwTier:
                    wordID_NNs = []
                    for x in glosswordID_NNs:
                        w_id = get_corr_word(igt, x)
                        if w_id:
                            wordID_NNs.append(w_id)
        NP_word_sets = []
        if w_psTier:
            # look for NPs that might include elements marked with gender for the included NN
            for i in w_psTier:
                if i.get_content() == 'NP':
                    nodes = i.get_attribute('children').split()
                    words = set()
                    while nodes:
                        child = nodes.pop()
                        if w_psTier[child].get_content() == 'NP':
                            # only allow NPs that aren't nested
                            words = set()
                            break
                        if w_psTier[child].get_attribute('children'):
                            nodes.extend(w_psTier[child].get_attribute('children').split())
                        else:
                            words.add(w_psTier[child].alignment)
                    if words:
                        NP_word_sets.append(words)

        # go through the gloss grams looking for genders
        for gloss in gTier:
            genderText = gloss.value().lower()
            m = all_gender_patt.match(genderText)
            if m:
                # this gloss gram is one of our gender grams
                # print('found gender gram: {}'.format(genderText))

                # there may be a person number in front, get just the gender
                genderText = m.group(2)

                # keep a count of genders found
                gender_counts[genderText] = gender_counts.get(genderText, 0) + 1

                # get the gloss word ID for the gw that contains this gender gram
                # glosswordID = xigt.ref.ids(genderItem.content)[0]
                glosswordID = get_corr_gword(igt, gTier, gloss.id)
                # TODO: what if there is more than one ID?
                # assume not found to start with
                idx = -1
                # is this gloss word (for this gender) associated with an NN?
                if (glosswordID in glosswordID_NNs):
                    # yes, this gloss word (for this gender) is connected to an NN
                    # we will need the index to get information out of those lists we created
                    idx = glosswordID_NNs.index(glosswordID)
                else:
                    # get the word ID for the gw that contains this gender gram
                    wordID = gwTier[glosswordID].alignment
                    for word_set in NP_word_sets:
                        if wordID in word_set:
                            # this word does occur in this NP word_set
                            intrsct = list(word_set.intersection(wordID_NNs))
                            if intrsct:
                                # and this does intersect with a word identified as a noun
                                idx = wordID_NNs.index(intrsct[0])
                            # only one word_set the wordID can be in, so done with loop
                            break

                if idx >= 0:
                    # we did find an associated noun
                    # get the gloss ID, and the gloss item for the correct NN
                    glossID = glossID_NNs[idx]
                    glossItem = gTier.get(glossID)
                    morphID = glossItem.get_attribute('alignment')
                    if morphID != None:
                        morphItem = igt.get_any(morphID)
                        lexeme = morphItem.value()
                        # remove (standard) punctuation at end/beginning of lexeme
                        lexeme = lexeme.strip(string.punctuation)
                        wordgloss = glossItem.value()

                        # find some way to get rid of unwanted parts of the gloss?
                        # for now,  remove everything after colon or semi-colon
                        # wordgloss = re.sub(r'[:;].+', '', wordgloss)

                        # store the lexeme and word gloss in the nouns dictionary
                        # using the gender as the main key
                        if not genderText in nouns.keys():
                            # start new gender
                            nouns[genderText] = {lexeme: {wordgloss: 1}}
                        elif not lexeme in nouns[genderText].keys():
                            # start new lexeme
                            nouns[genderText][lexeme] = {wordgloss: 1}
                        else:
                            # add to existing lexeme (wordgloss may or may not be new)
                            nouns[genderText][lexeme][wordgloss] = nouns[genderText][lexeme].get(wordgloss, 0) + 1
                else:
                    # no, this gloss word (for this gender) is not connected to an NN
                    # nothing else to do for now...
                    pass


if __name__ == '__main__':
    file_name = sys.argv[1]
    genders = infer_gender(file_name)