#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kristen Howell
# this contains helper functions for adding things to the lexicon that are not handled by mom (eg. detirminers)
# and changing the lexicon produced by mom to create subcategories (eg. pronouns from nouns and auxiliaries from verbs)


from XigtifiedToolbox import FeatDict
from mom import MatrixChoices
import re

known_features = FeatDict.FeatureDictionary().all_keys
feat_dict = FeatDict.FeatureDictionary().dictionaries

def infer_det_lex_items(determiners):
    dets = {}
    for igt, det in determiners:
        gw = igt.get('gw').get('g'+det).value()
        w = igt.get('w').get(det).value()
        w = w.rstrip('\"')
        w = w.lstrip('\"')
        person = ''
        number = ''
        gender = ''
        pernum = '' # KPH this is used by the gm for incl/excl
        case = ''
        feats = re.split('\.|-',gw)
        non_pngc_feats = feats.copy()
        for feat in feats:
            if feat.lower() in feat_dict['perNumFeatures']:
                person, number = feat_dict['perNumFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perNumInclFeatures']:
                p, n, i = feat_dict['perNumInclFeatures'][feat.lower()]
                p = re.sub('[a-z]', '', p)
                pernum = p+n+'_'+i
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perNumHeadFeatures']:
                person, number, head = feat_dict['perNumHeadFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perNumInclHeadFeatures']:
                p, n, i, head = feat_dict['perNumInclHeadFeatures'][feat.lower()]
                p = re.sub('[a-z]', '', p)
                pernum = p + n + '_' + i
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perGenFeatures']:
                person, gender = feat_dict['perGenFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perGenHeadFeatures']:
                person, gender, head = feat_dict['perGenHeadFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perFeatures']:
                person = feat_dict['perFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perHeadFeatures']:
                person, head = feat_dict['perHeadFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['numFeatures']:
                number = feat_dict['numFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['genFeatures']:
                gender = feat_dict['genFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['caseFeatures']:
                case = feat_dict['caseFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perNumGenFeatures']:
                person, number, gender = feat_dict['perNumGenFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)
            elif feat.lower() in feat_dict['perNumCaseFeatures']:
                person, number, case = feat_dict['perNumCaseFeatures'][feat.lower()]
                non_pngc_feats.remove(feat)


        pred = build_pred(non_pngc_feats)
        if w in dets:
            new_instance = True
            for inst in dets[w]:
                if not distinct_features({'person': person, 'number': number, 'pernum': pernum, 'gender': gender, 'case': case}, inst):
                    # KPH: we're combining the lex entries for every det that the new det is compatible with
                    # (compatible means same orth, compatible features). This may be too productive (lead to ambiguity),
                    # but I don't expect it to happen often. ea will tell
                    if person != '':
                        inst['person'].add(person)
                    if number != '':
                        inst['number'].add(number)
                    if pernum != '':
                        inst['pernum'].add(pernum)
                    if gender != '':
                        inst['gender'].add(gender)
                    if case != '':
                        inst['case'].add(case)
                    if pred != '':
                        inst['pred'].add(pred)
                    new_instance = False
            if new_instance:
                dets[w].append({'pred': set(), 'person': set(), 'number': set(), 'pernum': set(), 'gender': set(), 'case': set()})
                if person != '':
                    dets[w][-1]['person'].add(person)
                if number != '':
                    dets[w][-1]['number'].add(number)
                if pernum != '':
                    dets[w][-1]['pernum'].add(pernum)
                if gender != '':
                    dets[w][-1]['gender'].add(gender)
                if case != '':
                    dets[w][-1]['case'].add(case)
                if pred != '':
                    dets[w][-1]['pred'].add(pred)
        else:
            dets[w] = [{'pred': set(), 'person': set(), 'number': set(), 'pernum': set(), 'gender': set(), 'case': set()}]
            if person != '':
                dets[w][0]['person'].add(person)
            if number != '':
                dets[w][0]['number'].add(number)
            if pernum != '':
                dets[w][-1]['pernum'].add(pernum)
            if gender != '':
                dets[w][0]['gender'].add(gender)
            if case != '':
                dets[w][0]['case'].add(case)
            if pred != '':
                dets[w][0]['pred'].add(pred)

    choices = create_det_choices(dets)
    return choices

def build_pred(feats):
    if len(feats) == 0:
        return ''
    pred = feats[0]
    feats.remove(feats[0])
    while len(feats) > 0:
        pred += '.'
        pred += feats[0]
        feats.remove(feats[0])
    return pred

def distinct_features(new_feat_value_pairs, old_feat_value_pairs):
    # KPH The GM will create disjunctive types if we specify muliple values for a feature,
    # so we can go ahead and add multiples if eg. we see fem sometimes and masc others. But
    # if multiple features differ between two instances, eg. 3 masc and 1 fem, we want separate lex entires
    distinct = False
    diff_count = 0
    if new_feat_value_pairs['person'] != '' and old_feat_value_pairs['person'] != set():
        if new_feat_value_pairs['person'] not in old_feat_value_pairs['person']:
            diff_count += 1
    if new_feat_value_pairs['number'] != '' and old_feat_value_pairs['number'] != set():
        if new_feat_value_pairs['number'] not in old_feat_value_pairs['number']:
            diff_count += 1
    if new_feat_value_pairs['pernum'] != '' and old_feat_value_pairs['pernum'] != set():
        if new_feat_value_pairs['pernum'] not in old_feat_value_pairs['pernum']:
            diff_count += 1
    if new_feat_value_pairs['gender'] != '' and old_feat_value_pairs['gender'] != set():
        if new_feat_value_pairs['gender'] not in old_feat_value_pairs['gender']:
            diff_count += 1
    if new_feat_value_pairs['case'] != '' and old_feat_value_pairs['case'] != set():
        if new_feat_value_pairs['case'] not in old_feat_value_pairs['case']:
            diff_count += 1
    if diff_count > 1:
        distinct = True
    return distinct

def create_det_choices(dets):
    choices = MatrixChoices.ChoiceDict()
    det_count = 1
    for det in dets:
        for inst in dets[det]:
            choices.__setitem__('det'+str(det_count)+'_name', 'det'+str(det_count))
            choices.__setitem__('det'+str(det_count)+'_stem1_orth', det)
            # preds = list(inst['pred'])
            # if len(preds) == 0:
            #     full_pred = 'det_q_rel'
            # else:
            #     full_pred = preds[0]
            #     preds.remove(full_pred)
            #     while len(preds) != 0:
            #         full_pred += '_or_' + preds[0]
            #         preds.remove(preds[0])
            #     full_pred += '_q_rel'
            pred = '_exist_q_rel'
            choices.__setitem__('det'+str(det_count)+'_stem1_pred', pred)
            feat_count = 1
            if inst['person'] != set():
                feats = ', '.join([f for f in inst['person']])
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_name', 'person')
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_value', feats)
                feat_count += 1
            if inst['number'] != set():
                feats = ', '.join([f for f in inst['number']])
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_name', 'number')
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_value', feats)
                feat_count += 1
            if inst['pernum'] != set():
                feats = ', '.join([f for f in inst['pernum']])
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_name', 'pernum')
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_value', feats)
                feat_count += 1
            if inst['gender'] != set():
                feats = ', '.join([f for f in inst['gender']])
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_name', 'gender')
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_value', feats)
                feat_count += 1
            if inst['case'] != set():
                feats = ', '.join([f for f in inst['case']])
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_name', 'case')
                choices.__setitem__('det' + str(det_count) + '_feat' + str(feat_count) + '_value', feats)
                feat_count += 1
        det_count += 1
    return choices

def infer_pronouns(choices):
    noun_id = int(get_max_id(choices['noun'])) + 1
    noun_id_list = []
    for noun_type in choices['noun']:
        for stem in noun_type['stem']:
            if is_pronoun(stem['pred']):
                # copy the stem and create a new pronoun noun type, then delete the old stem
                new_pronoun_type = MatrixChoices.ChoiceDict()
                new_pronoun_type.full_key = 'noun' + str(noun_id)
                new_pronoun_type['name'] = 'noun' + str(noun_id)
                new_pronoun_type['pron'] = 'on'
                new_pronoun_type['stem'] = MatrixChoices.ChoiceList()
                new_pronoun_type['stem'].full_key = 'noun' + str(noun_id) + '_stem'
                stem_id = 1
                new_stem = MatrixChoices.ChoiceDict()
                new_stem.full_key = 'noun'+str(noun_id)+'_stem'+str(stem_id)
                new_stem['orth'] = stem['orth']
                new_stem['pred'] = '_pron_n_rel'
                new_pronoun_type['stem'].__setitem__(str(stem_id), new_stem)
                new_pronoun_type['feat'] = MatrixChoices.ChoiceList()
                feats = get_pronoun_feats(stem['pred'])
                feat_count = 1
                for feat in feats:
                    new_feat = MatrixChoices.ChoiceDict()
                    new_feat.full_key = 'feat' + str(feat_count)
                    new_feat['name'] = feat
                    new_feat['value'] = feats[feat]
                    new_pronoun_type['feat'].__setitem__(re.sub('feat', '', new_feat.full_key), new_feat)
                    #new_pronoun_type['feat'].__setitem__('noun' + str(noun_id) +new_feat.full_key, new_feat)
                    feat_count += 1
                del choices[stem.full_key]
                choices.__setitem__('noun' + str(noun_id), new_pronoun_type)
                noun_id_list.append(('noun' + str(noun_id), noun_type.full_key))
                noun_id += 1
    # add the pronoun type to the inputs of any pc that took the noun as an input
    for pronoun, noun in noun_id_list:
        for pc in choices['noun-pc']:
            if noun in pc['inputs']:
                pc['inputs'] += ', '+pronoun

    # if there are no longer stems in the noun_type, do I need to remove the noun_type
    return choices

def get_max_id(choices_list):
    max = 0
    for item in choices_list:
        key = item.full_key
        id = int(re.sub('[^0-9]', '', key))
        if id > max:
            max = id
    return max

def is_pronoun(noun_pred):
    pronouns = ['he', 'she', 'it', 'him', 'her', 'they', 'them']
    perFeatures = {}
    perFeatures.update(feat_dict['perFeatures'].copy())# kph why were num gen and case commented out?
    perFeatures.update(feat_dict['numFeatures'].copy())
    perFeatures.update(feat_dict['genFeatures'].copy())
    perFeatures.update(feat_dict['perGenFeatures'].copy())
    perFeatures.update(feat_dict['perNumFeatures'].copy())
    perFeatures.update(feat_dict['perNumGenFeatures'].copy())
    perFeatures.update(feat_dict['perGenHeadFeatures'].copy())
    perFeatures.update(feat_dict['perNumHeadFeatures'].copy())
    perFeatures.update(feat_dict['perNumCaseFeatures'].copy())
    perFeatures.update(feat_dict['perNumInclFeatures'].copy())
    perFeatures.update(feat_dict['perHeadFeatures'].copy())
    perFeatures.update(feat_dict['perNumInclHeadFeatures'].copy())
    perFeatures.update(feat_dict['caseFeatures'])
    pronoun = False
    preds = re.split('_|\.|-', noun_pred)
    preds.remove('n')
    preds.remove('rel')
    preds.remove('')
    if 'or' in preds:
        preds.remove('or')
    for pred in preds:
        if pred in pronouns:
            pronoun = True
    if not pronoun:
        if all([pred in known_features for pred in preds]):
            pronoun = True
            if not pronoun:
                for pred in preds:
                    if pred in perFeatures:
                        pronoun = True

    return pronoun

def get_pronoun_feats(pred):
    feats = {}
    pred = re.sub('_n_rel','',pred)
    grams = re.split('\.|-or-|_',pred)
    for gram in grams:
        if gram in feat_dict['caseFeatures']:
            case = feat_dict['caseFeatures'][gram]
            feats = add_feat_to_dict(feats, 'case', case)
        elif gram in feat_dict['perNumFeatures']:
            person, number = feat_dict['perNumFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
            feats = add_feat_to_dict(feats, 'number', number)
        elif gram in feat_dict['perGenFeatures']:
            person, gender = feat_dict['perGenFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
            feats = add_feat_to_dict(feats, 'gender', gender)
        elif gram in feat_dict['perNumGenFeatures']:
            person, number, gender = feat_dict['perNumGenFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
            feats = add_feat_to_dict(feats, 'number', number)
            feats = add_feat_to_dict(feats, 'gender', gender)
        elif gram in feat_dict['perGenHeadFeatures']:
            person, gender, head = feat_dict['perGenHeadFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
            feats = add_feat_to_dict(feats, 'gender', gender)
        elif gram in feat_dict['perNumHeadFeatures']:
            person, number, head = feat_dict['perNumHeadFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
            feats = add_feat_to_dict(feats, 'number', number)
        elif gram in feat_dict['perHeadFeatures']:
            person, head = feat_dict['perHeadFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
        elif gram in feat_dict['perFeatures']:
            person = feat_dict['perFeatures'][gram]
            feats = add_feat_to_dict(feats, 'person', person)
        elif gram in feat_dict['numFeatures']:
            number = feat_dict['numFeatures'][gram]
            feats = add_feat_to_dict(feats, 'number', number)
        elif gram in feat_dict['genFeatures']:
            gender = feat_dict['genFeatures'][gram]
            feats = add_feat_to_dict(feats, 'gender', gender)
        elif gram in feat_dict['perNumInclFeatures']:
            person, number, inclusive = feat_dict['perNumInclFeatures'][gram]
            person = re.sub('[a-z]','',person)
            pernum = person+number + '_' + inclusive
            feats = add_feat_to_dict(feats, 'pernum', pernum)
        elif gram in feat_dict['perNumInclHeadFeatures']:
            person, number, inclusive, head = feat_dict['perNumInclHeadFeatures'][gram]
            person = re.sub('[a-z]', '', person)
            pernum = person + number + '_' + inclusive
            feats = add_feat_to_dict(feats, 'pernum', pernum)
    return feats

def add_feat_to_dict(dict, name, value):
    if name in dict:
        if dict[name] != value:
            dict[name] = dict[name] + ', ' + value
    else:
        dict[name] = value
    return dict

def add_common_noun_features(choices):
    for noun_type in choices['noun']:
        feat_id = get_max_id(noun_type['feat'])
        choices.__setitem__(noun_type.full_key + '_det', 'opt')
        if noun_type['pron'] != 'on':
            if not lexical_rule_adds_person(choices, noun_type.full_key): #check to see if this is the sole input to a lexical rule that assigns person. if so and if that person value isn't third, don't add person to the noun or the grammar won't laod
                choices.__setitem__(noun_type.full_key + '_feat' + str(feat_id + 1) + '_name', 'person')
                choices.__setitem__(noun_type.full_key + '_feat' + str(feat_id + 1) + '_value', '3rd')
    return choices

def lexical_rule_adds_person(choices, noun_id):
    lexical_rule_clash = False
    for pc in choices['noun-pc']:
        inputs = pc['inputs'].strip().split(',')
        if len(inputs) == 1 and inputs[0] == noun_id:
            for lrt in pc['lrt']:
                for feat in lrt['feat']:
                    if feat['name'] == 'person':
                        if '3rd' not in feat['value']:
                            lexical_rule_clash = True
    return lexical_rule_clash

def add_auxiliaries(choices, auxiliaries, negation_auxiliaries):
    # currently, I'm not doing anything about complement features, or finiteness
    aux_id = 1
    aux_list = []
    verb_id_list = []
    aux_orth_list = set()
    max_verb_pc = get_max_id(choices['verb-pc'])
    agr_features = {}
    agr_features.update(feat_dict['subject_feature_vals'].copy())
    agr_features.update(feat_dict['object_feature_vals'].copy())
    agr_features.update(feat_dict['perFeatures'].copy())
    agr_features.update(feat_dict['numFeatures'].copy())
    agr_features.update(feat_dict['genFeatures'].copy())
    agr_features.update(feat_dict['perGenFeatures'].copy())
    agr_features.update(feat_dict['perNumFeatures'].copy())
    agr_features.update(feat_dict['perNumGenFeatures'].copy())
    agr_features.update(feat_dict['perNumInclFeatures'].copy())
    agr_features.update(feat_dict['perNumInclHeadFeatures'].copy())
    agr_features.update(feat_dict['perHeadFeatures'].copy())
    agr_features.update(feat_dict['perGenHeadFeatures'].copy())
    agr_features.update(feat_dict['perNumHeadFeatures'].copy())
    agr_features.update(feat_dict['perNumCaseFeatures'].copy())
    if 'adp' in choices:
        adp_case = True
    else:
        adp_case = False
    for verb_type in choices['verb']:
        for stem in verb_type['stem']:
            if stem['orth'] in auxiliaries:
                new_aux_type = MatrixChoices.ChoiceDict()
                new_aux_type.full_key = 'aux' + str(aux_id)
                new_aux_type['name'] = 'aux' + str(aux_id)
                new_aux_type['feat'] = MatrixChoices.ChoiceList()
                new_aux_type['feat'].full_key = 'aux' + str(aux_id) + '_feat'
                new_aux_type = copy_aux_features(new_aux_type, verb_type, aux_id)
                subj_feats = collect_subj_features(verb_type)
                stem_id = 1
                if stem['orth'] in negation_auxiliaries:
                    new_aux_type['name'] = 'neg' + str(aux_id)
                elif 'neg' in stem['pred']:
                    new_aux_type['name'] = 'neg' + str(aux_id)
                aux_orth_list.add(stem['orth'])
                new_aux_type = set_subj_case(new_aux_type, auxiliaries, stem['orth'])
                new_aux_type = set_stem(new_aux_type, stem, negation_auxiliaries, aux_id, stem_id)
                new_aux_type = set_aux_features(new_aux_type, aux_id, auxiliaries[stem['orth']]['aux_features']) # I'm a little suspicious of how this will work in practice.
                if any(feat in agr_features for feat in auxiliaries[stem['orth']]['aux_features']):
                    subj_feats.update(create_subj_feats(auxiliaries[stem['orth']]['aux_features'].copy()))


                aux_list.append(('aux' + str(aux_id), new_aux_type))
                if subj_feats != {} and choices['aux-comp'] != 's':
                    max_verb_pc += 1
                    new_verb_pc = create_verb_pc(subj_feats, aux_id, max_verb_pc)
                    choices['verb-pc'].__setitem__(re.sub('verb-pc', '', new_verb_pc.full_key), new_verb_pc)
                    verb_id_list.append(('verb-pc'+str(max_verb_pc), new_aux_type.full_key))
                else:
                    verb_id_list.append(('aux' + str(aux_id), new_aux_type.full_key))

                # if adp_case:
                #     aux_id += 1
                #     adp_aux_type = create_aux_with_adp_subj(new_aux_type, aux_id)
                #     aux_list.append(('aux' + str(aux_id), adp_aux_type))
                #     if subj_feats != {}:
                #         max_verb_pc += 1
                #         new_verb_pc = create_verb_pc(subj_feats, aux_id, max_verb_pc)
                #         choices['verb-pc'].__setitem__(re.sub('verb-pc', '', new_verb_pc.full_key), new_verb_pc)
                #         verb_id_list.append(('verb-pc' + str(max_verb_pc), new_aux_type.full_key))
                aux_id += 1
                # maybe don't do this if it looks like a real verb too
                # i don't think aux/verb homophony will hurt
                # del choices[stem.full_key]
    for aux in aux_orth_list:
        try:
            del auxiliaries[aux]
        except:
            pass
        negation_auxiliaries.discard(aux)

    if auxiliaries != {}:
        for aux in auxiliaries:
            new_aux_type = MatrixChoices.ChoiceDict()
            new_aux_type.full_key = 'aux' + str(aux_id)
            if aux in negation_auxiliaries:
                new_aux_type['name'] = 'neg' + str(aux_id)
            else:
                new_aux_type['name'] = 'aux' + str(aux_id)
            new_aux_type = set_subj_case(new_aux_type, auxiliaries, aux)
            new_aux_type['feat'] = MatrixChoices.ChoiceList()
            new_aux_type['feat'].full_key = 'aux' + str(aux_id) +'_feat'
            new_aux_type = set_aux_features(new_aux_type, aux_id, auxiliaries[aux]['aux_features'])
            new_stem = {'orth': aux, 'pred': ''}
            if aux in negation_auxiliaries:
                new_stem['pred'] = 'neg_rel'
            new_aux_type = set_stem(new_aux_type, new_stem, negation_auxiliaries, aux_id, 1)
            aux_list.append(('aux' + str(aux_id), new_aux_type))
            # look for subj features
            has_subj_feats = False
            if any(feat in agr_features for feat in auxiliaries[aux]['aux_features']):
                has_subj_feats = True
                max_verb_pc += 1
                subj_feats = create_subj_feats(auxiliaries[aux]['aux_features'])
                new_verb_pc = create_verb_pc(subj_feats, aux_id, max_verb_pc)
                choices['verb-pc'].__setitem__(re.sub('verb-pc', '', new_verb_pc.full_key), new_verb_pc)

            # create a copy that allows an adp subject if there are case-marking adpositions
            # if adp_case:
            #     aux_id += 1
            #     adp_aux_type = create_aux_with_adp_subj(new_aux_type, aux_id)
            #     aux_list.append(('aux' + str(aux_id), adp_aux_type))
            #     if has_subj_feats == True:
            #         max_verb_pc += 1
            #         new_verb_pc = create_verb_pc(subj_feats, aux_id, max_verb_pc)
            #         choices['verb-pc'].__setitem__(re.sub('verb-pc', '', new_verb_pc.full_key), new_verb_pc)
            negation_auxiliaries.discard(aux)
            aux_id += 1

    if negation_auxiliaries != set():
        for aux in negation_auxiliaries:
            new_aux_type = MatrixChoices.ChoiceDict()
            new_aux_type.full_key = 'aux' + str(aux_id)
            new_aux_type['name'] = 'neg' + str(aux_id)
            new_aux_type['subj'] = 'np-comp-case'
            new_aux_type['compfeature'] = MatrixChoices.ChoiceList()
            new_aux_type['compfeature'].full_key = 'aux' + str(aux_id) + '_compfeature'
            comp_feat = MatrixChoices.ChoiceDict()
            comp_feat.full_key = 'aux' + str(aux_id) + '_compfeature1'
            comp_feat['name'] = 'form'
            comp_feat['value'] = 'finite'
            new_aux_type['compfeature'].__setitem__(1, comp_feat)
            new_aux_type['stem'] = MatrixChoices.ChoiceList()
            new_aux_type['stem'].full_key = 'aux' + str(aux_id) + '_stem'
            new_stem = MatrixChoices.ChoiceDict()
            new_stem.full_key = 'stem1'
            new_stem['orth'] = aux
            new_stem['pred'] = 'neg_rel'
            new_aux_type['sem'] = 'add-pred'
            new_aux_type['stem'].__setitem__(re.sub('stem', '', new_stem.full_key), new_stem)
            aux_list.append(('aux' + str(aux_id), new_aux_type))
            # if adp_case:
            #     adp_aux_type = create_aux_with_adp_subj(new_aux_type, aux_id)
            #     aux_list.append(('aux' + str(aux_id), adp_aux_type))
            aux_id += 1


    for id, type in aux_list:
        choices.__setitem__(id, type)
        #choices[id] = type
        for stem in type['stem']:
            orth = stem['orth']
        # add the aux type to the inputs of any pc that took the verb as an input
    for aux, verb in verb_id_list:
        for pc in choices['verb-pc']:
            if verb in pc['inputs']:
                if pc.full_key != aux: # I added the pcs with aux suffix features to the verb_id_list. this statement checks to make sure that we don't create circular rules where those pcs are their own inputs
                    if choices['aux-comp'] == 's':
                        feat_lists = [lrt['feat'] for lrt in pc['lrt']]
                        feats = [f for fl in feat_lists for f in fl]
                        if not any([f for f in feats if f['head'] == 'subj']):
                            pc['inputs'] += ', ' + aux
                    else:
                        pc['inputs'] += ', ' + aux

    return choices

def copy_aux_features(new_aux_type, verb_type, aux_id):
    feat_count = 1
    for feat in verb_type['feat']:
        if feat['head'] == 'subj' or feat['head'] == 'obj':
            continue #we want to handle subj features differently, because customization doesn't allow them on auxes
        if feat['name'] != 'case':
            new_feat = MatrixChoices.ChoiceDict()
            new_feat.full_key = 'feat' + str(feat_count)
            new_feat['head'] = feat['head']
            new_feat['name'] = feat['name']
            new_feat['value'] = feat['value']
            new_aux_type['feat'].__setitem__(re.sub('feat', '', new_feat.full_key), new_feat)
            feat_count += 1

    return new_aux_type

def set_aux_features(new_aux_type, aux_id, features):
    feat_count = 1
    tense_feats = []
    aspect_feats = []
    mood_feats = []
    for feat in features:
        if feat in feat_dict['tenseFeatures']:
            tense_feats.append(feat_dict['tenseFeatures'][feat])
        if feat in feat_dict['aspectFeatures']:
            aspect_feats.append(feat_dict['aspectFeatures'][feat])
        if feat in feat_dict['moodFeatures']:
            mood_feats.append(feat_dict['moodFeatures'][feat])
    if tense_feats != []:
        feats = ', '.join(tense_feats)
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'feat' + str(feat_count)
        new_feat['head'] = 'verb'
        new_feat['name'] = 'tense'
        new_feat['value'] = feats
        new_aux_type['feat'].__setitem__(re.sub('feat', '', new_feat.full_key), new_feat)
        feat_count += 1
    if aspect_feats != []:
        feats = ', '.join(aspect_feats)
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'feat' + str(feat_count)
        new_feat['head'] = 'verb'
        new_feat['name'] = 'aspect'
        new_feat['value'] = feats
        new_aux_type['feat'].__setitem__(re.sub('feat', '', new_feat.full_key), new_feat)
        feat_count += 1
    if mood_feats != []:
        feats = ', '.join(mood_feats)
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'feat' + str(feat_count)
        new_feat['head'] = 'verb'
        new_feat['name'] = 'mood'
        new_feat['value'] = feats
        new_aux_type['feat'].__setitem__(re.sub('feat', '', new_feat.full_key), new_feat)
    new_aux_type['compfeature'] = MatrixChoices.ChoiceList()
    comp_feat = MatrixChoices.ChoiceDict()
    comp_feat.full_key = 'aux' + str(aux_id) + '_compfeature1'
    comp_feat['name'] = 'form'
    comp_feat['value'] = 'finite'
    new_aux_type['compfeature'].full_key = 'aux' + str(aux_id) + '_compfeature'
    new_aux_type['compfeature'].__setitem__(1, comp_feat)
    return new_aux_type


def collect_subj_features(verb_type):
    subj_feats = {}
    feat_count = 1
    for feat in verb_type['feat']:
        if feat['head'] == 'subj':
            if feat['name'] != 'case':
                subj_feats['feat'+str(feat_count)] = {}
                subj_feats['feat' + str(feat_count)]['head'] = feat['head']
                subj_feats['feat' + str(feat_count)]['name'] = feat['name']
                subj_feats['feat' + str(feat_count)]['value'] = feat['value']
                feat_count += 1
    return subj_feats

def set_subj_case(new_aux_type, auxiliaries, stem):
    if 'subj' not in new_aux_type:  # if this is the first stem we've seen
        if auxiliaries[stem]['subj_case'] == 'none':
            new_aux_type['subj'] = 'np'
        elif auxiliaries[stem]['subj_case'] == '':
            new_aux_type['subj'] = 'np'
        elif auxiliaries[stem]['subj_case'] == 'verb':
            new_aux_type['subj'] = 'np-comp-case'
        else:
            new_aux_type['subj'] = auxiliaries[stem]['subj_case']
    else:  # if we already set a subj value, we want to check for a conflict
        if new_aux_type['subj'] == 'np':
            if auxiliaries[stem]['subj_case'] == 'none':
                new_aux_type['subj'] = 'np'
            elif auxiliaries[stem]['subj_case'] == '':
                new_aux_type['subj'] = 'np'
            elif auxiliaries[stem]['subj_case'] == 'verb':
                new_aux_type['subj'] = 'np-comp-case'
            else:
                new_aux_type['subj'] = auxiliaries[stem]['subj_case']
        elif new_aux_type['subj'] != 'verb':  # if it's a conflict between quirky case and verb case, go with verb case
            if auxiliaries[stem]['subj_case'] == 'verb':
                new_aux_type['subj'] = 'np-comp-case'
        # else do nothing because it's already set to verb case
    return new_aux_type

def set_stem(new_aux_type, stem, negation_auxiliaries, aux_id, stem_id):
    new_stem = MatrixChoices.ChoiceDict()
    new_stem.full_key = 'stem' + str(stem_id)
    new_stem['orth'] = stem['orth']
    #new_aux_type['stem' + str(stem_id) + '_orth'] = stem['orth']
    pred = stem['pred'].replace('_v_rel', '')
    pred = pred.replace('_', '')
    # if there is something that's not a feature in the pred- keep it
    if 'stem' not in new_aux_type:
        new_aux_type['stem'] = MatrixChoices.ChoiceList()
        new_aux_type['stem'].full_key = 'aux' + str(aux_id) + '_stem'
    if stem['orth'] in negation_auxiliaries:
        new_stem['pred'] = 'neg_rel'
        #new_aux_type['aux' + str(aux_id) + '_stem' + str(stem_id) + '_pred'] = 'neg_rel'
        new_aux_type['sem'] = 'add-pred'
    elif pred in known_features and pred not in feat_dict['negFeatures']:
        new_aux_type['sem'] = 'no-pred'
    elif stem['pred'] == '':
        new_aux_type['sem'] = 'no-pred'
    else:
        new_stem['pred'] = stem['pred']
        #new_aux_type['aux' + str(aux_id) + '_stem' + str(stem_id) + '_pred'] = stem['pred']
        new_aux_type['sem'] = 'add-pred'
    new_aux_type['stem'].__setitem__(re.sub('stem', '', new_stem.full_key), new_stem)
    return new_aux_type

def create_verb_pc(subj_feats, aux_id, max_verb_pc):
    new_verb_pc = MatrixChoices.ChoiceDict()
    new_verb_pc.full_key = 'verb-pc' + str(max_verb_pc)
    new_verb_pc['name'] = 'verb-pc' + str(max_verb_pc)
    new_verb_pc['inputs'] = 'aux' + str(aux_id)
    new_verb_pc['order'] = 'prefix' # this is required by validation even though it's not an inflecting class
    new_verb_pc['obligatory'] = 'on'
    new_verb_pc['lrt'] = MatrixChoices.ChoiceList()
    new_verb_pc['lrt'].full_key = 'verb-pc' + str(max_verb_pc) + '_lrt'
    new_lrt = MatrixChoices.ChoiceDict()
    new_lrt.full_key = 'verb-pc' + str(max_verb_pc) + '_lrt1'
    new_lrt['name'] = 'verb-pc' + str(max_verb_pc) + '_lrt1'
    feat_count = 1
    new_lrt['feat'] = MatrixChoices.ChoiceList()
    new_lrt['feat'].full_key = 'verb-pc' + str(max_verb_pc) + '_lrt1_feat'
    for feat in subj_feats:
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'verb-pc' + str(max_verb_pc) + '_lrt1' + '_feat' + str(feat_count)
        new_feat['name'] = subj_feats[feat]['name']
        new_feat['value'] = subj_feats[feat]['value']
        new_feat['head'] = subj_feats[feat]['head']
        new_lrt['feat'].__setitem__(feat_count, new_feat)
        feat_count += 1
    new_lri = MatrixChoices.ChoiceDict()
    new_lri.full_key = 'verb-pc' + str(max_verb_pc) + '_lrt1' + '_lri1'
    new_lri['inflecting'] = 'no'
    new_lrt['lri'] = MatrixChoices.ChoiceList()
    new_lrt['lri'].full_key = 'verb-pc' + str(max_verb_pc) + '_lrt1_lri'
    new_lrt['lri'].__setitem__(1, new_lri)
    new_verb_pc['lrt'].__setitem__(1, new_lrt)
    return new_verb_pc

def create_aux_with_adp_subj(aux_type, aux_id): #KPH I think this is no longer needed and can be removed
    adp_aux_type = MatrixChoices.ChoiceDict()
    adp_aux_type.full_key = 'aux' + str(aux_id)
    if 'neg' in aux_type['name']:
        adp_aux_type['name'] = 'neg' + str(aux_id)
    else:
        adp_aux_type['name'] = 'aux' + str(aux_id)
    adp_aux_type['subj'] = 'adp'
    adp_aux_type['feat'] = MatrixChoices.ChoiceList()
    adp_aux_type['feat'].full_key = 'aux' + str(aux_id) + '_feat'
    feat_count = 1
    for feat in aux_type['feat']:
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'aux' + str(aux_id) + '_feat' + str(feat_count)
        new_feat['head'] = feat['head']
        new_feat['value'] = feat['value']
        new_feat['name'] = feat['name']
        adp_aux_type['feat'].__setitem__(re.sub('aux', '', new_feat.full_key), new_feat)
        feat_count += 1
    adp_aux_type['compfeature'] = MatrixChoices.ChoiceList()
    adp_aux_type['compfeature'].full_key = 'aux' + str(aux_id) + '_compfeature'
    feat_count = 1
    for feat in aux_type['compfeature']:
        new_feat = MatrixChoices.ChoiceDict()
        new_feat.full_key = 'aux' + str(aux_id) + '_compfeature1'
        new_feat['value'] = feat['value']
        new_feat['name'] = feat['name']
        adp_aux_type['compfeature'].__setitem__(1, new_feat)
        feat_count +=1
    adp_aux_type['stem'] = MatrixChoices.ChoiceList()
    adp_aux_type['stem'].full_key = 'aux' + str(aux_id) + '_stem'
    stem_id = 1
    for stem in aux_type['stem']:
        new_stem = MatrixChoices.ChoiceDict()
        new_stem.full_key = 'aux' + str(aux_id) + '_stem' + str(stem_id)
        new_stem['orth'] = stem['orth']
        if 'pred' in stem:
            new_stem['pred'] = stem['pred']
        adp_aux_type['stem'].__setitem__(stem_id, new_stem)
        stem_id += 1
    adp_aux_type['sem'] = aux_type['sem']
    return adp_aux_type

def create_subj_feats(features):
    subj_feats = {}
    feat_count = 1
    person_feats = []
    number_feats = []
    gender_feats = []
    pernum_feats = []
    for feat in features:
        if feat in feat_dict['perNumFeatures']:
            person, number = feat_dict['perNumFeatures'][feat]
            person_feats.append(person)
            number_feats.append(number)
        elif feat in feat_dict['perGenFeatures']:
            person, gender = feat_dict['perGenFeatures'][feat]
            person_feats.append(person)
            gender_feats.append(gender)
        elif feat in feat_dict['perNumGenFeatures']:
            person, number, gender = feat_dict['perNumGenFeatures'][feat]
            person_feats.append(person)
            number_feats.append(number)
            gender_feats.append(gender)
        elif feat in feat_dict['perGenHeadFeatures']:
            person, gender, head = feat_dict['perGenHeadFeatures'][feat]
            if head == 'subj':
                person_feats.append(person)
                gender_feats.append(gender)
        elif feat in feat_dict['perNumHeadFeatures']:
            person, number, head = feat_dict['perNumHeadFeatures'][feat]
            if head == 'subj':
                person_feats.append(person)
                number_feats.append(number)
        elif feat in feat_dict['perHeadFeatures']:
            person, head = feat_dict['perHeadFeatures'][feat]
            if head == 'subj':
                person_feats.append(person)
        elif feat in feat_dict['perFeatures']:
            person = feat_dict['perFeatures'][feat]
            person_feats.append(person)
        elif feat in feat_dict['numFeatures']:
            number = feat_dict['numFeatures'][feat]
            number_feats.append(number)
        elif feat in feat_dict['genFeatures']:
            gender = feat_dict['genFeatures'][feat]
            gender_feats.append(gender)
        elif feat in feat_dict['perNumInclFeatures']:
            person, number, inclusive = feat_dict['perNumInclFeatures'][feat]
            person = re.sub('[a-z]', '', person)
            pernum = person+number + '_' + inclusive
            pernum_feats.append(pernum)
        elif feat in feat_dict['perNumInclHeadFeatures']:
            person, number, inclusive, head = feat_dict['perNumInclHeadFeatures'][feat]
            person = re.sub('[a-z]', '', person)
            pernum = person + number + '_' + inclusive
            pernum_feats.append(pernum)

    if pernum_feats != []:
        feats = ', '.join(pernum_feats)
        subj_feats['feat' + str(feat_count)] = {'head': 'subj', 'value': feats, 'name': 'pernum'}
        feat_count += 1
    else:
        if person_feats != []:
            feats = ', '.join(person_feats)
            subj_feats['feat' + str(feat_count)] = {'head': 'subj', 'value': feats, 'name': 'person'}
            feat_count += 1
        if number_feats != []:
            feats = ', '.join(number_feats)
            subj_feats['feat' + str(feat_count)] = {'head': 'subj', 'value': feats, 'name': 'number'}
            feat_count += 1
    if gender_feats != []:
        feats = ', '.join(gender_feats)
        subj_feats['feat' + str(feat_count)] = {'head': 'subj', 'value': feats, 'name': 'gender'}
        feat_count += 1

        # for now, I'm not looking at case (like you would see on the second position clitic cluster/auxiliary in
        # basque, because I'm assuming that this is still constrained by the verb, and simply expressed in the clitic cluster
        # in other words, I don't think the auxiliary changes the case frame of the verb

    return subj_feats

def extract_adpositions(xc, choices, adp_tags, noun_tags, wo_choices):
    dets = set()
    if 'det' in choices:
        for det in choices['det']:
            for stem in det['stem']:
                dets.add(stem['orth'])
    adpositions = {}
    for igt in xc:
        pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
        if pos_tier:
            for pos in pos_tier:
                if pos.value() in adp_tags:
                    case = ''
                    if 'gw' in pos.attributes['alignment']:
                        gw = igt.get('gw').get(pos.attributes['alignment'])
                        if not igt.get('w').get(gw.attributes['alignment']):
                            continue
                        orth = igt.get('w').get(gw.attributes['alignment']).value()
                    else:
                        gw = igt.get('gw').get('g'+pos.id)
                        if not igt.get('w').get(pos.id):
                            continue
                        orth = igt.get('w').get(pos.id).value()
                    for gloss in re.split('-|\.', gw.value()):
                        if gloss.lower() in feat_dict['caseFeatures']:
                            case = feat_dict['caseFeatures'][gloss.lower()]

                    if case == '':
                        continue
                    closest_noun = find_nearest_np(gw.id, igt, noun_tags, dets)
                    if closest_noun == '':
                        continue
                    elif int(closest_noun[2:]) > int(gw.id[2:]):
                        position = 'before'
                    else:
                        position = 'after'

                    orth = orth.rstrip('[-=\?\.\!\,\:\;\(\)\'\"\`#]')
                    if orth not in adpositions:
                        adpositions[orth] = {'before': 0, 'after': 0, 'case': set()}
                    if position == 'before':
                        adpositions[orth]['before'] += 1
                    elif position == 'after':
                        adpositions[orth]['after'] += 1
                    adpositions[orth]['case'].add(case)

    count = 1
    for adp in adpositions:
        choices.__setitem__('adp'+str(count)+'_orth', adp)
        choices.__setitem__('adp' + str(count) + '_opt', 'on')
        if adpositions[adp]['before'] > adpositions[adp]['after']:
            choices.__setitem__('adp' + str(count) + '_order', 'before')
        elif adpositions[adp]['before'] > adpositions[adp]['after']:
            choices.__setitem__('adp' + str(count) + '_order', 'after')
        else:
            if wo_choices['word-order'] in ['vso', 'vos', 'svo' 'v-initial']:
                choices.__setitem__('adp' + str(count) + '_order', 'before')
            elif wo_choices['word-order'] in ['osv', 'sov', 'ovs', 'v-final']:
                choices.__setitem__('adp' + str(count) + '_order', 'after')
            else:
                choices.__setitem__('adp' + str(count) + '_order', 'after')
        choices.__setitem__('adp' + str(count) + '_feat1_name', 'case')
        cases = ', '.join(adpositions[adp]['case'])
        choices.__setitem__('adp' + str(count) + '_feat1_value', cases)
        count += 1

    return choices

def find_nearest_np(adp, igt, noun_tags, dets):
    np = ''
    pos_tier = igt.get('pos') if igt.get('pos') else igt.get('w-pos') if igt.get('w-pos') else igt.get('gw-pos')
    adp_id = int(re.sub('w|gw', '', adp))
    if pos_tier:
        if 'gw' in pos_tier.id:
            tier_id = 'gw'
        else:
            tier_id = 'w'
        previous = adp_id - 1
        if adp_id < len(igt.get('w'))-1:
            next = adp_id + 1
        else:
            next = adp_id
        while previous > 0 and next < len(igt.get('w'))+1:
            if igt.get('w').get('w'+str(previous)).value().lower() in dets and previous > 1:
                previous -= 1
            if igt.get('w').get('w'+str(next)).value().lower() in dets and next < len(igt.get('w'))+1:
                next += 1
            for pos in pos_tier:
                if pos.attributes['alignment'] == tier_id+str(previous) and pos.value() in noun_tags:
                    np = 'gw'+str(previous)
                elif pos.attributes['alignment'] == tier_id+str(next) and pos.value in noun_tags:
                    np = 'gw'+str(next)
            if previous > 1:
                previous -= 1
            if next < len(igt.get('w'))+1:
                next += 1
    return np
