#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kriten Howell
# inference for the sentential negation choices in the gm

from mom import MatrixChoices
from XigtifiedToolbox import FeatDict
from xigt.codecs import xigtxml
import sys
import re
from operator import itemgetter

KNOWN_FEATS = FeatDict.FeatureDictionary().all_keys

def negated_sentence(igt):
    # Return True if the English translation is negated, but not if there are multiple instances of negation
    # First check for not or 'nt in the sentence. Then see if it is preceded by exaclty one verb
    negation_count = 0
    is_negated = False
    translation = igt.get('tw')
    for word in translation:
        if word.value():
            if '\'nt' in word.value():
                negation_count += 1
                is_negated = True
            elif word.value() == 'not':
                negation_count += 1
                # first check if the word before neg is a verb
                prev_word_id = 'tw'+str(int(word.id[2:])-1)
                for pos in igt.get('tw-pos'):
                    if pos.attributes['alignment'] == prev_word_id:
                        if pos.value() == 'VERB':
                            is_negated = True
                            # then check if the word before that is a verb
                            prev_word_id = 'tw' + str(int(word.id[2:]) - 2)
                            for pos in igt.get('tw-pos'):
                                if pos.attributes['alignment'] == prev_word_id:
                                    if pos.value() == 'VERB':
                                        is_negated = False

    if negation_count > 1:
        is_negated = False
    return is_negated

def collect_negation_morphemes(igt):
    neg_glosswords = []
    neg_glosses = []
    glosses = igt.get('g')
    if glosses:
        for gloss in glosses:
            tokens = re.split(':|-|=|\.', gloss.value().lower())
            if 'neg' in tokens or 'not' in tokens:
                if all([tok in KNOWN_FEATS for tok in tokens]):
                    if 'gw' in gloss.id:
                        neg_glosswords.append(gloss)
                        neg_glosses.append(None)
                    else:
                        neg_glosses.append(gloss)
                        gw = igt.get('gw').get(gloss.attributes['segmentation'])
                        #gw_id = 'gw' + re.split('\.', gloss.id)[0][1:]
                        #gw = igt.get('gw').get(gw_id)
                        neg_glosswords.append(gw)
    return neg_glosswords, neg_glosses


def get_orth(igt, neg):
    orth = ''
    if neg:
        m_id = neg.attributes['alignment']
        orth = igt.get('m').get(m_id).value()
    return orth

def negation_bound_morpheme(igt, neg, orth, auxiliares):
    # Check to see if the neg gloss shows up on a verb. If it does, assume that it's a morpheme on the verb.
    # This assumption should be safe, because intent tags not as an adverb.
    neg_morph = False
    if 'gw' in neg.id:
        gw_id = neg.id
    # Here we are checking for an intent tier, because the linguist might gloss an neg aux as a verb, but INTENT won't
    if 'INTENT' in igt.get('gw-pos').attributes['data-provenance']:
        for gw_pos in igt.get('gw-pos'):
            if gw_pos.alignment == gw_id:
                if gw_pos.value() == 'VERB':
                    neg_morph = True
    # Neg could also be an affix on an auxiliary so check to see if a known aux is part of the word
    w = igt.get('gw').get(gw_id).attributes['alignment']
    morphemes = [m.value().lower() for m in igt.get('m') if m.attributes['segmentation'] == w]
    for morph in morphemes:
        if morph != orth and morph in auxiliares:
            neg_morph = True

    return neg_morph

def negation_auxiliary(igt, neg, orth, auxiliaries, igt2items):
    # if particle gets agreement or has TAM, it's an aux
    # if aux inference identified it as an aux, it's an aux
    neg_aux = False
    tokens = re.split(':|-|=|\.', neg.value().lower())
    feat_dict = FeatDict.FeatureDictionary().dictionaries
    del feat_dict['negFeatures']
    for gloss in tokens:
        for type in feat_dict:
            if gloss in feat_dict[type]:
                neg_aux = True
    # KPH: I was going to add a check to see if auxes are required in finite clauses, but I don't think projections
    # are good enough to do this reliably. Instead, I'm just checking to see if auxiliary inference picked the word up
    if not neg_aux:
        # if igt.id in igt2items and neg.attributes['alignment'] in igt2items[igt.id]:
        #     orth = str(igt2items[igt.id][neg.attributes['alignment']].root)
        # else:
        #     orth = igt.get('w').get(neg.attributes['alignment']).value().lower()
        if orth in auxiliaries:
            neg_aux = True
    return neg_aux

def update_adverb_patterns(adverb_patterns, igt, neg, orth):
    # find out if the neg adverb occurs before or after a vp or s by deciding if it intervenes between a verb and subjet
    # default to s if it does not intervene
    # v attachment is unlikely, so we don't worry about it
    subj_wid = ''
    verb_wid = ''
    neg_wid = neg.alignment
    head = ''
    position = ''
    attach = ''
    confident = False
    w_ds = igt.get('w-ds')
    if w_ds:
        for dep in w_ds:
            if dep.value() == 'neg':
                if 'dep' in dep.attributes and 'head' in dep.attributes:
                    if dep.attributes['dep'] == neg.alignment:
                        head = dep.attributes['head']
        for dep in w_ds:
            if dep.value() == 'nsubj':
                if 'head' in dep.attributes and 'dep' in dep.attributes:
                    if dep.attributes['head'] == head:
                        verb_wid = dep.attributes['head']
                        subj_wid = dep.attributes['dep']
                        confident = True

    if subj_wid != '' and verb_wid != '':
        if subj_wid < neg_wid < verb_wid:
            position = 'before'
            attach = 'vp'
        elif verb_wid < neg_wid < subj_wid:
            position = 'after'
            attach = 'vp'
        else:
            attach = 's'
            if neg_wid < verb_wid:
                position = 'before'
            else:
                position = 'after'
        pattern = position+'+'+attach
        if orth in adverb_patterns:
            adverb_patterns[orth]['count'] += 1
            if pattern in adverb_patterns[orth]['patterns']:
                adverb_patterns[orth]['patterns'][pattern]['count'] += 1
            else:
                adverb_patterns[orth]['patterns'][pattern] = {'count': 1, 'position': position, 'attach': attach, 'confident': confident}
        else:
            adverb_patterns[orth] = {'count': 1, 'patterns': {pattern: {'count': 1, 'position': position, 'attach': attach, 'confident': confident}}}

    # this approach is less reliable, so we will call it a guess and assign it lower confidence
    elif verb_wid == '' and head != '':
        verb_wid = head
        for dep in w_ds:
            if dep.value() == 'nsubj' and 'dep' in dep.attributes:
                if subj_wid == '':
                    subj_wid = dep.attributes['dep']
                elif (int(verb_wid[1:]) - int(subj_wid[1:])) > (int(verb_wid[1:]) - int (dep.attributes['dep'][1:])):
                        subj_wid = dep.attributes['dep']
        if subj_wid != '':
            if subj_wid < neg_wid < verb_wid:
                position = 'before'
                attach = 'vp'
            elif verb_wid < neg_wid < subj_wid:
                position = 'after'
                attach = 'vp'
            else:
                attach = 's'
                if neg_wid < verb_wid:
                    position = 'before'
                else:
                    position = 'after'
            pattern = position+'+'+attach
            if orth in adverb_patterns:
                adverb_patterns[orth]['count'] += 1
                if pattern in adverb_patterns[orth]['patterns']:
                    adverb_patterns[orth]['patterns'][pattern]['count'] += 1
                else:
                    adverb_patterns[orth]['patterns'][pattern] = {'count': 1, 'position': position,
                                                                                  'attach': attach,
                                                                                  'confident': confident}
            else:
                adverb_patterns[orth] = {'count': 1, 'patterns': {
                   pattern: {'count': 1, 'position': position, 'attach': attach,
                                              'confident': confident}}}

    return adverb_patterns

def find_best_adv_pattern(adverb_patterns):
    confident_patterns = []
    guessed_patterns = []
    s_count = 0
    vp_count = 0
    before = False
    after = False
    for orth in adverb_patterns:
        for pattern in adverb_patterns[orth]['patterns']:
            if adverb_patterns[orth]['patterns'][pattern]['confident']:
                confident_patterns.append((orth, adverb_patterns[orth]['patterns'][pattern]))
            else:
                guessed_patterns.append((orth, adverb_patterns[orth]['patterns'][pattern]))
    if confident_patterns != []:
        max_s = 0
        max_vp = 0
        for orth, pattern in confident_patterns:
            if pattern['attach'] == 's':
                if pattern['count'] > max_s:
                    max_s = pattern['count']
                    s_orth = orth
                s_count += pattern['count']
            elif pattern['attach'] == 'vp':
                if pattern['count'] > max_vp:
                    max_vp = pattern['count']
                    vp_orth = orth
                vp_count += pattern['count']
            if pattern['position'] == 'before':
                before = True
            elif pattern['position'] == 'after':
                after = True
        if s_count > vp_count:
            attach = 's'
            orth = s_orth
        else:
            attach = 'vp'
            orth = vp_orth
    else:
        max_s = 0
        max_vp = 0
        for orth, pattern in guessed_patterns:
            if pattern['attach'] == 's':
                if pattern['count'] > max_s:
                    max_s = pattern['count']
                    s_orth = orth
                s_count += pattern['count']
            elif pattern['attach'] == 'vp':
                if pattern['count'] > max_vp:
                    max_vp = pattern['count']
                    vp_orth = orth
                vp_count += pattern['count']
            if pattern['position'] == 'before':
                before = True
            elif pattern['position'] == 'after':
                after = True
        if s_count > vp_count:
            attach = 's'
            orth = s_orth
        else:
            attach = 'vp'
            orth = vp_orth
    if before and after:
        position = 'either'
    elif before:
        position = 'before'
    elif after:
        position = 'after'

    return position, attach, orth


def set_choices(choices, negation_patterns, adverb_patterns):
    if negation_patterns['simple']['count'] > negation_patterns['bipartite']['count']:
        choices.__setitem__('neg-exp', '1')
    elif negation_patterns['bipartite']['count'] > negation_patterns['simple']['count']:
        choices.__setitem__('neg-exp', '2')
    elif negation_patterns['zero']['count'] > negation_patterns['simple']['count'] and negation_patterns['zero']['count'] > negation_patterns['bipartite']['count']:
        choices.__setitem__('neg-exp', '0')
    else:
        choices.__setitem__('neg-exp', '1')
        print('theres a tie between two negation strategies')
    # simple negation
    if choices['neg-exp'] == '1':
        for strategy in negation_patterns['simple']['strategy']:
            if strategy == 'bound':
                choices.__setitem__('infl-neg', 'on')
            elif strategy == 'aux':
                choices.__setitem__('neg-aux', 'on')
            elif strategy == 'adv':
                choices.__setitem__('adv-neg', 'on')
                if adverb_patterns != {}:
                    position, attach, orth = find_best_adv_pattern(adverb_patterns)
                    #most_frequent_adv = sorted(adverb_patterns, key=lambda x: x[1]['count'], reverse=True)[0]
                    choices.__setitem__('neg-mod', attach)
                    choices.__setitem__('neg-order', position)
                    choices.__setitem__('neg-adv-orth', orth)
                else:
                    # if no pattern found, default to either side of vp
                    most_frequent_adv = sorted(negation_patterns['simple']['strategy']['adv']['orths'], key=lambda kv: kv[1], reverse=True)[0]
                    choices.__setitem__('neg-mod', 'vp')
                    choices.__setitem__('neg-order', 'either')
                    choices.__setitem__('neg-adv-orth', most_frequent_adv)
                    print('inference decided that a neg word was an adverb, but couldnt identify its position or attachment')

    # bipartite
    elif choices['neg-exp'] == '2':
        strats = negation_patterns['bipartite']['strategy']
        most_frequent_strategy = sorted(strats, key=lambda x: x[0])[0]
        if len(sorted(sorted(strats, key=lambda x: x[0]))) > 1:
            second_most_frequent_strategy = sorted(strats, key=lambda x: x[0])[1]
        else:
            second_most_frequent_strategy = None
        if most_frequent_strategy == 'aux+aux':
            if second_most_frequent_strategy:
                most_frequent_strategy = second_most_frequent_strategy
        if most_frequent_strategy == 'bound+bound':
            choices.__setitem__('infl-neg', 'on')
        elif most_frequent_strategy == 'bound+aux' or most_frequent_strategy == 'aux+bound':
            choices.__setitem__('infl-neg', 'on')
            choices.__setitem__('neg-aux', 'on')
            choices.__setitem__('neg-aux-index', '1')
            choices.__setitem__('bineg-type', 'infl-head')
            choices.__setitem__('neg1-type', 'fh')
            choices.__setitem__('neg2-type', 'b')
        elif most_frequent_strategy == 'bound+adv' or most_frequent_strategy == 'adv+bound':
            choices.__setitem__('infl-neg', 'on')
            choices.__setitem__('adv-neg', 'on')
            choices.__setitem__('bineg-type', 'infl-mod')
            choices.__setitem__('neg1-type', 'fm')
            choices.__setitem__('neg2-type', 'b')
            if adverb_patterns != {}:
                position, attach, orth = find_best_adv_pattern(adverb_patterns)
                choices.__setitem__('neg-mod-infl-mod-neg', attach)
                choices.__setitem__('neg-mod-order-infl-mod-neg', position)
                choices.__setitem__('neg-mod-orth', orth)
            else:
                most_frequent_adv = sorted(negation_patterns['bipartite']['strategy']['adv']['orths'], key=lambda kv: kv[1], reverse=True)[0]
                choices.__setitem__('neg-mod-infl-mod-neg', 'vp')
                choices.__setitem__('neg-mod-order-infl-mod-neg', 'either')
                choices.__setitem__('neg-mod-orth', most_frequent_adv)
                print('inference decided that a neg word was an adverb, but couldnt identify its position or attachment')
        elif most_frequent_strategy == 'aux+adv' or most_frequent_strategy == 'adv+aux':
            choices.__setitem__('neg-aux', 'on')
            choices.__setitem__('neg-aux-index', '1')
            choices.__setitem__('adv-neg', 'on')
            choices.__setitem__('bineg-type', 'head-mod')
            choices.__setitem__('neg1-type', 'fm')
            choices.__setitem__('neg2-type', 'fh')
            if adverb_patterns != {}:
                position, attach, orth = find_best_adv_pattern(adverb_patterns)
                choices.__setitem__('neg-mod-infl-mod-neg', attach)
                choices.__setitem__('neg-mod-order-infl-mod-neg', position)
                choices.__setitem__('neg-mod-orth', orth)
            else:
                most_frequent_adv = sorted(negation_patterns['bipartite']['strategy']['adv']['orths'], key=lambda kv: kv[1], reverse=True)[0]
                choices.__setitem__('neg-mod-infl-mod-neg', 'vp')
                choices.__setitem__('neg-mod-order-infl-mod-neg', 'either')
                choices.__setitem__('neg-mod-orth', most_frequent_adv)
                print('inference decided that a neg word was an adverb, but couldnt identify its position or attachment')
        elif most_frequent_strategy == 'adv+adv':
            choices.__setitem__('adv-neg', 'on')
            choices.__setitem__('bineg-type', 'mod-mod')
            choices.__setitem__('neg1-type', 'fm')
            choices.__setitem__('neg2-type', 'fm')
            if adverb_patterns != {}:
                position, attach, orth = find_best_adv_pattern(adverb_patterns)
                choices.__setitem__('neg1-mod', attach)
                choices.__setitem__('neg1-mod-order', position)
                choices.__setitem__('neg1-mod-orth', orth)
                if len(sorted(adverb_patterns, key=lambda x: x[1]['count'], reverse=True)) > 1:
                    most_frequent_adv = sorted(adverb_patterns, key=lambda x: x[1]['count'], reverse=True)[1]
                # otherwise, if there was only one, enter the it again
                choices.__setitem__('neg2-mod', attach)
                choices.__setitem__('neg2-mod-order', position)
                choices.__setitem__('neg2-mod-orth', most_frequent_adv)
            else:
                most_frequent_adv = sorted(negation_patterns['bipartite']['strategy']['adv']['orths'], key=lambda kv: kv[1], reverse=True)[0]
                choices.__setitem__('neg1-mod', 'vp')
                choices.__setitem__('neg1-mod-order', 'either')
                choices.__setitem__('neg1-mod-orth', most_frequent_adv)
                choices.__setitem__('neg2-mod', 'vp')
                choices.__setitem__('neg2-mod-order', 'either')
                choices.__setitem__('neg2-mod-orth', most_frequent_adv)
                print(
                    'inference decided that a neg word was an adverb, but couldnt identify its position or attachment')
        else:
            choices.__delitem__('neg-exp')
            if most_frequent_strategy != 'aux+aux':
                print('did i miss a bipartite combination?')
                print(most_frequent_strategy)
    return choices

def infer_sent_neg(xc, auxiliares, igt2items):
    choices = MatrixChoices.ChoiceDict()
    negation_patterns = {'zero': {'count': 0}, 'simple': {'count': 0, 'strategy': {}}, 'bipartite': {'count': 0, 'strategy': {}}} # a dictionary for keeping track of strategies until the end
    adverb_patterns = {}
    neg_aux_orths = set()
    # Main Routine
    for igt in xc: # for each igt in the corpus
        if not igt.get('gw-pos'):
            continue
        if negated_sentence(igt): # if the igt has sentential negation
            neg_glosswords, neg_glosses = collect_negation_morphemes(igt) # identify the 'neg' glosses and the word/morpheme they correspond to
            negation_morphemes = [] # a list of negation morphemes with their id, the type of negation and the orthography
            for neg_gw, neg_g in zip(neg_glosswords, neg_glosses):
                orth = get_orth(igt, neg_g).lower()
                neg_bound = negation_bound_morpheme(igt, neg_gw, orth, auxiliares) # is the negation morpheme an affix on a verb
                if neg_bound:
                    negation_morphemes.append([neg_gw, 'bound', orth])
                if not neg_bound:
                    neg_aux = negation_auxiliary(igt, neg_gw, orth, auxiliares, igt2items) # is the negation morpheme an auxiliary
                    if neg_aux:
                        if igt.id in igt2items and neg_gw.attributes['alignment'] in igt2items[igt.id]:
                            orth = str(igt2items[igt.id][neg_gw.attributes['alignment']].root)
                        negation_morphemes.append([neg_gw, 'aux', orth])
                        neg_aux_orths.add(orth)
                    else:# if it's not an auxiliary, let's call it an adverb
                        negation_morphemes.append([neg_gw, 'adv', orth])
                        adverb_patterns = update_adverb_patterns(adverb_patterns, igt, neg_gw, orth)
            # add the pattern in the igt to the patterns dictionary
            if len(neg_glosswords) == 0:
                negation_patterns['zero']['count'] += 1
            elif len(neg_glosswords) == 1:
                strategy = negation_morphemes[0][1]
                orth = negation_morphemes[0][2]
                negation_patterns['simple']['count'] += 1
                if strategy in negation_patterns['simple']['strategy']:
                    negation_patterns['simple']['strategy'][strategy]['count'] += 1
                    if orth in negation_patterns['simple']['strategy'][strategy]['orths']:
                        negation_patterns['simple']['strategy'][strategy]['orths'][orth] += 1
                    else:
                        negation_patterns['simple']['strategy'][strategy]['orths'][orth] = 1
                else:
                    negation_patterns['simple']['strategy'][strategy] = {'count': 1, 'orths': {orth: 1}}
            else:
                neg_type = 'bipartite'
                negation_morphemes = sorted(negation_morphemes,key=itemgetter(1))
                strategy = negation_morphemes[0][1]+'+'+negation_morphemes[1][1]
                orth = negation_morphemes[0][2]+'+'+negation_morphemes[1][2]
                negation_patterns['bipartite']['count'] += 1
                if strategy in negation_patterns['bipartite']['strategy']:
                    negation_patterns['bipartite']['strategy'][strategy]['count'] += 1
                    if orth in negation_patterns['bipartite']['strategy'][strategy]['orths']:
                        negation_patterns['bipartite']['strategy'][strategy]['orths'][orth] += 1
                    else:
                        negation_patterns['bipartite']['strategy'][strategy]['orths'][orth] = 1
                else:
                    negation_patterns['bipartite']['strategy'][strategy] = {'count': 1, 'orths': {orth: 1}}


    choices = set_choices(choices, negation_patterns, adverb_patterns)
    if not choices['neg-aux'] == 'on': # KPH if we didn't select neg aux as a strategy, we don't want to add these as special auxiliaries
        neg_aux_orths = set()
    return choices, neg_aux_orths

def write_choices_file(choices, file):
    # writes the chocies file
    with open(file, 'w', encoding="UTF-8") as f:
        matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
        f.write(str(matrix_choices.choices))

if __name__ == '__main__':
    corpus = open(sys.argv[1], encoding='utf-8')
    xc = xigtxml.load(corpus)
    #wo_choices = MatrixChoices.ChoicesFile(choices_file=sys.argv[2])
    wo_choices = {}
    negation_choices = infer_sent_neg(xc)
    write_choices_file(negation_choices, 'C:/Users/Kristen/Documents/code/agg/negation_choices_'+sys.argv[3])
