#!/usr/bin/env python3
# -*- coding: utf-8 -*
# Author: Kristen Howell
# this script infers the answers to the word order page and extracts lexical entries for detirminers and nouns
# with the appropriate choices

from mom import MatrixChoices
from XigtifiedToolbox import FeatDict
from xigt.codecs import xigtxml
import sys
from aggutils.find_root import find_root
import re

import sys
import math
from collections import Counter
from inference_scripts import infer_lex_items, infer_aux

# Axes: SV-VS (SV = 0, VS = 1)
#       OV-VO (OV = 0, VO = 1)
#       SO-OS (SO = 0, OS = 1)
#                              SV-VS OV-VO SO-OS
canonical_wo = [('sov',       (-1.0, -1.0, -1.0)),
                ('osv',       (-1.0, -1.0,  1.0)),
                ('svo',       (-1.0,  1.0, -1.0)),
               #('*v*',       (-1.0,  1.0,  1.0)), impossible
               #('*v*',       ( 1.0, -1.0, -1.0)), impossible
                ('ovs',       ( 1.0, -1.0,  1.0)),
                ('vso',       ( 1.0,  1.0, -1.0)),
                ('vos',       ( 1.0,  1.0,  1.0)),
                ('v-final',   (-1.0, -1.0,  0.0)),
                ('v-initial', ( 1.0,  1.0,  0.0)),
               #('v2',        ( 0.0,  0.0,  0.0)), # cannot tell v2 from free
                ('free',      ( 0.0,  0.0,  0.0))]#, # free needs a few points
 #               ('free',      ( 0.0,  0.0,  1.0)),
#                ('free',      ( 0.0,  0.0, -1.0))]

def binary_counts(cs):
    return Counter(
      sv=(cs['sv'] + cs['sov'] + cs['osv'] + cs['svo']),
      vs=(cs['vs'] + cs['vos'] + cs['vso'] + cs['ovs']),
      ov=(cs['ov'] + cs['sov'] + cs['osv'] + cs['ovs']),
      vo=(cs['vo'] + cs['svo'] + cs['vos'] + cs['vso']),
      so=(cs['sov'] + cs['svo'] + cs['vso']),
      os=(cs['osv'] + cs['vos'] + cs['ovs']))

def stddev(l):
    avg = float(sum(l))/len(l)
    return math.sqrt(sum((avg-x)**2 for x in l)/float(len(l)))

def euclidean(p1, p2):
    return math.sqrt(sum((x[0] - x[1])**2 for x in zip(p1,p2)))

def cosine(p1, p2):
    print(p1,p2)
    # invert so sorting works
    return -1 * float(sum(x[0] * x[1] for x in zip(p1, p2))) /\
                (math.sqrt(float(sum(x*x for x in p1))) *\
                 math.sqrt(float(sum(x*x for x in p2))) + 1)

def print_axes(cs, sv_vs, ov_vo, so_os, f):
    print('Counts:', file=f)
    l = (cs['sv'] + .1) / float(cs['vs'] + cs['sv'] + .1)
    r = (cs['vs'] + .1) / float(cs['vs'] + cs['sv'] + .1)
    print('SV {} VS ({} : {})'.format(format_bar(l, r), l, r), file=f)
    l = (cs['ov'] + .1) / float(cs['vo'] + cs['ov'] + .1)
    r = (cs['vo'] + .1) / float(cs['vo'] + cs['ov'] + .1)
    print('OV {} VO ({} : {})'.format(format_bar(l, r), l, r), file=f)
    l = (cs['so'] + .1) / float(cs['os'] + cs['so'] + .1)
    r = (cs['os'] + .1) / float(cs['os'] + cs['so'] + .1)
    print('SO {} OS ({} : {})'.format(format_bar(l, r), l, r), file=f)
    print('Scores:', file=f)
    l = abs(sv_vs) if sv_vs < 0 else 0
    r = 0 if sv_vs < 0 else abs(sv_vs)
    print('SV {} VS ({} : {})'.format(format_bar(l, r, '+'), l, r), file=f)
    l = abs(ov_vo) if ov_vo < 0 else 0
    r = 0 if ov_vo < 0 else abs(ov_vo)
    print('OV {} VO ({} : {})'.format(format_bar(l, r, '+'), l, r), file=f)
    l = abs(so_os) if so_os < 0 else 0
    r = 0 if so_os < 0 else abs(so_os)
    print('SO {} OS ({} : {})'.format(format_bar(l, r, '+'), l, r), file=f)


def format_bar(l, r, bar='='):
    return '[{}{}|{}{}]'.format('.' * int(.5 + 20 * (1 - l)),
                                bar * int(.5 + 20 * l),
                                bar * int(.5 + 20 * r),
                                '.' * int(.5 + 20 * (1 - r)))

def infer_sov(counts, dist=euclidean, scale=float):
    cs = binary_counts(counts)
    sv_vs = (scale(cs['vs']) - scale(cs['sv']) + .1) /\
            (scale(float(cs['vs'] + cs['sv'])) + .1)
    ov_vo = (scale(cs['vo']) - scale(cs['ov']) + .1) /\
            (scale(float(cs['vo'] + cs['ov'])) + .1)
    so_os = (scale(cs['os']) - scale(cs['so']) + .1) /\
            (scale(float(cs['os'] + cs['so'])) + .1)
    ranked = [(dist((sv_vs, ov_vo, so_os), cwo_pt), cwo)
              for (cwo, cwo_pt) in canonical_wo]
    print_axes(cs, sv_vs, ov_vo, so_os, sys.stderr)
    print(ranked, file=sys.stderr)
    wo = min(ranked)[1]
    # special case for free (ignore VOS and VSO)
    if wo == 'free' and (counts['svo'] + counts['ovs'] + counts['sv'] + counts['ov']) >\
                        (counts['sov'] + counts['osv'] + counts['vs'] + counts['vo'] + counts['v']) * 5:
        wo = 'v2'
    return wo

def infer_basic_wo(xc):
    choices = MatrixChoices.ChoiceDict()
    # KPH start by counting each observed order in the corpus (include intransitive and dropped arguments)
    observed_orders = {'svo': 0, 'sov': 0, 'vos': 0, 'ovs': 0, 'vso': 0, 'osv': 0, 'sv': 0, 'vs': 0, 'ov': 0, 'vo': 0, 'v': 0, 'total': 0}
    for igt in xc:
        if not igt.get('w-ds'):
            continue
        w_ds = igt.get('w-ds')
        root = [x.attributes['dep'] for x in w_ds if x.value() == 'root']
        if len(root) == 1:
            verb_id = root[0]
            verb = int(verb_id[1:])
            subj = 0
            obj = 0
            for ds in w_ds:
                if 'dep' in ds.attributes and 'head' in ds.attributes:
                    if ds.attributes['head'] == verb_id:
                        if ds.value() == 'nsubj':
                            subj = int(ds.attributes['dep'][1:])
                        elif ds.value() == 'dobj':
                            obj = int(ds.attributes['dep'][1:])
            if subj != 0 and obj != 0:
                observed_orders['total'] += 1
                if subj < verb < obj:
                    observed_orders['svo'] += 1
                elif subj < obj < verb:
                    observed_orders['sov'] += 1
                elif verb < obj < subj:
                    observed_orders['vos'] += 1
                elif obj < verb < subj:
                    observed_orders['ovs'] += 1
                elif verb < subj < obj:
                    observed_orders['vso'] += 1
                elif obj < subj < verb:
                    observed_orders['osv'] += 1
            elif subj != 0:
                observed_orders['total'] += 1
                if subj < verb:
                    observed_orders['sv'] += 1
                elif verb < subj:
                    observed_orders['vs'] += 1
            elif obj != 0:
                observed_orders['total'] += 1
                if obj < verb:
                    observed_orders['ov'] += 1
                if verb < obj:
                    observed_orders['vo'] += 1
            # else:
            #     observed_orders['v'] += 1
            # observed_orders['total'] += 1

    wo = infer_sov(observed_orders)
    if all(x == 0 for x in observed_orders.values()):
        wo = 'free'
    #
    # # I don' think the following is necessary (comes from 2013 paper)
    # axis_orders = {}
    # axis_orders['sv'] = observed_orders['svo'] + observed_orders['sov'] + observed_orders['osv'] + observed_orders['sv']
    # axis_orders['vs'] = observed_orders['vso'] + observed_orders['vos'] + observed_orders['ovs'] + observed_orders['vs']
    # axis_orders['ov'] = observed_orders['ovs'] + observed_orders['osv'] + observed_orders['sov'] + observed_orders['ov']
    # axis_orders['vo'] = observed_orders['vos'] + observed_orders['vso'] + observed_orders['svo'] + observed_orders['vo']
    # axis_orders['so'] = observed_orders['sov'] + observed_orders['svo'] + observed_orders['vso']
    # axis_orders['os'] = observed_orders['osv'] + observed_orders['vos'] + observed_orders['ovs']
    #
    # # KPH then map the observed orders with the canonical orders they are compatible with
    # # this means counting the single argument orders more than once
    # cannonical_orders = {}
    # cannonical_orders['svo'] = observed_orders['svo'] + observed_orders['sv'] + observed_orders['vo']
    # cannonical_orders['sov'] = observed_orders['sov'] + observed_orders['sv'] + observed_orders['ov']
    # cannonical_orders['vos'] = observed_orders['vos'] + observed_orders['vo'] + observed_orders['vs']
    # cannonical_orders['ovs'] = observed_orders['ovs'] + observed_orders['ov'] + observed_orders['vs']
    # cannonical_orders['vso'] = observed_orders['vso'] + observed_orders['vs'] + observed_orders['vo']
    # cannonical_orders['osv'] = observed_orders['osv'] + observed_orders['ov'] + observed_orders['ov']
    # cannonical_orders['v-initial'] = observed_orders['vso'] + observed_orders['vos'] + observed_orders['vo'] + observed_orders['vs']
    # cannonical_orders['v-final'] = observed_orders['sov'] + observed_orders['osv'] + observed_orders['ov'] + observed_orders['sv']
    # cannonical_orders['v2'] = observed_orders['svo'] + observed_orders['ovs'] + observed_orders['ov'] + observed_orders['sv'] # do other constituents need to be taken into account?
    #
    # if observed_orders['total'] != 0:
    #     SVO = float(cannonical_orders['svo']/observed_orders['total'])
    #     SOV = float(cannonical_orders['sov'] / observed_orders['total'])
    #     VOS = float(cannonical_orders['vos'] / observed_orders['total'])
    #     OVS = float(cannonical_orders['ovs'] / observed_orders['total'])
    #     VSO = float(cannonical_orders['vso'] / observed_orders['total'])
    #     OSV = float(cannonical_orders['osv'] / observed_orders['total'])
    #     VI = float(cannonical_orders['v-initial'] / observed_orders['total'])
    #     VF = float(cannonical_orders['v-final'] / observed_orders['total'])
    #     V2 = float(cannonical_orders['v2'] / observed_orders['total'])
    # else:
    #     SVO = 0
    #     SOV = 0
    #     VOS = 0
    #     OVS = 0
    #     VSO = 0
    #     OSV = 0
    #     VI = 0
    #     VF = 0
    #     V2 = 0
    # frequencies = [('svo', SVO), ('sov', SOV), ('vos', VOS), ('ovs', OVS), ('vso', VSO), ('osv', OSV), ('v-initial', VI), ('v-final', VF), ('v2', V2)]
    # # wo_by_frequency = sorted(frequencies, key=lambda x: x[1], reverse=True)
    # # free_orders = ['v-initial', 'v-final', 'v2']
    # # vi_set = set(['v-initial', 'vso', 'vos'])
    # # vf_set = set(['v-final', 'osv', 'ovs'])
    # # v2_set = set(['v2', 'svo', 'ovs'])
    # # top_3 = set([wo_by_frequency[0][0], wo_by_frequency[1][0], wo_by_frequency[2][0]])
    #
    # # if top_3 == vi_set:
    # #     wo = 'v-initial'
    # # elif top_3 == vf_set:
    # #     wo = 'v-final'
    # # elif top_3 == v2_set:
    # #     wo = 'v2'
    # # elif (wo_by_frequency[0][1]-wo_by_frequency[1][1] < .05) and (wo_by_frequency[1][1]-wo_by_frequency[2][1] < .05):
    # #     wo = 'free'
    # # elif (wo_by_frequency[0][0] in free_orders) and (wo_by_frequency[1][0] not in free_orders) and (wo_by_frequency[0][1]-wo_by_frequency[1][1] < .05):
    # #     wo = wo_by_frequency[1][0]
    # # else:
    # #     wo = wo_by_frequency[0][0]
    choices.__setitem__('word-order', wo)
    return choices

def collect_nouns(igt):
    nouns = []
    noun_tags = ['noun', 'n', 'NOUN', 'N', 'n.', 'N.']
    if igt.get('w-pos'):
        for pos in igt.get('w-pos'):
            if pos.value() in noun_tags:
                nouns.append(pos.attributes['alignment'])
    elif igt.get('gw-pos'):
        for pos in igt.get('gw-pos'):
            if pos.value() in noun_tags:
                if 'alignment' in igt.get('gw').get(pos.attributes['alignment']).attributes:
                    nouns.append(igt.get('gw').get(pos.attributes['alignment']).attributes['alignment'])
    return nouns

def is_det(gw):
    det = True
    det_list = ['the', 'a', 'this', 'that', 'these', 'those', 'no', 'every', 'each'] # KPH consider all and some (these may cause more ambiguity than they're worth)
    known_features = FeatDict.FeatureDictionary().all_keys
    lemma_count = 0
    grams = re.split('\.|-',gw)
    for gram in grams:
        if gram.islower() and gram not in known_features and gram not in det_list:
            lemma_count += 1
    if lemma_count > 0:
        det = False

    return det

def infer_det_noun_wo(xc, wo_choices):
    choices = MatrixChoices.ChoiceDict()
    orders = {'det-noun': 0, 'noun-det': 0, 'noun': 0, 'total': 0}
    unsupported_determiners = ['which', 'what'] #KPH This is a list of detirminers that aren't currently supported by inferred grammars. For now, we skip them
    determiners = []
    for igt in xc:
        if not igt.get('w-ds'):
            continue
        nouns = collect_nouns(igt)
        for noun in nouns:
            det = 0
            w_ds = igt.get('w-ds')
            for ds in w_ds:
                if 'dep' in ds.attributes and 'head' in ds.attributes:
                    if ds.value() == 'det':
                        if ds.attributes['head'] == noun:
                            det = int(ds.attributes['dep'][1:])
            if det == 0:
                orders['noun'] += 1
            else:
                gloss = igt.get('gw').get('gw' + str(det)).value()
                if gloss:
                    if any([x in gloss for x in unsupported_determiners]):
                        continue
                    if not is_det(gloss):
                        continue
                    else:
                        determiners.append((igt, 'w' + str(det)))
                else:
                    continue
                if det < int(noun[1:]):
                    orders['det-noun'] += 1
                elif int(noun[1:]) < det:
                    orders['noun-det'] += 1
                orders['total'] += 1
        pos_tier = igt.get('pos')
        if not pos_tier:
            pos_tier = igt.get('w-pos')
        if pos_tier:
            for pos in pos_tier:
                if pos.value() == 'det' or pos.value() == 'dem':
                    if (igt, pos.attributes['alignment']) not in determiners:
                        determiners.append((igt, pos.attributes['alignment']))
    if len(determiners) > 0:
        choices.__setitem__('has-dets', 'yes')
        if orders['det-noun'] > orders['noun-det']:
            choices.__setitem__('noun-det-order', 'det-noun')
        elif orders['noun-det'] > orders['det-noun']:
            choices.__setitem__('noun-det-order', 'noun-det')
        else:
            if len([x for x in determiners if x[1] == 'w1']) > 0:
                choices.__setitem__('noun-det-order', 'det-noun')
            elif len([x for x in determiners if int(x[1][1:]) > len(igt.get('w')) ]) > 0:
                choices.__setitem__('noun-det-order', 'noun-det')
            else:
                if wo_choices['word-order'] in ['vso', 'vos', 'svo' 'v-initial']:
                    choices.__setitem__('noun-det-order', 'noun-det')
                elif wo_choices['word-order'] in ['osv', 'sov', 'ovs', 'v-final']:
                    choices.__setitem__('noun-det-order', 'det-noun')
                else:
                    choices.__setitem__('noun-det-order', 'noun-det')
    else:
        choices.__setitem__('has-dets', 'no')
    choices.update(infer_lex_items.infer_det_lex_items(determiners))
    return choices

def infer_aux_wo(xc, verbs, case_system, igt2items, known_glosses, verb_tags, wo_choices):
    choices, auxiliaries, second_pos = infer_aux.infer_aux(xc, verbs, case_system, igt2items, known_glosses, wo_choices, verb_tags=verb_tags)
    return choices, auxiliaries, second_pos

def infer_word_order(xc, verbs, case_system, igt2items, verb_tags, known_glosses):
    choices = MatrixChoices.ChoiceDict()
    wo_choices = infer_basic_wo(xc)
    choices.update(wo_choices)
    choices.update(infer_det_noun_wo(xc,wo_choices))
    aux_wo, auxiliaries, second_pos = infer_aux_wo(xc, verbs, case_system, igt2items, known_glosses, verb_tags, wo_choices)
    choices.update(aux_wo)
    if second_pos:
        choices.__setitem__('word-order', 'v2')
    return choices, auxiliaries

def write_choices_file(choices, file):
    # writes the chocies file
    with open(file, 'w', encoding="UTF-8") as f:
        matrix_choices = MatrixChoices.ChoicesFile(choice_dict=choices)
        f.write(str(matrix_choices.choices))

if __name__ == '__main__':
    corpus = open(sys.argv[1], encoding='utf-8')
    xc = xigtxml.load(corpus)
    wo_choices, auxiliaries = infer_word_order(xc, verbs, case_system, igt2items, verb_tags)
    write_choices_file(wo_choices, 'C:/Users/Kristen/Documents/code/agg/wo_choices_'+sys.argv[2]+'_2013')
