#!/bin/bash
#  control file for the run-aggregation procedures
#  strategy is to setup condor submit files for 
## riples.condor 
## wo.condor
## case.condor
## lex.condor
## mom.condor
## merge.condor
## cleanup.condor
#  and then we use the condor_submit_dag of aggregation.dag in order to 
#  parallelize those that can be run in parallel

# The main thing this file has to do is check for an $INDIR and $FILEEXT and 
# create an $OUTDIR

# condor generic template
CONDOR_GEN='
Getenv = true
Log = /tmp/\$ENV(USER).condor.\$(Cluster)
Universe = vanilla
executable= %s 
arguments = %s 
output    = %s
error     = %s
+Research = True
Queue'


if [ $# -lt 1 ]; then
  echo "Usage: `basename $0` PATH/TO/INPUT-DIR FILE_EXT" >&2
  echo "  INPUT-DIR/ should contain plain text igt in odin-raw format" >&2
  echo "  FILE_EXT should give the last few letters of the file extention "
  echo "           for file in INPUT-DIR which contain the data, if omitted"
  echo "           all files in the INPUT-DIR will be processed as input data"

  exit 1 
else
  INDIR=`readlink -f $1`
  FILE_EXT="$2"
  OUTDIR=/workspace/agg$$
  AGGDIR=`pwd` # useful to get back here after cding around
  AGGCON="$OUTDIR/condor-files"
  mkdir $OUTDIR
  echo "$OUTDIR created for output" >&2
  mkdir $AGGCON
fi

# go ahead and make the igt and t files in OUTDIR (this should be a 
# dag PRE script, I guess)
for f in $INDIR/*$FILE_EXT; do
  scripts/odin2igt "$f" >> $OUTDIR/igt || exit 1
  scripts/odin2igt -t "$f" | sed -f scripts/tokenizer.sed >> $OUTDIR/t || exit 1
done

# make the riples.condor file
C_EXEC="$AGGDIR/igt2cfg"
C_ARGS="$OUTDIR/igt $OUTDIR/t $OUTDIR"
C_OUT="$OUTDIR/riples.out"
C_ERR="$OUTDIR/riples.err"
printf "$CONDOR_GEN" "$C_EXEC" "$C_ARGS" "$C_OUT" "$C_ERR" > $AGGCON/riples.condor

# the wo.condor file
C_EXEC="$AGGDIR/infer-wo"
C_ARGS="count2 $OUTDIR"
C_OUT="$OUTDIR/wo.out"
C_ERR="$OUTDIR/wo.err"
printf "$CONDOR_GEN" "$C_EXEC" "$C_ARGS" "$C_OUT" "$C_ERR" > $AGGCON/wo.condor

# the case.condor file
C_EXEC="$AGGDIR/infer-case"
C_ARGS="-c grams $OUTDIR"
C_OUT="$OUTDIR/case.out"
C_ERR="$OUTDIR/case.err"
printf "$CONDOR_GEN" "$C_EXEC" "$C_ARGS" "$C_OUT" "$C_ERR" > $AGGCON/case.condor

# the lex.condor file
C_EXEC="$AGGDIR/scripts/extract-lexicon"
C_ARGS="$OUTDIR/igt.all"
C_OUT="$OUTDIR/lex.out"
C_ERR="$OUTDIR/lex.err"
printf "$CONDOR_GEN" "$C_EXEC" "$C_ARGS" "$C_OUT" "$C_ERR" > $AGGCON/lex.condor

# the mom.condor file is different (inherited from David Wax, mainly)
# EMB 2015-03-23 Changing this one to make use of the $AGGDIR abstraction.
# Also: I note that $ENV(USER) and $(Cluster) are not getting interpreted,
# but removing \ doesn't help and in fact makes a non-sensicle condor.cmd file.

cat > "$AGGCON/mom.condor" <<EOF
Getenv = true
Log = /tmp/\$ENV(USER).condor.\$(Cluster)
Universe = vanilla
executable =  $AGGDIR/agg-svn/src/mom/code/driver.sh
getenv = true
+RequiresWholeMachine = true
Requirements = (Memory > 0 && TotalMemory >= (50*1024) )
arguments =  $OUTDIR/mom.conf
output     = $OUTDIR/mom.out
error      = $OUTDIR/mom.err
+Research = True
Queue
EOF

#TODO: apparently user has to provide mom with an iso
# the mom.conf file is also needed
# instead, we can take it from the meta lines of the odin file
ISO=`head $INDIR/*$FILE_EXT | $AGGDIR/scripts/get-iso.sh`
cat > "$OUTDIR/mom.conf" <<EOF
#General Header info
iso: $ISO
test ids: ids/test_list
data directory:$INDIR
data format:odin
file extension:$FILE_EXT
output id:test
output dir:$OUTDIR
parser:True
verb_re:\((VB|AUX|VBD|VBG|VBN|VBP|VBZ)\s\w:\d*\s*(\w*)\)
pred_re:None
match_group:2
pos tag:verb
pos_mark:None

#first step
align:Translation to Gloss Alignment
source:translation
target:gloss_feature
align iso:None
bidirectional:True
boost:match
boost multiplier:1

free memory:\$iso
exclude:True

#second step
align:Gloss to Language Alignment
align iso:\$iso
source:gloss
target:language
bidirectional:True
boost:align
boost multiplier:1

#third step
align:Gloss Verb to Language Verb Alignment
align iso:\$iso
source:gloss_word_morpheme
target:language_word_morpheme
bidirectional:True
boost:align
boost multiplier:1
tag:verb

#Output Choices
output:Choices Settings
compression:0.8
wordlist:\$choices_wordlist

#score:Mark Verb Translation
#source:gold/\$iso_dev.gold
#target:translation
#score iso:\$iso
#
#score:Mark Verb Gloss
#source:gold/\$iso_dev.gold
#target:gloss
#score iso:\$iso
#
#score:Mark Verb Language
#source:gold/\$iso_dev.gold
#target:language
#score iso:\$iso
EOF

# the merge.condor file
C_EXEC="$AGGDIR/scripts/merge_choices-mom2case"
C_ARGS="$OUTDIR/choices0 $OUTDIR/case.out"
C_OUT="$OUTDIR/merge.out"
C_ERR="$OUTDIR/merge.err"
printf "$CONDOR_GEN" "$C_EXEC" "$C_ARGS" "$C_OUT" "$C_ERR" > $AGGCON/merge.condor

# everything's ready, build the dag file and kick off
cat > "$AGGCON/aggregation.dag" <<EOF
# 
# File to setup run aggregation procedure as a condor_submit_dag

# this is the structure
# step 0: check args, get everything set up, run riples
#         
# step 1: in parallel:
#          infer-wo
#          infer-case
#          extract-lexicon
#          mom
# step 2: merge
#         merge_choices-mom2case
# step 3: cleanup

JOB 0 $AGGCON/riples.condor
JOB 1a $AGGCON/wo.condor
JOB 1b $AGGCON/case.condor
JOB 1c $AGGCON/lex.condor
JOB 1d $AGGCON/mom.condor DIR $AGGDIR/mom/code/
JOB 2 $AGGCON/merge.condor

PARENT 0 CHILD 1a 1b 1c 1d
PARENT 1a 1b 1c 1d CHILD 2
SCRIPT POST 1d $AGGDIR/scripts/extract-tar $OUTDIR/choicestest0.tar $OUTDIR
SCRIPT POST 2 $AGGDIR/scripts/clean-control $OUTDIR $AGGDIR
EOF

condor_submit_dag $AGGCON/aggregation.dag

