#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
eval-gender.py

This program evaluates a gender system by comparing it to a `choices' file.
Normally the gender system is one created by the infer-gender.py script
(based on enriched Interlinear Glossed Text (IGT) in Xigt format), and the
`choices' is one hand-built for the language.

Input(s):
gender_file - file in choices file format to evaluate
choices_file - file in choices file format to compare against
-options - command line options

Output:
evaluation results (sent to STDOUT)
   (Note that any errors or extra information are only sent to STDERR)

May 2015, Jeff Heath
  Initial development version
"""

import sys
import getopt
import re
import string


# a list of delimiters that can break up genders
gender_delim = string.punctuation
gender_delim = gender_delim.replace('_', '')
gender_delim = gender_delim.replace('-', '')


########################################################################
class choices_file:
	"""Class for loading choices file for gender comparison.
	
	This is a class which facilitates loading and storing information from a
	choices file as well as comparing the gender portions of different choices
	files.
	"""

	def __init__(self, filename):
		"""Constructor for a choices_file class object.
		
		Load information from the given choices file, specifically the names
		of the genders defined, and the all of the nouns defined in the lexicon,
		storing the orthography and predication of each by gender (if any).
		
		Args:
		  filename (str): file name/path for the choices file to load
		"""
		# regex's for finding specific data in choices files
		variable_patt = re.compile(r'\s*(\S+?)\s*=\s*(.+)\s*')
		gender_name_patt = re.compile(r'gender(\d+)_name', re.IGNORECASE)
		noun_feat_patt = re.compile(r'noun(\d+)_feat(\d+)_(.+)', re.IGNORECASE)
		noun_stem_patt = re.compile(r'noun(\d+)_stem(\d+)_(.+)', re.IGNORECASE)

		# define dictionary used to store the variable features
		choices_vars = {}
		# open and read the choices file as a dictionary of variable names/values
		try:
			f = open(filename, encoding='utf-8')
			for line in f:
				m = variable_patt.match(line)
				if m:
					choices_vars[m.group(1)] = m.group(2)
			f.close()
		except:
			print('Error loading choices file: {}'.format(filename), file=sys.stderr)
			sys.exit(1)
		
		# find the genders defined in the choices file
		self.genders = {}
		for k, v in choices_vars.items():
			# check if this is a gender name
			m = gender_name_patt.match(k)
			if m:
				# store under the gender name (gender number as value)
				self.genders[v] = m.group(1)
		
		# find all features and stems for noun types defined in the choices file
		nountypes = {}
		for k, v in choices_vars.items():
			# check if this is a noun feature
			line_type = 'features'
			m = noun_feat_patt.match(k)
			if not m:
				# not a noun feature, check if this is a noun stem
				line_type = 'stems'
				m = noun_stem_patt.match(k)
			# if either pattern matches, store the information (code common to both)
			if m:
				# get the noun type index
				nountype_idx = m.group(1)
				if nountype_idx not in nountypes:
					# we haven't seen this index yet, so add empty features/stems for it
					nountypes[nountype_idx] = {'features': [], 'stems': []}
				# store under noun type index, under line_type ('features' or 'stems')
				# store tuple of (line_type index, label, value)
				nountypes[nountype_idx][line_type].append( (m.group(2), m.group(3), v) )
		
		# now that we've collected all of this noun type info, put it in useful format
		self.nouns = {}
		for nountype_idx in nountypes:
			# add the nouns of this noun type to the noun dictionary, under the proper gender
			
			# first must identify the gender for this noun type
			# start with list of feature tuples (feature index, label, value)
			features = nountypes[nountype_idx]['features']
			# make sure that name/value pairs are next to each other, in that order
			features.sort()
			# default gender
			gender_list = ['']
			# look at the tuples a pair at a time
			for feat_pair in zip(features[0::2], features[1::2]):
				if feat_pair[0][1] == 'name' and feat_pair[0][2] == 'gender':
					# found the gender pair, use value from second element
					gender_text = feat_pair[1][2]
					# it could be a list of genders, get rid of any punctuation
					# (could be "m, f", "m/f", "[m,f]", "m&f")
					# but DON'T break up by underscores/dashes - could be used in gender name
					gender_text = re.sub('['+gender_delim+']', ' ', gender_text)
					gender_list = gender_text.split()
					# if this gender wasn't recorded previously, then add it to the list
					for gender in gender_list:
						if gender not in self.genders:
							# add this gender, using 'X' as the gender number
							self.genders[gender] = 'X'
					# found the gender, don't need to look further
					break
			
			# now that we have the gender_list for this noun type ([''] for none)
			# make sure there is a section in the nouns dictionary for the stems
			for gender in gender_list:
				if gender not in self.nouns:
					self.nouns[gender] = {}
			
			# now process the stems for this noun type
			# start with list of stem tuples (stem index, label, value)
			stems = nountypes[nountype_idx]['stems']
			# make sure that orth/pred pairs are next to each other, in that order
			stems.sort()
			# look at the tuples a pair at a time
			for stem_pair in zip(stems[0::2], stems[1::2]):
				if stem_pair[0][1] != 'orth' or stem_pair[1][1] != 'pred':
					print('Unexpected stem labels: {} {}'.format(stem_pair[0][1],
					                                             stem_pair[1][1]), file=sys.stderr)
				else:
					for gender in gender_list:
						if stem_pair[0][2] in self.nouns[gender]:
							# don't yet have the means to handle duplicate orthography values
							print('Duplicate stem orth: {}'.format(stem_pair[0][2]), file=sys.stderr)
							# get the two pred values
							pred1 = stem_pair[1][2]
							pred2 = self.nouns[gender][stem_pair[0][2]]
							# save the "lesser" one, just so that multiple runs get same value
							if pred1 < pred2:
								self.nouns[gender][stem_pair[0][2]] = pred1
						else:
							# save the orth and pred values
							self.nouns[gender][stem_pair[0][2]] = stem_pair[1][2]
		# as final step of constructor, print out genders and nouns found
		for gender in sorted(self.nouns):
			print("gender: '{}'".format(gender))
			for k, v in sorted(self.nouns[gender].items()):
				print('  {}: {}'.format(k, v))


	def get_genders(self):
		"""Returns the list of genders in this choices file.
		"""
		return self.genders


	def get_nouns(self, gender):
		"""Returns the dictionary of nouns for the given gender.
		"""
		
		if gender in self.nouns:
			nouns = self.nouns[gender]
		else:
			nouns = {}
		return nouns

	def compare_gender_numbers(self, compare_choices):
		"""Compare the number of genders with the given choices file.
		
		Print out a message whether there are the same number of genders in the two
		choices files. If the gender names are identical, print an additional
		message to that fact and return True, otherwise return False.

		Args:
		  compare_choices (choices_file object): choices file to compare
		Returns:
		  True, if the genders are identical
		"""
		# compare the number of genders
		print('  Evaluation genders: {}'.format(sorted(self.genders.keys())))
		compare_genders = compare_choices.get_genders()
		print('  Comparison genders: {}'.format(sorted(compare_genders.keys())))
		same_genders = False
		n_gen = len(self.genders)
		n_comp_gen = len(compare_genders)
		if n_gen != n_comp_gen:
			print('  Not the same number of genders ({} vs {})'.format(n_gen, n_comp_gen))
		else:
			print('  The same number of genders ({})'.format(n_gen))
			if sorted(self.genders.keys()) == sorted(compare_genders.keys()):
				print('  And in fact, the same gender names!')
				same_genders = True
		return same_genders


	def compare_nouns(self, compare_choices):
		"""Compare the list of nouns with given choices file.

		For each gender, print information about which nouns match between
		the choices files.

		Args:
		  compare_choices (choices_file object): choices file to compare
		"""
		# consider each of the genders
		for gender in sorted(self.genders):
			print("For gender '{}'".format(gender))
			# get the comparison nouns dictionary for this gender
			compare_nouns = compare_choices.get_nouns(gender)
			if gender in self.nouns:
				# make a copy of my nouns dictionary for this gender (we'll modify)
				my_nouns = self.nouns[gender].copy()
				# use set operations to compare
				my_nouns_set = set(my_nouns)
				compare_nouns_set = set(compare_nouns)
				# find the nouns that are the same, using intersection
				same_nouns_set = my_nouns_set.intersection(compare_nouns_set)
				same_pred = sum(1 for n in same_nouns_set if my_nouns[n] == compare_nouns[n])
				if len(same_nouns_set) != 0:
					print("  Identical nouns: {}, of which {} ({}%) have identical predications".
						  format(len(same_nouns_set), same_pred, int(same_pred*100/len(same_nouns_set))))
				else:
					print("  Identical nouns: 0, of which 0 (0%) have identical predications")
				print('    {}'.format(sorted(same_nouns_set)))
				# find the nouns that aren't in compare_choices, using difference
				diff_nouns_set = my_nouns_set.difference(compare_nouns_set)
				print("  Nouns not found in comparison choices file: {}".format(len(diff_nouns_set)))
				print('    {}'.format(sorted(diff_nouns_set)))
				# find the nouns in compare_choices that aren't in eval_choices, using difference
				diff_nouns_set = compare_nouns_set.difference(my_nouns_set)
				print("  Extra nouns found in comparison choices file: {}".format(len(diff_nouns_set)))
			else:
				# no nouns for this gender
				print("  Identical nouns: 0")
				print("  Nouns not found in comparison choices file: 0")
				print("  Extra nouns found in comparison choices file: {}".format(len(compare_nouns)))



def usage():
	"""Print command line usage information."""
	print('Usage: eval-gender.py [options] eval_choices compare_choices')
	print('  -h           : display this help')


# parse options and arguments
try:
	# option h doesn't have argument
	opts, cmdln_args = getopt.getopt(sys.argv[1:], 'h')
except getopt.GetoptError as err:
	print(str(err),file=sys.stderr)
	usage()
	sys.exit(1)

for opt, arg in opts:
	if opt == '-h':
		usage()
		sys.exit()

# make sure there are two command line parameters
if len(cmdln_args) != 2:
	raise(ValueError("Two parameters are required: gender_file choices_file"))

print('Processing the choices file to evaluate...')
eval_choices = choices_file(cmdln_args[0])
print('Processing the choices file used for comparison...')
compare_choices = choices_file(cmdln_args[1])

print('\nEvaluation of the number of genders:')
same_genders = eval_choices.compare_gender_numbers(compare_choices)

print('\nEvaluation of the nouns distinguished by gender:')
if same_genders:
	# we can just compare the noun lists directly
	eval_choices.compare_nouns(compare_choices)
else:
	print('  Noun comparison for differing genders is not yet implemented')

