

import sys
import getopt
import re
from xigt.codecs import xigtxml
from xigt_nltk_pstree import format_ptb_tree
from nltk.tree import ParentedTree
import nltk_tgrep
from collections import defaultdict


# Attemps to infer roots that exhibit valence alternations. Outputs a
# list of <root>,<n> pairs (one per line), where <n> is 0 (no alternations) or 1 (alternations).
# (This is simplified from a non-binary count of alternations.)

# Method descriptions:
#  * baseline: use # of agreement grams as proxy for valence
#  * grams: use presence/absence of explicit valence-changing grams as proxy
#  * finite: exclude verbs with nonfinite-marking grams
#  * nocop: exclude copular verbs in the translation tier
#  * twps: use translation phrase structure to infer valence
#
# Valid combinations (NOT checked):
#  * baseline
#  * baseline,finite
#  * grams
#  * grams,finite
#  * twps
#  * twps,nocop



def usage():
    """Print command line usage information."""
    print('Usage: find-valence-alternations.py [options] Xigt-input [Xigt-input ...]')
    print('  -h                       : display this help')
    print('  -m <meth1>[,<meth2>,...] : method(s) to use')
    print('      available methods: baseline,grams,finite,twps,nocop')
    print('      (default = baseline)')
    print('  -n <N>                   : stop after <N> instances')
    print('  -s                       : output summary counts')
    print('  -v                       : output verbose results')


class RootCounts:
    def __init__(self):
        self.arg_gram_cnt = 0
        self.val_gram_cnt = 0

class BreakLoop( Exception ):
    pass

def print_verbose(root, usages):
    print('===============')
    print('root:' + str(root))
    print('usages:' + str(len(usages)))
    for (igt, counts, usage) in sorted(usages, key=lambda u: u[1].arg_gram_cnt):
        print("{}\t{}\t{}".format(igt.id, counts.arg_gram_cnt, usage))
        print(igt.get('n')['n1'].value())
        print(igt.get('n')['n2'].value())
        print(igt.get('n')['n3'].value())
        print()

def clamp(minimum, x, maximum):
    return max(minimum, min(x, maximum))

# parse options and arguments
try:
    opts, cmdln_args = getopt.getopt(sys.argv[1:], 'hm:n:sv')
except getopt.GetoptError as err:
    print(str(err), file=sys.stderr)
    usage()
    sys.exit(1)

stop_n = None
methods = None
verbose = False
printSummary = False
for opt, arg in opts:
    if opt == "-h":
        usage()
        sys.exit()
    if opt == "-m":
        methods = arg.split(',')
        for method in methods:
            if method not in ['baseline', 'grams', 'finite', 'twps', 'nocop']:
                print('Unknown method:', method)
                usage()
                sys.exit(1)
    if opt == "-n":
        stop_n = int(arg)
    if opt == "-s":
        printSummary = True
    if opt == "-v":
        verbose = True

valence_grams = ['ANTIP', 'APPL', 'BEN', 'CAUS', 'INS', 'PASS', 'RECP', 'REFL']
nonfinite_grams = ['INF', 'CVB', 'NMLZ']
copulas = ['be', 'are', 'is', 'were', 'was']

use_wps = ("baseline" in methods) or ("grams" in methods)

for filename in cmdln_args:
    try:
        print('Processing Xigt file: ' + filename, file=sys.stderr)
        xigt_corpus = xigtxml.load(open(filename, encoding='utf-8'), mode='transient')
    except:
        print('Error loading Xigt corpus file: {}'.format(filename), file=sys.stderr)
        continue

    total = 0

    missing_tiers = 0
    parse_errors = 0

    roots = defaultdict(list)
    skipped_roots = []

    for igt in iter(xigt_corpus):
        total += 1
        # progress dots
        print('.', end="", flush=True, file=sys.stderr)

        # is there a w-ps?
        #   does it match?
        #   -> process it
        wpsTier = igt.get('w-ps')
        twpsTier = igt.get('tw-ps')
        if use_wps and wpsTier:
            # find all the VB* elements
            wVB_ids = [x.alignment for x in wpsTier if x.value()[:2] == 'VB']
            wTier = igt.get('w')
            mTier = igt.get('m')
            gTier = igt.get('g')

            for wVB in [wTier[x] for x in wVB_ids]:
                root = None
                include_root = False
                arg_gram_cnt = 0
                val_gram_cnt = 0
                refs = wVB.referrers()
                segs = refs.get('segmentation')

                if segs:

                    for seg in segs:
                        g_refs = mTier[seg].referrers().get('alignment')
                        gram = gTier[g_refs[0]].value().strip()
                        if mTier[seg].type == 'Root':
                            root = mTier[seg]
                            include_root = True
                        else:
                            # prefix/suffix
                            if "finite" in methods and gram in nonfinite_grams:
                                include_root = False
                                skipped_roots.append(root.get_content())
                                break

                            if re.match(r'\[?\d', gram):
                                arg_gram_cnt += 1
                            elif gram in valence_grams:
                                val_gram_cnt += 1
                            else:
                                pass

                if include_root:
                    counts = RootCounts()
                    counts.arg_gram_cnt = arg_gram_cnt
                    counts.val_gram_cnt = val_gram_cnt
                    roots[root.get_content()].append( (igt, counts, wVB.get_content()) )
        elif twpsTier:
            # use tw-ps
            try:
                ptb = format_ptb_tree(igt, 'tw-ps')
                tree = ParentedTree.fromstring(ptb)
                root = None

                # find the root
                vb = nltk_tgrep.tgrep_nodes(tree, "/VB/ > (/VP/ > /S|FRAG/)")
                aTier = igt.get('a')
                if vb and aTier:
                    twTier = igt.get('tw')
                    mTier = igt.get('m')
                    gTier = igt.get('g')
                    vb_items = [x for x in twTier if x.value() == vb[0][0]]

                    # find the aligned morpheme
                    if vb_items:
                        vb_targets = [x.attributes['target'] for x in aTier if x.attributes['source'] == vb_items[0].id]
                    else:
                        continue
                    if vb_targets:
                        m_id = gTier[vb_targets[0]].attributes['alignment']
                        root = mTier[m_id].value()
                    else:
                        continue

                    if 'nocop' in methods and vb_items[0].value() in copulas:
                        continue

                    sbj = nltk_tgrep.tgrep_nodes(tree, "/NP|NNP/ > (/S|FRAG/ < /VP/)")
                    obj = nltk_tgrep.tgrep_nodes(tree, "/NP|NNP/ > (/VP/ < /VB/)")
                    xobj = nltk_tgrep.tgrep_nodes(tree, "/NP|NNP/ $ (/NP|NNP/ > (/VP/ < /VB/))")

                    val_arg_cnt = 0
                    if sbj: val_arg_cnt += 1
                    if obj: val_arg_cnt += 1
                    if xobj: val_arg_cnt += len(xobj)
                    counts = RootCounts()
                    counts.val_arg_cnt = val_arg_cnt

                    roots[root].append((igt, counts, root))
                else:
                    continue

            except ValueError:   #ParentedTree throws this for bad trees
                parse_errors += 1
        else:
            missing_tiers += 1

        if stop_n and total >= stop_n:
            break

    # closeout for grams methods
    for root in skipped_roots:
        counts = RootCounts()
        roots[root].append(('<SKIPPED>', counts, root))

    # newline after progress dots
    print(file=sys.stderr)

    alternations_cnt = 0
    for (root, usages) in sorted(roots.items(), key=lambda u: u[0]):
        if len(usages) == 1:
            continue    # skip hapax roots
        valencies = set()
        for (_, counts, _) in usages:
            if "baseline" in methods:
                valencies.add(clamp(0, counts.arg_gram_cnt, 1))
            elif "grams" in methods:
                valencies.add(clamp(0, counts.val_gram_cnt, 1))
            elif "twps" in methods:
                valencies.add(clamp(0, counts.val_arg_cnt, 1))
        if not verbose:
            alternations = len(valencies) - 1
            print("{},{}".format(str(root), str(alternations)))
        if len(valencies) > 1:
            alternations_cnt += 1
            if verbose:
                print_verbose(root, usages)

    if printSummary:
        print()
        print("  total distinct roots: {}".format(len(roots)))
        print("w/valence alternations: {}".format(alternations_cnt))
        print("          parse errors: {}".format(parse_errors))
        print("         missing tiers: {}".format(missing_tiers))




